/***************************************************************************
 *   Copyright (C) 2017 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <math.h>
#include "common.h"
#include "macros.h"
#include "classifiers.h"

scalar ClassifierDatasetMargin(const t_Parameters *parm, t_Classifier classifier, scalar **train_data, int train_instances, int *train_labels, scalar **data, int instances, int *labels,
                                     int inputs, int classes, int *subspace, scalar *min_margin, t_Learner *learner, scalar *margin, int *subset, t_Ensemble *ensemble, bool single)
{
    t_Learner llearner;
    t_Ensemble lensemble;

    // Train classifier?
    if (single) {
        if (!learner)
            ClassifierTraining(parm, classifier, train_data, train_instances, train_labels, inputs, classes, NULL, subspace, subset, true, &llearner, NULL);
        else
            llearner = *learner;
    } else {
        if (!ensemble)
            ClassifierTraining(parm, classifier, train_data, train_instances, train_labels, inputs, classes, NULL, subspace, subset, true, NULL, &lensemble);
        else
            lensemble = *ensemble;
    }

    // Get margin
    scalar ave_margin = 0.0, instance_margin;
    *min_margin = 10.0E20;
    for (int k = 0; k < instances; k++) {
        instance_margin = ClassifierMargin(parm, classifier, data[k], inputs, classes, NULL, subspace, labels[k], train_data, train_instances, train_labels, subset, &llearner, &lensemble);

        if (margin)
            margin[k] = instance_margin;

        ave_margin += instance_margin;

        if (instance_margin < *min_margin)
            *min_margin = instance_margin;
    }

    if (!learner)
        ClassifierFree(parm, &llearner, classifier, classes, inputs, NULL);

    return ave_margin/instances;
}
