/***************************************************************************
 *   Copyright (C) 20089 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "macros.h"
#include "typex.h"
#include "classifiers.h"
#include "data.h"
#include "ensemble.h"

t_ClassModel *TrainMultiLayerPerceptron(const t_Parameters *parm, scalar **data, int instances, int classes, int inputs, int *labels, int *subspace, int *subset, int **index)
{
    t_ClassModel *classifier;
    scalar **desired, **t_data;

    MALLOC(classifier, t_ClassModel, 1, "TrainMultiLayerPerceptron: classifier");

    AllocateScalarMatrix(&desired, instances, classes);

    // Training data
    MALLOC(t_data, scalar *, instances, "TrainMultiLayerPerceptron: t_data");
    int next = 0;
    for (int i = 0; i < instances; i++)
        if (!subset || subset[i]) {
            for (int j = 0; j < classes; j++)
                desired[next][j] = (labels[i] == j);
            t_data[next] = data[i];
            next++;
        }

    AllocateClassifier(parm, classifier, inputs, parm->nodes, true, instances, classes, 0);
    GenerateNewClassifier(parm, classifier, true, next, inputs, classes);
    for (int i = 0; i < inputs; i++)
        classifier->subspace[i] = (scalar) subspace[i];
    BP(parm, classifier, 0, parm->n_epochs, parm->epoch_size, t_data, next, desired, inputs, classes, index);

    FreeScalarMatrix(&desired, instances);
    FREE(t_data);

    return classifier;
}

scalar MLPMargin(const t_Parameters *parm, t_ClassModel *mlp, scalar *instance, int classes, int label, int *index)
{
    scalar outputs[classes];

    int category = NetworkActualOutput(parm, mlp, instance, outputs, classes, index);

    // Wrong classification? Return negative margin
    if (category != label)
        return outputs[label] - outputs[category];

    // Otherwise get second maximum
    int second = (category != 0) ? 0 : 1;
    for (int i = 0; i < classes; i++)
        if (i != category && outputs[i] > outputs[second])
            second = i;

    return outputs[category] - outputs[second];
}
