/***************************************************************************
*   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
*   npedrajas@uco.es                                                      *
*                                                                         *
*   This program is FREE software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/
/*
 * Input/Output functions.
 */
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <stdbool.h>
#include <string.h>
#include "common.h"
#include "macros.h"
#include <string.h>

#define MAX_LINE_LENGTH 1000000

// Read a subset of instances from a file using direct access
void ReadSubsetWithDirectAccess(scalar **data, int *labels, int instances, int inputs, char *data_file, long int **address, int train_instances, int *list)
{
    FILE *in_file;
    char *line;
    int feature, label;
    scalar value;

    MALLOC(line, char, MAX_LINE_LENGTH, "ReadSubsetWithDirectAccess: line");

    // If first time read index file
    if (*address == NULL) {
        FILE *index;
        char name[1000];

        sprintf(name, "%s.idx", data_file);
        if ((index = fopen(name, "r")) == NULL)
            OnErrorExit(ERR_FILE, "Cannot open index file");

        MALLOC(*address, long int, train_instances, "ReadSubsetWithDirectAccess: *address");
        for (int i = 0; i < train_instances; i++)
            if (!fscanf(index, "%ld", &(*address)[i]))
                OnErrorExit(ERR_READ, "Error reading index file");
        fclose(index);
    }

    // Check data file is in sparse format
    if ((in_file = fopen(data_file, "r")) == NULL) {
        sprintf(line, "Cannot open data file '%s'", data_file);
        OnErrorExit(ERR_FILE, line);
    }
    if (fgets(line, MAX_LINE_LENGTH, in_file) == NULL)
        OnErrorExit(ERR_READ, "Error reading data file");
    if (strncasecmp(line, "[SPARSE]", 8))
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Direct access only available for SPARSE format");
    // If a number after "[SPARSE]" is the index of the first feature, otherwise assume 1
    int shift = 1;
    if (strlen(line) > 9)
        shift = atoi(line+8);

    // Space for data must be allocated
    for (int i = 0; i < instances; i++) {
        // Jump to instance
        fseek(in_file, (*address)[list[i]], SEEK_SET);

        for (int j = 0; j < inputs; j++)
            data[i][j] = 0.0;

        // Read line
        if (fgets(line, MAX_LINE_LENGTH, in_file) == NULL)
            OnErrorExit(1, "ReadSubsetWithDirectAccess: Problems reading file");

        // Read pairs feature:value separated by spaces or tabs until reaching label
        int offset = 0;
        while (GetNextPair(line, &offset, &feature, &value, &label)) {
            if (feature-shift >= inputs || feature-shift < 0)
                OnErrorExit(ERR_READ, "Bad read feature");
            data[i][feature-shift] = value;
        }

        // Label
        labels[i] = label;
    }

    fclose(in_file);

    FREE(line);
}

