/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "common.h"
#include "macros.h"
#include "svm.h"

void LoadGlobalParameters(t_Parameters *parm, char *f_name, char *program)
{
    extern char *isalgorithm[], *fsalgorithm[], *ensembles[], *partitions[], *performance[], /**threshold_methods[],*/ *fralgorithm[], *votes_recording[], *evolutionary_scheme[], *proximity[],
           *discretizer[], *search[], *estimation[], *margin_method[], *sampling[], *learner[], *implementation[], *feature_subset_evaluation[], *isfsalgorithm[], *nbcinterval[], *ps_strategy[],
           *samplingmethod[], *ml_method[], *ml_instance_selection[], *mlpartition[], *merge_method[], *class_imbalance_method[], *ml_kperlabel_method[], *external[], *split_measure[],
           *ml_ensemble_methods[], *ml_class_imbalance_method[], *ml_ext_base_method[], *ml_ext_method[], *smote_variant[], *subset_combination[], *subset_evaluation[], *ml_feature_selection[],
           *ml_resampling[], *label_propagation[];

    extern int n_isalgorithm, n_fsalgorithm, n_ensembles, n_partitions, n_performance, n_fralgorithm, n_votes_recording, n_evolutionary_scheme, n_proximity, n_discretizer, n_class_imbalance_method,
           n_search, n_estimation, n_margin_method, n_sampling, n_classifier, n_implementation, n_feature_subset_evaluation, n_isfsalgorithm, n_nbcinterval, n_samplingmethod, n_ml_ext_method,
           n_mlmethod, n_ml_instance_selection, n_mlpartition, n_merge_method, n_ml_kperlabel_method, n_ml_ensemble_methods, n_ml_class_imbalance_method, n_ml_ext_base_method, n_split_measure,
           n_smote_variant, n_subset_combination, n_subset_evaluation, n_ml_feature_selection, n_base_classifier, n_ml_base_classifier, n_ml_resampling, n_ps_strategy, n_label_propagation;

    FILE *f;
    char *undersampling[] = {"Random", "Cluster based"}, *transfer[] = {"Log", "HypTan", "Sin", "Linear"};
    int n_undersampling = 2, n_transfer = 4;

    SetDefaultsParameterValues(parm);

    // Open file
    if ((f = fopen(f_name, "r")) == NULL)
        OnErrorExit(ERR_PARAM, "Cannot open parm file");

    // Verifies program if not NULL
    // Must be: PROGRAM: "program name"
    if (program) {
        char line[MAX_LENGTH];
        fgets(line, MAX_LENGTH, f);
        if (strncmp(line+9, program, strlen(program))) {
            printf("Expected program: %s Found: %s (%s)\n", program, line+9, line);
            OnErrorExit(ERR_PARAM, "Parameters not for this program");
        }
    }

    fprintf(parm->out_file, "Parameters (* = default)\n───────────────────────────\n");

    // General
    LoadString(f, "Data directory", "Data/MachineLearning/Classification", parm->datad, parm->out_file);
    parm->evaluate_standard_knn = LoadBoolean(f, "Evaluate standard k-NN", false, parm->out_file);
    parm->fasta_file = LoadBoolean(f, "FASTA file", false, parm->out_file);
    parm->huge_dataset = LoadBoolean(f, "Huge dataset", false, parm->out_file);
    parm->obtain_auc = LoadBoolean(f, "Obtain AUC", true, parm->out_file);
    parm->only_save_selection = LoadBoolean(f, "Only save selection", false, parm->out_file);
    parm->output_2class_measures = LoadBoolean(f, "Output 2 class measures", true, parm->out_file);
    parm->print_verbose = LoadBoolean(f, "Verbose standard error printing", false, parm->out_file);
    parm->save_outputs = LoadBoolean(f, "Save outputs", false, parm->out_file);
    parm->seed = LoadInteger(f, "Seed", 0, parm->out_file);
    parm->bias = LoadBoolean(f, "Bias", false, parm->out_file);;
    parm->show_progress = LoadInteger(f, "Show progress", 0, parm->out_file);
    parm->force_sparse_format = LoadBoolean(f, "Force sparse format", false, parm->out_file);
    parm->sparse_format = LoadBoolean(f, "Sparse format", parm->force_sparse_format, parm->out_file);
    parm->typify = LoadBoolean(f, "Typify", true, parm->out_file);
    parm->use_all_data = LoadBoolean(f, "Use all data", false, parm->out_file);
    parm->verbose = LoadInteger(f, "Verbose", -1, parm->out_file);
    parm->remove_features_after_selection = LoadBoolean(f, "Remove features after selection", false, parm->out_file);
    parm->remove_constant_features = LoadBoolean(f, "Remove constant features", false, parm->out_file);
    parm->evaluate_margin = LoadBoolean(f, "Evaluate margin", true, parm->out_file);
    parm->time_all = LoadBoolean(f, "Time all", false, parm->out_file);
    parm->minimum_info = LoadBoolean(f, "Minimum information", false, parm->out_file);
    parm->precision_for_zero = LoadReal(f, "Precision for zero", 0.0, parm->out_file);

    // Classifier evaluation
    parm->evaluate_with[SVM] = LoadBoolean(f, "Evaluate with SVM", false, parm->out_file);
    parm->evaluate_with[C45] = LoadBoolean(f, "Evaluate with C4.5", false, parm->out_file);
    parm->evaluate_with[kNN] = LoadBoolean(f, "Evaluate with k-NN", false, parm->out_file);
    parm->evaluate_with[MLP] = LoadBoolean(f, "Evaluate with MLP", false, parm->out_file);
    parm->evaluate_with[ELM] = LoadBoolean(f, "Evaluate with ELM", false, parm->out_file);
    parm->evaluate_with[NBC] = LoadBoolean(f, "Evaluate with NBC", false, parm->out_file);   // Internal implementation
    parm->evaluate_with[DT_CA] = LoadBoolean(f, "Evaluate with DT(Accuracy)", false, parm->out_file);   // Personal implementation
    parm->evaluate_with[DT_C45] = LoadBoolean(f, "Evaluate with DT(ID3)", false, parm->out_file);
    parm->evaluate_with[DT_CART] = LoadBoolean(f, "Evaluate with DT(CART)", false, parm->out_file);
    parm->evaluate_with[DT_C45] = LoadBoolean(f, "Evaluate with DT(C4.5)", false, parm->out_file);
    parm->evaluate_with[IBT] = LoadBoolean(f, "Evaluate with IBT", false, parm->out_file);
    parm->evaluate_with[RF] = LoadBoolean(f, "Evaluate with RF", false, parm->out_file);
    parm->evaluate_with[NN] = LoadBoolean(f, "Evaluate with 1-NN", false, parm->out_file);
    parm->evaluate_with[GNBC] = LoadBoolean(f, "Evaluate with GaussianNBC", false, parm->out_file);
    parm->evaluate_with[LR] = LoadBoolean(f, "Evaluate with LR", false, parm->out_file);
    parm->evaluate_with[RForest] = LoadBoolean(f, "Evaluate with RForest", false, parm->out_file);  // Own implementation
    parm->evaluate_with[ExtraTree] = LoadBoolean(f, "Evaluate with ExtraTree", false, parm->out_file);  // Own implementation

    // Ensemble evaluation
    parm->evaluate_with[OVO] = LoadBoolean(f, "Evaluate with OVO", false, parm->out_file);
    parm->evaluate_with[AdaBoostM1] = LoadBoolean(f, "Evaluate with AdaBoost.M1", false, parm->out_file);
    parm->evaluate_with[MadaBoost] = LoadBoolean(f, "Evaluate with MadaBoost", false, parm->out_file);
    parm->evaluate_with[MultiBoost] = LoadBoolean(f, "Evaluate with MultiBoost", false, parm->out_file);
    parm->evaluate_with[Bagging] = LoadBoolean(f, "Evaluate with Bagging", false, parm->out_file);
    parm->evaluate_with[GentleBoost] = LoadBoolean(f, "Evaluate with GentleBoost", false, parm->out_file);
    parm->evaluate_with[ReweightBoost] = LoadBoolean(f, "Evaluate with ReweightBoost", false, parm->out_file);
    parm->evaluate_with[FloatBoost] = LoadBoolean(f, "Evaluate with FloatBoost", false, parm->out_file);
    parm->evaluate_with[FeatureBoost] = LoadBoolean(f, "Evaluate with FeatureBoost", false, parm->out_file);
    parm->evaluate_with[ARCING] = LoadBoolean(f, "Evaluate with Arcing", false, parm->out_file);

    // External classifiers
    parm->evaluate_with_external[RandomForest] = LoadBoolean(f, "Evaluate with Random Forest", false, parm->out_file);
    parm->evaluate_with_external[XGBoost] = LoadBoolean(f, "Evaluate with XGBoost", false, parm->out_file);
    parm->evaluate_with_external[NaiveBayes] = LoadBoolean(f, "Evaluate with Naive Bayes", false, parm->out_file);
    parm->evaluate_with_external[DeepLearning] = LoadBoolean(f, "Evaluate with Deep Learning", false, parm->out_file);
    parm->evaluate_with_external[DecisionTree] = LoadBoolean(f, "Evaluate with Decision Tree", false, parm->out_file);
    parm->evaluate_with_external[PLSDA] = LoadBoolean(f, "Evaluate with PLS-DA", false, parm->out_file);
    parm->evaluate_with_external[SVC] = LoadBoolean(f, "Evaluate with SVC", false, parm->out_file);
    parm->evaluate_with_external[LDA] = LoadBoolean(f, "Evaluate with LDA", false, parm->out_file);
    parm->evaluate_with_external[QDA] = LoadBoolean(f, "Evaluate with QDA", false, parm->out_file);
    parm->evaluate_with_external[GPML] = LoadBoolean(f, "Evaluate with GPML", false, parm->out_file);
    parm->evaluate_with_external[Ridge] = LoadBoolean(f, "Evaluate with Ridge Regression", false, parm->out_file);
    parm->evaluate_with_external[ERT] = LoadBoolean(f, "Evaluate with Extremely Randomized Trees", false, parm->out_file);
    parm->evaluate_with_external[KRR] = LoadBoolean(f, "Evaluate with Kernel Ridge Regression", false, parm->out_file);
    parm->evaluate_with_external[AdaBoostR] = LoadBoolean(f, "Evaluate with AdaBoost.R", false, parm->out_file);
    parm->evaluate_with_external[AdaBoostD] = LoadBoolean(f, "Evaluate with AdaBoost.D", false, parm->out_file);

    // External learner for scikit Boosting
    parm->external_base = LoadEnumerate(f, "Boosting external learner", external, N_EXTERNAL, (int) DecisionTree, parm->out_file);

    // Training data
    parm->training_max_size = LoadInteger(f, "Maximum number of instances for training", 0, parm->out_file);
    parm->max_instances = LoadInteger(f, "Maximum number of instances for learning", 0, parm->out_file);
    parm->evaluate_train_error = LoadBoolean(f, "Evaluate training error", false, parm->out_file);
    parm->sample_using_kmeans = LoadBoolean(f, "Sample using k-Means", false, parm->out_file);                       // Only apply for boostfs
    parm->stratified = LoadBoolean(f, "Stratified sampling", true, parm->out_file);
    parm->sample_using_kmeans_subset_size = LoadInteger(f, "Sample using k-Means subset size", 0, parm->out_file);   // Only apply for boostfs

    // Parallel version of 1-NN
    parm->test_1nn_slaves = LoadInteger(f, "Test 1-NN slaves", 0, parm->out_file);

    // What to do
    parm->feature_selection = LoadBoolean(f, "Feature selection", false, parm->out_file);      // Feature selection?
    parm->instance_selection = LoadBoolean(f, "Instance selection", false, parm->out_file);      // Instance selection?

    // Features
    parm->randomize_inputs_for_fs = LoadBoolean(f, "Randomize inputs for feature selection", false, parm->out_file);

    // Data tests
    parm->duplicate_instances = LoadBoolean(f, "Duplicate instances", false, parm->out_file);
    parm->duplicate_instances_with_noise = LoadBoolean(f, "Duplicate instances with noise", false, parm->out_file);
    if (parm->duplicate_instances_with_noise)
        parm->duplicate_instances = true;
    parm->duplicate_features = LoadBoolean(f, "Duplicate features", false, parm->out_file);
    parm->duplicate_features_ratio = LoadInteger(f, "Duplicate features ratio", 1, parm->out_file);
    parm->save_selection = LoadBoolean(f, "Save selection", false, parm->out_file);

    // Use more than one independent training set
    parm->maximum_training_set_size = LoadInteger(f, "Maximum training set size", 0, parm->out_file);

    // Boost-based IS method (and some FS)
    parm->boost_based_filter_noise_k = LoadInteger(f, "Noise filtering k", 1, parm->out_file);
    parm->second_chances = LoadInteger(f, "Second chances", 1, parm->out_file);
    parm->use_averaged_accuracy_for_stop = LoadBoolean(f, "Use averaged accuracy for stop", true, parm->out_file);
    parm->evaluate_with_loo = LoadBoolean(f, "Evaluate with leave one out", false, parm->out_file);
    parm->search_heuristic = LoadEnumerate(f, "Heuristic search", search, n_search, (int) SFS, parm->out_file);
    parm->plus_l = LoadInteger(f, "Plus l", 3, parm->out_file);
    parm->minus_r = LoadInteger(f, "Minus r", 1, parm->out_file);
    parm->multiboost_reinit_iterations = LoadInteger(f, "eMultiBoost reinitialization iterations", 10, parm->out_file);
    parm->boostedis_ensemble_method = LoadEnumerate(f, "Boosting-Based IS ensemble method", ensembles, n_ensembles, (int) eMultiBoost, parm->out_file);
    parm->bbis_select_best_instance_from_every_class = LoadBoolean(f, "Boosting-Based IS select best instance from every class", true, parm->out_file);
    parm->bbis_instances_per_iteration = LoadInteger(f, "Boosting-Based maximum instances per iteration", 0, parm->out_file);

    // MABUSE FS
    parm->mabuse_classifier = LoadEnumerate(f, "MABUSE FS learner", learner, n_classifier, (int) C45, parm->out_file);
    parm->mabuse_max_instances = LoadInteger(f, "MABUSE FS maximum instances per training", 0, parm->out_file);
    parm->mabuse_feat_per_iteration = LoadInteger(f, "MABUSE FS/SIFS2 maximum features per iteration", 0, parm->out_file);
    parm->mabuse_ensemble_method = LoadEnumerate(f, "MABUSE FS ensemble method", ensembles, n_ensembles, (int) AdaBoost, parm->out_file);
    parm->mabuse_performance_measure = LoadEnumerate(f, "MABUSE FS Performance measure", performance, n_performance, (int) pfKappa, parm->out_file);
    parm->mabuse_max_losing_streak = LoadInteger(f, "MABUSE FS/SIFS2 maximum losing streak", 10, parm->out_file);
    parm->second_chances = LoadInteger(f, "MABUSE FS second chances", 2, parm->out_file);
    parm->mabuse_feature_life = LoadInteger(f, "MABUSE FS feature life", 1, parm->out_file);   // Number of times a feature can be removed before ignoring it
    parm->mabuse_filter_weighted_sampling_ratio = LoadReal(f, "MABUSE FS filter weighted sampling ratio", 1.0, parm->out_file);
    parm->mabuse_condit_excl_with_margin = LoadBoolean(f, "MABUSE FS conditional exclusion with margin", false, parm->out_file);
    parm->mabuse_folds = LoadInteger(f, "MABUSE FS folds", 1, parm->out_file);
    parm->mabuse_ranking_initial_selection_percentage = LoadReal(f, "MABUSE FS ranking initial selection percentage", 0.0, parm->out_file);
    parm->mabuse_sort_by_weighted_mi_limit = LoadInteger(f, "MABUSE FS/SIFS2 sort by weighted MI limit", 0, parm->out_file);
    parm->mabuse_min_selected_percent = LoadReal(f, "MABUSE FS minimum selected percentage", 0.0, parm->out_file);
    parm->mabuse_use_rankings_for_perf = LoadBoolean(f, "MABUSE use rankings for performance", false, parm->out_file);
    parm->rank_w = LoadBoolean(f, "MABUSE average ranking weight", 0.0, parm->out_file);
    parm->mabuse_roulette_selection = LoadBoolean(f, "MABUSE roulette selection", false, parm->out_file);

    // IS+FS floating search. Reusing some MABUSE parameters
    parm->mabuse_max_losing_streak = LoadInteger(f, "MABUSE FS/SIFS2 maximum losing streak", 10, parm->out_file);
    parm->mabuse_feat_per_iteration = LoadInteger(f, "MABUSE FS/SIFS2 maximum features per iteration", 0, parm->out_file);
    parm->mabuse_sort_by_weighted_mi_limit = LoadInteger(f, "MABUSE FS/SIFS2 sort by weighted MI limit", 0, parm->out_file);
    parm->sifs2_convergence_with_only_acc = LoadBoolean(f, "SIFS2 convergence using only accuracy", false, parm->out_file);
    parm->sifs2_best_with_only_acc = LoadBoolean(f, "SIFS2 best using only accuracy", true, parm->out_file);
    parm->sifs2_evaluator = (t_Classifier) LoadEnumerate(f, "SIFS2 fitness evaluator", learner, n_classifier, (int) kNN, parm->out_file);
    parm->sifs2_sampling_size = LoadInteger(f, "SIFS2 stratified sampling size", 0, parm->out_file);
    parm->sifs2_stratum_size = LoadInteger(f, "SIFS2 stratum size", 0, parm->out_file);
    parm->sifs2_when_stratified_evaluate_with_training_set = LoadBoolean(f, "SIFS2 when stratified evaluate with whole training set", false, parm->out_file);
    parm->sifs2_is_additive = LoadBoolean(f, "SIFS2 stratified additive instance selection", false, parm->out_file);
    parm->sifs2_stratified_incremental_selection = LoadBoolean(f, "SIFS2 stratified incremental selection", true, parm->out_file);
    parm->sifs2_sampling_size_for_performance = LoadInteger(f, "SIFS2 stratified sampling size for performance evaluation", 0, parm->out_file);
    parm->sifs2_add_group_size = LoadInteger(f, "SIFS2 features added group size", 1, parm->out_file);
    parm->sifs2_random_is_alg = LoadBoolean(f, "SIFS2 random instance selection algorithm", false, parm->out_file);

    // Fitness
    parm->ired_w = LoadReal(f, "Instance reduction weight", 0.0, parm->out_file);
    parm->fred_w = LoadReal(f, "Feature reduction weight", 0.0, parm->out_file);
    parm->exp_redfeature = LoadBoolean(f, "Exponential reduction feature", false, parm->out_file);
    parm->exp_redinstance = LoadBoolean(f, "Exponential reduction instance", false, parm->out_file);
    parm->exp_reduction = LoadBoolean(f, "Exponential reduction", false, parm->out_file);
    parm->fitness_as_ratio = LoadBoolean(f, "Fitness as ratio", false, parm->out_file);
    parm->cv_for_fitness_as_ratio = LoadBoolean(f, "CV for fitness as ratio", true, parm->out_file);

    // IS+FS algorithm
    parm->isfsalgorithm = (t_ISFSAlgorithm) LoadEnumerate(f, "Instance/feature selection algorithm", isfsalgorithm, n_isfsalgorithm, (int) CHCGA, parm->out_file);
    parm->random_fs_algorithm = LoadBoolean(f, "Random FS algorithm", false, parm->out_file);
    parm->random_fs_algorithm_low_redundancy = LoadBoolean(f, "Random FS algorithm low redundancy methods", false, parm->out_file);
    parm->random_fs_algorithm_fast = LoadBoolean(f, "Random FS algorithm fast methods", false, parm->out_file);
    parm->random_isalgorithm = LoadBoolean(f, "Random IS algorithm", false, parm->out_file);

    // Multi-selection of instances
    parm->multiselection = LoadBoolean(f, "Multi-selection of instances", false, parm->out_file);                       // Allow selecting an instance more than once
    parm->multiselection_as_local_k = LoadBoolean(f, "Multi-selection of instances as local k", false, parm->out_file);
    parm->multiselect_only_minority_class = LoadBoolean(f, "Multi-select only minority class", false, parm->out_file);  // Apply class-imbalance aware methods when available
    parm->multiis_adjust_initial_percentage = LoadBoolean(f, "Multi-selection adjust initial percentage", false, parm->out_file);

    // Pre-sample data before doing anything
    parm->presample = LoadBoolean(f, "Pre-sample", false, parm->out_file);
    parm->presample_ratio = LoadReal(f, "Pre-sample ratio", 0.0, parm->out_file);
    parm->presample_subset_size = LoadInteger(f, "Pre-sample subset size", 1000, parm->out_file);
    parm->presample_stratified = LoadBoolean(f, "Pre-sample stratified", true, parm->out_file);

    // Class-imbalanced when applicable
    parm->ci = LoadBoolean(f, "Class-imbalanced", false, parm->out_file);
    parm->oversampling_ratio = LoadReal(f, "Oversampling ratio", 1.0, parm->out_file);

    // BalanceCascade
    parm->ex_un_s_i = LoadInteger(f, "BalanceCascade s_i", 5, parm->out_file);
    parm->ex_un_s_i = LoadInteger(f, "EasyEnsemble s_i", 5, parm->out_file);

    // Random selection
    parm->randomsel_withreplacement = LoadBoolean(f, "Random selection with replacement", false, parm->out_file);
    parm->adapt_ratio = LoadBoolean(f, "Adapt ratio", true, parm->out_file);
    parm->min_ratio = LoadReal(f, "Random selection minimum ratio", 0.01, parm->out_file);
    parm->max_ratio = LoadReal(f, "Random selection maximum ratio", 0.25, parm->out_file);
    parm->randomsel_ratio = LoadReal(f, "Random selection ratio", 0.1, parm->out_file);

    // Ensemble
    parm->t = LoadInteger(f, "T", 1, parm->out_file);
    parm->resample = LoadBoolean(f, "Resampling", true, parm->out_file);
    parm->alpha_acc = LoadReal(f, "Alpha", 0.5, parm->out_file);
    parm->dynamic_threshold = LoadBoolean(f, "Dynamic threshold", true, parm->out_file);
    parm->maximum_number_of_thresholds = LoadInteger(f, "Maximum number of thresholds", 0, parm->out_file);
    parm->boostis_classifier = LoadEnumerate(f, "Classifier", learner, n_classifier, (int) kNN, parm->out_file);
    parm->votes_threshold_measure = LoadEnumerate(f, "Votes evaluation measure", performance, n_performance, (int) pfAccuracy, parm->out_file);
    parm->ensemble_method = LoadEnumerate(f, "Ensemble method", ensembles, n_ensembles, (int) NoBoost, parm->out_file);
    parm->ensemble_size = LoadInteger(f, "Ensemble size", 50, parm->out_file);
    parm->votes_recording_method = LoadEnumerate(f, "Votes recording", votes_recording, n_votes_recording, (int) ICPOS, parm->out_file);
    parm->random_is_algo = LoadBoolean(f, "Random IS algorithm", false, parm->out_file);
    parm->boostis_considerlist = LoadBoolean(f, "Use consider list", false, parm->out_file);
    parm->add_k_neighbors = LoadInteger(f, "Add k neighbors", 0, parm->out_file);
    parm->boosting_resampling_method = LoadEnumerate(f, "Boosting resampling method", sampling, n_sampling, (int) WithoutReplacement, parm->out_file);
    parm->update_w_n_iterarions = LoadInteger(f, "Update w iterations", 1, parm->out_file);
    parm->th_eval_n_iterarions = LoadInteger(f, "Threshold evaluation iterations", 1, parm->out_file);
    parm->alpha_boost_set_to_one = LoadBoolean(f, "Alpha set to 1", false, parm->out_file);
    parm->subsampling_ratio = LoadReal(f, "Subsampling ratio", 0.1, parm->out_file);
    parm->weigthed_sampling_for_training = LoadBoolean(f, "Weighted sampling for training", false, parm->out_file);
    parm->boostfs_classifier = LoadEnumerate(f, "BoostFS Classifier", learner, n_classifier, (int) C45, parm->out_file);
    parm->evaluator = LoadEnumerate(f, "Evaluator", learner, n_classifier, (int) C45, parm->out_file);
    parm->ensemble_implementation = LoadEnumerate(f, "Ensemble implementation", implementation, n_implementation, (int) Resampling, parm->out_file);
    parm->select_fs_randomly = LoadBoolean(f, "Select feature selection algorithm randomly", false, parm->out_file);

    // Ensemble for IS+FS
    parm->boostisfs_classifier = LoadEnumerate(f, "Classifier", learner, n_classifier, (int) kNN, parm->out_file);
    parm->max_isel_thresholds = LoadInteger(f, "Maximum instance selection thresholds", 0, parm->out_file);
    parm->max_fsel_thresholds = LoadInteger(f, "Maximum feature selection thresholds", 0, parm->out_file);
    parm->max_ifsel_thresholds = LoadInteger(f, "Maximum IS+FS selection thresholds", 0, parm->out_file);
    parm->votes_evaluator_classifier = LoadEnumerate(f, "Votes evaluator learner", learner, n_classifier, (int) kNN, parm->out_file);
    parm->alpha_redins = LoadReal(f, "Threshold evaluation instance reduction weight", 1.0, parm->out_file);
    parm->alpha_redfeat = LoadReal(f, "Threshold evaluation feature reduction weight", 1.0, parm->out_file);
    parm->eval_exp_redfeature = LoadBoolean(f, "Exponential evaluation of feature reduction", false, parm->out_file);
    parm->eval_exp_redinstance = LoadBoolean(f, "Exponential evaluation of instance reduction", false, parm->out_file);
    parm->eval_folds = LoadInteger(f, "Evaluation folds", 10, parm->out_file);

    // Selecting from bites
    parm->isfs_n_bites = LoadInteger(f, "IS+FS number of bites", 1, parm->out_file);

    // Graph voting approach
    // parm->threshold_evaluation = (t_ThresholdEvaluation) LoadEnumerate(f, "Threshold evaluation", threshold_methods, N_THRESHOLDS_METHODS, (int) Majority, parm->out_file);
    parm->feature_subset_evaluation = (t_FeatureSubsetEvaluation) LoadEnumerate(f, "Feature subset evaluation", feature_subset_evaluation, n_feature_subset_evaluation, (int) ClassModel,
                                                                                parm->out_file);
    parm->feature_subset_combination = LoadEnumerate(f, "Feature subset combinations", subset_combination, n_subset_combination, (int) Thresholding, parm->out_file);
    parm->max_adjacency_matrix_size = LoadReal(f, "Maximum adjacency matrix size (GB)", 16.0, parm->out_file); // In GB
    parm->use_edge_list_for_graph = LoadBoolean(f, "Use edge list for vote graph", false, parm->out_file); // If false use sparse matrix when dense matrix too large
    parm->class_perf_as_percentage = LoadBoolean(f, "Classification performance as percentage", false, parm->out_file);
    parm->select_with_sorted_graph = LoadBoolean(f, "Select with sorted graph", false, parm->out_file);
    parm->save_graph = LoadBoolean(f, "Save graph", false, parm->out_file);

    /* // Combination of IS+FS algorithms
    parm->evaluate_with_threshold_methods[Thresholding] = LoadBoolean(f, "Evaluate with threshold method Thresholding", false, parm->out_file);
    parm->evaluate_with_threshold_methods[MFDR] = LoadBoolean(f, "Evaluate with threshold method MFDR", false, parm->out_file);
    parm->evaluate_with_threshold_methods[VOOR] = LoadBoolean(f, "Evaluate with threshold method VOOR", false, parm->out_file);
    parm->evaluate_with_threshold_methods[MFE] = LoadBoolean(f, "Evaluate with threshold method MFE", false, parm->out_file);
    parm->evaluate_with_threshold_methods[CF] = LoadBoolean(f, "Evaluate with threshold method CF", false, parm->out_file);
    parm->evaluate_with_threshold_methods[LOOR] = LoadBoolean(f, "Evaluate with threshold method LOOR", false, parm->out_file);
    parm->evaluate_with_threshold_methods[RIIC] = LoadBoolean(f, "Evaluate with threshold method RIIC", false, parm->out_file);
    parm->evaluate_with_threshold_methods[Union] = LoadBoolean(f, "Evaluate with threshold method Union", false, parm->out_file);
    parm->evaluate_with_threshold_methods[Intersection] = LoadBoolean(f, "Evaluate with threshold method Intersection", false, parm->out_file);
    parm->evaluate_with_threshold_methods[Majority] = LoadBoolean(f, "Evaluate with threshold method Majority", false, parm->out_file);
    parm->evaluate_with_threshold_methods[GraphMinorThreshold] = LoadBoolean(f, "Evaluate with threshold method Graph Minor threshold", false, parm->out_file);
    parm->evaluate_with_threshold_methods[GraphMajorMinorThreshold] = LoadBoolean(f, "Evaluate with threshold method Graph Major-Minor threshold", false, parm->out_file);
    parm->evaluate_with_threshold_methods[GraphMinorThresholdIFEdges] = LoadBoolean(f, "Evaluate with threshold method Graph Minor threshold IF edges", false, parm->out_file);
    parm->evaluate_with_threshold_methods[GraphMajorMinorThresholdIFEdges] = LoadBoolean(f, "Evaluate with threshold method Graph Major-Minor threshold IF edges", false, parm->out_file);
    parm->evaluate_with_threshold_methods[GraphMinorConnectedSubgraph] = LoadBoolean(f, "Evaluate with threshold method Graph Minor connected subgraph", false, parm->out_file);
    parm->evaluate_with_threshold_methods[GraphMajorMinorConnectedSubgraph] = LoadBoolean(f, "Evaluate with threshold method Graph Major-Minor connected subgraph", false, parm->out_file);
    parm->evaluate_with_threshold_methods[GraphMinorThresholdChain] = LoadBoolean(f, "Evaluate with threshold method Graph Minor threshold chain", false, parm->out_file);
    parm->evaluate_with_threshold_methods[GraphMajorMinorThresholdChain] = LoadBoolean(f, "Evaluate with threshold method Graph Major-Minor threshold chain", false, parm->out_file);
    parm->evaluate_with_threshold_methods[GraphMinorConnectedSubgraphIFEdges] = LoadBoolean(f, "Evaluate with threshold method Graph Minor connected subgraph IF edges", false, parm->out_file);
    parm->evaluate_with_threshold_methods[GraphMajorMinorConnectedSubgraphIFEdges] = LoadBoolean(f, "Evaluate with threshold method Graph Major-Minor connected subgraph IF edges", false,
            parm->out_file);
    parm->evaluate_with_threshold_methods[GraphMinorThresholdChainIFEdges] = LoadBoolean(f, "Evaluate with threshold method Graph Minor threshold chain IF edges", false, parm->out_file);
    parm->evaluate_with_threshold_methods[GraphMajorMinorThresholdChainIFEdges] = LoadBoolean(f, "Evaluate with threshold method Graph Major-Minor threshold chain IF edges", false, parm->out_file); */

    // Boosting FS
    parm->sampling_ratio = LoadReal(f, "Sampling ratio", 0.0, parm->out_file);
    parm->boostfs_w_acc = LoadReal(f, "Accuracy weight for threshold", 1.0, parm->out_file);
    parm->boostfs_w_gme = LoadReal(f, "G-mean weight for threshold", 0.0, parm->out_file);
    parm->boostfs_w_kap = LoadReal(f, "Kappa weight for threshold", 1.0, parm->out_file);
    parm->boostfs_w_auroc = LoadReal(f, "auROC weight for threshold", 0.0, parm->out_file);
    parm->boostfs_w_auprc = LoadReal(f, "auPRC weight for threshold", 0.0, parm->out_file);
    parm->boostfs_w_red = LoadReal(f, "Reduction weight for threshold", 1.0, parm->out_file);
    parm->evaluation_subset_size = LoadInteger(f, "Evaluation subset size", parm->sampling_size, parm->out_file);
    parm->boost_ci_initialization = LoadBoolean(f, "BoostFS CI weight initialization", false, parm->out_file);
    parm->fixed_number_of_thresholds = LoadBoolean(f, "Fixed number of thresholds", false, parm->out_file);
    parm->max_n_thresholds = LoadInteger(f, "Maximum number of thresholds", 0, parm->out_file);
    LoadString(f, "Performance file", "None", parm->performance_file, parm->out_file);
    parm->save_performance = LoadBoolean(f, "Save performance", false, parm->out_file);

    // FS wrapper methods
    parm->wrapper_fs_classifier = LoadEnumerate(f, "Wrapper FS learner", learner, n_classifier, (int) kNN, parm->out_file);

    // eReweightBoost
    parm->reweightboost_r = LoadInteger(f, "eReweightBoost r", 5, parm->out_file);

    // Stratified approach
    parm->external_partitioning_approach = LoadBoolean(f, "External partitioning approach", false, parm->out_file);
    parm->external_subset_size = LoadInteger(f, "External subset size", 1000, parm->out_file);
    parm->internal_partitioning_approach = LoadBoolean(f, "Internal partitioning approach", false, parm->out_file);
    parm->internal_subset_size = LoadInteger(f, "Internal subset size", 1000, parm->out_file);
    parm->internal_partitioning_method = (t_Partition) LoadEnumerate(f, "Internal partitioning method", partitions, n_partitions, (int) pRANDOM, parm->out_file);
    parm->external_partitioning_method = (t_Partition) LoadEnumerate(f, "External partitioning method", partitions, n_partitions, (int) pRANDOM, parm->out_file);
    parm->external_stratified_approach = LoadBoolean(f, "External stratified partition", false, parm->out_file);
    parm->internal_stratified_approach = LoadBoolean(f, "Internal stratified partition", false, parm->out_file);
    parm->use_pcomponents = LoadBoolean(f, "Use PCs", false, parm->out_file);

    // Feature subspace approach
    parm->internal_subspace_size = LoadInteger(f, "Internal subspace size", 0, parm->out_file);
    parm->external_subspace_size = LoadInteger(f, "External subspace size", 0, parm->out_file);

    // Learning
    parm->isalgorithm = (t_ISAlgorithm) LoadEnumerate(f, "Instance selection algorithm", isalgorithm, n_isalgorithm, (int) DROP3, parm->out_file);
    parm->sampling_size = LoadInteger(f, "Sampling size", 0, parm->out_file);
    parm->discretizer = (t_Discretizer) LoadEnumerate(f, "Discretizer", discretizer, n_discretizer, (int) dFAYYAD, parm->out_file);

    // Hyper-parameters
    parm->cv_hyperparameters = LoadBoolean(f, "Cross-validate hyper-parameters", false, parm->out_file);
    parm->cv_hyperparameters_for_testing = LoadBoolean(f, "Cross-validate hyper-parameters for testing", false, parm->out_file);
    parm->cv_external_hyperparameters_for_testing = LoadBoolean(f, "Cross-validate external hyper-parameters for testing", false, parm->out_file);
    parm->cv_k_for_testing = LoadBoolean(f, "Cross-validate k for testing", false, parm->out_file);
    parm->cv_performance_measure  = LoadEnumerate(f, "Cross-validation performance measure", performance, n_performance, (int) pfAverageAccuracy, parm->out_file);
    parm->cv_sampling_method = LoadEnumerate(f, "Sampling for cross-validation of hyper-parameters", samplingmethod, n_samplingmethod, (int) sStratified, parm->out_file);
    parm->cv_sampling_size = LoadInteger(f, "Sampling size for cross-validation of hyper-parameters", 0, parm->out_file);

    // k-NN
    parm->k = LoadInteger(f, "k-NN k", 1, parm->out_file);
    parm->min_k = LoadInteger(f, "Minimum k", 1, parm->out_file);
    parm->max_k = LoadInteger(f, "Maximum k", 25, parm->out_file);
    parm->adaptive_distance_measure = LoadBoolean(f, "Adaptive distance measure", false, parm->out_file);
    parm->knn_with_weighted_votes = LoadBoolean(f, "k-NN with weighted votes", false, parm->out_file);
    parm->symmetric_knn = LoadBoolean(f, "Symmetric k-NN", false, parm->out_file);
    parm->kmeans_max_iterations = LoadInteger(f, "k-Means maximum iterations", 100, parm->out_file);
    parm->evolutionary_local_k = LoadBoolean(f, "Evolutionary local k", false, parm->out_file);
    parm->optimal_local_k = LoadBoolean(f, "Optimal local k", false, parm->out_file);
    parm->edited_knn = LoadBoolean(f, "Edited k-NN", false, parm->out_file);
    parm->show_average_nn_margin = LoadBoolean(f, "Show average NN margin", false, parm->out_file);

    // Gaussian NBC
    parm->gnbc_use_logs = LoadBoolean(f, "Gaussian NBC use logs", true, parm->out_file);

    // Internal Naive Bayes learner
    parm->NBC_discretizer = LoadEnumerate(f, "Naive Bayes learner discretization", nbcinterval, n_nbcinterval, (int) FANDI, parm->out_file);
    parm->NBC_intervals = LoadInteger(f, "Naive Bayes learner intervals", 100, parm->out_file);

    // Random forest
    parm->rf_ntree = LoadInteger(f, "Random forest trees", 100, parm->out_file);
    parm->rf_percent = LoadReal(f, "Random forest percentage of features", 0.5, parm->out_file);
    parm->keep_files = LoadBoolean(f, "External classifiers keep files", false, parm->out_file);
    parm->use_ranger_for_rf = LoadBoolean(f, "Use 'ranger' for random forest", false, parm->out_file);

    // CHC
    parm->bit_mutation = LoadReal(f, "Bit mutation", 0.1, parm->out_file);
    parm->chc_wrapper = (t_Classifier) LoadEnumerate(f, "CHC wrapper", learner, n_classifier, (int) kNN, parm->out_file);
    parm->ebus_gs_gm = LoadBoolean(f, "EBUS GS GM", false, parm->out_file);          // Follows S. García and F. Herrera, "", Evol. Comp, 2009
    parm->LOO_for_knn_performance = LoadBoolean(f, "LOO for k-NN performance", true, parm->out_file);
    parm->evolutionary_scheme = LoadEnumerate(f, "Evolutionary scheme", evolutionary_scheme, n_evolutionary_scheme, (int) schemeCHCGA, parm->out_file);
    parm->generations = LoadInteger(f, "Generations", 1000, parm->out_file);
    parm->individuals = LoadInteger(f, "Individuals", 100, parm->out_file);
    parm->init_fs_percent = LoadReal(f, "Initial feature percentage", 0.5, parm->out_file);
    parm->init_percent = LoadReal(f, "Initial percentage", 0.25, parm->out_file);
    parm->max_evaluations = LoadInteger(f, "Maximum evaluations", 0, parm->out_file);
    parm->ok_w = LoadReal(f, "Classification weight", 0.5, parm->out_file);
    parm->performance_measure = LoadEnumerate(f, "Performance measure", performance, n_performance, (int) pfAccuracy, parm->out_file);
    parm->reduction_w = LoadReal(f, "Reduction weight", 0.0, parm->out_file);
    parm->reinit_distance0 = LoadBoolean(f, "Reinitialize at distance 0", true, parm->out_file);
    parm->time_limit = LoadInteger(f, "Time limit", 0, parm->out_file);
    parm->chc_max_reinit_attempts = LoadInteger(f, "CHC maximum re-initialization attempts", 0, parm->out_file);

    // Differential evolution
    parm->de_np = LoadInteger(f, "DE NP", 100, parm->out_file);
    parm->de_f = LoadReal(f, "DE F", 0.8, parm->out_file); // Usually in [0, 2]
    parm->de_cr = LoadReal(f, "DE CR", 0.9, parm->out_file);
    parm->multiis_and_k_with_DE = LoadBoolean(f, "DE Multi-IS and k", false, parm->out_file);
    parm->de_evolve_threshold = LoadBoolean(f, "DE evolve threshold", false, parm->out_file);
    parm->de_separate_thresholds = LoadBoolean(f, "DE separate thresholds", false, parm->out_file);
    parm->de_local_search_is_probability = LoadReal(f, "DE local search IS probability", 0.0, parm->out_file);
    parm->de_local_search_k_probability = LoadReal(f, "DE local search k probability", 0.0, parm->out_file);
    parm->de_undersample_mut_prob = LoadReal(f, "DE undersampling mutation probability", 0.0, parm->out_file);

    // Democracy
    // parm->democratic_approach = LoadBoolean(f, "Democratic approach", false, parm->out_file);
    // parm->democratic_rounds = LoadInteger(f, "Democratic rounds", 10, parm->out_file);
    parm->subset_size = LoadInteger(f, "Subset size", 0, parm->out_file);
    parm->min_subset_size = LoadInteger(f, "Minimum subset size", 0, parm->out_file);
    parm->n_electors = LoadInteger(f, "Electors", 10, parm->out_file);
    parm->weight_votes = LoadBoolean(f, "Weight votes", false, parm->out_file);
    parm->votes_evaluator = LoadEnumerate(f, "Votes evaluator", learner, n_classifier, (int) kNN, parm->out_file);
    parm->get_best_round = LoadBoolean(f, "Get best round", false, parm->out_file);
    parm->angle_step = LoadReal(f, "Angle step", 1.0, parm->out_file);
    parm->angle_step = MAX(0.001, parm->angle_step);        // Minimum value
    parm->fusion_subsets = LoadBoolean(f, "Fusion subsets", false, parm->out_file);
    parm->fusion_from_first = LoadBoolean(f, "Fusion from first", false, parm->out_file);
    parm->percent_error_estimation = LoadReal(f, "Percent for error estimation", 0.1, parm->out_file);
    parm->get_threshold_with_all = LoadBoolean(f, "Get threshold with all", true, parm->out_file);

    // Sampling parameters
    parm->sampling = LoadBoolean(f, "Feature sampling", false, parm->out_file);
    parm->instance_partition = LoadBoolean(f, "Instance partition", false, parm->out_file);
    // parm->alpha = LoadReal(f, "Alpha", 0.75, parm->out_file);
    parm->partition = (t_Partition) LoadEnumerate(f, "Partition", partitions, n_partitions, (int) pRANDOM, parm->out_file);
    parm->n_thresholds = LoadInteger(f, "Number of thresholds", 0, parm->out_file);
    parm->keep_class_distribution = LoadBoolean(f, "Keep class distribution", true, parm->out_file);
    parm->subspace_size = LoadInteger(f, "Subspace size", 5, parm->out_file);
    parm->subsampling_for_evaluation = LoadBoolean(f, "Subsampling for evaluation", false, parm->out_file);

    // Democratic evaluation
    parm->democratic_evaluation_of_threshold = LoadBoolean(f, "Democratic evaluation of threshold", true, parm->out_file);
    parm->demo_eval_rounds = LoadInteger(f, "Democratic evaluation rounds", 1, parm->out_file);
    parm->democratic_evaluation_of_threshold_with_cv = LoadBoolean(f, "Democratic evaluation of threshold with CV", true, parm->out_file);

    // Memetic mutation: IS
    parm->rnn_is_mutation = LoadReal(f, "RNN IS mutation", 0.0, parm->out_file);
    parm->drop3_mutation = LoadReal(f, "DROP3 mutation", 0.0, parm->out_file);
    parm->icf_mutation = LoadReal(f, "ICF mutation", 0.0, parm->out_file);
    parm->ib3_mutation = LoadReal(f, "IB3 mutation", 0.0, parm->out_file);
    parm->mss_mutation = LoadReal(f, "MSS mutation", 0.0, parm->out_file);
    parm->rnnstd_mutation = LoadReal(f, "RNN standard mutation", 0.0, parm->out_file);
    parm->addI_mutation = LoadReal(f, "Add IS mutation", 0.0, parm->out_file);
    parm->boost_based_is_mutation = LoadReal(f, "Boost-based IS mutation", 0.0, parm->out_file);

    // Memetic mutation: FS
    parm->rfe_mutation = LoadReal(f, "RFE mutation", 0.0, parm->out_file);
    parm->rnn_fs_mutation = LoadReal(f, "RNN FS mutation", 0.0, parm->out_file);
    parm->focus_mutation = LoadReal(f, "FOCUS mutation", 0.0, parm->out_file);
    parm->addF_mutation = LoadReal(f, "Add FS mutation", 0.0, parm->out_file);

    // Memetic mutation: IS+FS
    parm->rnn_isfs_mutation = LoadReal(f, "RNN IS+FS mutation", 0.0, parm->out_file);
    parm->imoea_mutation = LoadReal(f, "IMOEA mutation", 0.0, parm->out_file);
    parm->addiorf_mutation = LoadReal(f, "Add IS+FS mutation", 0.0, parm->out_file);
    parm->isfs_cache_size = LoadReal(f, "IS+FS maximum cache size (GB)", 4.0, parm->out_file);

    // RMHC-FP1 IS+FS
    parm->rmhc_fp1_iterations = LoadInteger(f, "RMHC-PF1 iterations", 1000, parm->out_file);

    // Asymmetric AdaBoost
    parm->asymmetric_adaboost_k = LoadReal(f, "Asymmetric AdaBoost K", 0, parm->out_file);

    if (parm->ensemble_method == AdaCost || parm->ensemble_method == AdaC1 || parm->ensemble_method == AdaC3) {
        // AdaCost
        parm->C_plus = LoadReal(f, "C+", 1.0, parm->out_file);
        parm->C_minus = LoadReal(f, "C-", 0.0, parm->out_file);
    } else if (parm->ensemble_method == AdaC2 || parm->ensemble_method == CSB0 || parm->ensemble_method == CSB1 || parm->ensemble_method == CSB2) {
        // AdaC2
        parm->C_plus = LoadReal(f, "C+", 0.0, parm->out_file);
        parm->C_minus = LoadReal(f, "C-", 1.0, parm->out_file);
    } else {
        parm->C_plus = LoadReal(f, "C+", 0.0, parm->out_file);
        parm->C_minus = LoadReal(f, "C-", 0.0, parm->out_file);
    }

    // Democracy
    parm->comp_votes = LoadBoolean(f, "Compare votes", false, parm->out_file);
    parm->comp_clas = LoadBoolean(f, "Compare class", false, parm->out_file);
    parm->feature_partition = LoadBoolean(f, "Feature partition", false, parm->out_file);
    parm->ind_filter = LoadBoolean(f, "Individual filter", true, parm->out_file);
    parm->random_sampling_for_threshold_evaluation = LoadBoolean(f, "Random undersampling for threshold evaluation", false, parm->out_file);
    parm->random_sampling_for_threshold_evaluation_ratio = LoadReal(f, "Random undersampling for threshold evaluation ratio", 0.1, parm->out_file);
    parm->random_sampling_for_threshold_evaluation_min_size = LoadInteger(f, "Random undersampling for threshold evaluation minimum size", 10000, parm->out_file);
    parm->random_sampling_for_threshold_evaluation_max_size = LoadInteger(f, "Random undersampling for threshold evaluation maximum size", 10000, parm->out_file);

    // Extreme learning machine
    parm->elm_n_hidden = LoadInteger(f, "ELM Hidden nodes", 100, parm->out_file);
    parm->elm_tf = (t_Transfer) LoadEnumerate(f, "ELM Transfer function", transfer, n_transfer, (int) HypTan, parm->out_file);
    parm->elm_a = LoadReal(f, "ELM A", 1.0000, parm->out_file);
    parm->elm_b = LoadReal(f, "ELM B", 0.6666, parm->out_file);

    // Votes threshold
    parm->validation_percentage = LoadReal(f, "Validation percentage", 0.1, parm->out_file);
    parm->use_validation_data_for_votes_threshold = LoadBoolean(f, "Use validation data for votes threshold", false, parm->out_file);

    // CHC IS+FS used as FS
    parm->only_accept_if_better = LoadBoolean(f, "Only accept mutation if better", true, parm->out_file);

    // Cache
    parm->isfs_use_cache = LoadBoolean(f, "IS+FS use cache", true, parm->out_file);
    parm->isfs_stepwise_cache = LoadBoolean(f, "IS+FS stepwise cache", false, parm->out_file);
    parm->is_use_cache = LoadBoolean(f, "IS use cache", true, parm->out_file);
    parm->fs_use_cache = LoadBoolean(f, "FS use cache", true, parm->out_file);

    // IS+FS mutation
    parm->isfsmutation_random_order = LoadBoolean(f, "IS+FS mutation random order", false, parm->out_file); // Features/instances are mixed
    parm->isfs_inputs_first = LoadBoolean(f, "IS+FS inputs first", false, parm->out_file);
    parm->isfs_use_fitness = LoadBoolean(f, "IS+FS use fitness", true, parm->out_file);

    // FOCUS mutation
    parm->max_features_for_focus = LoadInteger(f, "Maximum features Focus", 10, parm->out_file);

    // IMOEA mutation
    parm->imoea_seconds = LoadInteger(f, "IMOEA seconds", 10, parm->out_file);

    // Generational genetic algorithm
    parm->gga_elitism = LoadReal(f, "GGA elitism", 0.1, parm->out_file);
    parm->gga_crossover_prob = LoadReal(f, "GGA crossover probability", 0.6, parm->out_file);
    parm->gga_mutation = LoadReal(f, "GGA mutation", 0.1, parm->out_file);

    // Early stopping?
    parm->early_stopping = LoadBoolean(f, "Early stopping", false, parm->out_file);
    parm->percent_for_early_stopping = LoadReal(f, "Percent for early stopping", 0.25, parm->out_file);

    // ICF/DROP3/Boost based/CHC
    parm->filter_noise = LoadBoolean(f, "Filter noise", parm->isalgorithm == CHC ? false : true, parm->out_file);

    // LVQ pruning
    parm->Npc = LoadInteger(f, "LVQ Npc", 25, parm->out_file);

    // Generalized CNN
    parm->rho = LoadReal(f, "GCNN rho", 0.99, parm->out_file);
    parm->cnn_ci = LoadBoolean(f, "CNN class-imbalanced dataset", false, parm->out_file);

    // RMHC
    parm->rmhc_m = LoadInteger(f, "RMHC m", 3, parm->out_file);
    parm->rmhc_iterations = LoadInteger(f, "RMHC iterations", 1000, parm->out_file);

    // HMNEI
    parm->hmnei_epsilon = LoadReal(f, "HMNEI epsilon", 0.1, parm->out_file);

    // RNGE
    parm->rnge_pg = LoadEnumerate(f, "Proximity graph", proximity, n_proximity, (int) Gabriel, parm->out_file);

    // CV classifiers
    parm->cv_parameters = LoadBoolean(f, "Validate parameters", false, parm->out_file);
    parm->cv_val_percent = LoadReal(f, "Cross-validation percentage", 0.1, parm->out_file);

    // SMOTE
    parm->smote_undersampling_ratio = LoadReal(f, "SMOTE undersampling ratio", 1.0, parm->out_file);
    parm->smote_k = LoadInteger(f, "Smote k", 5, parm->out_file);
    parm->smote_N = LoadInteger(f, "Smote N", 2, parm->out_file);
    parm->smote_variant = LoadEnumerate(f, "SMOTE variant", smote_variant, n_smote_variant, (int) vSMOTE, parm->out_file);

    // EvoSMOTE
    parm->evosmote_eval_new_instances = LoadBoolean(f, "EvoSMOTE evaluate new instances", false, parm->out_file);
    parm->evosmote_only_minority = LoadBoolean(f, "EvoSMOTE use only minority class", true, parm->out_file);
    parm->evosmote_use_majority_threshold = LoadInteger(f, "EvoSMOTE use majority class threshold", 5, parm->out_file);
    parm->evosmote_undersamplig_mutation_rate = LoadReal(f, "EvoSMOTE undersamplig mutation rate", 0.0, parm->out_file);
    parm->evosmote_oss_mutation_rate = LoadReal(f, "EvoSMOTE OSS mutation rate", 0.0, parm->out_file);
    parm->is_mutation = LoadReal(f, "EvoSMOTE instance selection mutation rate", 0.0, parm->out_file);
    parm->evosmote_cnntl_mutation_rate = LoadReal(f, "EvoSMOTE CNNTL mutation rate", 0.0, parm->out_file);
    parm->evosmote_min_N = LoadInteger(f, "EvoSMOTE minimum N", 1, parm->out_file);
    parm->evosmote_max_N = LoadInteger(f, "EvoSMOTE maximum N", 10, parm->out_file);
    // parm->evosmote_undersample_minority = LoadBoolean(f, "EvoSMOTE undersample also minority class", false, parm->out_file);
    parm->evosmote_wrapper_eval_interval = LoadInteger(f, "EvoSMOTE  wrapper evaluation interval", 0, parm->out_file);
    parm->symmetric_selection = LoadBoolean(f, "Instance selection symmetric evolution", false, parm->out_file);
    parm->evosmote_roulette_initialization = LoadBoolean(f, "EvoSMOTE roulette initialization", false, parm->out_file);
    parm->best_individual_fitness_improvement_generations = LoadInteger(f, "Best individual fitness improvement generations", 0, parm->out_file);
    parm->reinitalize_percent_addition = LoadReal(f, "Best individual fitness improvement addition", 0.1, parm->out_file);
    parm->evosmote_noise_mutation_rate = LoadReal(f, "EvoSMOTE noise mutation rate", 0.0, parm->out_file);
    parm->evosmote_noise_mutation_k = LoadInteger(f, "EvoSMOTE noise mutation rate k", 5, parm->out_file);
    parm->evosmote_one_w = LoadBoolean(f, "EvoSMOTE use only one weight", false, parm->out_file);

    parm->accuracy_w = LoadReal(f, "Accuracy fitness weight", 0.0, parm->out_file);
    parm->gmean_w = LoadReal(f, "G-mean fitness weight", 0.0, parm->out_file);
    parm->f1_w = LoadReal(f, "F1 fitness weight", 0.0, parm->out_file);
    parm->auroc_w = LoadReal(f, "auROC fitness weight", 0.0, parm->out_file);
    parm->kappa_w = LoadReal(f, "Kappa fitness weight", 0.0, parm->out_file);
    parm->recall_macro_w = LoadReal(f, "Recall macro fitness weight", 0.0, parm->out_file);
    parm->recall_macro_g_mean_w = LoadReal(f, "Recall macro g-mean fitness weight", 0.0, parm->out_file);
    parm->average_accuracy_w = LoadReal(f, "Average accuracy fitness weight", 0.0, parm->out_file);
    parm->added_w = LoadReal(f, "Oversampled instances fitness weight", 0.0, parm->out_file);
    parm->cc_w = LoadReal(f, "Correlation coefficient fitness weight", 0.0, parm->out_file);
    parm->recall_w = LoadReal(f, "Sensitivity fitness weight", 0.0, parm->out_file);
    parm->specificity_w = LoadReal(f, "Specificity fitness weight", 0.0, parm->out_file);
    parm->precision_w = LoadReal(f, "Precision fitness weight", 0.0, parm->out_file);
    parm->npv_w = LoadReal(f, "Negative predictive value fitness weight", 0.0, parm->out_file);
    parm->balance_w = LoadReal(f, "Balance weight", 0.0, parm->out_file);

    // Class-imbalance method
    parm->imbalance_method = LoadEnumerate(f, "Class-imbalance method", class_imbalance_method, n_class_imbalance_method, (int) iNone, parm->out_file);

    // Random selection
    parm->ratio = LoadReal(f, "Random selection ratio", 0.1, parm->out_file);
    parm->with_replacement = LoadBoolean(f, "Sampling with replacement", true, parm->out_file);

    // LVF
    parm->lvf_iter = LoadInteger(f, "LVF Iterations", 100000, parm->out_file);

    // Floating algorithms
    parm->floating_attempts = LoadInteger(f, "Floating algorithms attempts", 3, parm->out_file);

    // K-S CBF
    parm->kscbf_k = LoadInteger(f, "K-S CBF k", 10, parm->out_file);
    parm->kscbf_alpha = LoadReal(f, "K-S CBF alpha", 0.05, parm->out_file);

    // G-flip
    parm->gflip_max_iterations = LoadInteger(f, "G-flip maximum iterations", 1000, parm->out_file);

    // Simba
    parm->simba_T = LoadInteger(f, "Simba T", 0, parm->out_file);

    // iSimba
    parm->isimba_lambda = LoadReal(f, "iSimba lambda", 0.1, parm->out_file);
    parm->isimba_T = LoadInteger(f, "iSimba T", 0, parm->out_file);
    parm->isimba_delta = LoadReal(f, "iSimba delta", 0.1, parm->out_file);

    // Genetic Algorithm
    parm->elite = LoadReal(f, "Elitism percentage", 0.0, parm->out_file);
    parm->init_setprob = LoadReal(f, "Initial set probability", 0.5, parm->out_file);
    parm->mutation = LoadReal(f, "Mutation", 0.1, parm->out_file);
    parm->ga_alpha = LoadReal(f, "GA alpha", 0.75, parm->out_file);

    // Class-imbalance datasets
    parm->random_undersampling = LoadBoolean(f, "Random undersampling", false, parm->out_file);
    parm->random_undersampling_ratio = LoadReal(f, "Random undersampling ratio", 1.0, parm->out_file);
    parm->positive_class = LoadInteger(f, "Positive class", -1, parm->out_file);

    // Ranking selection algorithms
    parm->feature_ranking = LoadBoolean(f, "Feature ranking", false, parm->out_file);
    parm->rank_to_fs_threshold_evaluation = LoadEnumerate(f, "Threshold evaluation", subset_evaluation, n_subset_evaluation, (int) Voor, parm->out_file);
    parm->fralgorithm = LoadEnumerate(f, "Feature ranking algorithm", fralgorithm, n_fralgorithm, (int) frMI, parm->out_file);
    parm->cub_lambda = LoadReal(f, "Unbalanced correlation score lambda", 3.0, parm->out_file);
    parm->alpha_red = LoadReal(f, "Threshold evaluation reduction weight", 0.5, parm->out_file);
    parm->alpha_ok = LoadReal(f, "Threshold evaluation accuracy weight", 0.5, parm->out_file);
    parm->coeff_as_geom_mean = LoadBoolean(f, "Coefficient as geometric mean", false, parm->out_file);
    parm->coeff_as_harm_mean = LoadBoolean(f, "Coefficient as harmonic mean", false, parm->out_file);

    // SVM
    parm->kernel = LoadInteger(f, "Kernel", 2, parm->out_file);                // LINEAR, POLY, RBF, SIGMOID, PRECOMPUTED, WD, WDS
    parm->svm_type = LoadInteger(f, "SVM type", 0, parm->out_file);            // 0 = C-SVC, 1 = nu-SVc, 2 = one-class SVM, 3 = epsilon-SVR, 4 = nu-SVR
    if (parm->kernel == WDS || parm->kernel == WD) {
        parm->gamma = LoadReal(f, "Sigma (WDS)", 0.5, parm->out_file);     // sigma parameter in WDS kernel
        parm->svm_d = LoadInteger(f, "Degree (WD)", 12, parm->out_file);   // d parameter in WD and WDS kernels
    } else {
        parm->gamma = LoadReal(f, "Gamma", 1.0, parm->out_file);
        parm->svm_d = LoadInteger(f, "Polynomial degree", 3, parm->out_file);
    }
    parm->svm_c = LoadReal(f, "C", 1.0, parm->out_file);
    parm->svm_r = LoadReal(f, "R", 0.0, parm->out_file);
    parm->tolerance = LoadReal(f, "Tolerance", 0.001, parm->out_file);
    parm->cache_size = LoadInteger(f, "SVM Cache size", 1000000, parm->out_file);
    parm->cv_with_string_kernel = LoadBoolean(f, "Cross-validation with string kernel", false, parm->out_file);

    // C4.5
    parm->soften_thresholds = LoadBoolean(f, "Soften thresholds", false, parm->out_file);
    parm->prune_tree = LoadBoolean(f, "Prune tree", false, parm->out_file);
    TRIALS = LoadInteger(f, "Trials", 1, parm->out_file);

    // RForest
    parm->rforest_split_measure = LoadEnumerate(f, "RForest split measure", split_measure, n_split_measure, (int) imGainRatio, parm->out_file);

    // Neural Network
    parm->nodes = LoadInteger(f, "Nodes", 100, parm->out_file);
    parm->nodes++;
    parm->init_weight = LoadReal(f, "Initial weight", 0.1, parm->out_file);
    parm->transfer_o = (t_Transfer) LoadEnumerate(f, "Transfer output", transfer, n_transfer, (int) Log, parm->out_file);
    parm->transfer_h = (t_Transfer) LoadEnumerate(f, "Transfer hidden", transfer, n_transfer, (int) HypTan, parm->out_file);
    parm->n_epochs = LoadInteger(f, "Number of epochs", 10000, parm->out_file);
    parm->epoch_size = LoadInteger(f, "Epoch size", 10, parm->out_file);
    parm->eta = LoadReal(f, "Eta", 0.5, parm->out_file);
    parm->alpha = LoadReal(f, "Alpha", 0.1, parm->out_file);
    parm->lambda = LoadReal(f, "Lambda", 0.0001, parm->out_file);
    parm->flat_spot = LoadReal(f, "Flat spot", 0.1, parm->out_file);
    parm->a_output = LoadReal(f, "A output", 1.0, parm->out_file);
    parm->b_output = LoadReal(f, "B output", 0.6666, parm->out_file);
    parm->a = LoadReal(f, "A", 1.0000, parm->out_file);
    parm->b = LoadReal(f, "B", 0.6666, parm->out_file);
    parm->asymptote = LoadReal(f, "Asymptote", 0.2, parm->out_file);
    parm->max_nodes = LoadInteger(f, "Maximum nodes", 100, parm->out_file);

    // Local decision trees
    parm->dt_min_samples_split = LoadInteger(f, "DT minimum samples to split", 10, parm->out_file);
    parm->dt_max_depth = LoadInteger(f, "DT maximum depth", 50, parm->out_file);
    parm->dt_grow_full_tree = LoadBoolean(f, "DT grow full tree", false, parm->out_file);   // Only for IBT
    parm->ibt_split_measure = LoadEnumerate(f, "IBT split measure", split_measure, n_split_measure, (int) imGainRatio, parm->out_file);
    parm->print_dt = LoadBoolean(f, "DT print tree", false, parm->out_file);

    // Base learner for any multi-classifier
    parm->base_learner = LoadEnumerate(f, "Multi-classifier base learner", learner, n_base_classifier, (int) DT_C45, parm->out_file);

    // Internally boosted trees
    parm->ibt_ens_method = LoadEnumerate(f, "IBT ensemble method", ensembles, n_ensembles, (int) NoBoost, parm->out_file);
    parm->ibt_alpha_method = LoadEnumerate(f, "IBT alpha method", ensembles, n_ensembles, (int) parm->ibt_ens_method, parm->out_file);
    parm->ibt_no_stop = LoadBoolean(f, "IBT do not stop until complete", true, parm->out_file);
    parm->ibt_do_not_use_internal_nodes = LoadBoolean(f, "IBT do not use internal nodes", false, parm->out_file);
    parm->ibt_multiboost_I = LoadInteger(f, "IBT eMultiBoost I", 10, parm->out_file);
    parm->ibt_fixed_node_weight = LoadBoolean(f, "IBT fixed node weight", false, parm->out_file);
    parm->ibt_predict_with_probabilities = LoadBoolean(f, "IBT predict with probabilities", false, parm->out_file);

    // Feature selection
    parm->fsalgorithm = LoadEnumerate(f, "Feature selection algorithm", fsalgorithm, n_fsalgorithm, (int) fsLVF, parm->out_file);
    parm->random_undersampling_for_training = LoadBoolean(f, "Random undersampling for training", false, parm->out_file);
    parm->use_g_measure = LoadBoolean(f, "Use G measure", false, parm->out_file);
    parm->cv_folds = LoadInteger(f, "CV folds", 3, parm->out_file);

    // ReliefF
    parm->m = LoadInteger(f, "ReliefF m", 0, parm->out_file);
    parm->relief_k = LoadInteger(f, "ReliefF k", 10, parm->out_file);
    parm->relief_threshold = LoadReal(f, "ReliefF threshold", 0.05, parm->out_file);
    parm->relief_alpha = LoadReal(f, "ReliefF alpha", 0.0, parm->out_file);

    // Evolutionary IS+Local k
    parm->evolutionary_local_k_and_is = LoadBoolean(f, "Evolutionary local k+IS", false, parm->out_file);
    parm->localk_mutation = LoadReal(f, "Local k mutation rate", 0.0, parm->out_file);
    parm->localk_mutation_with_no_selection = LoadReal(f, "Local k mutation with no selection", 0.0, parm->out_file);

    // Local k
    parm->localk_repetitions = LoadInteger(f, "Local k repetitions", 1, parm->out_file);
    parm->localk_lambda = LoadInteger(f, "Local k lambda", 3, parm->out_file);
    parm->localk_alpha_acc = LoadReal(f, "Local k alpha accuracy", 1.0, parm->out_file);
    parm->localk_alpha_reg = LoadReal(f, "Local k alpha regularization", 0.0, parm->out_file);
    parm->localk_alpha_mar = LoadReal(f, "Local k alpha margin", 0.0, parm->out_file);
    parm->localk_error_estimation_method = LoadEnumerate(f, "Local k error estimation method", estimation, n_estimation, (int) kFOLDCV, parm->out_file);
    parm->localk_nfolds = LoadInteger(f, "Folds for training error estimation", 10, parm->out_file);
    parm->localk_memetic_mutation_ratio = LoadReal(f, "Local k memetic mutation ratio", 0.1, parm->out_file);     // Only for GGA scheme
    parm->localk_memetic_bit_mutation_ratio = LoadReal(f, "Local k memetic bit mutation ratio", 0.25, parm->out_file);
    parm->localk_min_k = LoadInteger(f, "Minimum local k", 1, parm->out_file);
    parm->localk_max_k = LoadInteger(f, "Maximum local k", 25, parm->out_file);
    parm->localk_bs_n = LoadInteger(f, "Number of bootstrapping samples", 10, parm->out_file);
    parm->localk_holdout_percent = LoadReal(f, "Percentage for hold out method", 0.1, parm->out_file);
    parm->localk_reevaluate = LoadBoolean(f, "Reevaluate individuals each generation", false, parm->out_file);
    parm->margin_method = LoadEnumerate(f, "Margin method", margin_method, n_margin_method, (int) NormalizedVotesDifference, parm->out_file);
    parm->populate_number_of_neighbors = LoadInteger(f, "Artificial data number of neighbors", 1, parm->out_file);
    parm->populate_copies_per_instance = LoadInteger(f, "Artificial data copies per instance", 1, parm->out_file);
    parm->localk_populate = LoadBoolean(f, "Populate for optimal k", false, parm->out_file);
    parm->localk_no_neigbors_assign_nearest_value = LoadBoolean(f, "Local k assign nearest value to no neighbors instances", false, parm->out_file);
    LoadString(f, "k values for local k", "3", parm->k_values_for_local_k, parm->out_file);
    parm->leave_no_neighbors_unchanged = LoadBoolean(f, "Leave no neighbors unchanged", false, parm->out_file);    // For evolutionary purposes

    // FCBF
    parm->fcbf_delta = LoadReal(f, "FCBF Delta", -1.0, parm->out_file);

    // Logistic Regression
    parm->lr_epochs = LoadInteger(f, "LR number of epochs", 1000, parm->out_file);
    parm->lr_alpha = LoadReal(f, "LR learning coefficient", 0.01, parm->out_file);

    // mRMR
    parm->mrmr_n = LoadInteger(f, "mRMR n", 50, parm->out_file);
    parm->mrmr_evaluator = LoadEnumerate(f, "mRMR evaluator", learner, n_classifier, (int) kNN, parm->out_file);
    parm->mrmr_search = LoadEnumerate(f, "mRMR search heuristic", search, n_search, (int) SFS, parm->out_file);

    // wMI
    parm->wmi_epsilon = LoadReal(f, "wMI epsilon", 0.0, parm->out_file);

    // RANDOM
    parm->selection_probability = LoadReal(f, "Random FS selection probability", 0.1, parm->out_file);

    // FAST
    parm->fast_theta = LoadReal(f, "FAST Theta", -1.0, parm->out_file);

    // MICIBM
    parm->micibm_k = LoadInteger(f, "MICIBM k", -1, parm->out_file);

    // UFS
    parm->ufs_epsilon = LoadReal(f, "UFS epsilon", 0.1, parm->out_file);
    parm->ufs_R2max = LoadReal(f, "UFS R2max", 0.5, parm->out_file);

    // Undersampling
    parm->pre_undersampling = LoadBoolean(f, "Pre-undersampling", false, parm->out_file);
    parm->undersampling_method = LoadEnumerate(f, "Undersampling method", undersampling, n_undersampling, (int) umRandom, parm->out_file);

    // Random sampling before doing anything?
    parm->random_sampling = LoadBoolean(f, "Random sampling", false, parm->out_file);
    parm->random_sampling_ratio = LoadReal(f, "Random sampling ratio", 0.1, parm->out_file);

    // Factor to multiply each LS probability for existing individuals
    parm->ls_factor_old_individuals = LoadReal(f, "LS factor old individuals", 0.0, parm->out_file);
    parm->ls_only_use_selected = LoadBoolean(f, "LS only use selected", true, parm->out_file);

    /**************************************************/
    /************** Multi-Label methods ***************/
    /**************************************************/
    // Multi-label approach
    parm->ml_method = LoadEnumerate(f, "Multi-Label method", ml_method, n_mlmethod, (int) MLkNN, parm->out_file);
    parm->data_partition_method = LoadEnumerate(f, "Data partition method", mlpartition, n_mlpartition, (int) NoDataPartition, parm->out_file);
    parm->print_base_training_error = LoadBoolean(f, "ML print base learner training error", false, parm->out_file);

    parm->ml_undersample_binary_learners = LoadBoolean(f, "Multi-label undersample binary learners", false, parm->out_file);

     // Hyper-parameters
    parm->cv_ml_hyperparameters = LoadBoolean(f, "Cross-validate multi-label hyper-parameters", false, parm->out_file);

    // BR
    parm->br_model = LoadEnumerate(f, "BR classification model", learner, n_classifier, (int) DT_C45, parm->out_file);

    // ML-TREE
    parm->mltree_model = LoadEnumerate(f, "ML-TREE classification model", learner, n_classifier, (int) DT_C45, parm->out_file);

    // ML-FOREST
    parm->mlforest_ntrees = LoadInteger(f, "ML-FOREST number of trees", 50, parm->out_file);
    parm->use_mlcut = LoadBoolean(f, "ML-FOREST use MLCut", false, parm->out_file);

    // DBR
    parm->dbr_model = LoadEnumerate(f, "DBR classification model", learner, n_classifier, (int) DT_C45, parm->out_file);

    // STA
    parm->sta_model = LoadEnumerate(f, "STA classification model", learner, n_classifier, (int) DT_C45, parm->out_file);

    // NS
    parm->ns_model = LoadEnumerate(f, "NS classification model", learner, n_classifier, (int) DT_C45, parm->out_file);

    // Classifier chains
    parm->cc_model = LoadEnumerate(f, "CC classification model", learner, n_classifier, (int) DT_C45, parm->out_file);
    parm->ml_cc_random_order = LoadBoolean(f, "CC random order of labels", false, parm->out_file);

    // RAkEL
    parm->rakel_model = LoadEnumerate(f, "RAkEL classification model", learner, n_classifier, (int) DT_C45, parm->out_file);
    parm->rakel_k = LoadInteger(f, "RAkEL k", 3, parm->out_file);
    parm->rakel_m = LoadInteger(f, "RAkEL m", 0, parm->out_file);

    // Label Powerset
    parm->lp_model = LoadEnumerate(f, "LP classification model", learner, n_classifier, (int) DT_C45, parm->out_file);

    // Calibrated Label Ranking
    parm->clr_model = LoadEnumerate(f, "CLR classification model", learner, n_classifier, (int) DT_C45, parm->out_file);

    // HOMER
    parm->homer_k = LoadInteger(f, "HOMER k", 3, parm->out_file);   // Number of clusters
    parm->homer_model = LoadEnumerate(f, "HOMER classification model", learner, n_classifier, (int) DT_C45, parm->out_file);
    parm->bkmeans_iter = LoadInteger(f, "HOMER balanced k-means iterations", 100, parm->out_file);
    parm->homer_max_depth = LoadInteger(f, "HOMER maximum depth", 10, parm->out_file);

    // PEVUS
    parm->pevus_remove_null = LoadBoolean(f, "PEVUS remove null label instances", false, parm->out_file);
    parm->pevus_random_init = LoadBoolean(f, "PEVUS random initialization", false, parm->out_file);

    // Classifier evaluation
    parm->evaluate_ml_with[MLkNN] = LoadBoolean(f, "Evaluate ML with ML-kNN", false, parm->out_file);
    parm->evaluate_ml_with[BR] = LoadBoolean(f, "Evaluate ML with BR", false, parm->out_file);
    parm->evaluate_ml_with[CC] = LoadBoolean(f, "Evaluate ML with CC", false, parm->out_file);
    parm->evaluate_ml_with[LAMLkNN] = LoadBoolean(f, "Evaluate ML with LAML-kNN", false, parm->out_file);
    parm->evaluate_ml_with[MLksNN] = LoadBoolean(f, "Evaluate ML with ML-k'sNN", false, parm->out_file);
    parm->evaluate_ml_with[LAMLksNN] = LoadBoolean(f, "Evaluate ML with LAML-k'sNN", false, parm->out_file);
    parm->evaluate_ml_with[RAkEL_d] = LoadBoolean(f, "Evaluate ML with RAkEL-D", false, parm->out_file);
    parm->evaluate_ml_with[RAkEL_o] = LoadBoolean(f, "Evaluate ML with RAkEL-O", false, parm->out_file);
    parm->evaluate_ml_with[LP] = LoadBoolean(f, "Evaluate ML with LP", false, parm->out_file);
    parm->evaluate_ml_with[CLR] = LoadBoolean(f, "Evaluate ML with CLR", false, parm->out_file);
    parm->evaluate_ml_with[HOMER_B] = LoadBoolean(f, "Evaluate ML with HOMER-B", false, parm->out_file);
    parm->evaluate_ml_with[HOMER_K] = LoadBoolean(f, "Evaluate ML with HOMER-K", false, parm->out_file);
    parm->evaluate_ml_with[HOMER_R] = LoadBoolean(f, "Evaluate ML with HOMER-R", false, parm->out_file);
    parm->evaluate_ml_with[MLDT] = LoadBoolean(f, "Evaluate ML with ML-DT", false, parm->out_file);
    parm->evaluate_ml_with[MLBASICBP] = LoadBoolean(f, "Evaluate ML with ML-BASICBP", false, parm->out_file);
    parm->evaluate_ml_with[BPMLL] = LoadBoolean(f, "Evaluate ML with BP-MLL", false, parm->out_file);
    parm->evaluate_ml_with[MLRF] = LoadBoolean(f, "Evaluate ML with ML-RF", false, parm->out_file);
    parm->evaluate_ml_with[DBR] = LoadBoolean(f, "Evaluate ML with DBR", false, parm->out_file);
    parm->evaluate_ml_with[STA] = LoadBoolean(f, "Evaluate ML with STA", false, parm->out_file);
    parm->evaluate_ml_with[NS] = LoadBoolean(f, "Evaluate ML with NS", false, parm->out_file);
    parm->evaluate_ml_with[ECC] = LoadBoolean(f, "Evaluate ML with ECC", false, parm->out_file);
    parm->evaluate_ml_with[IBLRML] = LoadBoolean(f, "Evaluate ML with IBLR-ML", false, parm->out_file);
    parm->evaluate_ml_with[IBLRMLplus] = LoadBoolean(f, "Evaluate ML with IBLR-ML+", false, parm->out_file);
    parm->evaluate_ml_with[PCT] = LoadBoolean(f, "Evaluate ML with PCT", false, parm->out_file);
    parm->evaluate_ml_with[EPCT] = LoadBoolean(f, "Evaluate ML with E-PCT", false, parm->out_file);
    parm->evaluate_ml_with[RFPCT] = LoadBoolean(f, "Evaluate ML with RF-PCT", false, parm->out_file);
    parm->evaluate_ml_with[TSVM] = LoadBoolean(f, "Evaluate ML with TSVM", false, parm->out_file);
    parm->evaluate_ml_with[TSCCM] = LoadBoolean(f, "Evaluate ML with TSCCM", false, parm->out_file);
    parm->evaluate_ml_with[TSPCCM] = LoadBoolean(f, "Evaluate ML with TSPCCM", false, parm->out_file);
    parm->evaluate_ml_with[MLTREE] = LoadBoolean(f, "Evaluate ML with ML-TREE", false, parm->out_file);
    parm->evaluate_ml_with[MLFOREST] = LoadBoolean(f, "Evaluate ML with ML-FOREST", false, parm->out_file);
    parm->evaluate_ml_with[MLRS] = LoadBoolean(f, "Evaluate ML with MLRS", false, parm->out_file);
    parm->evaluate_ml_with[MLRSLC] = LoadBoolean(f, "Evaluate ML with MLRS-LC", false, parm->out_file);
    parm->evaluate_ml_with[OPCT] = LoadBoolean(f, "Evaluate ML with OPCT", false, parm->out_file);
    parm->evaluate_ml_with[SMLKNN] = LoadBoolean(f, "Evaluate ML with Stacked ML-kNN", false, parm->out_file);
    parm->evaluate_ml_with[DMLKNN] = LoadBoolean(f, "Evaluate ML with DML-kNN", false, parm->out_file);
    parm->evaluate_ml_with[MrKNN] = LoadBoolean(f, "Evaluate ML with Mr.KNN", false, parm->out_file);
    parm->evaluate_ml_with[DWKNND] = LoadBoolean(f, "Evaluate ML with MLC-DWkNN.D", false, parm->out_file);
    parm->evaluate_ml_with[DWKNNM] = LoadBoolean(f, "Evaluate ML with MLC-DWkNN.M", false, parm->out_file);
    parm->evaluate_ml_with[DWKNNZ] = LoadBoolean(f, "Evaluate ML with MLC-DWkNN.Z", false, parm->out_file);
    parm->evaluate_ml_with[FRONEC1H] = LoadBoolean(f, "Evaluate ML with FRONEC-1.H", false, parm->out_file);
    parm->evaluate_ml_with[FRONEC2H] = LoadBoolean(f, "Evaluate ML with FRONEC-2.H", false, parm->out_file);
    parm->evaluate_ml_with[FRONEC3H] = LoadBoolean(f, "Evaluate ML with FRONEC-3.H", false, parm->out_file);
    parm->evaluate_ml_with[FRONEC1LD] = LoadBoolean(f, "Evaluate ML with FRONEC-1.Ld", false, parm->out_file);
    parm->evaluate_ml_with[FRONEC2LD] = LoadBoolean(f, "Evaluate ML with FRONEC-2.Ld", false, parm->out_file);
    parm->evaluate_ml_with[FRONEC3LD] = LoadBoolean(f, "Evaluate ML with FRONEC-3.Ld", false, parm->out_file);
    parm->evaluate_ml_with[FVkNN] = LoadBoolean(f, "Evaluate ML with FV-kNN", false, parm->out_file);
    parm->evaluate_ml_with[NSSC] = LoadBoolean(f, "Evaluate ML with NSSC", false, parm->out_file);
    parm->evaluate_ml_with[EDT] = LoadBoolean(f, "Evaluate ML with E-DT", false, parm->out_file);
    parm->evaluate_ml_with[EXTRAPCT] = LoadBoolean(f, "Evaluate ML with Extra-PCT", false, parm->out_file);
    parm->evaluate_ml_with[PRUNEDSET] = LoadBoolean(f, "Evaluate ML with PRUNEDSETS", false, parm->out_file);
    parm->evaluate_ml_with[EPS] = LoadBoolean(f, "Evaluate ML with EPS", false, parm->out_file);
    parm->evaluate_ml_with[PPT] = LoadBoolean(f, "Evaluate ML with PPT", false, parm->out_file);
    parm->evaluate_ml_with[PPTN] = LoadBoolean(f, "Evaluate ML with PPT-n", false, parm->out_file);
    parm->evaluate_ml_with[PPTC] = LoadBoolean(f, "Evaluate ML with PPT-c", false, parm->out_file);
    parm->evaluate_ml_with[EXTRAMLDT] = LoadBoolean(f, "Evaluate ML with Extra-MLDT", false, parm->out_file);
    parm->evaluate_ml_with[ZEROML] = LoadBoolean(f, "Evaluate ML with ZERO", false, parm->out_file);

    // ML ensembles
    parm->evaluate_ml_with[BAGGING] = LoadBoolean(f, "Evaluate ML with Bagging", false, parm->out_file);
    parm->evaluate_ml_with[REALADABOOSTRE] = LoadBoolean(f, "Evaluate ML with RealAdaBoost.RE", false, parm->out_file);
    parm->evaluate_ml_with[DISCRETEADABOOSTRE] = LoadBoolean(f, "Evaluate ML with DiscreteAdaBoost.RE", false, parm->out_file);
    parm->evaluate_ml_with[REALADABOOSTMH] = LoadBoolean(f, "Evaluate ML with RealAdaBoost.MH", false, parm->out_file);
    parm->evaluate_ml_with[DISCRETEADABOOSTMH] = LoadBoolean(f, "Evaluate ML with DiscreteAdaBoost.MH", false, parm->out_file);

    // External learner evaluation
    parm->evaluate_ml_with_external[ExMLkNN] = LoadBoolean(f, "Evaluate ML with ML-kNN-py", false, parm->out_file);
    parm->evaluate_ml_with_external[RAkELd] = LoadBoolean(f, "Evaluate ML with RAkELd", false, parm->out_file);
    parm->evaluate_ml_with_external[RAkELo] = LoadBoolean(f, "Evaluate ML with RAkELo", false, parm->out_file);
    parm->evaluate_ml_with_external[MLARAM] = LoadBoolean(f, "Evaluate ML with MLARAM", false, parm->out_file);
    parm->evaluate_ml_with_external[ClassifierChain] = LoadBoolean(f, "Evaluate ML with Classifier Chain", false, parm->out_file);
    parm->evaluate_ml_with_external[BinaryRelevance] = LoadBoolean(f, "Evaluate ML with Binary Relevance", false, parm->out_file);
    parm->evaluate_ml_with_external[LabelPowerset] = LoadBoolean(f, "Evaluate ML with Label Powerset", false, parm->out_file);
    parm->evaluate_ml_with_external[BRkNNa] = LoadBoolean(f, "Evaluate ML with BRkNNa", false, parm->out_file);
    parm->evaluate_ml_with_external[BRkNNb] = LoadBoolean(f, "Evaluate ML with BRkNNb", false, parm->out_file);
    parm->evaluate_ml_with_external[MLTSVM] = LoadBoolean(f, "Evaluate ML with MLTSVM", false, parm->out_file);
    parm->evaluate_ml_with_external[MLMLP] = LoadBoolean(f, "Evaluate ML with MLMLP", false, parm->out_file);
    parm->evaluate_ml_with_external[DL] = LoadBoolean(f, "Evaluate ML with Deep Learning", false, parm->out_file);
    parm->evaluate_ml_with_external[BOOMER] = LoadBoolean(f, "Evaluate ML with BOOMER", false, parm->out_file);
    parm->evaluate_ml_with_external[BCC] = LoadBoolean(f, "Evaluate ML with BCC", false, parm->out_file);
    parm->evaluate_ml_with_external[CDN] = LoadBoolean(f, "Evaluate ML with CDN", false, parm->out_file);
    parm->evaluate_ml_with_external[HASEL] = LoadBoolean(f, "Evaluate ML with HASEL", false, parm->out_file);
    parm->evaluate_ml_with_external[PSt] = LoadBoolean(f, "Evaluate ML with PSt", false, parm->out_file);
    parm->evaluate_ml_with_external[CT] = LoadBoolean(f, "Evaluate ML with CT", false, parm->out_file);
    parm->evaluate_ml_with_external[MCC] = LoadBoolean(f, "Evaluate ML with MCC", false, parm->out_file);
    parm->evaluate_ml_with_external[NSR] = LoadBoolean(f, "Evaluate ML with NSR", false, parm->out_file);
    parm->evaluate_ml_with_external[SCC] = LoadBoolean(f, "Evaluate ML with SCC", false, parm->out_file);
    parm->evaluate_ml_with_external[PS] = LoadBoolean(f, "Evaluate ML with PS", false, parm->out_file);
    parm->evaluate_ml_with_external[LC] = LoadBoolean(f, "Evaluate ML with LC", false, parm->out_file);
    parm->evaluate_ml_with_external[CDT] = LoadBoolean(f, "Evaluate ML with CDT", false, parm->out_file);
    parm->evaluate_ml_with_external[DBPNN] = LoadBoolean(f, "Evaluate ML with DBPNN", false, parm->out_file);
    parm->evaluate_ml_with_external[FW] = LoadBoolean(f, "Evaluate ML with FW", false, parm->out_file);
    parm->evaluate_ml_with_external[Maniac] = LoadBoolean(f, "Evaluate ML with Maniac", false, parm->out_file);
    parm->evaluate_ml_with_external[PCC] = LoadBoolean(f, "Evaluate ML with PCC", false, parm->out_file);
    parm->evaluate_ml_with_external[PMCC] = LoadBoolean(f, "Evaluate ML with PMCC", false, parm->out_file);

    // ML ensembles
    parm->ml_ensemble_t = LoadReal(f, "ML ensemble threshold", 0.5, parm->out_file);
    parm->ml_ensemble_threshold_by_resubs = LoadBoolean(f, "ML ensemble set threshold by resubstitution", true, parm->out_file);

    // CDN
    parm->cdn_I = LoadInteger(f, "CDN iterations", 1000, parm->out_file);
    parm->cdn_Ic = LoadInteger(f, "CDN collection iterations", 100, parm->out_file);

    // NSR
    parm->nsr_p = LoadInteger(f, "NSR P", 2, parm->out_file);
    parm->nsr_n = LoadInteger(f, "NSR N", 2, parm->out_file);

    // ML-ESO
    parm->mleso_unsel_nocopy = LoadBoolean(f, "ML-ESO do not copy unselected instances", false, parm->out_file);
    parm->mleso_donot_eval_copies = LoadBoolean(f, "ML-ESO do not evaluate copies", true, parm->out_file);
    parm->ml_eso_os_ratio = LoadReal(f, "ML-ESO random oversampling ratio", 0.5, parm->out_file);
    parm->ml_eso_us_ratio = LoadReal(f, "ML-ESO random undersampling ratio", 0.5, parm->out_file);
    parm->chc_random_evaluation_ratio = LoadReal(f, "ML-ESO random evaluation ratio", 1.0, parm->out_file);
    parm->mleso_one_w = LoadBoolean(f, "ML-ESO use only one weight", false, parm->out_file);
    parm->label_propagation = LoadEnumerate(f, "ML-ESO label propagation", label_propagation, n_label_propagation, (int) EVOL, parm->out_file);
    parm->mleso_loo = LoadBoolean(f, "ML-ESO leave one out", true, parm->out_file);

    // Multi-label ensembles
    parm->ml_ensemble_size = LoadInteger(f, "ML ensemble size", 50, parm->out_file);
    parm->ml_resampling_method = LoadEnumerate(f, "ML resampling method", ml_resampling, n_ml_resampling, (int) ResamplingPerLabel, parm->out_file);
    parm->ml_ensemble_method = LoadEnumerate(f, "ML ensemble method", ml_ensemble_methods, n_ml_ensemble_methods, (int) MLNone, parm->out_file);
    parm->ml_base_learner = LoadEnumerate(f, "ML base classifier", ml_method, n_ml_base_classifier, (int) MLDT, parm->out_file);

    // Mr.KNN
    parm->mrknn_f = LoadReal(f, "Mr.KNN f", 2.0, parm->out_file);
    parm->mrknn_loo = LoadBoolean(f, "Mr.KNN loo", true, parm->out_file);

    // Option MLDT
    parm->optiondt_epsilon = LoadReal(f, "Option PCT epsilon", 0.1, parm->out_file);
    parm->optiondt_O = LoadInteger(f, "Option PCT O", 10, parm->out_file);
    parm->optiondt_l = LoadInteger(f, "Option PCT l", 2, parm->out_file);

    // MLRS
    parm->mlrs_s = LoadReal(f, "MLRS s", 1.0, parm->out_file);
    parm->mlrs_k = LoadInteger(f, "MLRS k", 10, parm->out_file);
    parm->mlrs_beta = LoadReal(f, "MLRS beta", 0.9, parm->out_file);

    // Pruned sets method
    parm->ps_b = LoadInteger(f, "PS b", 2, parm->out_file);
    parm->ps_p = LoadInteger(f, "PS p", 1, parm->out_file);
    parm->ps_labels_bound = LoadInteger(f, "PS labels upper limit", 10, parm->out_file);
    parm->ps_strategy = LoadEnumerate(f, "PS strategy", ps_strategy, n_ps_strategy, (int) A, parm->out_file);
    parm->ps_model = LoadEnumerate(f, "PS classification model", learner, n_classifier, (int) DT_C45, parm->out_file);

    // Pruned sets ensembles
    parm->eps_size = LoadInteger(f, "EPS size", 30, parm->out_file);
    parm->eps_t = LoadReal(f, "EPS t", 0.3, parm->out_file);

    // PPT combined
    parm->pptc_k = LoadInteger(f, "PPT-c k", 5, parm->out_file);
    parm->pptc_t = LoadReal(f, "PPT-c t", 0.2, parm->out_file);
    parm->pptc_no_loss = LoadBoolean(f, "PPT-c with no loss", true, parm->out_file);

    // Base method for external ML classifiers
    parm->ml_ext_base_method = LoadEnumerate(f, "ML external base method", ml_ext_base_method, n_ml_ext_base_method, (int) DecisionTree, parm->out_file);

    // External learner of evaluating individuals at a certain number of iterations
    parm->evolution_external_classifier = LoadEnumerate(f, "ML external fitness evaluator", ml_ext_method, n_ml_ext_method, (int) BinaryRelevance, parm->out_file);
    parm->ml_external_eval_interval = LoadInteger(f, "ML external fitness evaluation interval", 0, parm->out_file);

    // ML-DT
    parm->mldt_adaptive_threshold = LoadBoolean(f, "ML-DT adaptive threshold", false, parm->out_file);

    // ML-RF and other ensembles of trees
    parm->mlrf_t = LoadReal(f, "ML-RF threshold", 0.5, parm->out_file);
    parm->mlrf_real_votes = LoadBoolean(f, "ML-RF real votes", false, parm->out_file);

    // ML-TREE
    parm->mltree_lambda = LoadReal(f, "ML-TREE lambda", 0.5, parm->out_file);
    parm->mltree_depth = LoadInteger(f, "ML-TREE maximum depth", 5, parm->out_file);

    // MLMLP
    parm->mlmlp_hidden_nodes = LoadInteger(f, "MLMLP hidden nodes", 0, parm->out_file); // 0 means number of inputs

    // ML-kNN
    parm->mlknn_k = LoadInteger(f, "ML-kNN k", 3, parm->out_file);
    parm->mlknn_s = LoadReal(f, "ML-kNN s", 1.0, parm->out_file);
    parm->use_neighbors_cache = LoadBoolean(f, "Use neighbors cache", true, parm->out_file);

    // SML-kNN
    parm->smlknn_k = LoadInteger(f, "SML-kNN second layer k", 10, parm->out_file);

    // DML-kNN
    parm->dmlknn_delta = LoadInteger(f, "DML-kNN delta", 3, parm->out_file);

    // LAML-kNN
    parm->lamlknn_n_clusters = LoadInteger(f, "LAML-kNN number of clusters", 10, parm->out_file);
    parm->lamlknn_min_n_clusters = LoadInteger(f, "LAML-kNN minimum number of clusters", 1, parm->out_file);
    parm->lamlknn_max_n_clusters = LoadInteger(f, "LAML-kNN maximum number of clusters", 1, parm->out_file);

    // LAML-k'sNN
    parm->lamlksnn_is = LoadBoolean(f, "LAML-k'sNN instance selection", false, parm->out_file);

    // TS architecture
    parm->tsa_t = LoadReal(f, "ML TSA threshold", -0.4, parm->out_file);

    // ECC
    parm->ecc_T = LoadInteger(f, "ECC size", 10, parm->out_file);
    parm->ecc_with_bagging = LoadBoolean(f, "ECC trained with bagging", true, parm->out_file);

    // IBLR
    parm->iblr_k = LoadInteger(f, "IBLR number of neighbors", 10, parm->out_file);

    // ML IS evolution
    parm->chc_init_prob = LoadReal(f, "ML CHC initialization probability", 0.25, parm->out_file);
    parm->chc_minimum_distance_factor = LoadReal(f, "ML CHC minimum distance factor", 4.0, parm->out_file);
    parm->ml_chc_random_mut = LoadReal(f, "ML CHC random mutation", 0.0, parm->out_file);
    parm->ml_return_best_performing = LoadBoolean(f, "ML IS return overall best performing", false, parm->out_file);
    parm->ml_k_memetic_mutation_ratio = LoadReal(f, "k ML memetic mutation ratio", 0.0, parm->out_file);

    // Multi-selection
    parm->ml_multiselection_mutation_ratio = LoadReal(f, "ML Multi-selection mutation ratio", 0.0, parm->out_file);
    parm->ml_max_ioversampling_ratio = LoadInteger(f, "ML maximum intelligent oversampling ratio", 0, parm->out_file);

    // ML Class-imbalance method
    parm->ml_imbalance_method = LoadEnumerate(f, "ML Class-imbalance method", ml_class_imbalance_method, n_ml_class_imbalance_method, (int) MLNoneS, parm->out_file);
    parm->ml_ros_p = LoadReal(f, "ML Random oversampling m/M ratio", 0.1, parm->out_file);
    parm->ml_rus_p = LoadReal(f, "ML Random undersampling M/m ratio", 10.0, parm->out_file);

    // PRUS and PROS
    parm->partial_ci_samples = LoadBoolean(f, "ML PRUS/PROS applied to every sample", false, parm->out_file);

    // PRUS
    parm->prus_relative_ratio = LoadBoolean(f, "ML PRUS ratio is relative", false, parm->out_file);

    // PROS
    parm->pros_relative_ratio = LoadBoolean(f, "ML PROS ratio is relative", false, parm->out_file);

    // Cooperative coevolution
    parm->ml_cc_mutate_elite = LoadBoolean(f, "ML mutate elite", false, parm->out_file);
    parm->ml_cc_local_opt_best = LoadBoolean(f, "ML CC local optimization of best individual", false, parm->out_file);

    // Instance selection algorithm
    parm->ml_instance_selection = LoadEnumerate(f, "ML instance selection algorithm", ml_instance_selection, n_ml_instance_selection, (int) NoInstanceSelection, parm->out_file);
    parm->ml_relative_multiselection = LoadBoolean(f, "ML relative multi-selection", false, parm->out_file);
    parm->sbc_nc = LoadReal(f, "SBC Nc", 2.0, parm->out_file);
    parm->initialize_all_minority_class = LoadBoolean(f, "Initialize all minority class", false, parm->out_file);

    // Feature selection algorithm
    parm->ml_feature_selection = LoadEnumerate(f, "ML feature selection algorithm", ml_feature_selection, n_ml_feature_selection, (int) NoFeatureSelection, parm->out_file);

    // Binary classifiers
    parm->binary_classifier = LoadEnumerate(f, "Binary learner", learner, n_classifier, (int) kNN, parm->out_file);
    parm->k = LoadInteger(f, "k-NN k", 3, parm->out_file);

    // BP-MLL
    parm->bpmll_alpha = LoadReal(f, "BP-MLL alpha", 0.1, parm->out_file);
    parm->bpmll_nodes = LoadReal(f, "BP-MLL hidden nodes", 1.0, parm->out_file);

    // Subpopulation fitness
    parm->subpop_fitness_w = LoadReal(f, "Subpopulation fitness weight fitness", 0.0, parm->out_file);
    parm->subpop_isolated_w = LoadReal(f, "Subpopulation fitness weight isolated", 0.0, parm->out_file);
    parm->subpop_difference_w = LoadReal(f, "Subpopulation fitness weight difference", 0.0, parm->out_file);
    parm->subpop_substitution_w = LoadReal(f, "Subpopulation fitness weight substitution", 0.0, parm->out_file);
    parm->subpop_reduction_w = LoadReal(f, "Subpopulation fitness weight reduction", 0.0, parm->out_file);
    parm->subpop_onelabel_w = LoadReal(f, "Subpopulation fitness weight one label", 0.0, parm->out_file);

    parm->ml_cc_subpop_random_fitness = LoadBoolean(f, "Subpopulation random fitness", false, parm->out_file);

    parm->subpop_perf_measure = LoadEnumerate(f, "Subpopulation performance measure", performance, n_performance, (int) pfGMean, parm->out_file);
    parm->subpop_binary_classifier = LoadEnumerate(f, "Subpopulation binary learner", learner, n_classifier, (int) kNN, parm->out_file);

    parm->ml_init_with_undersampling = LoadBoolean(f, "Initialize ML instance selection with undersampling", false, parm->out_file);
    parm->ml_init_with_roulette = LoadBoolean(f, "Initialize ML instance selection with roulette", false, parm->out_file);
    parm->ml_init_with_cardinality = LoadBoolean(f, "Initialize ML instance selection with cardinality", false, parm->out_file);
    parm->ml_evolve_undersampled_subpops = LoadBoolean(f, "Evolve undersampled subpopulations", false, parm->out_file);

    // Use hash
    parm->ml_use_hash_table = LoadBoolean(f, "ML Use hash table", false, parm->out_file);

    // 3-way instance selection
    parm->three_way_is = LoadBoolean(f, "Three way instance selection", false, parm->out_file);
    parm->three_way_minus1_prob = LoadReal(f, "Three way -1 initial probability", 0.1, parm->out_file);
    parm->cheat_neg_to_pos_only = LoadBoolean(f, "Three way only negative to positive", false, parm->out_file);
    parm->reinitalize_after_gs_no_improvement = LoadInteger(f, "Reinitalize after these number of G's with no improvement", 0, parm->out_file);

    // Subpopulation memetic mutation
    parm->subpop_drop3_mut = LoadReal(f, "Subpopulation memetic Drop3 mutation", 0.0, parm->out_file);
    parm->subpop_oss_mut = LoadReal(f, "Subpopulation memetic OSS mutation", 0.0, parm->out_file);

    // Partial selection
    parm->partial_selection_mutation = LoadReal(f, "Partial selection mutation", 0.0, parm->out_file);
    parm->partial_selection_mutation_bit_ratio = LoadReal(f, "Partial selection mutation bit ratio", 0.1, parm->out_file);
    parm->partial_is_mutation = LoadReal(f, "Partial selection single-label instance selection mutation", 0.0, parm->out_file);
    parm->biased_hux_ratio = LoadReal(f, "Biased HUX 1 ratio", 0.0, parm->out_file);

    // Evaluation weights
    parm->reduction_ml_w = LoadReal(f, "Evaluation weight reduction", 0.0, parm->out_file);
    parm->reduction_ml_ms_w = LoadReal(f, "Evaluation weight multi-selection reduction", 0.0, parm->out_file);
    parm->reduction_perlabel_w = LoadReal(f, "Evaluation weight per label reduction", 0.0, parm->out_file);
    parm->eb_subset_accuracy_w = LoadReal(f, "Evaluation weight EB subset accuracy", 0.0, parm->out_file);
    parm->eb_hamming_loss_w = LoadReal(f, "Evaluation weight EB Hamming loss", 0.0, parm->out_file);
    parm->eb_jaccard_coeff_w = LoadReal(f, "Evaluation weight EB Jaccard coefficient", 1.0, parm->out_file);
    parm->eb_accuracy_w = LoadReal(f, "Evaluation weight EB accuracy", 0.0, parm->out_file);
    parm->eb_precision_w = LoadReal(f, "Evaluation weight EB precision", 0.0, parm->out_file);
    parm->eb_recall_w = LoadReal(f, "Evaluation weight EB recall", 0.0, parm->out_file);
    parm->eb_f1_w = LoadReal(f, "Evaluation weight EB F1", 0.0, parm->out_file);
    parm->lb_accuracy_macro_w = LoadReal(f, "Evaluation weight LB accuracy macro", 0.0, parm->out_file);
    parm->lb_precision_macro_w = LoadReal(f, "Evaluation weight LB precision macro", 0.0, parm->out_file);
    parm->lb_recall_macro_w = LoadReal(f, "Evaluation weight LB recall macro", 0.0, parm->out_file);
    parm->lb_f1_macro_w = LoadReal(f, "Evaluation weight LB F1 macro", 0.0, parm->out_file);
    parm->lb_accuracy_micro_w = LoadReal(f, "Evaluation weight LB accuracy micro", 0.0, parm->out_file);
    parm->lb_precision_micro_w = LoadReal(f, "Evaluation weight LB precision micro", 0.0, parm->out_file);
    parm->lb_recall_micro_w = LoadReal(f, "Evaluation weight LB recall micro", 0.0, parm->out_file);
    parm->lb_f1_micro_w = LoadReal(f, "Evaluation weight LB F1 micro", 0.0, parm->out_file);
    parm->eb_one_error_w = LoadReal(f, "Evaluation weight EB one error", 0.0, parm->out_file);
    parm->eb_coverage_w = LoadReal(f, "Evaluation weight EB coverage", 0.0, parm->out_file);
    parm->eb_ranking_loss_w = LoadReal(f, "Evaluation weight EB ranking loss", 0.0, parm->out_file);
    parm->eb_average_precision_w = LoadReal(f, "Evaluation weight EB average precision", 0.0, parm->out_file);
    parm->lb_auc_macro_w = LoadReal(f, "Evaluation weight LB AUC macro", 0.0, parm->out_file);
    parm->lb_auc_micro_w  = LoadReal(f, "Evaluation weight LB AUC micro", 0.0, parm->out_file);
    parm->eb_auc_w  = LoadReal(f, "Evaluation weight LB AUC instance", 0.0, parm->out_file);
    parm->eb_f1_alt_w = LoadReal(f, "Evaluation weight LB F1 instance", 0.0, parm->out_file);
    parm->n_predicted_ml_w = LoadReal(f, "Evaluation weight number of predicted labels", 0.0, parm->out_file);
    parm->regularize_k_w = LoadReal(f, "Evaluation k regularization weight", 0.0, parm->out_file);
    parm->oversampling_ml_ms_w = LoadReal(f, "Evaluation oversampling weight", 0.0, parm->out_file);

    parm->gauss_reduction = LoadBoolean(f, "Gaussian reduction", false, parm->out_file);

    parm->combine_metrics_gmean = LoadBoolean(f, "Combine metrics using G-mean", false, parm->out_file);

    parm->ml_cc_combine_best_subinviduals = LoadBoolean(f, "ML CC combine best subindividuals", false, parm->out_file);   // Population is not evolved

    // Cooperative coevolution
    parm->ml_cc_pop_evol_scheme = LoadEnumerate(f, "ML CC population evolutionary scheme", evolutionary_scheme, n_evolutionary_scheme, (int) schemeCHCGA, parm->out_file);
    parm->ml_return_overall_best = LoadBoolean(f, "ML CC return overall best", false, parm->out_file);
    parm->population_size = LoadInteger(f, "Population size", 100, parm->out_file);
    parm->chc_pop_init_prob = LoadReal(f, "Population CHC initialization probability", 0.5, parm->out_file);
    parm->subpop_size = LoadInteger(f, "Subpopulations size", 100, parm->out_file);
    parm->chc_subpop_init_prob = LoadReal(f, "Subpopulations CHC initialization probability", 0.25, parm->out_file);
    parm->chc_subpop_minimum_distance_factor = LoadInteger(f, "Subpopulation CHC minimum distance factor", (int) 2.0/parm->chc_subpop_init_prob, parm->out_file);
    parm->np = LoadInteger(f, "Np", 10, parm->out_file);
    parm->ssga_mutation = LoadReal(f, "SSGA Mutation", 0.01, parm->out_file);
    parm->nsp = LoadInteger(f, "Nsp", 100, parm->out_file);
    parm->reevaluate_individuals = LoadBoolean(f, "Reevaluate individuals each Nsp generation", false, parm->out_file);
    parm->merge_method = LoadEnumerate(f, "Merge method", merge_method, n_merge_method, (int) QuorumSensing, parm->out_file);
    parm->quorum_sensing_threshold_min = LoadReal(f, "Minimum quorum sensing threshold", 0.0, parm->out_file);
    parm->quorum_sensing_threshold_max = LoadReal(f, "Maximum quorum sensing threshold", 1.0, parm->out_file);
    parm->ml_cc_weighted_qs = LoadBoolean(f, "ML CC weighted subpopulations for quorum sensing", false, parm->out_file);
    parm->ml_onesubpop_per_cycle = LoadBoolean(f, "Evolve only one subpopulation per cycle", false, parm->out_file);
    parm->ml_select_subpop_by_relevance = LoadBoolean(f, "Select subpopulation by relevance", false, parm->out_file);
    parm->ml_allow_negative_subpops = LoadBoolean(f, "ML CC allow negative subpopulation individual index", false, parm->out_file);
    parm->ml_min_k = LoadInteger(f, "ML-kNN minimum k", 1, parm->out_file);
    parm->ml_max_k = LoadInteger(f, "ML-kNN maximum k", 100, parm->out_file);
    parm->mlksnn_min_k = LoadInteger(f, "ML-k'sNN minimum k", 1, parm->out_file);
    parm->mlksnn_max_k = LoadInteger(f, "ML-k'sNN maximum k", 100, parm->out_file);
    parm->regularize_mlksnn = LoadBoolean(f, "Regularize ML-k'sNN", false, parm->out_file);
    parm->ml_elitism = LoadReal(f, "ML elitism", 0.0, parm->out_file);

    // ML multi-selection
    parm->ml_is_set_k_cv = LoadBoolean(f, "ML IS k fixed by CV", false, parm->out_file);
    parm->ml_use_hux = LoadBoolean(f, "ML use HUX", true, parm->out_file);

    // ML-k'sNN
    parm->mlksnn_memetic_mut = LoadReal(f, "ML-k'sNN memetic mutation", 0.0, parm->out_file);
    parm->mlksnn_ls_mut = LoadReal(f, "ML-k'sNN local search mutation", 0.0, parm->out_file);
    parm->mlksnn_ls_steps = LoadInteger(f, "ML-k'sNN local search mutation steps", 100, parm->out_file);
    parm->ml_loo = LoadBoolean(f, "ML evaluate with leave-one-out", true, parm->out_file);
    parm->mlksnn_loo = LoadBoolean(f, "Evaluate ML-k'sNN with leave one out", parm->ml_loo, parm->out_file);

    parm->cc_memetic_mutation = LoadReal(f, "ML CC memetic mutation", 0.0, parm->out_file);
    parm->cc_memetic_mutation_for_subpop = LoadReal(f, "ML CC memetic mutation for subpopulations", 0.0, parm->out_file);
    parm->cc_memetic_mutation_for_subpop_bit_ratio = LoadReal(f, "ML CC memetic mutation for subpopulations bit ratio", 0.1, parm->out_file);

    parm->negatives_ml_w = LoadReal(f, "Evaluation weight negative subpopulations", 0.0, parm->out_file);
    parm->ml_ins_sel_per_label = LoadBoolean(f, "ML instance selection per label", false, parm->out_file);
    parm->ml_kperlabel_method = LoadEnumerate(f, "ML-k'sNN method", ml_kperlabel_method, n_ml_kperlabel_method, (int) PerpendicularSearch, parm->out_file);
    parm->kperlabel_local_opt_ratio = LoadReal(f, "ML-k'sNN local optimization ratio", 0.0, parm->out_file);
    parm->kperlabel_rubber_band_penalty = LoadReal(f, "ML-k'sNN rubber band penalty", 0.0, parm->out_file);
    parm->kperlabel_different_ks_penalty = LoadReal(f, "ML-k'sNN different k's penalty", 0.0, parm->out_file);
    parm->mlksnn_rounds = LoadInteger(f, "ML-k'sNN rounds", 1, parm->out_file);
    parm->with_replacement = LoadBoolean(f, "ML-k'sNN sampling with replacement", true, parm->out_file);
    parm->mlksnn_ties_closer_to_cv_k = LoadBoolean(f, "ML-k'sNN resolve ties with closer to cross-validated k", true, parm->out_file);
    parm->mlksnn_use_window = LoadBoolean(f, "ML-k'sNN use window for best K", false, parm->out_file);
    parm->mlksnn_window_size = LoadInteger(f, "ML-k'sNN window size", 5, parm->out_file);
    parm->mlksnn_per_label_sampling = LoadBoolean(f, "ML-k'sNN per label sampling", false, parm->out_file);
    parm->mlksnn_update_ada_weight_with_current_best = LoadBoolean(f, "ML-k'sNN update AdaboostMH weights with current best", false, parm->out_file);
#ifdef ITERATIVE_SAMPLING_FOR_MLKsNN
    parm->kperlabel_sampling_ratio_step = LoadReal(f, "ML-k'sNN sampling ratio step", 1.0, parm->out_file);
#endif
    parm->ml_ins_sel_per_label_same_neighbors = LoadBoolean(f, "ML instance selection per label with the same neighbors", false, parm->out_file);
    parm->mlknn_k_per_label_stepwise_attempts = LoadInteger(f, "ML-k'sNN stepwise attempts", 1, parm->out_file);
    parm->ml_k_penalty_from_cv = LoadReal(f, "ML-k'sNN penalty from global CV valued", 0.0, parm->out_file);

    // Open MP
    parm->omp_allow_nested_loops = LoadBoolean(f, "OpenMP allow nested loops", false, parm->out_file);

    fclose(f);

    fprintf(parm->out_file, "───────────────────────────\n───────────────────────────\n");

    /**************************************************/
    /************** Incompatibilities MULTI-LABEL *****/
    /**************************************************/
    if ((parm->ml_allow_negative_subpops || parm->cc_memetic_mutation > ZERO || parm->ml_cc_local_opt_best || parm->cc_memetic_mutation_for_subpop)
        && parm->ml_instance_selection == CooperativeCoevolution && parm->ml_ins_sel_per_label)
        OnErrorExit(ERR_PARAM, "Partial cooperative selection cannot evolve with negative subpopulations");

    /**************************************************/
    /************** Incompatibilities *****************/
    /**************************************************/
    if (parm->isalgorithm == RNN || parm->isalgorithm == IB3 || parm->isalgorithm == MSS) {
        fprintf(parm->out_file, "k set to 1\n");
        parm->k = 1;
    }

    if (parm->isalgorithm == PSA && abs(parm->k) == 1) {
        OnErrorExit(ERR_PARAM, "Bad k for PSA");
    }

    if (parm->huge_dataset && !parm->external_partitioning_approach)
        OnErrorExit(ERR_PARAM, "Huge dataset must be stratified");

    if (parm->isalgorithm == BoostBased) {
        if (parm->ensemble_method != AdaBoost && parm->ensemble_method != eMultiBoost && parm->ensemble_method != eFloatBoost && parm->ensemble_method != eMadaBoost && parm->ensemble_method !=
            eGentleBoost && parm->ensemble_method != Wagging && parm->ensemble_method != eReweightBoost)
            OnErrorExit(ERR_PARAM, "Boost based instance selection only implemented for AdaBoost, eFloatBoost or eMultiBoost");
        if (parm->ensemble_method == eFloatBoost)
            parm->search_heuristic = SFFS;
        if (parm->ensemble_method == eReweightBoost)
            parm->search_heuristic = SFS;
    }

    if (parm->k == 1 && parm->multiselection && !parm->evolutionary_local_k_and_is)
        printf("WARNING: With k=1 multi-instance selection MIGHT make no sense\n");

    if (parm->rnn_isfs_mutation > 0.0 && parm->isfs_inputs_first && parm->isfsmutation_random_order)
        printf("Inputs first IGNORED if random order\n");

    //     if (parm->multiselect_only_minority_class && !parm->multiis_and_k_with_DE && (parm->performance_measure != pfGMean || !parm->multiselection))
    //         OnErrorExit(ERR_PARAM, "Multi-select only minority class must be used with G mean and multi-selection");

    if (parm->multiis_and_k_with_DE)
        parm->multiselection = true;

    if (parm->evolutionary_local_k) {
        parm->multiselection = parm->feature_selection = parm->external_partitioning_approach = parm->instance_selection = parm->random_sampling = false;
        if (parm->localk_error_estimation_method == BS || parm->localk_error_estimation_method == E632)
            parm->localk_nfolds = 2; // For cache size
        if ((parm->localk_error_estimation_method == LOO || parm->localk_error_estimation_method == RESUBS) && parm->localk_reevaluate)
            parm->localk_reevaluate = false;

        if (parm->localk_memetic_mutation_ratio > 0.0 && (parm->evolutionary_scheme == schemeCHCGA || parm->evolutionary_scheme == schemeSSGA))
            OnErrorExit(ERR_PARAM, "No memetic mutation with CHC or SS genetic algorithms");
    }

    if (parm->evolutionary_local_k_and_is) {
        parm->evolutionary_local_k = parm->feature_selection = parm->random_sampling = false;
        parm->drop3_mutation = parm->icf_mutation = parm->ib3_mutation = parm->mss_mutation = parm->rnnstd_mutation = parm->addI_mutation = 0.0;
    }

    if (parm->optimal_local_k) {
        if (parm->evolutionary_local_k)
            OnErrorExit(ERR_PARAM, "Evolutionary and optimal local k incompatible");

        if (parm->localk_min_k < 1)
            OnErrorExit(ERR_PARAM, "Optimal local k minimum must be at least 1");

        parm->multiselection = parm->feature_selection = parm->instance_selection = parm->random_sampling = parm->localk_populate = false;
    }

    // if (parm->external_partitioning_approach && parm->n_electors)
    //     OnErrorExit(ERR_PARAM, "Stratified and democratic approaches incompatible, choose one");

    if (parm->performance_measure == pfauROC && parm->localk_alpha_mar > 0.0) {
        printf("For auROC estimation margin weight is set to 0\n");
        parm->localk_alpha_mar = 0.0;
    }

    if (parm->symmetric_knn && parm->knn_with_weighted_votes)
        OnErrorExit(ERR_PARAM, "Weighted votes and symmetric k-NN are incompatible");

    if (parm->use_all_data)
        parm->instance_selection = parm->feature_selection = false;

    if (parm->chc_wrapper != kNN)
        parm->democratic_evaluation_of_threshold_with_cv = false;

    if (parm->votes_evaluator != parm->chc_wrapper && parm->isalgorithm == CHC)
        printf("WARNING: CHC wrapper and votes evaluator differ\n");

    // Only partition available for huge datasets
    if (parm->huge_dataset)
        parm->internal_partitioning_method = parm->external_partitioning_method = pRANDOM;

    if (parm->multiselection && parm->isalgorithm != CHC)
        printf("WARNING: Democratic multiple-instance selection MIGHT not be implemented for this selector\n");

#ifdef SCALAR_is_DOUBLE
    if (parm->huge_dataset)
        OnErrorExit(ERR_PARAM, "Huge dataset only compatible with float type");
#endif

    if (parm->isalgorithm == IB3 || parm->isalgorithm == MSS) {
        parm->resample = true;
        parm->k = 1;
        fprintf(parm->out_file, "Resampling => TRUE\nk => 1\n");
    }

    if (parm->isalgorithm == RNN) {
        parm->resample = true;
        fprintf(parm->out_file, "Resampling => TRUE\n");
    }

    // Random algorithm only make sense with Ada Boost
    if (parm->isalgorithm == RANDOMSELECTION && (parm->ensemble_method == Wagging || parm->ensemble_method == eBagging || parm->ensemble_method ==
            RandomSubsampling))
        OnErrorExit(ERR_PARAM, "Random algorithm only make sense with boosting");

    if (parm->ensemble_method == eBagging) {
        parm->resample = true;
        fprintf(parm->out_file, "Resampling => TRUE (bagging method)\n");
    }

    if (parm->ensemble_method == Wagging) {
        parm->resample = false;
        fprintf(parm->out_file, "Resampling => FALSE\n");
        // Only valid for algorithms accepting weights
        if (parm->isalgorithm != DROP3 && parm->isalgorithm != CHC)
            OnErrorExit(ERR_PARAM, "Wagging not valid without weights");
    }

    if (parm->ensemble_method == RandomSubsampling) {
        parm->resample = true;
        fprintf(parm->out_file, "Resampling => TRUE (random subsampling)\n");
    }

    if (parm->boostis_considerlist) {
        if (parm->isalgorithm != CNN && parm->isalgorithm != DROP3 && parm->isalgorithm != ICF && parm->isalgorithm != CHC)
            OnErrorExit(ERR_PARAM, "Consider list only implemented for CNN, Drop3, CHC and ICF");

        if (parm->ensemble_method == Wagging)
            OnErrorExit(ERR_PARAM, "Wagging not compatible with consider list");

        if (parm->resample) {
            parm->resample = false;
            printf("Resampling => FALSE (consider list is true)\n");
        }
    }

    if (parm->boostis_considerlist && (parm->external_partitioning_approach || parm->internal_partitioning_approach))
        OnErrorExit(ERR_PARAM, "Consider list and partitioning approach are incompatibles");

    if (parm->isalgorithm == ICF)
        parm->filter_noise = true;

    if (parm->individuals % 2)
        parm->individuals++;

    if (parm->internal_partitioning_approach && parm->ensemble_method == eFloatBoost)
        printf("WARNING: eFloatBoost does not use internal partitioning for weight updating\n");

    if (parm->external_partitioning_approach && (parm->th_eval_n_iterarions != 1 || parm->update_w_n_iterarions != 1))
        parm->th_eval_n_iterarions = parm->update_w_n_iterarions = 1;

    if (parm->add_k_neighbors && parm->boosting_resampling_method == WithReplacement)
        OnErrorExit(ERR_PARAM, "Add k and sampling with replacement are incompatible");

    if (parm->ensemble_implementation == Reweighting)
        OnErrorExit(ERR_PARAM, "Reweighting not yet implemented");

    if (parm->use_validation_data_for_votes_threshold)
        OnErrorExit(ERR_PARAM, "Use of validation data not yet implemented");

    if ((parm->ensemble_method == BalanceCascade || parm->ensemble_method == EasyEnsemble || parm->ensemble_method == AsymmetricAdaBoost) && parm->random_undersampling) {
        parm->random_undersampling = false;
        printf("Random undersampling deactivated for BalanceCascade, EasyEnsemble or AsymmetricAdaBoost\n");
    }

    if (parm->random_undersampling)
        // Already undersampled
        parm->random_undersampling_for_training = false;

    // Inconsistency between learner and evaluator, always correct if evaluator is eIncM
    if ((parm->fsalgorithm == fsGA) && ((parm->evaluator  == eC45 && parm->classifier != C45) || (parm->evaluator  == eSVM && parm->classifier != SVM) || (parm->evaluator == ekNN &&
                                        parm->classifier != kNN)))
        OnErrorExit(ERR_PARAM, "LoadGlobalParameters: Inconsistency between learner and evaluator\n");

    if (parm->ensemble_method != NoBoost && parm->fsalgorithm != fsMABUSE && parm->isalgorithm != BoostBased && parm->t == 1 && !parm->external_partitioning_approach &&
        !parm->internal_partitioning_approach) {
        printf("WARNING: Ensemble method makes no sense with T = 1 and no partitions, IGNORED\n");
        parm->ensemble_method = NoBoost;
    }

    if (parm->boostisfs_classifier != parm->votes_evaluator_classifier && parm->t > 1) {
        if (parm->verbose > 0)
            printf("Votes evaluator and boosting learner must be the same, CORRECTED\n");
        parm->boostisfs_classifier = parm->votes_evaluator_classifier;
    }

    if (parm->coeff_as_geom_mean && parm->coeff_as_harm_mean)
        printf("WARNING: Geometric mean and harmonic mean cannot be both true\n");

    // Forced configuration
    /* if (!parm->instance_selection)
        parm->isalgorithm = NoIS;

    if (!parm->feature_selection) {
        parm->fsalgorithm = fsNONE;
        parm->random_fs_algorithm = parm->random_fs_algorithm_low_redundancy = false;
    }*/

    if (parm->maximum_training_set_size && parm->feature_selection)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "More than one training set for feature selection");

    if (parm->mlknn_k <= 0)
        OnErrorExit(ERR_PARAM, "ML-kNN k <= 0");
}

