/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "common.h"
//#include "global.h"
#include "extern.i"
#include "macros.h"

#define C_TO_BOOLEAN(c) ((((c) == 't') || ((c) == 'T')) ? true : false)

int LoadInteger(FILE *f, char *label, int def_value, FILE *out_file)
{
    char line[100], new_label[100], *token;
    bool found = false;
    int value;

    // Start at the beginning of the file
    rewind(f);

    strcpy(new_label, label);
    strcat(new_label, ":");

    // Search for label
    if (fgets(line, 100, f) == NULL && !feof(f))
        OnErrorExit(ERR_READ, "LoadInteger");
    while (!feof(f) && !found) {
        if (!strncasecmp(line, new_label, strlen(new_label)))
            found = true;
        else if (fgets(line, 100, f) == NULL && !feof(f))
            OnErrorExit(ERR_READ, "LoadInteger");
    }

    // If found assign value
    if (found) {
        token = strchr(line, ':') + 1;
        value = atoi(token);
        fprintf(out_file, "%s = %d\n", label, value);
    }
    // If not assign default value
    else {
        value = def_value;
        fprintf(out_file, "%s = %d (*)\n", label, value);
    }
    return (value);
}

bool LoadBoolean(FILE *f, char *label, bool def_value, FILE *out_file)
{
    char line[100], new_label[100], *token;
    bool found = false, value;

    // Start at the beginning of the file
    rewind(f);

    strcpy(new_label, label);
    strcat(new_label, ":");


    // Search for label
    if (fgets(line, 100, f) == NULL && !feof(f))
        OnErrorExit(ERR_READ, "LoadBoolean");
    while (!feof(f) && !found) {
        if (!strncasecmp(line, new_label, strlen(new_label)))
            found = true;
        else if (fgets(line, 100, f) == NULL && !feof(f))
            OnErrorExit(ERR_READ, "LoadBoolean");
    }

    // If found assign value
    if (found) {
        token = strchr(line, ':') + 2;
        value = C_TO_BOOLEAN(token[0]);
        fprintf(out_file, "%s = %s\n", label, value ? "true" : "false");
    }
    // If not assign default value
    else {
        value = def_value;
        fprintf(out_file, "%s = %s (*)\n", label, value ? "true" : "false");
    }
    return (value);
}

scalar LoadReal(FILE *f, char *label, scalar def_value, FILE *out_file)
{
    char line[100], new_label[100], *token;
    bool found = false;
    scalar value;

    // Start at the beginning of the file
    rewind(f);

    strcpy(new_label, label);
    strcat(new_label, ":");

    // Search for label
    if (fgets(line, 100, f) == NULL && !feof(f))
        OnErrorExit(ERR_READ, "LoadReal");
    while (!feof(f) && !found) {
        if (!strncasecmp(line, new_label, strlen(new_label)))
            found = true;
        else if (fgets(line, 100, f) == NULL && !feof(f))
            OnErrorExit(ERR_READ, "LoadReal");
    }

    // If found assign value
    if (found) {
        token = strchr(line, ':') + 1;
        value = (scalar) atof(token);
        fprintf(out_file, "%s = %lf\n", label, (double) value);
    }
    // If not assign default value
    else {
        value = def_value;
        fprintf(out_file, "%s = %lf (*)\n", label, (double) value);
    }
    return (value);
}

int LoadEnumerate(FILE *f, char *label, char *values[], int n, int def_value, FILE *out_file)
{
    char line[100], new_label[100], *token;
    bool found = false;
    int value = 0;

    // Start at the beginning of the file
    rewind(f);

    strcpy(new_label, label);
    strcat(new_label, ":");

    // Search for label
    if (fgets(line, 100, f) == NULL && !feof(f))
        OnErrorExit(ERR_READ, "LoadEnumerate");
    while (!feof(f) && !found) {
        if (!strncasecmp(line, new_label, strlen(new_label)))
            found = true;
        else if (fgets(line, 100, f) == NULL && !feof(f))
            OnErrorExit(ERR_READ, "LoadEnumerate");
    }

    // If found assign value
    if (found) {
        token = strchr(line, ':') + 2;

        // Check for syntax error
        if (token == NULL) {
            value = def_value;
            fprintf(out_file, "%s = %s (*)\n", label, values[value]);
            return def_value;
        }

        // Remove initial spaces
        while (token[0] == ' ')
            token++;

        // Remove trailing spaces and new line
        Trim(token);

        bool ok;
        int i;
        for (i = 0, ok = false; i < n && !ok; i++) {
            // if (!strcasecmp(strtok(token, "\n"), values[i]/*, strlen(values[i])*/)) {
            if (!strcasecmp(token, values[i])) {
                value = i;
                ok = true;
            }
        }
        if (!ok)
            value = def_value;

        fprintf(out_file, "%s = %s\n", label, values[value]);
    }
    // If not assign default value
    else {
        value = def_value;
        fprintf(out_file, "%s = %s (*)\n", label, values[value]);
    }

    return value;
}

void LoadString(FILE *f, char *label, char *def_value, char *value, FILE *out_file)
{
    char line[100], new_label[100], *token;
    bool found = false;

    // Start at the beginning of the file
    rewind(f);

    strcpy(new_label, label);
    strcat(new_label, ":");

    // Search for label
    if (fgets(line, 100, f) == NULL && !feof(f))
        OnErrorExit(ERR_READ, "LoadString");
    while (!feof(f) && !found) {
        if (!strncasecmp(line, new_label, strlen(new_label)))
            found = true;
        else if (fgets(line, 100, f) == NULL && !feof(f))
            OnErrorExit(ERR_READ, "LoadString");
    }

    // If found assign value
    if (found) {
        token = strchr(line, '\"') + 1;
        // Check for syntax error
        if (token == NULL) {
            strcpy(value, def_value);
            fprintf(out_file, "%s = \"%s\" (*)\n", label, value);
            return;
        }
        strcpy(value, token);

        // Substitute last " by \0
        token = strchr(value, '\"');
        if (token == NULL) {
            strcpy(value, def_value);
            fprintf(out_file, "%s = \"%s\" (*)\n", label, value);
            return;
        }
        token[0] = '\0';

        fprintf(out_file, "%s = \"%s\"\n", label, value);
    }
    // If not assign default value
    else {
        strcpy(value, def_value);
        fprintf(out_file, "%s = \"%s\" (*)\n", label, value);
    }
}

// Remove training spaces from string
void Trim(char *string)
{
    int i = strlen(string) - 1;

    while ((string[i] == ' ' || string[i] == '\n') && i > 0) {
        string[i] = '\0';
        i--;
    }
}
