/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// Generic macros
#pragma once

#include <malloc.h>
#include <stdlib.h>
#include <alloca.h>
#include <stdbool.h>
#include <unistd.h>
#include <string.h>
#include "common.h"

/* MACROS. */
/* Generic definitions. */
#define YES             1
#define NO              0
#define ZERO            0.000001
#define ZERO2           0.000000000001
#define INFINITE        10.0e20

// Show test messages
#define TEST    false

// Minimum size of memory allocation in bytes
#define MIN_ALLOCATION  10

// Relief algorithm: maximum ((unsigned int) number) of features that allows to use a cache (so there is enough memory)
#define MAX_FEAT_MEM 25

/* Macros*/
#define SHOWPROGRESS(show, format, ...) {               \
    if (show)                         \
        fprintf(stdout, format, ##__VA_ARGS__); \
    fflush(stdout);  \
}

#define SAVEPROGRESS(show, format, ...) {               \
    if (show)                         \
        fprintf(parm->out_file, format, ##__VA_ARGS__); \
    fflush(parm->out_file);  \
}

#define VERBOSE(show, file, format, ...) {               \
    if (show)                         \
        fprintf(file, format, ##__VA_ARGS__); \
    fflush(file);  \
}

#define SHOWTIME(show, file, format, ...) {               \
    if (show)                         \
        fprintf(stderr, format, ##__VA_ARGS__); \
    fflush(stderr);  \
    if (file && show)                         \
        fprintf(file, format, ##__VA_ARGS__); \
    fflush(stderr);  \
}

#define EQUAL(x, y)      (fabs((scalar) x - (scalar) y) < ZERO)

#define RANDOM(a, b)     (((((double) random()-1.0)/(RAND_MAX))*((b)-(a)))+(a))       /* a <= x  < b */

#define MIN(a, b)        (((a)>(b)) ? (b) : (a))
#define MAX(a, b)        (((a)<(b)) ? (b) : (a))

#define SQR(x)           ((x)*(x))

#define MALLOC(pt, type, number, str) {{ \
    if ((((unsigned int) number) != 0)) { \
        if (((pt) = (type *) malloc (((size_t) (MAX(sizeof (type) *((((unsigned int) number))), MIN_ALLOCATION))))) == NULL) \
            OnErrorExit (ERR_MEMORY, (str));} \
        else \
            pt = NULL; \
}}

#define CALLOC(pt, type, number, str) { if (((unsigned int) number)) { \
                                        if (((pt) = (type *) \
                                           calloc (sizeof (type),((unsigned int) number))) \
                                           == NULL) \
                                          OnErrorExit (ERR_MEMORY, (str)); } \
                                        else \
                                          pt = NULL;}

#define REALLOC(pt, type, number, str) {{ \
    if ((((unsigned int) number))) { \
        if (((pt) = (type *) realloc (pt, sizeof (type) *(((unsigned int) number)))) == NULL) \
            OnErrorExit (ERR_MEMORY, (str)); }\
        else \
            pt = NULL; \
}}

#define ALLOCA(pt, type, number, str)   {if (((unsigned int) number)) { \
                                        if (((pt) = (type *) \
                                           alloca (MAX(MIN_ALLOCATION, sizeof (type) *(((unsigned int) number))))) \
                                           == NULL) \
                                          OnErrorExit (ERR_MEMORY, (str)); }\
                                        else \
                                          pt = NULL;}

#define BOUND(m, value, M)              {if (value < (m)) value = (m); else if (value > (M)) value = (M);}

#define ASSERT(pred, err, str)          {if (!(pred)) OnErrorExit (err, str);}

#define ABS(x)                          ((x)>=0 ? (x) : -(x))

// #define FREE(x)                         {if ((x)==NULL) OnErrorExit(ERR_MEMORY, "Attempt to free null pointer"); free (x); (x) = NULL;}
#define FREE(x)                         {if ((x)!=NULL) {free (x); (x) = NULL;} else {OnErrorExit(ERR_MEMORY, "Attempt to free NULL pointer");}}

static inline int WIDTH(int x)
{
    bool power_of_ten;
    double exponent;

    power_of_ten = !modf(log10(fabs((scalar) x)), &exponent);

    int width = (int) ceil(log10((x))) + (power_of_ten ? 1 : 0);

    // Limit [1, 10]
    BOUND(1, width, 10);

    return width;
}

static inline bool ISSKYNET()
{
    char name[100];

    gethostname(name, 100);

    return !strncmp(name, "skynet", 6);
}


static inline void PRINTCHARAFTERSPACES(int spaces, char character, int count, FILE *fp)
{
    for (int i = 0; i < spaces; i++)
        fprintf(fp, " ");
    for (int i = 0; i < count; i++)
        fprintf(fp, "%c", character);
    fprintf(fp, "\n");
}
