/***************************************************************************
 *   Copyright (C) 2017 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FItnESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <math.h>
#include "common.h"
#include "typex.h"

#define ZERO 0.000001

void PrintTwoClassMeasures(scalar tp, scalar tn, scalar fp, scalar fn, FILE *out_file, char *text, t_PerformanceInfo *perf)
{
    perf->specificity = (tn + fp) > ZERO ? tn/(tn + fp) : 0.0;
    perf->sensitivity = (tp + fn) > ZERO ? tp/(tp + fn) : 0.0;
    perf->precision = (tp + fp) > ZERO ? tp/(tp + fp) : 0.0;
    perf->false_neg_rate = (tp+fn > ZERO) ?  fn/(tp+fn) : 0.0;
    perf->ccr = 0.5 * (perf->specificity + perf->sensitivity);
    perf->mcc = ((tp)+(fp))*((tp)+(fn))*((tn)+(fp))*((tn)+(fn)) > 0.0 ? ((tn)*(tp)-(fn)*(fp))/sqrt(((tp)+(fp))*((tp)+(fn))*((tn)+(fp))*((tn)+(fn))) : 0.0;
    perf->f_measure = perf->f1_score = perf->sensitivity+perf->precision > ZERO ? (2.0*perf->sensitivity*perf->precision)/(perf->sensitivity+perf->precision) : 0.0;
    perf->npv = tn+fn > ZERO ? tn/(tn+fn) : 0.0;
    perf->fpr = perf->fall_out = (fp + tn) > ZERO ? fp/(fp + tn) : 0.0;
    perf->ppv = perf->precision;
    perf->fdr = 1.0 - perf->ppv;
    perf->tp_rate = perf->recall = perf->poc = perf->sensitivity;
    perf->tn_rate = perf->specificity;
    perf->tp = tp;
    perf->tn = tn;
    perf->fp = fp;
    perf->fn = fn;
    perf->gmean = pow(perf->sensitivity * perf->specificity, 0.5);
    perf->bac = (perf->specificity+perf->sensitivity)/2.0;
    perf->cc = sqrt((tp+fn)*(tn+fp)*(tp+fp)*(tn+fn)) > 0.0 ? ((tp*tn)-(fn*fp))/sqrt((tp+fn)*(tn+fp)*(tp+fp)*(tn+fn)) : 0.0;
    perf->informedness = perf->tp_rate+perf->specificity-1.0;
    perf->markedness = perf->ppv+perf->npv-1.0;
    perf->acp = ((tp/(tp+fn))+(tp/(tp+fp))+(tn/(tn+fp))+(tn/(tn+fn))) > 0.0 ? 0.25*((tp/(tp+fn))+(tp/(tp+fp))+(tn/(tn+fp))+(tn/(tn+fn))) : 0.0;
    perf->ac = (perf->acp-0.5)*2.0;

    if (text && out_file) {
        // Print tokens
        fprintf(out_file, "@%s TP: %.0lf\n", text, tp);
        fprintf(out_file, "@%s FN: %.0lf\n", text, fn);
        fprintf(out_file, "@%s TN: %.0lf\n", text, tn);
        fprintf(out_file, "@%s FP: %.0lf\n", text, fp);
        // Specificity or tn rate
        fprintf(out_file, "@%s Specificity: %lf\n", text, perf->specificity);
        fprintf(out_file, "@%s TN-rate: %lf\n", text, perf->tn_rate);
        // Sensitivity, true positive rate, recall or probability of detection
        fprintf(out_file, "@%s Sensitivity: %lf\n", text, perf->sensitivity);
        fprintf(out_file, "@%s Recall: %lf\n", text, perf->recall);
        fprintf(out_file, "@%s TP-rate: %lf\n", text, perf->tp_rate);
        fprintf(out_file, "@%s POC: %lf\n", text, perf->poc);
        fprintf(out_file, "@%s G-mean: %lf\n", text, perf->gmean);
        // F measure or F1 score
        fprintf(out_file, "@%s F-measure: %lf\n", text, perf->f_measure);
        fprintf(out_file, "@%s F1-score: %lf\n", text, perf->f1_score);
        // Precision or positive predictive value (PPV)
        fprintf(out_file, "@%s Precision: %lf\n", text, perf->precision);
        fprintf(out_file, "@%s PPV: %lf\n", text, perf->ppv);
        // Negative predictive value (NPV)
        fprintf(out_file, "@%s NPV: %lf\n", text, perf->npv);
        // fp rate or fall out
        fprintf(out_file, "@%s FP-rate (-): %lf\n", text, perf->fpr);
        fprintf(out_file, "@%s Fall-out: %lf\n", text, perf->fall_out);
        fprintf(out_file, "@%s FN-rate (-): %lf\n", text, perf->false_neg_rate);
        // False discovery rate
        fprintf(out_file, "@%s FDR: %lf\n", text, perf->fdr);
        fprintf(out_file, "@%s BAC: %lf\n", text, perf->bac);
        fprintf(out_file, "@%s CCR: %lf\n", text, perf->ccr);
        // Matthews correlation coefficient (MCC)
        fprintf(out_file, "@%s MCC: %lf\n", text, perf->mcc);
        // Corrected classification rate
        fprintf(out_file, "@%s CC: %lf\n", text, perf->cc);
        // Informedness
        fprintf(out_file, "@%s Informedness: %lf\n", text, perf->informedness);
        // Markedness
        fprintf(out_file, "@%s Markedness: %lf\n", text, perf->markedness);
        fprintf(out_file, "@%s ACP: %lf\n", text, perf->acp);
        fprintf(out_file, "@%s AC: %lf\n", text, perf->ac);
        fflush(out_file);
    }
}
