/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <pvm3.h>
#include <unistd.h>
#include <string.h>
#include <stdbool.h>
#include "common.h"
#include "macros.h"

// Skynet nodes and number of cores for PVM scheduling
// Machines: 79 + Skynet
// Cores: 760

int PVMSpawn(int n_tasks, char *task, int *tids, char *args[], int *guardian, char *out_file, int show_progress)
{
    char *machines[] = {"skynetx01", "skynetx02", "skynetx03", "skynetx04", "skynetx05", "skynetx06", "skynetx07", "skynetx08", "skynetx09", "skynetx10", "skynetx11",
                        "skynetx12", "skynetx13", "skynetx14", "skynetx15", "skynetx16",
                        "skynet401", "skynet402", "skynet403", "skynet404", "skynet405", "skynet406", "skynet407", "skynet408", "skynet409", "skynet410", "skynet411",
                        "skynet412", "skynet413", "skynet414", "skynet415", "skynet416",
                        "skynet501", "skynet502", "skynet503", "skynet504", "skynet505", "skynet506", "skynet507", "skynet508", "skynet509", "skynet510", "skynet511",
                        "skynet512", "skynet513", "skynet514", "skynet515", "skynet516",
                        "skynet101", "skynet102", "skynet103", "skynet104", "skynet105", "skynet106", "skynet107", "skynet108", "skynet109", "skynet110", "skynet111",
                        "skynet201", "skynet202", "skynet203", "skynet204", "skynet205", "skynet206",              "skynet208", "skynet209", "skynet210", "skynet211",
                        "skynet301", "skynet302", "skynet303", "skynet304", "skynet305",              "skynet307", "skynet308", "skynet309", "skynet310", "skynet311"
                       };
    int cores[] = {16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
                   8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
                   8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
                   8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
                   8, 8, 8, 8, 8, 8,    8, 8, 8, 8,
                   8, 8, 8, 8, 8,    8, 8, 8, 8, 8
                  };

    static int next_machine = 0, next_core = 0;
    char hostname[1000];
    int numt, n_machines = 79;

    // Get hostname, if not in skynet spawn all tasks at once
    gethostname(hostname, 1000);
    if (strcmp(hostname, "skynet.uco.es")) {
        numt = pvm_spawn(task, args, PvmTaskDefault, NULL, n_tasks, tids);
        // numt = pvm_spawn(task, args, PvmTaskDefault+PvmTaskDebug, NULL, n_slaves, tids);

        if (numt != n_tasks) {
            SHOWPROGRESS(show_progress > 0, "%d should slaves should be started, only %d actually done\n", n_tasks, numt);
            OnErrorExit(ERR_PVM, "Processes not spawned correctly");
        }
    } else {
        // Send process in a Round-Robin fashion to Skynet nodes
        numt = 0;
        for (int i = 0; i < n_tasks; i++) {
            // Check running hosts

            bool one_alive = false;
            for (int j = 0; j < n_machines && !one_alive; j++) {
                if (pvm_mstat(machines[next_machine]) == PvmOk)
                    one_alive = true;
                else {
                    SHOWPROGRESS(show_progress > 0, "\nHost %s down\n", machines[next_machine]);
                    next_machine = (next_machine+1) % n_machines;
                }
            }

            if (!one_alive)
                OnErrorExit(ERR_PVM, "All hosts down");

            SHOWPROGRESS(show_progress > 0, "Task: %*d/%d\tMachine: %s\r", WIDTH(n_tasks), i+1, n_tasks, machines[next_machine]);

            int ok = pvm_spawn(task, args, PvmTaskHost, machines[next_machine], 1, &tids[i]);

            if (ok != 1) {
                SHOWPROGRESS(show_progress > 0, "Slave %d/%d '%s' not started in node %s\n", i+1, n_tasks, task, machines[next_machine]);
                OnErrorExit(ERR_PVM, "Processes not spawned correctly");
            }

            // Next core or machine
            numt++;
            next_core++;
            if (next_core == cores[next_machine]) {
                next_core = 0;
                next_machine = (next_machine+1) % n_machines;
            }
        }

        SHOWPROGRESS(show_progress > 0, "%c", '\n');
    }


    // Start guardian and send info
    if (guardian)
        *guardian = InitGuardian(numt, tids, task, out_file);

    return numt;
}
