 /***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#pragma once

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <stdbool.h>
#include "abandoned.h"


typedef unsigned char byte;

// Type for neighbors cache
typedef unsigned int cache_t;

// Scalar can be {double, float}
#define SCALAR_IS_DOUBLE
// #define SCALAR_IS_FLOAT

// For PVM
#ifdef SCALAR_IS_FLOAT
typedef float scalar;
#define pvm_pkscalar    pvm_pkfloat
#define pvm_upkscalar   pvm_upkfloat
#define SCALAR_FORMAT   "%f"
#else
typedef double scalar;
#define pvm_pkscalar    pvm_pkdouble
#define pvm_upkscalar   pvm_upkdouble
#define SCALAR_FORMAT   "%lf"
#endif

#define MAX_FILENAME_LENGTH 4096
#define MAX_LENGTH 31268
#define FILENAME_LENGTH 8192

#define MAXIMUM_CACHE_SIZE      256.0 // In gigabytes

// Structure for ordering the individuals
typedef struct {
    int being;
    scalar fitness;
} RANK;

// Methods for obtaining the intervals for NBC
typedef enum {
    EWD, EFD, PKID, WPKID, FANDI
} t_NBCInterval;

typedef struct {
    int origin, destination;
    scalar value;
} t_Edge;

// Structure for sorting features by their weight
typedef struct {
    scalar weight;
    scalar w;
    int index;
    int ranking;
} t_Feat;

// Classifier models
#define N_CLASSIFIERS   27
#define N_BASE_CLASSIFIERS 17
typedef enum {
    SVM = 0, kNN = 1, C45 = 2, ELM = 3, MLP = 4, NBC = 5, DT_CA = 6, IBT = 7, RF = 8, NN = 9, GNBC = 10, DT_CART = 11, DT_C45 = 12, DT_ID3 = 13, LR = 14, RForest = 15, ExtraTree = 16,
    // Non-base classifiers
    OVO = 17, AdaBoostM1 = 18, MadaBoost = 19, MultiBoost = 20, Bagging = 21, GentleBoost = 22, ReweightBoost = 23, FloatBoost = 24, FeatureBoost = 25, ARCING = 26
} t_Classifier;

// Pending: RealAdaBoost, DiscreteAdaBoost, RandomSubsampling, BalanceCascade, EasyEnsemble, AsymmetricAdaBoost, AdaCost, AdaC1, AdaC2, AdaC3, CSB0, CSB1, CSB2, BoostingProj,
//          RandomProj, AdaBoostMH, FreundNN, RotationForest, LearningFromBites, FilterBoost

// Need a classifier that takes weights: Wagging

// External classifiers for evaluation only
#define N_EXTERNAL  15
typedef enum {
    NaiveBayes = 0, DecisionTree = 1, RandomForest = 2, XGBoost = 3, PLSDA = 4,  DeepLearning = 5, SVC = 6, LDA = 7, QDA = 8, GPML = 9, Ridge = 10, ERT = 11, KRR = 12, AdaBoostR= 13, AdaBoostD = 14
} t_ExternalClassifier;

// Evaluators
typedef enum {
    eSVM, ekNN, eC45, eIncM
} t_Evaluator;

// Structure for ordering
typedef struct {
    int index;
    scalar Hi, SUic;
} t_EntropyInfo;

typedef struct {
    int feature;
    scalar score;
} t_RankInfo;

typedef enum {
    imAccuracy = 0, imGini = 1 /* CART */, imGainRatio = 2 /* C4.5 */, imInformationGain = 3 /* ID3 */    // Information gain is based on entropy
} t_DTSplitMeasure;

// Instance selection algorithms
typedef enum {
    NoIS, DROP2, DROP3, ICF, CHC, MSS, RNN, IB3, RANDOMSELECTION, PSRCG, RC45, CFilter, PSA, CNN, CNNTL, OSS, SMOTE, ENTROPY, Undersampling, GCNN, BoostBased, RMHC, HMNEI, RNGE, LVQPRU,
    SimulatedAnnealing, kNNCHC, MCS
} t_ISAlgorithm;
#define RANDOMIS        7       // Algorithms to consider for random selection

// Feature selection algorithms
typedef enum {
    fsNONE, fsGA, fsLVF, fsFOCUS, fsRELIEF, fsRFE, fsCLASSIFIER, fsRANDOM, fsABB, fsBB, fsFCBF, fsFOCUSSF, fsFAST, fsCFSSF, fsSetCover, fsQBB, fsSFFS, fsSBFS, fsMICIBM, fsCHC, fsRanking, fsUFS,
    fsKSCBF, fsMABUSE, fsGFlip, fsBMIFS, fswMI, fsiSimba, fsmRMR
} t_FSAlgorithm;

typedef enum {
    frCUB, frFisher, frIG, frRFE, frMI, frCHISQ, frSE, frDSE, frReliefF, frSVDEi, frEDSE, frGR, frFScore, frtTest, frWilcoxon, frSOMIFS, frSimba, friSimba
} t_FRAlgorithm;

// Instance+feature selection algorithms
typedef enum {
     NoISFS, CHCGA, IMOEA, ISplusFS, FSplusIS, SIFS2, RMHC_PF1
} t_ISFSAlgorithm;

// Partition methods for scaling
typedef enum {
    pRANDOM, pBOARD, pEQUAL, pRPROTOS, pOneInput, pGrandTour, p2DGrandTour, pkMeans, pkClustering, pDecoupled, peBagging
} t_Partition;

// Ensembling/Boosting feature selection
typedef enum {
    NoBoost, AdaBoost, RealAdaBoost, DiscreteAdaBoost, Wagging, eBagging, RandomSubsampling, eGentleBoost, eMadaBoost, eReweightBoost, eMultiBoost, eFloatBoost, BalanceCascade, EasyEnsemble,
    AsymmetricAdaBoost, AdaCost, AdaC1, AdaC2, AdaC3, CSB0, CSB1, CSB2, Arcing, BoostingProj, RandomProj, AdaBoostMH, eFeatureBoost, FreundNN, RotationForest, LearningFromBites, FilterBoost
} t_EnsembleMethod;

// Performance measures
typedef enum {
    pfAccuracy, pfGMean, pfKappa, pfauROC, pfauPRC, pfRecallMacro, pfRecallMacroGMean, pfMargin, pfAverageAccuracy, pfPrecisionMacro, pfF1Macro, pfCC, pfF1
} t_PerformanceMeasure;

#define N_FEATURE_SUBSET_COMBINATION    4
typedef enum {
    Thresholding, Union, Intersection, Majority
} t_FeatureSubsetCombination;

#define N_FEATURE_SUBSET_EVALUATION 7
typedef enum {
    ClassModel, Mfdr, Voor, Mfe, Cf, Loor, Riic
} t_FeatureSubsetEvaluation;

typedef struct {
    scalar average_accuracy, error_rate, precision_micro, recall_micro, f1_score_micro, precision_macro, recall_macro, f1_score_macro, accuracy, recall_macro_g_mean, kappa, cen;

    // Two classes
    scalar specificity, sensitivity, precision, false_neg_rate, ccr, mcc, recall, tp_rate, tn_rate, poc, f_measure, f1_score, ppv, npv, fpr, fall_out, fdr, acp, tp, tn, fp, fn, gmean, bac, cc,
        informedness, markedness, ac, auroc, auprc;
} t_PerformanceInfo;

typedef enum {
    LOO, kFOLDCV, RESUBS, BS, E632, HOLDOUT, ArtificialData
} t_AccuracyEstimationMethod;

typedef enum {
    WithReplacement, WithoutReplacement, Subsample, CISubsample
} t_BoostingResamplingMethod;

// Methods for weighting classifiers (in principle only used for AdaBoost)
typedef enum {
    StdAda, Constant
} t_ClassWeights;

// Combination of networks
typedef enum {
    Sum, Voting
} t_Combination;

typedef enum {
    TrainingStratum, RandomStratum, WholeDataset
} t_DemoClassifierWeight;

// Discretizers
typedef enum {
    dFAYYAD, dWPKID, dNONE
} t_Discretizer;

typedef enum {
    Resampling, Reweighting, ResamplingWithoutReplacement
} t_EnsembleImplementation;

// Evolution scheme for the evolutionary approach
typedef enum {
    schemeCHCGA, schemeSSGA, schemeSSMA, schemeGGA
} t_EvolutionaryScheme;

// Types of heuristic search
typedef enum {
    SFS, SBS, LRS, SFFS, SFBS
} t_HeuristicSearch;

// Methods for dealing with imbalance two-class problems
typedef enum {
    iNone, iOversampling, iUndersampling, iSMOTE, iEvoSMOTE, iROSE, iADASYN, iBorderlineSMOTE, iKMeansSMOTE, iSVMSMOTE, iSMOTETomek, iSMOTEENN, iIHT, iNCR, iNearMiss, iSMOTEVariant
} t_ImbalanceMethod;
#define N_CLASS_IMBALANCE_METHODS   16

typedef enum {
    vSMOTE, vSMOTE_TomekLinks, vSMOTE_ENN, vBorderline_SMOTE1, vBorderline_SMOTE2, vADASYN, vAHC, vLLE_SMOTE, vdistance_SMOTE, vSMMO, vpolynom_fit_SMOTE, vStefanowski, vADOMS,
    vSafe_Level_SMOTE, vMSMOTE, vDE_oversampling, vSMOBD, vSUNDO, vMSYN, vSVM_balance, vTRIM_SMOTE, vSMOTE_RSB, vProWSyn, vSL_graph_SMOTE, vNRSBoundary_SMOTE, vLVQ_SMOTE,
    vSOI_CJ, vROSE, vSMOTE_OUT, vSMOTE_Cosine, vSelected_SMOTE, vLN_SMOTE, vMWMOTE, vPDFOS, vIPADE_ID, vRWO_sampling, vNEATER, vDEAGO, vGazzah, vMCT, vADG, vSMOTE_IPF,
    vKernelADASYN, vMOT2LD, vV_SYNTH, vOUPS, vSMOTE_D, vSMOTE_PSO, vCURE_SMOTE, vSOMO, vISOMAP_Hybrid, vCE_SMOTE, vEdge_Det_SMOTE, vCBSO, vE_SMOTE, vDBSMOTE, vASMOBD,
    vAssembled_SMOTE, vSDSMOTE, vDSMOTE, vG_SMOTE, vNT_SMOTE, vLee, vSPY, vSMOTE_PSOBAT, vMDO, vRandom_SMOTE, vISMOTE, vVIS_RST, vGASMOTE, vA_SUWO, vSMOTE_FRST_2T,
    vAND_SMOTE, vNRAS, vAMSCO, vSSO, vNDO_sampling, vDSRBF, vGaussian_SMOTE, vkmeans_SMOTE, vSupervised_SMOTE, vSN_SMOTE, vCCR, vANS, vcluster_SMOTEv
} t_SMOTEVariant;

// Margin calculation for k-NN
typedef enum {
    VotesDifference, AverageNeighborhoodMargin, NormalizedVotesDifference, SmoothedVotesDifference, NormalizedAverageNeighborhoodMargin, OrderedVotesDifference,
    NormalizedOrderedVotesDifference, NearestNeighbourEnemyDistance
} t_kNNMargin;

// For generating next NLP projection we can consider the previous network
// or the whole ensemble
typedef enum {
    Ensemble, Classifier, ClassMargin, EnsMargin, ClassiferAndNN, AdaWeights
} t_NLPPrevious;

typedef enum {
    FirstMinority, FirstMajority, BothSimultenously
} t_OligoThresholdEvaluation;

typedef enum {
    Nopreprocess, Scale, Typify
} t_Preprocess;

typedef enum {
    BPStd, EDBD, RProp, TryAll, LLE, NDA, HDA, RCGA, NWFE, EDA, NLEDA, SCCA, SISOMAP, SCCA_GRNN, PCA
} t_ProjAlgorithm;
#define N_BP_ALGORITHMS 3

typedef enum {
    Gabriel, RelativeNeighborhood
} t_ProximityGraph;

typedef enum {
    Simple, Friedman, Kuo
} t_Regularization;

// Subspace method
typedef enum {
    AllInputs, RSM, ConservativeBoosted, Aggressive1Boosted, Aggressive2Boosted, VeryAggressiveBoosted, Lazarevic, Adamenn
} t_SubspaceMethod;

// Transfer function
typedef enum {
    Log, HypTan, Sin, Linear
} t_Transfer;

// Sampling methods
typedef enum {
    sNone, sKMeans, sStratified
} t_SamplingMethod;

typedef enum {
    umRandom, umClusterBased
} t_UndersamplingMethod;

// For boosting instance selection
typedef enum {
    ICPOS, ICPOSNEG, CCMAJ, CCONCE
} t_VotesRecordingMethod;

/////////////////////////
// Multi-label methods //
/////////////////////////
#define N_ML_CLASSIFIERS    62
#define N_ML_BASE_CLASSIFIERS 57
typedef enum {
    MLkNN = 0, BR = 1, CC = 2, LAMLkNN = 3, MLksNN = 4, LAMLksNN = 5, RAkEL_o = 6, RAkEL_d = 7, LP = 8, CLR = 9, HOMER_B = 10, HOMER_K = 11, HOMER_R = 12, MLDT = 13, MLBASICBP = 14, BPMLL = 15,
    MLRF = 16, DBR = 17, STA = 18, NS = 19, ECC = 20, IBLRML = 21, IBLRMLplus = 22, PCT = 23, EPCT = 24, RFPCT = 25, TSVM = 26, TSCCM = 27, TSPCCM = 28, MLTREE = 29, MLFOREST = 30, MLRS = 31,
    MLRSLC = 32, OPCT = 33, SMLKNN = 34, DMLKNN = 35, MrKNN = 36, DWKNND = 37, DWKNNM = 38, DWKNNZ = 39, FRONEC1H = 40, FRONEC2H = 41, FRONEC3H = 42, FRONEC1LD = 43, FRONEC2LD = 44, FRONEC3LD = 45,
    FVkNN = 46, NSSC = 47, EDT = 48, EXTRAPCT = 49, PRUNEDSET = 50, EPS = 51, PPT = 52, PPTN = 53, PPTC = 54, EXTRAMLDT = 55, ZEROML = 56,
    // Ensembles
    BAGGING = 57, REALADABOOSTRE = 58, DISCRETEADABOOSTRE = 59, REALADABOOSTMH = 60, DISCRETEADABOOSTMH = 61
} t_MLClassifier;

#define N_ML_EXT_CLASSIFIERS    29
typedef enum {
    LabelPowerset = 0, ClassifierChain = 1, RAkELd = 2, RAkELo = 3, MLARAM = 4, BinaryRelevance = 5, ExMLkNN = 6, BRkNNa = 7, BRkNNb = 8, MLTSVM = 9, MLMLP = 10, DL = 11, BOOMER = 12, BCC = 13,
    CDN = 14, HASEL = 15, PSt = 16, CT = 17, MCC = 18, NSR = 19, SCC = 20, PS = 21, LC = 22, CDT = 23, DBPNN = 24, FW = 25, Maniac = 26, PCC = 27, PMCC = 28
} t_MLExternalClassifier;

#define N_ML_EXT_BASE_CLASSIFIERS    9
typedef enum {
    LogisticRegression = 0, KNeighborsClassifier = 1, RandomForestClassifier = 2, DecisionTreeClassifier = 3, GaussianNB = 4, SVMLinear = 5, SVMGaussian = 6, AdaBoostClassifier = 7,
    MLPerceptron = 8,
    //  Methods not considered:   QuadraticDiscriminantAnalysis
} t_MLExternalBaseClassifier;

typedef enum {
    And, Or, MajorityVoting, QuorumSensing
} t_MergeMethod;

#define N_ML_INSTANCE_SELECTION 5
typedef enum {
    NoInstanceSelection, CooperativeCoevolution, CHCIS, LAMLkNNIS, DEIS
} t_MLInstanceSelection;

#define N_ML_FEATURE_SELECTION 2
typedef enum {
    NoFeatureSelection, CHCFS
} t_MLFeatureSelection;

typedef enum {
    NoDataPartition, RandomDataPartition, Clustering
} t_MLDataPartitionMethod;

typedef enum {
    Evolutionary, SingleLabel, GreedyPerLabel, PerpendicularSearch, DifferentialEvolution, LAMLksNNEvolutionary
} t_MLKPerLabelMethod;

typedef enum {
    MLNone, MLeBagging, MLAdaBoostMH
} t_MLEnsembleMethod;

// Methods for dealing with imbalance ML problems
#define N_ML_CI_METHOD  11
typedef enum {
    MLNoneS, MLRandomOS, MLSmote, MLESO, MLRandomUS, PRUS, PROS, PRUOS, PROUS, PEVUS, REPOPLABELS
} t_MLImbalanceMethod;

typedef enum {
    ResamplingPerLabel, ResamplingPerInstance, ResamplingPartial
} t_MLResampling;

typedef enum {
    A, B
} t_PS_Strategy;

// Methods for propagating labels in ML-ESO
#define N_LABEL_PROPAGATION 6
typedef enum {
    EVOL, AND, OR, COPY, MINORITY, MINORITYOR
} t_LabelPropagation;

// Types common to different programs
typedef struct {
    int ensemble_size, nda_k, demo_max_features, cache_size, external_subset_size, localk_min_k, kscbf_k, populate_number_of_neighbors, random_sampling_for_threshold_evaluation_max_size, isimba_T,
        random_sampling_for_threshold_evaluation_min_size, plus_l, is_threshold_step, minus_r, seed, smote_N, scca_proj_iterations, max_evaluations, demo_eval_rounds, populate_copies_per_instance,
        max_n_minority_thresholds, cv_folds, sampling_size, max_inputs, Npc, lvf_iter, relief_k, micibm_k, fasta_seq_length, generations, easy_si, localk_repetitions, isfs_n_bites, maximum_threads,
        projection_subspace_minimum_size, scca_iterations, inputs, ex_un_s_i, duplicate_features_ratio, projection_subspaces, rmhc_m, leave_no_neighbors_unchanged,
        n_subpops, verbose, max_k, k_t, add_k_neighbors, positive_class, update_w_n_iterarions, individuals, multiboost_reinit_iterations, localk_lambda, is_epochs, kmeans_max_iterations, smote_k,
        k_bits, minimum_instances, n_thresholds, nsp, internal_subspace_size, min_iterations, m, max_isel_thresholds, nlp_epochs, boost_based_filter_noise_k, t, eval_folds, chc_max_reinit_attempts,
        imoea_seconds, elm_n_hidden, evaluation_subset_size, reweightboost_r, subset_size, seq_offset, max_features_for_focus, maximum_number_of_thresholds, epoch_size, rmhc_fp1_iterations,
        max_fsel_thresholds, max_dimension, projection_subspace_size, presample_subset_size, np, k, localk_bs_n, seq_length, n_epochs, min_k, nodes, instances_per_subpop, max_n_thresholds, lr_epochs,
        localk_nfolds, knn_folds, dimension, mrmr_n, stratum_size, easy_t, floating_attempts, second_chances, localk_max_k, min_subset_size, external_subspace_size, evosmote_noise_mutation_k,
        subpop_size, rmhc_iterations, max_nodes, subspace_size, eval_subset_size, select_with_sorted_graph,  th_eval_n_iterarions, n_electors, chc_size, maximum_training_set_size, max_instances, knn,
        fasta_seq_offset, grnn_nodes, rf_ntree, dt_max_depth, internal_subset_size, fs_threshold_step, demo_max_strata, nlp_attempts, svm_type, kernel, svm_d, svm_hktype, svm_offset, eda_mu,
        eda_lambda, eda_rho, eda_cycles_max, lfb_subset_size, lfb_bite_size, lfb_iterations, fb_mt, min_post_k, test_1nn_slaves, gflip_max_iterations, simba_T, mabuse_max_instances, show_progress,
        mabuse_feat_per_iteration, mabuse_ensemble_method, dt_min_samples_split, boostedis_ensemble_method, sample_using_kmeans_subset_size, mabuse_max_losing_streak, mabuse_feature_life, ps_b, ps_p,
        mabuse_folds, NBC_intervals, cv_sampling_size, training_max_size, ibt_multiboost_I, max_ifsel_thresholds, bbis_instances_per_iteration, sifs2_sampling_size, mabuse_sort_by_weighted_mi_limit,
        sifs2_sampling_size_for_performance, sifs2_stratum_size, sifs2_add_group_size, evosmote_min_N, evosmote_max_N, evosmote_wrapper_eval_interval, best_individual_fitness_improvement_generations,
        evosmote_use_majority_threshold, reinitalize_after_gs_no_improvement, rakel_k, rakel_m, homer_k, bkmeans_iter, chc_generations;

    scalar boostfs_w_kap, b, C_minus, ga_alpha, ufs_R2max, a_output, eta_max, asymptote, performance_w, de_cr, angle_step, addF_mutation, de_local_search_is_probability, flat_spot, reduction_w,
        rotation_class_selec_prob, val_percentage, alpha_acc, fred_w, sampling_ratio, gamma_alpha, localk_mutation, eta0, projection_a_output, rnn_isfs_mutation, localk_memetic_mutation_ratio,
        scca_precision, cv_val_percent, bit_mutation, hda_lambda, rfe_mutation, svm_c, drop3_mutation, asymmetric_adaboost_k, threshold, minority_is_ratio, majority_fs_ratio, max_ratio, isimba_delta,
        ls_factor_old_individuals, theta, maximum_to_train, rnnstd_mutation, cub_lambda, a, subspace_prob, sigma, boostfs_w_auroc, lfb_p, presample_ratio, phi_alpha, lle_max_dimension,
        focus_mutation, minority_fs_ratio, boostrsm_dim, subsampling_ratio, boost_based_is_mutation, init_weight, localk_memetic_bit_mutation_ratio, projection_b_output, rsm_setprob, gamma_eta,
        wmi_epsilon, projection_asymptote, min_ratio, val_percent, relieff_threshold, localk_holdout_percent, rnn_is_mutation, selection_probability, ired_w, alpha_max, ib3_mutation, is_mutation,
        max_adjacency_matrix_size, isfs_cache_size, random_sampling_ratio, fitness_w, projection_eta_hidden, percent_error_estimation, kappa_eta, mutation_prob, elite, boostrsm_overlap, lr_alpha,
        init_fs_percent, percent_for_early_stopping, alpha, de_undersample_mut_prob, delta_max, ok_w, ratio, fb_delta, kscbf_alpha, randomsel_ratio, majority_is_ratio, projection_b, npv_w,
        oversampling_ratio, validation_percentage, eta_minus, nlp_bias_input, projection_a, de_local_search_k_probability, scale_max, gamma, init_setprob, init_percent, fcbf_delta, fast_theta,
        kappa_alpha, alpha0, gga_crossover_prob, lle_min_dimension, instance_weight_limit, alpha_ok, addI_mutation, de_np, random_addition_percent, hmnei_epsilon, tolerance, subs_w, minimum_to_train,
        *c45_weights, mutation, gga_mutation, localk_alpha_acc, local_perf_w, boostfs_w_auprc, boostfs_w_red, phi_eta, C_plus, addiorf_mutation, alpha_red, relief_threshold, icf_mutation, lle_alpha,
        eta_plus, projection_alpha_hidden, scale_min, gga_elitism, de_f, time_all, localk_mutation_with_no_selection, nlp_lambda, fb_epsilon, lambda, rho, diff_w, precision_for_zero, bpmll_nodes,
        random_sampling_for_threshold_evaluation_ratio, one_probability, projection_limit_weight, isimba_lambda, rnn_fs_mutation, hda_nu, dimension_scale, boostfs_w_acc, eda_sigma, svm_r,
        alpha_redfeat, subpop_red_w, alpha_redins, ufs_epsilon, min_votes, boostfs_w_gme, imoea_mutation, boostrsm_acc, delta_min, mss_mutation, relief_alpha, localk_alpha_reg, eta, rnn_mutation,
        delta0, subpop_mutation, b_output, localk_alpha_mar, random_undersampling_ratio, elm_a, elm_b, rf_percent, mabuse_filter_weighted_sampling_ratio, mabuse_ranking_initial_selection_percentage,
        mabuse_min_selected_percent, rank_w, biased_hux_ratio, evosmote_undersamplig_mutation_rate, evosmote_oss_mutation_rate, evosmote_cnntl_mutation_rate, average_accuracy_w, balance_w,
        recall_macro_g_mean_w, recall_macro_w, f1_w, gmean_w, kappa_w, accuracy_w, added_w, cc_w, recall_w, specificity_w, precision_w, smote_undersampling_ratio, reinitalize_percent_addition,
        evosmote_noise_mutation_rate, evosmote_one_w, auroc_w, three_way_minus1_prob, bpmll_alpha, chc_random_evaluation_ratio;

    bool LOO_for_knn_performance, adapt_ratio, adaptive_distance_measure, alpha_boost_set_to_one, bagging_projection, balance_laststep, bias, boost_ci_initialization, boost_is, boostis_considerlist,
        br_adaboost_resampling, ci, class_perf_as_percentage, cnn_ci, coeff_as_geom_mean, coeff_as_harm_mean, comp_clas, comp_votes, cross_validate_dimension, fs_use_cache, fitness_as_ratio,
        cv_parameters, cv_with_string_kernel, de_evolve_threshold, de_separate_thresholds, decouple, demo_eval_exp_redfeature, demo_eval_exp_redinstance, evaluate_margin,
        democratic_evaluation_of_threshold, democratic_evaluation_of_threshold_with_cv, duplicate_features, duplicate_instances, duplicate_instances_with_noise, dynamic_threshold, early_stopping,
        ebus_gs_gm, eda_use_only_missclass, eda_wrapper, edited_knn, eval_exp_redfeature, eval_exp_redinstance, evaluate_oligois_votes_with_auc, evaluate_standard_knn, evaluate_train_error,
        evaluate_undersampling, evaluate_with_loo, evolutionary_local_k, evolutionary_local_k_and_is, evolve_k, evolve_subspace, exp_redfeature, exp_redinstance, exp_reduction, print_dt,
        external_stratified_approach, fasta_file, weight_votes, feature_partition, cv_k_for_testing, feature_selection, filter_noise, fixed_number_of_thresholds, force_balanced, force_undersampled,
        fusion_from_first, fusion_subsets, get_best_round, get_threshold_with_all, grnn_cluster, huge_dataset, ind_filter, initially_marked, instance_partition, instance_selection, ibt_no_stop,
        internal_partitioning_approach, internal_stratified_approach, is_use_cache, isfs_inputs_first, isfs_use_cache, isfs_use_fitness, isfsmutation_random_order, keep_class_distribution,
        knn_weight_by_distance, knn_with_weighted_votes, laststep_ebus_gs_gm, localk_no_neigbors_assign_nearest_value, localk_populate, cv_hyperparameters, localk_reevaluate, ls_only_use_selected,
        multiis_adjust_initial_percentage, multiis_and_k_with_DE, multiselect_only_minority_class, multiselection, nlp_project_from_previous, noise_edition, obtain_auc, only_accept_if_better,
        only_save_selection, optimal_local_k, output_2class_measures, oversampling, perform_last_iteration, minimum_info, perform_laststep_is, populate, post_validate_k, pre_undersampling, presample,
        presample_stratified, print_verbose, projection, projection_bias, projection_disjoint_subspaces, prune_tree, random_fs_algorithm, random_fs_algorithm_fast, random_fs_algorithm_low_redundancy,
        random_is_algo, random_isalgorithm, random_sampling, random_sampling_for_threshold_evaluation, random_undersampling, random_undersampling_for_training, randomize_inputs_for_fs,
        randomsel_withreplacement, rcga_random_projection, rcga_undersample, rcga_use_accuracy, rcga_use_auc_measure, reinit_distance0, remove_constant_features, remove_features_after_selection,
        resample, rprop_undo, sampling, save_graph, save_kappa_error_data, save_margins, save_outputs, save_performance, save_projs, save_selection, scale_projection, scca_already_projected,
        select_fs_randomly, selection_minority, separate_inputs, show_average_nn_margin, soften_thresholds, sparse_data, sparse_format, stratified, sifs2_convergence_with_only_acc, use_ranger_for_rf,
        subsampling_for_evaluation, symmetric_knn, typify, undersample_laststep, undersampling, use_all_data, keep_files, use_averaged_accuracy_for_stop, use_brent, sample_using_kmeans,
        use_edge_list_for_graph, use_g_mean, use_g_measure, use_grnn, use_pcomponents, use_validation_data_for_votes_threshold, validate_k, validation_until_end, weigthed_sampling_for_training,
        with_replacement, mabuse_roulette_selection, cv_hyperparameters_for_testing, mabuse_condit_excl_with_margin, cv_external_hyperparameters_for_testing, sifs2_best_with_only_acc, gnbc_use_logs,
        external_partitioning_approach, mabuse_use_rankings_for_perf, evaluate_with[N_CLASSIFIERS], evaluate_with_external[N_EXTERNAL], ibt_do_not_use_internal_nodes, dt_grow_full_tree,
        ibt_fixed_node_weight, ibt_predict_with_probabilities, force_sparse_format, sifs2_random_is_alg, ml_cc_random_order, three_way_is,
        isfs_stepwise_cache, bbis_select_best_instance_from_every_class, sifs2_when_stratified_evaluate_with_training_set, sifs2_stratified_incremental_selection, sifs2_is_additive,
        evosmote_eval_new_instances, multiselection_as_local_k, evosmote_only_minority, evosmote_roulette_initialization, mldt_adaptive_threshold, feature_ranking;

    FILE *out_file, *comp_file1, *comp_file2;

    short TRIALS;

    time_t time_limit;

    char datad[MAX_FILENAME_LENGTH], performance_file[MAX_FILENAME_LENGTH], k_values_for_local_k[MAX_LENGTH], additional_datad[MAX_FILENAME_LENGTH];

    t_AccuracyEstimationMethod localk_error_estimation_method;
    t_BoostingResamplingMethod boosting_resampling_method;
    t_Classifier boostfs_classifier, chc_wrapper, votes_evaluator, boostis_classifier, votes_evaluator_classifier, boostisfs_classifier, classifier, base_learner, mabuse_classifier, sifs2_evaluator,
        wrapper_fs_classifier, mrmr_evaluator;
    t_ClassWeights class_weight;
    t_Combination combination;
    t_DemoClassifierWeight class_weight_method;
    t_Discretizer discretizer;
    t_EnsembleImplementation ensemble_implementation, ensemble_technique;
    t_EnsembleMethod ensemble_method, ibt_ens_method, ibt_alpha_method;
    t_Evaluator evaluator;
    t_EvolutionaryScheme evolutionary_scheme;
    t_FRAlgorithm fralgorithm;
    t_FSAlgorithm fsalgorithm;
    t_FeatureSubsetEvaluation feature_subset_evaluation;
    t_HeuristicSearch search_heuristic, mrmr_search;
    t_ImbalanceMethod imbalance_method;
    t_ISAlgorithm isalgorithm, laststep_isalgorithm;
    t_ISFSAlgorithm isfsalgorithm;
    t_kNNMargin margin_method;
    t_NLPPrevious nlp_previous;
    t_OligoThresholdEvaluation oligo_threshold_evaluation;
    t_Partition internal_partitioning_method, external_partitioning_method, partition;
    t_PerformanceMeasure performance_measure, votes_threshold_measure, mabuse_performance_measure, cv_performance_measure;
    t_Preprocess scale_inputs;
    t_ProjAlgorithm projection_algorithm;
    t_ProximityGraph rnge_pg;
    t_Regularization projection_regularization;
    t_SubspaceMethod subspace_method;
    t_Transfer transfer_h, transfer_o, projection_transfer_o, projection_transfer_h, projection_transfer, elm_tf;
    t_UndersamplingMethod undersampling_method;
    t_VotesRecordingMethod votes_recording_method;
    t_NBCInterval NBC_discretizer;
    t_SamplingMethod cv_sampling_method;
    t_DTSplitMeasure ibt_split_measure, rforest_split_measure;
    t_MLExternalClassifier evolution_external_classifier;
    t_SMOTEVariant smote_variant;
    t_ExternalClassifier external_base;
    t_FeatureSubsetCombination feature_subset_combination;
    t_FeatureSubsetEvaluation rank_to_fs_threshold_evaluation;
    t_PS_Strategy ps_strategy;
    t_LabelPropagation label_propagation;

    // Multi-label methods
    int mlknn_k, population_size, chc_subpop_minimum_distance_factor, ml_min_k, ml_max_k, mlknn_k_per_label_stepwise_attempts, mlksnn_min_k, mlksnn_max_k, mlksnn_rounds, mlksnn_window_size, cdn_I,
        lamlknn_n_clusters, lamlknn_min_n_clusters, lamlknn_max_n_clusters, ml_external_eval_interval, ml_max_ioversampling_ratio, mlmlp_hidden_nodes, ecc_T, iblr_k, mlforest_ntrees, optiondt_O,
        optiondt_l, dmlknn_delta, ml_ensemble_size, eps_size, ps_labels_bound, pptc_k, smlknn_k, homer_max_depth, cdn_Ic, nsr_p, nsr_n;

    scalar mlknn_s, chc_subpop_init_prob, ssga_mutation, subpop_fitness_w, subpop_isolated_w, subpop_difference_w, subpop_substitution_w, subpop_reduction_w, subpop_onelabel_w, ml_k_penalty_from_cv,
        eb_subset_accuracy_w, eb_hamming_loss_w, eb_accuracy_w, eb_precision_w, eb_recall_w, eb_f1_w, lb_accuracy_macro_w, lb_precision_macro_w, lb_recall_macro_w, lb_f1_macro_w, mlksnn_ls_mut,
        lb_accuracy_micro_w, lb_precision_micro_w, lb_recall_micro_w, lb_f1_micro_w, eb_one_error_w, eb_coverage_w, eb_ranking_loss_w, eb_average_precision_w, lb_auc_macro_w, lb_auc_micro_w,
        ml_rus_p, chc_init_prob, reduction_ml_w, chc_pop_init_prob, quorum_sensing_threshold_max, quorum_sensing_threshold_min, eb_auc_w, eb_f1_alt_w, subpop_drop3_mut, subpop_oss_mut, pptc_t,
        ml_elitism, sbc_nc, negatives_ml_w, cc_memetic_mutation, ml_chc_random_mut, cc_memetic_mutation_for_subpop, cc_memetic_mutation_for_subpop_bit_ratio, n_predicted_ml_w, eps_t, mlrf_t,
        reduction_perlabel_w, mlksnn_memetic_mut, partial_selection_mutation, partial_selection_mutation_bit_ratio, partial_is_mutation, kperlabel_local_opt_ratio, kperlabel_rubber_band_penalty,
        kperlabel_different_ks_penalty, ml_ros_p, mlksnn_ls_steps, ml_multiselection_mutation_ratio, reduction_ml_ms_w, ml_k_memetic_mutation_ratio, chc_minimum_distance_factor, ml_ensemble_t,
        eb_jaccard_coeff_w, regularize_k_w, oversampling_ml_ms_w, mltree_lambda, tsa_t, mlrs_s, mlrs_k, mlrs_beta, optiondt_epsilon, mltree_depth, mrknn_f, ml_eso_os_ratio, ml_eso_us_ratio;

    bool reevaluate_individuals, use_neighbors_cache, ml_undersample_binary_learners, ml_onesubpop_per_cycle, ml_use_hash_table, ml_select_subpop_by_relevance, ml_allow_negative_subpops, ml_loo,
        ml_return_overall_best, ml_init_with_undersampling, ml_evolve_undersampled_subpops, ml_ins_sel_per_label, ml_cc_subpop_random_fitness, ml_cc_mutate_elite, ml_cc_local_opt_best, pptc_no_loss,
        ml_cc_weighted_qs, ml_ins_sel_per_label_same_neighbors, ml_cc_combine_best_subinviduals, ml_return_best_performing, ml_relative_multiselection, mlksnn_loo, lamlksnn_is, ml_is_set_k_cv,
        combine_metrics_gmean, mlksnn_ties_closer_to_cv_k, regularize_mlksnn, mlksnn_use_window, mlksnn_per_label_sampling, mlksnn_update_ada_weight_with_current_best, ml_init_with_roulette,
        ml_init_with_cardinality, initialize_all_minority_class, symmetric_selection, ml_use_hux, cv_for_fitness_as_ratio, evaluate_ml_with[N_ML_CLASSIFIERS], gauss_reduction, use_mlcut, mleso_loo,
        evaluate_ml_with_external[N_ML_EXT_CLASSIFIERS], mrknn_loo, cheat_neg_to_pos_only, mleso_unsel_nocopy, mleso_donot_eval_copies, ecc_with_bagging, cv_ml_hyperparameters,
        print_base_training_error, pevus_remove_null, pevus_random_init, prus_relative_ratio, pros_relative_ratio, partial_ci_samples, mlrf_real_votes, mleso_one_w, ml_ensemble_threshold_by_resubs;

    t_MLClassifier ml_method, ml_base_learner;
    t_MLExternalBaseClassifier ml_ext_base_method;
    t_MergeMethod merge_method;
    t_Classifier binary_classifier, br_model, cc_model, subpop_binary_classifier, rakel_model, lp_model, clr_model, homer_model, dbr_model, sta_model, ns_model, mltree_model, ps_model;
    t_MLInstanceSelection ml_instance_selection;
    t_MLFeatureSelection ml_feature_selection;
    t_MLDataPartitionMethod data_partition_method;
    t_PerformanceMeasure subpop_perf_measure;
    t_EvolutionaryScheme ml_cc_pop_evol_scheme;
    t_MLKPerLabelMethod ml_kperlabel_method;
    t_MLEnsembleMethod ml_ensemble_method;
    t_MLImbalanceMethod ml_imbalance_method;
    t_MLResampling ml_resampling_method;

    // OpenMP
    bool omp_allow_nested_loops;

// Abandoned
#ifdef ITERATIVE_SAMPLING_FOR_MLKsNN
    scalar kperlabel_sampling_ratio_step;
#endif

} t_Parameters;
