/***************************************************************************
 *   Copyright (C) 2017 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#pragma once

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "macros.h"
#include "common.h"

// Structure for sorting the values
typedef struct {
    scalar value;
    int label;
} t_info;

// Prototypes
/* subsampling.c */
void StratifiedRandomSubsampling(scalar ***, int *, int **, scalar, int, int);
void TemporalUndersampleDataForTraining(scalar **, int, int, int *, int, scalar ***, int **, int *, int);
/* stratified.c */
void DataPartition(const t_Parameters *, scalar **, int, int *, int *, int *, scalar ****, int **, int ***, int, int ***, int **, int ****, int, int, t_Partition, bool, int ***);
void StratifiedRandomSample(scalar **, int, int *, scalar ***, int **, int, int, int **, int **, int ***);
void StratifiedPartition(int, int *, int, int, int *);
void StratifyDataset(scalar **, int, int *, int **, int);
/* partclass.c */
void RandomPartitionKCD(int *, int *, int, int, int *, int, int);
void PartitionateEqualKCD(int *, int *, scalar **, int, int, int *, int, int);
void PartitionateOneInputKCD(const t_Parameters *, int *, int *, scalar **, int, int, int *, int, int, t_Partition);
/* treepart.c */
void PartitionateEqual(int *, int *, scalar **, int, int, int *, int, int, bool);
void PartitionateBoard(int *, int *, scalar **, int, int, int);
void PartitionateRandomPrototypes(int *, int *, scalar **, int, int, int, int **);
void FusionPartitions(int *, int *, int *, int, int *);
void RandomPartition(int *, int *, int, int);
void PartitionateOneInput(const t_Parameters *, int *, int *, scalar **, int, int, int *, int, int, t_Partition);
void MixSubsets(int *, int *, int *, int, int);
/* kmeans.c */
void DataDivisionkMeans(const t_Parameters *, int *, int **, scalar **, int, int *, int, int, int **, int);
void RemoveEmptyClusters(int *, int *, int *, int);
void kMeans(int *, int *, scalar **, int, int, int, int, int **, int *);
scalar DunnIndex(scalar **, int, int, int *);
void kMeansStratifiedSample(const t_Parameters *, scalar ***, int, int **, scalar ***, int **, int *, int, int, int, int **, bool);
/* bigrandtour.c */
void BidimensionalGrandTour(const t_Parameters *, int *, int *, scalar **, int, int *, int *, int, int);
void gram_schmidt(int, scalar *, scalar *);
void getVec2D(int OP, ...);
/* sampling.c */
void CalculateCentroids(scalar **, int, scalar **, int, int, int *, int *, int **);
void kMeansSample(const t_Parameters *, scalar ***, int, int **, scalar ***, int **, int *, int, int, int, int **, bool);
void StratifiedSample(scalar **, int, int *, scalar **, int *, int, int);
/* discretize.c */
void WPKIDMethod(scalar **, int, int, scalar **, int *, int **);
void FayyadAndIraniMDLMethod(scalar **, int, int, int *, int, scalar **, int *, int *, int **);
int CompareInfo(const void *x, const void *y);
void DiscretizeFeature(t_info *, int, int, int *, scalar **);
scalar EvaluateCutPointCIF(scalar, t_info *, int, int, int *, int *, scalar *, scalar *);
bool EvaluateMDLCriterion(t_info *, int, int, scalar, scalar, scalar, int, int);
/* sparse.c*/
void GetSubspaceData(scalar **, int **, scalar **, int, int **, int, int *, int);
/* libsvm.c */
void SaveLibSVMData(scalar **, int **, int *, int, char *);
