/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdbool.h>
#include "typex.h"
#include "macros.h"

void GetSubspaceData(scalar **subs_data, int **subs_index, scalar **data, int instances, int **index, int subs_size, int *subspace, int features)
{
    // Get binary vector for faster run
    bool *s;
    int *translation;

    MALLOC(s, bool, features, "GetSubspaceData: s");
    MALLOC(translation, int, features, "GetSubspaceData: translation");

    for (int i = 0; i < features; i++)
        s[i] = false;

    for (int i = 0; i < subs_size; i++) {
        s[subspace[i]] = true;
        translation[subspace[i]] = i+1;
    }

    for (int i = 0; i < instances; i++) {
        // Count
        int c = 0;
        for (int j = 0; index[i][j] != -1; j++)
            if (s[index[i][j]-1])
                c++;

        if (!c) {
            // Create dummy feature to avoid null pointer
            REALLOC(subs_data[i], scalar, 1, "GetSubspaceData: subs_data[i]");
            REALLOC(subs_index[i], int, 2, "GetSubspaceData: subs_index[i]");

            subs_data[i][0] = 0.0;
            subs_index[i][0] = 1;
            subs_index[i][1] = -1;
        } else {
            // Reallocate
            REALLOC(subs_data[i], scalar, c, "GetSubspaceData: subs_data[i]");
            REALLOC(subs_index[i], int, c+1, "GetSubspaceData: subs_index[i]");

            c = 0;
            for (int j = 0; index[i][j] != -1; j++)
                if (s[index[i][j]-1]) {
                    subs_data[i][c] = data[i][j];
                    subs_index[i][c] = translation[index[i][j]-1];
                    c++;
                }
            subs_index[i][c] = -1;
        }
    }

    FREE(s);
    FREE(translation);
}
