/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"
#include "ensemble.h"

t_Boosting *TrainBagging(const t_Parameters *parm, scalar **data, int instances, int *labels, int inputs, int classes, int **index, int *subset, int *subspace)
{
    t_Boosting *boost;
    scalar **l_data;
    int *l_labels,  **l_index = NULL, *l_subset = NULL;

    MALLOC(l_data, scalar *, instances, "TrainAdaBoost: l_data");
    MALLOC(l_labels, int, instances, "TrainAdaBoost: l_labels");
    if (subset)
        MALLOC(l_subset, int, instances, "TrainAdaBoost: l_subset");
    if (index)
        MALLOC(l_index, int *, instances, "TrainAdaBoost: l_index");

    // Initialize ensemble
    MALLOC(boost, t_Boosting, 1, "TrainAdaBoost: boost");
    MALLOC(boost->alpha, scalar, parm->ensemble_size, "TrainAdaBoost: boost->alpha");
    MALLOC(boost->bcs, t_Learner, parm->ensemble_size, "TrainAdaBoost: boost->bcs");

    boost->T = parm->ensemble_size;
    boost->base = parm->base_learner;
    FillVector(boost->alpha, NULL, boost->T, 1.0);

    // First round with original data
    for (int i = 0; i < instances; i++) {
        l_data[i] = data[i];
        l_labels[i] = labels[i];
        if (subset)
            l_subset[i] = subset[i];
        if (index)
            l_index[i] = index[i];
    }

    for (int t = 0; t < boost->T; t++) {
        ClassifierTraining(parm, boost->base, l_data, instances, l_labels, inputs, classes, l_index, subspace, l_subset, true, &(boost->bcs[t]), NULL);

        // Resample for next round
        for (int i = 0; i < instances; i++) {
            int next = RANDOM(0, instances);
            l_data[i] = data[next];
            l_labels[i] = labels[next];
            if (subset)
                l_subset[i] = subset[next];
            if (index)
                l_index[i] = index[next];
        }
    }

    FREE(l_data);
    FREE(l_labels);
    if (subset)
        FREE(l_subset);
    if (index)
        FREE(l_index);

    return boost;
}
