/***************************************************************************
 *   Copyright (C) 2007 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Data types

#pragma once

#include <stdbool.h>
#include "defns.i"
#include "types.i"
#include "typex.h"
#include "svm.h"
#include "common.h"
#include "classifiers.h"

// #define AUTOTEST             // Set auto-test code

// Weights penalty function
#define PENALTY(x)              (x)

#define MAX_WEIGHT      +1.0E+5
#define MIN_WEIGHT      -1.0E+5

#define MIN_ALPHA   0.0001
#define MAX_ALPHA   10.0

// Small values
#define NEARZERO2       0.000000001
#define NEARZERO        0.0001

// For rotation of vectors
#define SET     1
#define INIT    2
#define NEXT    3
#define CLEAR   4

// // Measures of diversity
// typedef enum {
//     Q, Kappa
// } t_DiversityMeasure;

// Structure for ordering the individuals
typedef struct {
    int being;
    scalar value;
} t_Rank;

// Function prototypes
// /* adaboostmh.c */
// void GenerateAdaBoostMHWeights(const t_Parameters *, t_ClassModel *, scalar **, int, t_OldEnsemble, int, int, scalar **, int, scalar **);
// /* auc.c */
// scalar ObtainAUCForClassifier(const t_Parameters *, t_OldEnsemble, scalar **, int, int *, int *, int, bool, int, int, int, char *, int, int **);
// int CompareDoubleInverted(const void *, const void *);
// void GetPositiveAndNegativeInstancesOutputsEnsemble(const t_Parameters *, t_OldEnsemble, scalar **, int, int *, scalar *, scalar *, int, int, int, int, int **);
// scalar f(scalar, scalar **, int);
// scalar trapzd(scalar **, scalar, scalar, int, int);
// scalar qtrap(scalar **, scalar, scalar, int);
// scalar ObtainAUPRC(const t_Parameters *, t_OldEnsemble, scalar **, int, int *, int, bool, int, char *, int, int **);
// /* balance-cascade.c */
// void ConstructBalanceCascade(const t_Parameters *, t_OldEnsemble *, int, int, int, int, int *, scalar **, scalar **, int, int **);
// /* bites.c */
// t_OldEnsemble ConstructLearningFromBitesEnsemble(t_Parameters *, int, scalar **, int, int *, int, int);
// bool EnsembleVotingWithMask(const t_Parameters *, t_OldEnsemble, int, scalar *, int, int, int, int, bool *);
// /* boostrsm.c */
// void ObtainBestSubspace(const t_Parameters *, t_ClassModel *, scalar *, t_OldEnsemble, int, scalar **, int, int *, int, int, scalar **, int);
// int HammingDistance(int *, int *, int);
// void HuxCrossover(int *, int *, int *, int *, int);
// scalar EvaluateIndividualRSM(const t_Parameters *, int *, int, scalar **, int, int *, int *, scalar *, t_OldEnsemble, int, int, int, scalar *, scalar *, t_ClassModel, scalar **, int);
// int NthOrderedSelection(scalar *, int, int);
// int CompareValue(const void *v, const void *);
// /* bp_adamh.c */
// void BPAdaBoostMH(const t_Parameters *, t_ClassModel *, scalar, int, int, int, scalar **, scalar **, int, int, scalar **);
/* bp.c */
void BP(const t_Parameters *, t_ClassModel *, int, int, int, scalar **, int, scalar **, int, int, int **);
// /* bp_nlp.c */
// void BPProjectNetwork(const t_Parameters *, t_ClassModel *, scalar *, int, int, int, scalar **, int, int, scalar **);
// /* brent.c */
// scalar Brent(const t_Parameters *, t_GRNN *, scalar, scalar, scalar cx, scalar, int, int);
// scalar EvaluateGRNNSigma(t_GRNN *, scalar, int, int);
// void Order3Values(scalar *, scalar *, scalar *);
// /* c45.c */
void TrainC45ClassifierEnsemble(const t_Parameters *, t_ClassModel *, int, scalar **, int, int *, bool, t_OldEnsemble, int, int);
scalar C45OuputEnsemble(const t_Parameters *, t_ClassModel *, scalar *, scalar *, int, t_OldEnsemble, int, int);
// /* chc_classifier.c */
// void CHCInstanceSelection(const t_Parameters *, t_ClassModel *, scalar *, bool, t_OldEnsemble, int, int, scalar **, int *, int, int);
// scalar EvaluateInstanceSelectionIndividual(const t_Parameters *, int *, scalar *, t_ClassModel *, t_OldEnsemble, int, int, int, int);
// /* chc_is.c */
// void kNNCHCInstanceSelection(const t_Parameters *, t_ClassModel *, scalar *, bool, t_OldEnsemble, int, scalar **, int *, int, int, int);
// scalar kNNEvaluateInstanceSelectionIndividual(const t_Parameters *, int *, t_ClassModel *, scalar *, scalar **, int, int *, scalar **, int);
// void RNNMutationCHCIS(const t_Parameters *, int *, t_ClassModel *, scalar *, scalar **, int, scalar **, int *, int);
// int ObtainkCVForIs(t_Parameters *, t_ClassModel *, int, int *, scalar **, int);
// /* classifier.c */
// scalar ClassifierOutput(const t_Parameters *, t_ClassModel *, scalar *, scalar *, t_OldEnsemble, int, int, int, int, t_Classifier, int *);
// void ClassifierTraining(const t_Parameters *, t_ClassModel *, int, scalar **, int, int *, bool, t_OldEnsemble, int, int, int, scalar **, int, t_Classifier, int, int **);
void AllocateClassifier(const t_Parameters *, t_ClassModel *, int, int, bool, int, int, int);
void GenerateNewClassifier(const t_Parameters *, t_ClassModel *, bool, int, int, int);
int NetworkActualOutput(const t_Parameters *, t_ClassModel *, scalar *, scalar *, int, int *);
// scalar EvaluateClassifier(const t_Parameters *, t_ClassModel *, scalar *, scalar **, int, int *, int, t_OldEnsemble, int, int, int, int **);
// void TestWeights(const t_Parameters *, t_ClassModel *, int);
// scalar GetPatternMargin(const t_Parameters *, t_ClassModel *, scalar *, int, int, t_OldEnsemble, int, int, int, int *);
// void ReallocateNetwork(t_ClassModel *, int, int);
void FreeClassifier(const t_Parameters *, t_ClassModel *, int);
// /* cv.c */
// void SetCrossValidatedParametersEnsemble(t_Parameters *, t_ClassModel *, int, scalar **, int, int *, scalar **, t_OldEnsemble, int, int, int, int, int **);
// /* cvknn.c */
// int ObtainkCrossValidationEnsemble(const t_Parameters *, t_ClassModel *, int *, int, t_OldEnsemble, int, scalar **, int, int *, int, int);
// scalar EvaluatekEnsemble(int *, t_ClassModel *, scalar *, scalar **, int, int last, int *, int, int, bool);
// /* demo-auc.c */
// scalar ObtainAUCForDemocraticEnsemble(const t_Parameters *, t_DemocraticEnsemble, scalar **, int, int *, int, bool, int, int, int, int, int, char *);
// void GetPositiveAndNegativeInstancesOutputsEnsembleForDemocraticEnsemble(const t_Parameters *, t_DemocraticEnsemble, scalar **, int, int *, scalar *, scalar *, int, int, int, int, int, int);
// /* demo-boosting.c */
// void DemocraticUpdateAdaBoostWeights(const t_Parameters *, scalar *, scalar *, scalar **, int, int, int, int *, t_DemocraticEnsembleMember, int, int, int t, char *);
// /* democracy.c */
// void DemocraticEnsembleOutput(const t_Parameters *, t_DemocraticEnsemble, int, int, int, scalar *, scalar *, int, int);
// int DemocraticEnsembleCategory(const t_Parameters *, t_DemocraticEnsemble, int, int, int, scalar *, int, int);
// scalar EvaluateDemocraticEnsemble(const t_Parameters *, t_DemocraticEnsemble, int, int, int, scalar **, int, int *, int, int);
// scalar EvaluateDemocraticEnsemble2Class(const t_Parameters *, t_DemocraticEnsemble, int, scalar **, int, int *, scalar *, scalar *, scalar *, int, int, int, int);
// void DemocraticContingencyTable(const t_Parameters *, t_DemocraticEnsemble, int, scalar **, int, int *, int, int, int **, int, int);
// void DemocraticEnsembleMemberOutput(const t_Parameters *, t_DemocraticEnsembleMember, int, int, scalar *, scalar *, int, int);
// int DemocraticEnsembleMemberCategory(const t_Parameters *, t_DemocraticEnsembleMember, int, int, scalar *, int, int);
// scalar EvaluateDemocraticEnsembleMember(const t_Parameters *, t_DemocraticEnsembleMember, int, int, int, scalar **, int, int *, int, int);
// scalar GenerateAdaBoostClassifierWeight(const t_Parameters *, t_ClassModel *, scalar *, int, int, scalar **, int *, int);
// /* demo-ensemble.c */
// void print_null(const char *s);
// scalar GetDatamemorySize(char *file);
// /* demo-kappa.c */
// void DemocraticEnsembleSaveKappaErrorData(const t_Parameters *, t_DemocraticEnsemble, int, scalar **, int, int *, int, int, bool, int);
// scalar DemocraticEnsembleKappaError(const t_Parameters *, t_DemocraticEnsemble, scalar **, int, int *, int, int C, int, int);
// /* drop3.c */
// void DecrementalReductionOptimizationProcedure3(const t_Parameters *, t_ClassModel *, scalar *, bool, int, scalar **, int *, int);
// void GetkNearestNeighbourInEnsembleCached(t_ClassModel *, int, int *, scalar **, int *, int);
// void OrderedListDrop3(scalar *, int *, int);
// /* easy-.c */
// void ConstructEasyEnsemble(const t_Parameters *, t_OldEnsemble *, int, int, int, int, int *, scalar **, scalar **, int, int **);
// /* eda.c */
// void EvolutionaryDiscriminantAnalysis(t_Parameters *, t_ClassModel *, scalar *, int, int, int, int *, scalar **);
// void DiscreteRecombination(const t_Parameters *, scalar **, scalar ***, int, int, int);
// scalar EvaluateEDAIndividual(const t_Parameters *, scalar **, scalar **, scalar **, int, int *, int, int, int);
// void LinearProjection(scalar *, scalar *, scalar **, int, int);
// /* edbd_nlp.c */
// void EDBDProjectNetwork(const t_Parameters *, t_ClassModel *, scalar *, int, int, int, scalar **, int, int, scalar **);
// /* .c */
// void EnsembleOutput(const t_Parameters *, t_OldEnsemble, int, scalar *, scalar *, int, int, int, t_Combination, int *);
// bool EnsembleTestClassifierOutput(const t_Parameters *, t_OldEnsemble, int, scalar *, int, int, int, int, t_Combination, int *);
// int EnsembleClassifyPattern(const t_Parameters *, t_OldEnsemble, int, scalar *, int, int, int, t_Combination, int *);
// scalar EvaluateEnsemble(const t_Parameters *, t_OldEnsemble, int, scalar **, int, int *, int, int, t_Combination, int **);
// void ContingecyTable(const t_Parameters *, t_OldEnsemble, int, scalar **, int, int *, int, int, int **, t_Combination, int **);
// scalar EnsembleMSEError(const t_Parameters *, t_OldEnsemble, int, scalar **, int, int, int, t_Combination, int **);
// scalar EnsembleMargin(const t_Parameters *, t_OldEnsemble, int, scalar *, int, int, int, int, t_Combination, int *);
// scalar EvaluateEnsemble2ClassProblem(const t_Parameters *, t_OldEnsemble, int, scalar **, int, int *, scalar *, scalar *, scalar *, int, int, t_Combination, scalar *, scalar *, scalar *, scalar *,
//                                      scalar *, scalar *, int, int **);
// /* featureboost.c */
// void eFeatureBoostEnsemble(const t_Parameters *, t_ClassModel *, int, scalar **, int, int *, int, scalar **, int);
// scalar Deemphasize(const t_Parameters *, scalar *, int t, int, int, scalar, t_ClassModel *, scalar *, t_OldEnsemble, int, scalar **, int, int);
// void Importance(const t_Parameters *, t_ClassModel *, scalar *, int, t_OldEnsemble, int, scalar **, int, int);
// scalar Marginalize(const t_Parameters *, scalar *, int, int, t_ClassModel *, t_OldEnsemble, int, scalar **, int, int);
// /* filter.c */
// t_OldEnsemble ConstructFilterBoostEnsemble(t_Parameters *, int, scalar **, int, int *, int);
// scalar GetEdge(int, scalar, scalar, scalar);
// bool Filter(scalar, scalar, scalar **, int *, int, t_OldEnsemble, int t, int, int *);
// /* freundnn.c */
// void FreundNNBoost(const t_Parameters *, t_ClassModel *, int, scalar **, int *, int, int, int);
// scalar EvaluateInstanceSubset(int *, int, scalar *, int, int *, scalar **, int);
// /* grnn.c */
// void ConstructGRNN(const t_Parameters *, t_GRNN **, int, int, int, char *, scalar *, scalar *, scalar **, int, scalar **);
// void OutputGRNN(t_GRNN *, scalar *, scalar *, int, int);
// scalar SqEuclideanDistanceSubspace(scalar *, scalar *, int, scalar *);
// void ConstructRandomGRNN(const t_Parameters *, t_GRNN **, int, int);
// /* hda.c */
// void GenerateHDAProjection(const t_Parameters *, t_ClassModel *, scalar *, bool *, t_OldEnsemble, int, int, int, scalar **, int *, int, scalar **, int);
// /* huge-.c */
// void ReadPartitionFromFile(scalar **, int *, int, int, int, char *, int *list);
// int CompareCoordinate(const void *, const void *);
// /* icf.c */
// void IterativeCaseFiltering(const t_Parameters *, t_ClassModel *, scalar *, bool, int, scalar **, int *, int);
// void ComputeReachable(scalar *, scalar *, int *, scalar **, scalar *, int, int *);
// /* is.c */
// void ObtainWeightedInstanceSelection(const t_Parameters *, t_ClassModel *, scalar *, bool, t_OldEnsemble, int, scalar **, int *, int, int, int);
// void ReducedNearestNeighbors(t_ClassModel *, scalar *, bool, int, scalar **, int *, int);
// void SimulatedAnnealingInstanceSelection(const t_Parameters *, t_ClassModel *, scalar *, bool, t_OldEnsemble, int, scalar **, int, int *, int, int);
// /* kappa.c */
// scalar KappaStatistic(int *, int *, int, int);
// scalar AverageKappa(int **, int, int, int, int);
// void SaveKappaErrorData(const t_Parameters *, t_OldEnsemble, int, scalar **, int, int *, int, int, int, int **);
// /* kappa-e.c */
// scalar KappaError(int **, const t_Parameters *, t_OldEnsemble *, scalar **, int, int *, int, int C, int, int **);
// /* kmeans.c */
// void kMeansGRNN(scalar **, int *, int, int, scalar **, int, scalar *, int);
// /* knn.c */
void TrainkNNClassifier(const t_Parameters *, t_ClassModel *, int, scalar **, int, int *, t_OldEnsemble, int, int, int);
// scalar kNNOutput(const t_Parameters *, t_ClassModel *, scalar *, scalar *, int, t_OldEnsemble, int, int);
// int NearestNeighbourEnsemble(t_ClassModel *, scalar *, int *, scalar *);
// int kNearestNeighbourInEnsemble(t_ClassModel *, scalar *, int *, scalar *, int);
// scalar EuclideanDistanceSubspaceScalar(scalar *v, scalar *, int, scalar *);
// int kNearestNeighbourInEnsembleCached(t_ClassModel *, int, int *, scalar **, int);
// /* lazarevic.c */
// void ObtainLazarevicMostRelevantAttributes(const t_Parameters *, scalar *, scalar *, int, int, int, scalar **, scalar **);
// /* math.c */
scalar TransferFunction(scalar, t_Transfer, scalar, scalar);
scalar HyperbolicTangent(scalar, scalar, scalar);
scalar Logistic(scalar, scalar, scalar);
scalar Sine(scalar, scalar, scalar);
scalar Normal(scalar, scalar);
scalar Exponential(void);
void NormalizeVector(scalar *, int);
void NormalizeVectorToN(scalar *, int);
// /* nda.c */
// void GenerateNDAProjection(const t_Parameters *, t_ClassModel *, scalar *, bool *, t_OldEnsemble, int, int, scalar **, int, int *, int, scalar **, int);
// void ScaleProjection(const t_Parameters *, t_ClassModel *, scalar, scalar ubound, int, int, int, t_OldEnsemble, int, scalar **);
// /* nleda.c */
// void EvolutionaryNonLinearDiscriminantAnalysis(const t_Parameters *, t_ClassModel *, scalar *, int, int, int, int *, scalar **);
// scalar EvaluateNLEDAIndividual(scalar **, scalar **, int *, int, scalar *, int, scalar **, int, int *, int);
// void NonLinearProjection(scalar *, scalar *, scalar **, int, int);
// /* nonlinear.c */
// void GenerateNonLinearProjection(const t_Parameters *, t_ClassModel *, scalar *, bool *, t_OldEnsemble, int, int, scalar **, int *, int, scalar **, int);
// void CheckVariance(scalar **, int, int, bool *);
// void Populate(scalar ***, int **, int *, int, scalar **, int, int *, int, scalar *, int, int);
// scalar EuclideanDistanceDifferentSubspaces(scalar *, scalar *, int, scalar *);
// void FilterNonLinearProjection(const t_Parameters *, char *, int, int);
// /* nwfe.c */
// void GenerateNWFEProjection(const t_Parameters *, t_ClassModel *, scalar *, bool *, t_OldEnsemble, int, int, scalar **, int *, int, int, scalar **, int);
// /* partition.c */
// void PartitionateOneInputEnsembleStratified(const t_Parameters *, scalar **, int, int, int *, int, int, scalar ***, int *, int **, int **);
// void PartitionateOneInputEnsemble(const t_Parameters *, scalar **, int, int, int *, int, scalar ***, int *, int **, int **);
// /* project.c */
// void GenerateProjection(t_Parameters *, t_ClassModel *, t_ClassModel *[3], t_OldEnsemble, int, scalar *, bool *, int, int, int *, int, scalar **, scalar **, bool *, int);
// scalar GenerateBPProjection(const t_Parameters *, t_ClassModel *, t_ClassModel *[3], t_OldEnsemble, int, scalar *, int *best, int, int, int *, int, scalar **, scalar **);
// bool TestProjection(const t_Parameters *, t_ClassModel *, scalar *, int, int);
// void BPProjectionOutput(const t_Parameters *, t_ClassModel *, scalar *, scalar *, int);
// scalar EvaluateProjection(const t_Parameters *, t_ClassModel *, scalar **, int *, int, scalar *, int);
// void GenerateProjectionWeights(const t_Parameters *, t_ClassModel *, t_OldEnsemble, int, scalar *, int *, int, int *, int, int, scalar **);
// void ProjectInstance(const t_Parameters *, scalar *, scalar *, t_ClassModel *, t_OldEnsemble, int, int);
// void ProjectInstanceIteratively(const t_Parameters *, t_OldEnsemble, scalar *, scalar *, int, int);
// void ProjectInstanceOnce(const t_Parameters *, scalar *, scalar *, t_ClassModel *, int);
// /* proj_eval.c */
// scalar Distance(sca/*lar *, scalar *, int);
// scalar Accuracy1NN(char *, int, int, int *, int);
// scalar WrapperEvaluat*/ion(const t_Parameters *, char *, int , int v, int *, scalar *, t_OldEnsemble, int, int, scalar **, int);
// /* projnet.c */
// void ConstructProjectionNet(const t_Parameters *, t_ProjNet **, int, int, int, char *, scalar *, scalar *, scalar **, scalar **, int);
// void OutputProjectionNet(t_ProjNet *, scalar *, scalar *, int);
// void OutputProjectionIndividualNet(t_ProjNet *, scalar *, scalar *);
// void TrainProjectionNet(t_ProjNet *, int, int, scalar **, int);
// scalar EvaluateProjectionNet(t_ProjNet *, scalar **, int);
// /* random.c */
// void RandomPartitionBites(scalar **, int, int *, int, scalar ***, int *, int **);
// /* rcga.c */
// void ObtainSupervisedProjection(const t_Parameters *, t_ClassModel *, scalar *, t_OldEnsemble, int, int, int *, int, scalar **, scalar **, int, bool *, int);
// void BLXAlphaMatrix(scalar **father, scalar **, scalar **, scalar **, int, int);
// void NonUniformMutationMatrix(scalar **, int, int, scalar);
// scalar EvaluateProjectionIndividual(const t_Parameters *, scalar **, scalar **, int, int *, scalar *, scalar *, t_OldEnsemble, int, int, t_ClassModel, int, scalar **, scalar **, int, int *);
// void InitializeIndividual(const t_Parameters *, scalar **, int, int);
// /* reinit.c */
// scalar AverageDifference(scalar *, scalar *, int);
// scalar AverageVotingMargin(const t_Parameters *, t_OldEnsemble, int, scalar **, int, int *, scalar *, scalar *, int, int);
// void WeightsHistogram(t_OldEnsemble, int, int, int);
// /* rotation.c */
// void GeneratePCAProjection(const t_Parameters *, t_ClassModel *, scalar *, bool *, int, int, int, int *, scalar **);
// /* rprop_nlp.c */
// void RP/*ropProjectNetwork(const t_Parameters *, t_ClassModel *, scalar *, int, int, int, scalar **, int, int, scalar **);
// scalar Sign(scalar);*/
// /* sampling.c */
// void GenerateWaggingSample(scalar *, int);
// void GenerateeBaggingSample(scalar *, int);
scalar Poisson(void);
// void GenerateArcingSample(const t_Parameters *, scalar *, t_OldEnsemble, int, int, scalar *, int *, scalar **, int, int, int **);
// void GenerateAdaSample(const t_Parameters *, t_ClassModel *, scalar *, scalar *, scalar *, int, t_OldEnsemble, int, int, scalar **, int *, int, t_OldEnsembleMethod, int **);
// void GenerateeGentleBoostSample(const t_Parameters *, t_ClassModel *, scalar *, scalar *, int, t_OldEnsemble, int, scalar **, int *, int, int, int **);
// /* scale.c */
// t_scale *ScaleOutputData(scalar **, int, int, int, scalar, scalar);
// /* scca.c */
// void GetSCCAProjection(const t_Parameters *, t_ClassModel *, scalar *, scalar *, int, int);
// void GenerateSCCAProjection(const t_Parameters *, t_ClassModel *, scalar *, bool *, t_OldEnsemble, int, int, scalar **, int *, int, scalar **, int);
// scalar **SCCAProjection(t_Parameters *, t_ClassModel *, scalar **, int, int);
// void FreeSCCAProjection(scalar ***, int, bool);
// /* subspaces.c */
// void CheckSubspacesSize(const t_Parameters *, int *, int, int);
// /* svmlocal.c */
void TrainSVMClassifierEnsemble(const t_Parameters *, t_ClassModel *, int, scalar **, int, int *, t_OldEnsemble, int, int, int **);
scalar SVMOuputEnsemble(const t_Parameters *, t_ClassModel *, scalar *, scalar *, int, t_OldEnsemble, int, int, int *);
scalar GetSVMMarginEnsemble(const t_Parameters *, t_ClassModel *, scalar *, int, int, t_OldEnsemble, int, int, int *);
// /* yuleQ.c */
// scalar YuleQStatistic(bool *, bool *, int);
// scalar AverageYuleQ(bool **, int, int, int);
/* adaboost.c */
t_Boosting *TrainAdaBoost(const t_Parameters *, scalar **, int, int *, int, int, int **, int *, int *, bool, bool, bool, bool, bool);
void GenerateAdaSample(const t_Parameters *, t_Learner *, scalar *, scalar *, scalar *, int, int, scalar **, int *, int, t_Classifier, int **, int *, int *, bool);
void FreeAdaBoost(t_Boosting *, const t_Parameters *, int, int);
int AdaBoostOutput(const t_Parameters *, t_Boosting *, scalar *, int, int, int *, int *, scalar **, int, int *, int *, int, int **, scalar *);
scalar AdaBoostMargin(const t_Parameters *, t_Boosting *, scalar *, int, int, int, int *, int *, scalar **, int, int *, int *, int **, bool);
void GenerateArcingSample(const t_Parameters *, t_Boosting *, scalar *, int, int, scalar **, int *, int, t_Classifier, int **, int, int *, int *);
/* ensemble.c */
bool IsEnsemble(t_Classifier);
/* bagging.c */
t_Boosting *TrainBagging(const t_Parameters *, scalar **, int, int *, int, int, int **, int *, int *);
/* gentleboost.c */
void UpdateGentleBoostWeights(const t_Parameters *, t_Classifier, int, int, int **, int *, scalar **, int, int *, int *, scalar *, t_Learner *);
/* reweight.c */
int ReweightBoostOutput(const t_Parameters *, t_Boosting *, scalar *, int, int, int *, int *, scalar **, int, int *, int *, int, int **, scalar *);
void GenerateReweightBoostSample(const t_Parameters *, t_Boosting *, scalar *, scalar *, scalar *, int, int, scalar **, int *, int, int **, int *, int *, int);
/* floatboost.c */
t_Boosting *TrainFloatBoost(const t_Parameters *, scalar **, int, int *, int, int, int **, int *, int *);
void UpdateFloatBoostWeights(const t_Parameters *, int, int, int **, int *, scalar **, int, int *, int *, scalar *, t_Ensemble *);
/* featureboost.c */
void Importance(const t_Parameters *, t_Learner *, scalar *, int, scalar **, int, int, int, int **, int *, int *, t_Classifier);
scalar Marginalize(const t_Parameters *, scalar *, int, int, t_Learner *, int, scalar **, int, int, int **, int *, int *, t_Classifier);
scalar Deemphasize(const t_Parameters *, scalar *, int, int, int, scalar, t_Learner *, int *, int, scalar **, int, int, int **, int *, int *, t_Classifier);
t_FeatureBoost *TrainFeatureBoostEnsemble(const t_Parameters *, int, scalar **, int, int *, int, int **, int *);
int FeatureBoostOutput(const t_Parameters *, t_FeatureBoost *, scalar *, int, int, int *, scalar **, int, int *, int *, int, int **, scalar *);
void FreeFeatureBoost(t_FeatureBoost *, const t_Parameters *, int, int);
scalar FeatureBoostMargin(const t_Parameters *, t_FeatureBoost *, scalar *, int, int, int, int *, scalar **, int, int *, int *, int **);
