/***************************************************************************
 *   Copyright (C) 2007 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <stdlib.h>
#include "ensemble.h"
#include "macros.h"


scalar TransferFunction(scalar x, t_Transfer t, scalar a, scalar b)
{
    if (t == Log)
        return Logistic(x, a, b);
    else if (t == HypTan)
        return HyperbolicTangent(x, a, b);
    else if (t == Linear)
        return x;
    else
        return Sine(x, a, b);
}

scalar HyperbolicTangent(scalar x, scalar a, scalar b)
{
    if (x < -45.0)
        return -a;
    else if (x > 45.0)
        return a;
    else
        return a*tanh(b*x);
}

scalar Logistic(scalar x, scalar a, scalar b)
{
    if (x < -45.0)
        return 0.0;
    else if (x > 45.0)
        return a;
    else
        return a/ (1.0 + exp(-b*x));
}

scalar Sine(scalar x, scalar a, scalar b)
{
    return a*sin(b*x);
}

// Return an exponentially distributed value of unit mean
scalar Exponential()
{
    scalar dum;

    dum = -log10(RANDOM(0.000000001, 1.000000001));

    // Flip the sign at random
    if (RANDOM(0.0, 1.0) < 0.5)
        dum = -dum;

    return dum;
}

// Normalize vector to module 1

void NormalizeVector(scalar *v, int n)
{
    scalar sum = 0.0;

    for (int i = 0; i < n; i++)
        sum += v[i];

    if (sum < NEARZERO)
        return;

    for (int i = 0; i < n; i++)
        v[i] /= sum;
}

// Normalize vector to N

// void NormalizeVectorToN(scalar *v, int n)
// {
//     scalar sum = 0.0;
//
//     for (int i = 0; i < n; i++)
//         sum += v[i];
//
//     if (sum < NEARZERO)
//         return;
//
//     sum /= n;
//
//     for (int i = 0; i < n; i++)
//         v[i] /= sum;
// }
