#!	/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Decision trees
"""

import sys
import os
import numpy as np
from sklearn.tree import DecisionTreeClassifier
from sklearn.metrics import confusion_matrix
from sklearn import tree
#from sklearn.tree.export import export_text

try:
	import argparse
	import pandas as pd
except ImportError:
	sys.exit(0)

parser = argparse.ArgumentParser()
parser.add_argument("trainfile", help="Train dataset", type=str)
parser.add_argument("testfile", help="Test dataset", type=str)
parser.add_argument("-c", help="Print confusion matrix", action='store_true')
parser.add_argument("-s", help="Save tree to file", action='store_true')
args = parser.parse_args()

# Read train file
if not os.path.isfile(args.trainfile):
    print("Train dataset file '%s' does not exist" % args.trainfile)
    sys.exit()

dtrain = pd.read_csv(args.trainfile, header=None, sep=" ")
r, c = dtrain.shape

# Read test file
if not os.path.isfile(args.testfile):
    print("Test dataset file '%s' does not exist" % args.testfile)
    sys.exit()

dtest = pd.read_csv(args.testfile, header=None, sep=" ")

# Create tree
clf = DecisionTreeClassifier()

# Train the model on training data
clf.fit(dtrain.loc[:, dtrain.columns != c-1], dtrain[c-1])

nodes = clf.tree_.node_count

# Save tree
if args.s:
    r = export_text(clf, decimals=4)
    foutput = args.testfile+'.tree'
    with open(foutput, "w") as text_file:
        text_file.write(r)

# Save predictions
preds = clf.predict(dtest.drop(c-1, axis=1))
np.savetxt(args.testfile+".preds", preds, fmt='%d')

# Save probabilities for margin
probs = clf.predict_proba(dtest.drop(c-1, axis=1))
np.savetxt(args.testfile+".probs", probs)

if args.c:
	cm = confusion_matrix(dtest[c-1], preds)
	print(cm)

# Save number of nodes
f = open(args.testfile+".nodes", "w")
f.write("Number of nodes: %g\n"  % nodes)
f.close()

sys.exit(1)
