#!	/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Deep Learning
"""

import sys
import os
import numpy as np
from keras.models import Sequential
from keras.layers import Dense
from sklearn.preprocessing import LabelEncoder
from keras.utils import np_utils
from sklearn.metrics import confusion_matrix
import warnings
import logging
import os

warnings.filterwarnings("ignore")
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'  # FATAL
logging.getLogger('tensorflow').setLevel(logging.FATAL)

try:
	import argparse
	import pandas as pd
except ImportError:
	sys.exit(0)

parser = argparse.ArgumentParser()
parser.add_argument("trainfile", help="Train dataset", type=str)
parser.add_argument("testfile", help="Test dataset", type=str)
parser.add_argument("-l", help="Layer nodes 20,8,10,...", type=str, required=True)
args = parser.parse_args()

# Read train file
if not os.path.isfile(args.trainfile):
    print("Train dataset file '%s' does not exist" % args.trainfile)
    sys.exit()

dtrain = pd.read_csv(args.trainfile, header=None, sep=" ")
r, c = dtrain.shape
classes = len(np.unique(dtrain[c-1]))

# Read test file
if not os.path.isfile(args.testfile):
    print("Test dataset file '%s' does not exist" % args.testfile)
    sys.exit()

dtest = pd.read_csv(args.testfile, header=None, sep=" ")

# Construct  network
model = Sequential()

# Add layers
layers = [int(d) for d in args.l.split(',')]

# Input layers
model.add(Dense(layers[0], activation='tanh', input_shape=(c-1,)))

# Add hidden layers
for nodes in layers[1:]:
    model.add(Dense(nodes, activation='tanh'))

# Add an output layer
model.add(Dense(classes, activation='sigmoid'))

# Compile model
model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])

# Fit model
model.fit(dtrain.loc[:, dtrain.columns != c-1], np_utils.to_categorical(dtrain[c-1]), epochs=100, batch_size=1, verbose=0)

# Save predictions
probs = model.predict(dtest.drop(c-1, axis=1))
preds = [np.argmax(prob) for prob in probs]
np.savetxt(args.testfile+".preds", preds, fmt='%d')

# Save probabilities for margin
np.savetxt(args.testfile+".probs", probs)
# print (confusion_matrix(dtest[c-1], preds))

sys.exit(1)
