/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "typex.h"
#include "common.h"
#include "data.h"
#include "external.h"

// Stub for external classifier confusion matrix

void GetExternalClassifierConfusionMatrix(const t_Parameters *parm, t_ExternalClassifier method, scalar **test_data, int test_instances, int *test_labels, int *subset, int *subspace, int classes,
        int inputs, scalar **train_data, int train_instances, int *train_labels, int **test_index, int **train_index, int **M, char *out_file_name, scalar *average_margin, scalar *minimum_margin,
        scalar *margin, int minority, scalar *sc_pos, scalar *sc_neg)
{
    scalar **t_data = NULL;
    int *t_labels = NULL, t_instances;

    // Sampling?
    if (parm->training_max_size && parm->training_max_size < train_instances) {
        t_instances = parm->training_max_size;
        StratifiedRandomSample(train_data, train_instances, train_labels, &t_data, &t_labels, t_instances, classes, NULL, NULL, NULL);

        train_data = t_data;
        train_labels = t_labels;
        train_instances = t_instances;
    }

    if (method != RandomForest || !parm->use_ranger_for_rf)
        GetConfusionMatrixWithPython(parm, test_data, test_instances, test_labels, subset, subspace, classes, inputs, train_data, train_instances, train_labels, test_index, train_index, M,
                                     out_file_name, average_margin, minimum_margin, margin, method, minority, sc_pos, sc_neg);
    else
        // Ranger implementation of RFs
        GetRandomForestConfusionMatrix(parm, test_data, test_instances, test_labels, subset, subspace, classes, inputs, train_data, train_instances, train_labels, test_index, train_index, M,
                                       out_file_name, average_margin, minimum_margin, margin);

    if (t_data) {
        FREE(t_data);
        FREE(t_labels);
    }
}
