#!	/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Gaussian process for machine learning
"""

import sys
import os
import numpy as np
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.gaussian_process.kernels import RBF


try:
	import argparse
	import pandas as pd
except ImportError:
	sys.exit(0)

parser = argparse.ArgumentParser()
parser.add_argument("trainfile", help="Train dataset", type=str)
parser.add_argument("testfile", help="Test dataset", type=str)
args = parser.parse_args()

# Read train file
if not os.path.isfile(args.trainfile):
    print("Train dataset file '%s' does not exist" % args.trainfile)
    sys.exit()

dtrain = pd.read_csv(args.trainfile, header=None, sep=" ")
r, c = dtrain.shape

# Read test file
if not os.path.isfile(args.testfile):
    print("Test dataset file '%s' does not exist" % args.testfile)
    sys.exit()

dtest = pd.read_csv(args.testfile, header=None, sep=" ")

kernel = 1.0 * RBF()
y = dtrain[c-1]
X = dtrain.drop(c-1, axis=1)

gpc = GaussianProcessClassifier(kernel=kernel).fit(X, y)

# Get probabilities and save them for margin
probs = gpc.predict_proba(dtest.drop(c-1, axis=1))
np.savetxt(args.testfile+".probs", probs)

# Get predictions from probabilities
preds = gpc.predict(dtest.drop(c-1, axis=1))
np.savetxt(args.testfile+".preds", preds, fmt='%d')

sys.exit(1)
