#!	/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Partial Least Squares discriminant analysis
"""

import sys
import os
import numpy as np
from sklearn.cross_decomposition import PLSRegression


try:
	import argparse
	import pandas as pd
except ImportError:
	sys.exit(0)

parser = argparse.ArgumentParser()
parser.add_argument("trainfile", help="Train dataset", type=str)
parser.add_argument("testfile", help="Test dataset", type=str)
args = parser.parse_args()

# Read train file
if not os.path.isfile(args.trainfile):
    print("Train dataset file '%s' does not exist" % args.trainfile)
    sys.exit()

dtrain = pd.read_csv(args.trainfile, header=None, sep=" ")
r, c = dtrain.shape

# Read test file
if not os.path.isfile(args.testfile):
    print("Test dataset file '%s' does not exist" % args.testfile)
    sys.exit()

dtest = pd.read_csv(args.testfile, header=None, sep=" ")

# Create dummy outputs in decoded form
classes = len(np.unique(dtrain[c-1]))
decoded_y = np.zeros((len(dtrain[0]), classes))

for i in range(0, len(dtrain[0])):
    decoded_y[i, int(dtrain[c-1][i])] += 1

plsda = PLSRegression().fit(X=dtrain.drop(c-1, axis=1), Y=decoded_y)

# Get probabilities and save them for margin
probs = plsda.predict(dtest.drop(c-1, axis=1))
np.savetxt(args.testfile+".probs", probs)

# Get predictions from probabilities
preds = np.argmax(probs, axis=1)
np.savetxt(args.testfile+".preds", preds, fmt='%d')

sys.exit(1)
