#!	/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun  7 18:50:27 2019

@author: frodo
"""

import xgboost as xgb
import sys
import os
import numpy as np

try:
	import argparse
except ImportError:
	sys.exit(0)

parser = argparse.ArgumentParser()
parser.add_argument("trainfile", help="Train dataset (libsvm format)", type=str)
parser.add_argument("testfile", help="Test dataset (libsvm format)", type=str)
parser.add_argument("classes", help="Number of classes", type=int)

args = parser.parse_args()

# Read train file
if not os.path.isfile(args.trainfile):
    print("Train dataset file '%s' does not exist" % args.trainfile)
    sys.exit()

# Read test file
if not os.path.isfile(args.testfile):
    print("Test dataset file '%s' does not exist" % args.testfile)
    sys.exit()

dtrain = xgb.DMatrix(args.trainfile, silent=True)
dtest = xgb.DMatrix(args.testfile, silent=True)

# specify parameters
param = {
    'max_depth': 5,  # the maximum depth of each tree
    'eta': 0.3,  # the training step for each iteration
    'n_estimators': 50, # number of trees
    'silent': 1,  # logging mode - quiet
    'objective': 'multi:softprob',  # error evaluation for multiclass training
    'num_class': args.classes}  # the number of classes that exist in this datset
num_round = 10  # the number of training iterations

# Train model
bst = xgb.train(param, dtrain, num_round)

# make prediction
preds = bst.predict(dtest)

# extracting most confident predictions
best_preds = np.asarray([np.argmax(line) for line in preds])

np.savetxt(args.testfile+".preds", best_preds, fmt='%d')
np.savetxt(args.testfile+".probs", preds)

sys.exit(1)
