/***************************************************************************
 *   Copyright (C) 2013 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FItnESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <typex.h>
#include "fs.h"
#include "defns.i"
#include "types.i"
#include "extern.i"
#include "macros.h"
#include "common.h"
#include "svm.h"
#include "is.h"
#include "data.h"
#include "classifiers.h"

// Margin for a classifier ensemble

scalar GetEnsembleInstanceMargin(const t_Parameters *parm, t_Classifier classifier, scalar *pattern, int label, scalar **data, int *labels, int **subspaces, int instances, int k, int classes,
                                 int inputs, t_Learner **learners, int size, scalar *alpha, bool *correct, int first)
{
    scalar votes[classes], margin, sum = 0.0;

    // Combination in the ensemble always by means of majority voting
    ZeroVector(votes, NULL, classes);

    for (int i = first; i < size; i++) {
        votes[ClassifierOutput(parm, classifier, pattern, inputs, classes, NULL, subspaces[i], data, instances, labels, NULL, learners[i], label, NULL, -1, NULL, NULL)] += alpha[i];
        sum += alpha[i];
    }

    // Get maximum number of votes
    int category = 0;
    for (int i = 1; i < classes; i++)
        if (votes[i] > votes[category])
            category = i;

    // Correctly classified?
    if (category == label) {
        *correct = true;

        int second_best = (category == 0 ? 1 : 0);

        for (int i = 0; i < classes; i++)
            if (votes[second_best] < votes[i] && i != category)
                second_best = i;

        margin = (votes[category] - votes[second_best])/sum;
    } else {
        *correct = false;

        margin = (votes[label] - votes[category])/sum;
    }

    return margin;
}


scalar GetEnsembleAccuracy(const t_Parameters *parm, t_Classifier classifier, scalar **data, int *labels, int **subspaces, int instances, int k, int classes, int inputs, t_Learner **learners,
                           int size, scalar *alpha, bool *ok, int first, bool use_g_measure, int sampling_size, int minority)
{
    scalar votes[classes], accuracy = 0.0, tp = 0.0, tn = 0.0, fp = 0.0, fn = 0.0;


    for (int j = 0; j < instances; j++) {
        // Combination in the ensemble always be means of majority voting
        for (int i = 0; i < classes; i++)
            votes[i] = 0.0;

        for (int i = first; i < size; i++)
            votes[ClassifierOutput(parm, classifier, data[j], inputs, classes, NULL, subspaces[i], data, instances, labels, NULL, learners[i], labels[j], NULL, -1, NULL, NULL)] += alpha[i];

        // Get maximum number of votes
        int category = 0;
        for (int i = 1; i < classes; i++)
            if (votes[i] > votes[category])
                category = i;

        // Correctly classified?
        if (category == labels[j]) {
            accuracy += 1.0;
            if (ok != NULL)
                ok[j] = true;
            if (labels[j] == minority)
                tp++;
            else
                tn++;
        } else {
            if (ok != NULL)
                ok[j] = false;
            if (labels[j] == minority)
                fn++;
            else
                fp++;
        }
    }

    if (use_g_measure) {
        scalar specificity = (tn + fp) > ZERO ? tn/(tn + fp) : 0.0;
        scalar sensitivity = (tp + fn) > ZERO ? tp/(tp + fn) : 0.0;

        accuracy = sqrt(specificity*sensitivity);
    } else
        accuracy /= instances;

    return accuracy;
}
