/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <stdbool.h>
#include "macros.h"
#include "svm.h"
#include "filter.h"
#include "fs.h"
#include "is.h"
#include "data.h"

// For the genetic algorithm
scalar EvaluateSubspace(const t_Parameters *parm, scalar **data, int instances, int *labels, int *subspace, scalar *ok_test, int inputs, int classes, bool use_g_measure)
{
    scalar ok = 0.0, **limits;
    int *intervals;

    switch (parm->evaluator) {
        case eIncM:
            if (use_g_measure)
                OnErrorExit(1, "Not yet implemented");
            else {
                // No training classifier
                //Memory allocation for auxiliary data structure
                MALLOC(limits, scalar *, inputs, "EvaluateSubspace: limits");
                //Set limits to NULL
                for (int k = 0; k < inputs; k++)
                    limits[k] = NULL;
                MALLOC(intervals, int, inputs, "EvaluateSubspace: intervals");

                // Discretize dataset
                switch (parm->discretizer) {
                    case dFAYYAD:
                        FayyadAndIraniMDLMethod(data, instances, classes, labels, inputs, limits, intervals, NULL, NULL);
                        break;

                    case dWPKID:
                        WPKIDMethod(data, instances, inputs, limits, intervals, NULL);
                        break;

                    case dNONE:
                        //Do nothing
                        break;
                }

                ok = InconsistCount(inputs, data, instances, subspace, labels, classes, limits, intervals);
                ok = 1 - ok;
                // If not null evaluate testing error
                if (ok_test != NULL)
                    *ok_test = 0;

                for (int k = 0; k < inputs; k++)
                    FREE(limits[k]);
                FREE(limits);
                FREE(intervals);
            }
            break;

        default:
            OnErrorExit(ERR_UNSPECTED_RESULT, "Should not be here");
    }

    return ok;
}
