/****************************************************************************
* The algorithm implemented here can be found in the book titled            *
*    FEATURE SELECTION FOR KNOWLEDGE DISCOVERY by Huan Liu &                *
* Hiroshi Motoda published by Kluwer in 1998, ISBN 0-7923-8198-X            *
*                                                                           *
* The program developed by:                                                 *
*                                                                           *
*                       Md. Farhad Hussain                                  *
*                           DISCS, NUS                                      *
*                    Kent Ridge, S'pore 119260                              *
*                  email: farhad@iscs.nus.edu.sg                            *
*                          liuh@iscs.nus.edu.sg                             *
*****************************************************************************/
#include <setjmp.h>
#include <time.h>
#include "filter.h"

/*------------------------------Global definitions---------------------------*/
long BestIncon;                   /* Contains the lowest inconsistency count */

/*---------------------------------------------------------------------------*/
TREENODE *InitRoot(int COL)
{
    TREENODE *Node;

    Node = (TREENODE *)malloc(sizeof(TREENODE));
    SUCCESS(Node);
    Node->mask = (char *)malloc(COL - 1);
    SUCCESS(Node->mask);
    IncludeAllFeatures(Node->mask, COL);
    return Node;
}
/*---------------------------------------------------------------------------*/
long BBNumOfNewSubSet(TREENODE *Node, int COL)
{

    return NumOfFeatures(Node->mask, COL);
}
/*---------------------------------------------------------------------------*/
long BBInconCal(TREENODE *Node, int ROW, scalar **DataSet, int COL)
{
    long InconCnt;

    LoadInstances(Node->mask, ROW, DataSet, COL);
    InconCnt = InconsistancyCount();
    FreeHashTable();
    return InconCnt;
}
/*---------------------------------------------------------------------------*/
void BBPrintSubSet(TREENODE *Node, int COL)
{

    DisplayFeatures(Node->mask, COL);
    printf("\n");
}
/*---------------------------------------------------------------------------*/
TREENODE *InitChildNode(TREENODE *Node, long SubSetNo, int COL)
{
    TREENODE *ChildNode;
    long FeatureNo = 0, i;

    ChildNode = (TREENODE *)malloc(sizeof(TREENODE));
    SUCCESS(ChildNode);
    ChildNode->mask = (char *)malloc(COL - 1);
    SUCCESS(ChildNode->mask);
    for (i = 0; i < COL - 1; i++) {
        ChildNode->mask[i] = Node->mask[i];

        if (ChildNode->mask[i]) {
            FeatureNo ++;
            if (FeatureNo == SubSetNo)
                ChildNode->mask[i] = 0;
        }
    }
    return ChildNode;
}
/*---------------------------------------------------------------------------*/
void FreeNode(TREENODE *Node)
{

    free(Node->mask);
    free(Node);
}
/*---------------------------------------------------------------------------*/
BOOL B_and_B(TREENODE *Node, int ROW, scalar **DataSet, int COL, jmp_buf env_buffer, time_t time_limit, time_t init_time)
{
    long Incon, NumOfSubSet, SubSetNo, FailCount = 0;
    TREENODE *ChildNode;
    extern TREENODE *ResultNode;

    if ((!time_limit || (time(NULL) - init_time < time_limit)) &&
        (Incon = BBInconCal(Node, ROW, DataSet, COL)) <= BestIncon && (NumOfSubSet = BBNumOfNewSubSet(Node, COL))) {

        BestIncon = Incon;
        /* Run B and B for its child nodes */

        for (SubSetNo = 1; SubSetNo <= NumOfSubSet; SubSetNo++) {
            ChildNode = InitChildNode(Node, SubSetNo, COL);
            if (B_and_B(ChildNode, ROW, DataSet, COL, env_buffer, init_time, time_limit) == FAIL)
                FailCount ++;
            FreeNode(ChildNode);
        }

        if (FailCount == NumOfSubSet || (time_limit && (time(NULL) - init_time > time_limit))) {  /* All children node have higher Beta */
            ResultNode = Node;
            longjmp(env_buffer, 1);
        }
    } else {
        ResultNode = Node;
        return FAIL;
    }

    ResultNode = Node;
    return TRUE;
}
