/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef _FILTER_H
#define _FILTER_H

#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <setjmp.h>
#include "uthash.h"
#include "fs.h"
#include "typex.h"
#include "classifiers.h"

#define FOREVER for( ; ; )
#define STRBUF 256                             /* Default string buffer size */
#define SEPARATOR(a) (a == ' ' || a == ',') ? 1 : 0
#define SUCCESS(a) if(a == NULL) { printf("\nUnsuccessful Allocation"); exit(0);}

#define HASHTBLSIZE 101

#define BOOL char
#define FAIL 0
#define TRUE 1
#define FALSE 0

typedef struct tupleclass {
    long ClassValue;                            /* Class of the instance */
    long ClassCount;                       /* Frequency within the class */
    struct tupleclass *Next;               /* Ptr to the next TUPLECLASS */
} TUPLECLASS;

typedef struct instance {
    long tuple;             /* Contains index of the instance in DataSet */
    long ClassCount;             /* Number of distinct class of instance */
    struct instance *Down;               /* Pointer to the next instance */
    TUPLECLASS *Right;                     /* Pointer to the tuple class */
} INSTANCE;

typedef struct hashtable {
    INSTANCE *Down;                     /* Pointer to the first instance */
    long Count;                           /* Number of distinct instance */
} HASHTABLE;

typedef struct treenode {
    char *mask;                    /* Contains attributes for evaluation */
} TREENODE;

typedef struct attriblist {
    char *mask;                      /* Contains selected attributes */
    struct attriblist *Next;                  /* Next selected attribute */
} ATTRIBLIST;

// Hash table functions
typedef struct my_struct {
    char *id;                    /* key */
    int *labels;
    UT_hash_handle hh;         /* makes this structure hash table */
} t_Hash;

// Prototypes
scalar ClassEntropy(int, int *, int );
scalar Entropy(scalar **, int, int, scalar **, int);
scalar ConditionalClassEntropy(scalar **, int, int, scalar **, int, int, int *);
scalar ConditionalEntropy(scalar **, int, int, int, scalar **, int, int);
void FocusAlg(const t_Parameters *, int *, int, scalar **, int, int *, int, time_t);
void FocusSF(const t_Parameters *, int *, int, scalar **, int, int *, int);
scalar InconsistCount(int, scalar **, int, int *, int *, int, scalar **, int *);
void add_element(t_Hash **ind, char *user_id, int, int, int);
t_Hash *find_element(t_Hash *ind, char *user_id);
int get_inconsistent_elements(t_Hash *ind, int);
void delete_element(t_Hash **ind, t_Hash *);
void delete_all(t_Hash **ind);
void print_elements(t_Hash *ind, FILE *, int);
void LvfAlg(const t_Parameters *, int *, int, scalar **, int, int *, int, int);
void ReliefAlg(const t_Parameters *, int *, int, scalar **, int, int *, int , int, int, scalar, scalar, t_RankInfo *);
void RFEFS(const t_Parameters *, t_Feat *, int, scalar **, int, int *);
void CorrelationBasedFilterSequentialForwardSelection(const t_Parameters *, scalar **, int, int, int, int *, int *);
void UnsupervisedForwardSelection(const t_Parameters *, scalar **, int, int, int *);
void FastclusteringBasedFeatureSubsetSelection(const t_Parameters *, scalar **, int, int, int, int *, int *, scalar);
void FastCorrelationBasedFilter(const t_Parameters *, scalar **, int, int, int, int *, int *, scalar);
void SetCover(const t_Parameters *, int *S, int, scalar **, int, int *, int);
void QuickBranchAndBound(const t_Parameters *, int *, int, scalar **, int, int *, int);
void BranchAndBound(scalar **, int, int *, int *, int, time_t);
void KolmogorovSmirnovCorrelationBasedFilter(const t_Parameters *, scalar **, int, int, int, int *, int *);
int CompareDescending(const void *, const void *);
void Prim(scalar **, t_Edge **, int, int *);
scalar MeritS(scalar **, int *, int, int *, int, t_EntropyInfo *, int *, scalar **, t_Discretizer);
scalar FeatureClassCorrelation(scalar **, int, int *, int);
scalar FeatureFeatureCorrelation(scalar **, int, int, int);
scalar EvaluateUsingMahalanobisDistance(int, int, int *, scalar **, scalar **);
void CovarianceMatrix(scalar **, int, int, scalar **);
scalar MaskedMahalanobisDistance(scalar *, scalar *, scalar **, int *, int);
scalar ColumnVectorDotProduct(scalar **, int, scalar *, int);
scalar ColumnLength(scalar **, int, int);
scalar LengthOrtogonalProjection(scalar **, int, scalar **, int, int);
int CompareScore(const void *, const void *);

// Adaptive Branch and Bound
void GenerateRandomFeatures(char *, unsigned long, int);      /* Generates feature subset */
void InitRandomize(unsigned long);                      /* Initializes random seed */
void FreeDataBuffer();                  /* Release the DataSet buffer */
void CreateHashTableBB();
long InconsistancyCount();
long NumOfFeatures(char *, int);
void DisplayFeatures(char *, int);
void FreeHashTable();
void IncludeAllFeatures(char *, int);
void LoadInstances(char *, int, scalar **, int);
long GetHashIndex(long, char *, scalar **, int);
long GetClassValue(long, scalar **, int);
void InsertInstanceInHashTable(long, char *, scalar **, int);
void AttachClass(INSTANCE *, long);
void InsertInstanceAtLoc(long, long, char *, scalar **, int);
char TupleMatch(INSTANCE *, long, char *, scalar **, int);
void DisplayHashTable();
long GetNumberofAttributes(char *);
long GetNumberofTuples(char *);
void AllocateDataSpace(long, long);
void GetDataIntoBuffer(char *);
void DisplayDataBuffer();
void InitABB(int, scalar **, int);
char *GetChildSubSet(ATTRIBLIST *, long, int);
long HammingDistanceCharCharChar(char *, char *, int);
void InsertInPrunedList(char *, int);
char InsertInChildLevel(char *, int, scalar **, int);
BOOL Legitimated(char *, int);
long NumOfNewSubSet(ATTRIBLIST  *, int);
long InconCal(char *mask, int, scalar **, int);
void PrintSubSet(ATTRIBLIST  *, int);
void GenerateChildSubSets(int, scalar **, int);
void PrintParentList(int);
void PrintPrunedList(int);
void RunABB(scalar **, int, int, time_t);
void PrintChildList(int);
void AutomaticBranchAndBound(scalar **, int, int *, int *, int, time_t);

// Branch and Bound
TREENODE *InitRoot(int);
TREENODE *InitChildNode(TREENODE *, long, int);
long BBNumOfNewSubSet(TREENODE *, int);
long BBInconCal(TREENODE *, int, scalar **, int);
void BBPrintSubSet(TREENODE *, int);
BOOL B_and_B(TREENODE *, int, scalar **, int, jmp_buf, time_t, time_t);
void FreeNode(TREENODE *);
TREENODE *InitRoot(int);

/* floating.c */
void SequentialForwardFloatingSelection(const t_Parameters *, scalar **, int, int, int, int *, int *);
void SequentialBackwardFloatingSelection(const t_Parameters *, scalar **, int, int, int, int *, int *);
scalar EvaluateUsingMahalanobisDistance(int, int, int *, scalar **, scalar **);
void CovarianceMatrix(scalar **, int, int, scalar **);
scalar MaskedMahalanobisDistance(scalar *, scalar *, scalar **, int *, int);

/* mitra.c */
void MaximalInformationCompressionIndexBasedMethod(int *, scalar **, int, int, int *);
void MaximalInformationCompressionIndex(scalar **, scalar **, int, int);
void GetKNeighbors(int, scalar **, int, int, int *, scalar *);

/* g-flip.c */
scalar InstanceMargin(scalar *, int, scalar **, int, int *, int, int *, scalar *);
scalar MarginBasedSubsetEvaluation(int *, scalar **, int, int, int *, scalar *);
void GreedyFeatureFlip(const t_Parameters *, scalar **, int, int, int *, int *, int);

/* isimba.c */
void ImprovedIterativeSearchMarginBasedAlgorithm(const t_Parameters *, scalar **, int, int, int *, t_RankInfo *, int *, int);
void CenterMissHit(scalar **, int, int, int, int *, scalar **, scalar **);

#endif /*_FILTER_H */
