#include "filter.h"
#include "fs.h"

/*---------------------------Global variables--------------------------------*/
HASHTABLE *HashTable;                    /* Global pointer to the hash table */


/*---------------------------------------------------------------------------*/
void CreateHashTableBB()
{
    long i;

    HashTable = (HASHTABLE *)malloc(HASHTBLSIZE * sizeof(HASHTABLE));
    SUCCESS(HashTable);
    for (i = 0 ; i < HASHTBLSIZE; i++) {
        HashTable[i].Down = NULL;
        HashTable[i].Count = 0;
    }
}
/*---------------------------------------------------------------------------*/
long GetHashIndex(long instance, char *mask, scalar **DataSet, int COL)
{
    long i, index = 0, wt = 1;

    for (i = 0; i < COL - 1; i ++) {
        if (mask[i])
            index = index + abs((long)DataSet[instance][i] * wt++);
    }
    return index % HASHTBLSIZE;
}
/*---------------------------------------------------------------------------*/
long GetClassValue(long instance, scalar **DataSet, int COL)
{

    return (long)DataSet[instance][COL - 1];
}
/*---------------------------------------------------------------------------*/
void InsertInstanceInHashTable(long instance, char *mask, scalar **DataSet, int COL)
{
    long index;
    //     INSTANCE *instanceloc;

    index = GetHashIndex(instance, mask, DataSet, COL);
    InsertInstanceAtLoc(index, instance, mask, DataSet, COL);
}
/*---------------------------------------------------------------------------*/
void AttachClass(INSTANCE *InstPTR, long ClassValue)
{
    TUPLECLASS *TupleClassPTR, *TupleFwdClassPTR = NULL, *TupleBckClassPTR = NULL;

    InstPTR->ClassCount ++;

    if (InstPTR->Right == NULL) {                     /* No class attached yet */
        TupleClassPTR = (TUPLECLASS *)malloc(sizeof(TUPLECLASS));
        SUCCESS(TupleClassPTR);
        TupleClassPTR->ClassCount = 1;
        TupleClassPTR->ClassValue = ClassValue;
        TupleClassPTR->Next = NULL;
        InstPTR->Right = TupleClassPTR;
        return;
    }

    TupleFwdClassPTR = InstPTR->Right->Next;
    TupleBckClassPTR = InstPTR->Right;
    while (TupleFwdClassPTR != NULL &&
           TupleBckClassPTR->ClassValue > ClassValue) {
        TupleFwdClassPTR = TupleFwdClassPTR->Next;
        TupleBckClassPTR = TupleBckClassPTR->Next;
    }

    if (TupleBckClassPTR->ClassValue == ClassValue) {         /* Class matched */
        TupleBckClassPTR->ClassCount++;
        InstPTR->ClassCount --;                           /* Duplicate class */
        return;
    }

    if (TupleBckClassPTR->ClassValue < ClassValue &&
        TupleBckClassPTR == InstPTR->Right) {
        TupleClassPTR = (TUPLECLASS *)malloc(sizeof(TUPLECLASS));
        SUCCESS(TupleClassPTR);
        TupleClassPTR->ClassCount = 1;
        TupleClassPTR->ClassValue = ClassValue;
        TupleClassPTR->Next = TupleBckClassPTR;
        InstPTR->Right = TupleClassPTR;
        return;
    }

    if (TupleBckClassPTR->ClassValue > ClassValue) {
        TupleClassPTR = (TUPLECLASS *)malloc(sizeof(TUPLECLASS));
        SUCCESS(TupleClassPTR);
        TupleClassPTR->ClassCount = 1;
        TupleClassPTR->ClassValue = ClassValue;
        TupleClassPTR->Next = NULL;
        TupleBckClassPTR->Next = TupleClassPTR;
        return;
    }

    TupleFwdClassPTR = TupleBckClassPTR;
    TupleBckClassPTR = InstPTR->Right;
    while (TupleBckClassPTR->Next != TupleFwdClassPTR)
        TupleBckClassPTR = TupleBckClassPTR->Next;

    TupleClassPTR = (TUPLECLASS *)malloc(sizeof(TUPLECLASS));
    SUCCESS(TupleClassPTR);
    TupleClassPTR->ClassCount = 1;
    TupleClassPTR->ClassValue = ClassValue;
    TupleClassPTR->Next = TupleFwdClassPTR;
    TupleBckClassPTR->Next = TupleClassPTR;
    return;
}
/*---------------------------------------------------------------------------*/
void InsertInstanceAtLoc(long index, long instance, char *mask, scalar **DataSet, int COL)
{
    INSTANCE *InstFwdPTR, *InstBckPTR, *InstPTR;

    HashTable[ index ].Count ++;                        /* Adding new instance */

    if (HashTable[ index ].Down == NULL) {       /* No entry in the hash table */
        HashTable[ index ].Down = (INSTANCE *)malloc(sizeof(INSTANCE));
        SUCCESS(HashTable[index].Down);
        InstPTR = HashTable[ index ].Down;
        InstPTR->ClassCount = 0;
        InstPTR->Right = NULL;
        InstPTR->Down = NULL;
        InstPTR->tuple = instance;
        AttachClass(InstPTR, GetClassValue(instance, DataSet, COL));
        return;
    }

    InstFwdPTR = HashTable[ index ].Down->Down;
    InstBckPTR = HashTable[ index ].Down;
    while (InstFwdPTR != NULL && !TupleMatch(InstBckPTR, instance, mask, DataSet, COL)) {
        InstFwdPTR = InstFwdPTR->Down;
        InstBckPTR = InstBckPTR->Down;
    }

    if (TupleMatch(InstBckPTR, instance, mask, DataSet, COL)) {
        AttachClass(InstBckPTR, GetClassValue(instance, DataSet, COL));
        HashTable[ index ].Count --;                  /* Duplicate  instance */
        return;
    }

    InstPTR = (INSTANCE *)malloc(sizeof(INSTANCE));
    SUCCESS(InstPTR);
    InstPTR->ClassCount = 0;
    InstPTR->Right = NULL;
    InstPTR->Down = NULL;
    InstPTR->tuple = instance;
    AttachClass(InstPTR, GetClassValue(instance, DataSet, COL));
    InstBckPTR->Down = InstPTR;

    return;
}
/*---------------------------------------------------------------------------*/
char TupleMatch(INSTANCE *InstPTR, long Inst2, char *mask, scalar **DataSet, int COL)
{
    long Inst1, i;

    Inst1 = InstPTR->tuple;
    for (i = 0 ; i < COL - 1; i++) {
        if (mask[i] && DataSet[Inst1][i] != DataSet[Inst2][i])
            return 0;
    }
    return 1;
}
/*---------------------------------------------------------------------------*/
void DisplayHashTable()
{
    INSTANCE *InstPTR;
    TUPLECLASS *ClassPTR;
    long i, count, Total;

    Total = 0;
    for (i = 0; i < HASHTBLSIZE; i++) {
        InstPTR = HashTable[i].Down;
        while (InstPTR != NULL) {
            count = 0;
            ClassPTR = InstPTR->Right;
            while (ClassPTR != NULL) {
                count = count + ClassPTR->ClassCount;
                ClassPTR = ClassPTR->Next;
            }
            Total = Total + count;
            printf("\n Instance %ld exists %ld times", InstPTR->tuple, count);
            InstPTR = InstPTR->Down;
        }
    }
    printf("\nTotal %ld Instances In The Hash Table", Total);
}
/*---------------------------------------------------------------------------*/
long InconsistancyCount()
{
    INSTANCE *InstPTR;
    TUPLECLASS /**ClassPTR,*/ *InconPTR;
    long i, count, LargestCnt;

    count = 0;
    for (i = 0; i < HASHTBLSIZE; i++) {
        InstPTR = HashTable[i].Down;
        while (InstPTR != NULL) {
            InconPTR = InstPTR->Right;
            LargestCnt = 0;
            while (InconPTR != NULL) {
                count = count + InconPTR->ClassCount;
                if (InconPTR->ClassCount > LargestCnt)
                    LargestCnt = InconPTR->ClassCount;
                InconPTR = InconPTR->Next;
            }
            count = count - LargestCnt;
            InstPTR = InstPTR->Down;
        }
    }
    return count;
}
/*---------------------------------------------------------------------------*/
long NumOfFeatures(char *mask, int COL)
{
    long i, count = 0;

    for (i = 0; i < COL - 1; i++)
        if (mask[i])
            count++;
    return count;
}
/*---------------------------------------------------------------------------*/
void DisplayFeatures(char *mask, int COL)
{
    long i, count = 0;

    printf("\n");
    for (i = 0; i < COL - 1; i++)
        if (mask[i]) {
            printf("1 ");
            count++;
        } else
            printf("0 ");
    printf(" Cnt = %ld", count);
}
/*---------------------------------------------------------------------------*/
void FreeClassMemory(INSTANCE *InstPTR)
{
    TUPLECLASS *ClassPTR, *NextClassPTR;
    long i = 0;

    ClassPTR = InstPTR->Right;
    for (i = 0; i < InstPTR->ClassCount; i ++) {
        NextClassPTR = ClassPTR->Next;
        free(ClassPTR);
        ClassPTR = NextClassPTR;
    }
}
/*---------------------------------------------------------------------------*/
void FreeHashTable()
{
    INSTANCE *InstPTR, *NextInstPTR;
    long i = 0, j = 0;

    for (i = 0; i < HASHTBLSIZE; i++) {
        if ((InstPTR = HashTable[i].Down) != NULL) {
            InstPTR = HashTable[i].Down;
            for (j = 0; j < HashTable[i].Count; j++) {
                NextInstPTR = InstPTR->Down;
                FreeClassMemory(InstPTR);
                free(InstPTR);
                InstPTR = NextInstPTR;
            }
            HashTable[i].Down = NULL;
            HashTable[i].Count = 0;
        }
    }
}
/*---------------------------------------------------------------------------*/
void IncludeAllFeatures(char *mask, int COL)
{
    long i;

    for (i = 0; i < COL - 1; i++)
        mask[ i ] = 1;

}
/*---------------------------------------------------------------------------*/
void LoadInstances(char *mask, int ROW, scalar **DataSet, int COL)
{
    long inst;

    for (inst = 0; inst < ROW; inst++)
        InsertInstanceInHashTable(inst, mask, DataSet, COL);
}
/*---------------------------------------------------------------------------*/
