/*
 * Strategies of selection of individuals from the population.
 */
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <time.h>
#include <string.h>
#include "global.h"
#include "macros.h"
#include <math.h>
#include "filter.h"


void add_element(t_Hash **ind, char *user_id, int label, FILE *f)
{
    t_Hash *s, *find;

    find = find_element(*ind, user_id);
    //If element_id is new: create its entry in the hash table
    if (find == NULL)  {
        //fprintf(f, "New element in hash table, id:  %s, label: %d\n", user_id, label);
        s = malloc(sizeof(t_Hash));
        MALLOC(s->id, char, (gl_inputs * 3 + 1), "InconsistCount: add_element");
        CALLOC(s->labels, int, gl_outputs, "InconsistCount: add_element");
        strcpy(s->id, user_id);
        HASH_ADD_KEYPTR(hh, *ind, s->id, strlen(s->id), s);
        s->labels[label]++;
        //for (int j = 0; j < gl_outputs; j++)
        //fprintf(f, "%d ", s->labels[j]);
    }
    //If element_id is already in the hash table: increase the counter of its label
    else  {
        //fprintf(f, "New label %d in previous element id: %s", label, find->id);
        find->labels[label]++;
        //fprintf(f, "\n");
        //for (int j = 0; j < gl_outputs; j++)
        //fprintf(f, "%d ", find->labels[j]);
        //printf("\nHash table composed by:\n");
        //print_elements(*ind);
    }
}

t_Hash *find_element(t_Hash *ind, char *user_id)
{
    t_Hash *s;

    HASH_FIND_STR(ind, user_id, s);    /* s: output pointer */
    return s;
}

int get_inconsistent_elements(t_Hash *ind)
{
    t_Hash *s;
    int max = 0, acc_max = 0;

    for (s = ind; s != NULL; s = s->hh.next) {
        max = 0;
        //printf("\nuser id -> %s: ", s->id);
        for (int j = 0; j < gl_outputs; j++)  {
            //printf("%d ", s->labels[j]);
            //Get the majority class for the previous pattern
            if (max < s->labels[j])
                max = s->labels[j];
        }
        //printf("\nMaximum: %d", max);
        //Addition of the number of instances classified in the majority class
        acc_max += max;
    }
    return acc_max;
}

void delete_element(t_Hash **ind, t_Hash *user)
{
    HASH_DEL(*ind, user);     /* user: pointer to delete */
    free(user->id);
    free(user->labels);
    free(user);
}

void delete_all(t_Hash **ind)
{
    t_Hash *current_user;

    while (*ind) {
        current_user = *ind;          /* grab pointer to first item */
        HASH_DEL(*ind,current_user);     /* delete it (ind advances to next) */
        free(current_user->id);
        free(current_user->labels);
        free(current_user);            /* free it */
    }
}

void print_elements(t_Hash *ind, FILE *f)
{
    t_Hash *s;

    for (s=ind; s != NULL; s=s->hh.next) {
        //fprintf(f, "\nuser id -> %s: ", s->id);
        for (int j = 0; j < gl_outputs; j++)  {
            fprintf(f, "%d ", s->labels[j]);
        }
    }
}
