/***************************************************************************
 *   Copyright (C) 2009 by Aida de Haro-García                             *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/*
 * Strategies of selection of individuals from the population.
 */
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <time.h>
#include <string.h>
#include "macros.h"
#include <math.h>
#include "filter.h"


scalar InconsistCount(int n_features, scalar **data, int n, int *sel_feat, int *labels, int classes, scalar **limits, int *intervals)
{
    int i, k, aux_index = 0, acc_max = 0;
    scalar incons_res = 0.0;
    char feat_ind[32], *indx = NULL;
    //Create hash table
    t_Hash *ind = NULL;

    for (k = 0; k < n_features; k++) {
        // Check if too many intervals and exit
        if (intervals[k] >= 10000) {
            printf("%d intervals for feature %d\n", intervals[k], k+1);
            OnErrorExit(1, "InconsistCount: Too many intervals\n");
        }
    }

    MALLOC(indx, char, (n_features * 4 + 1), "InconsistCount");

    for (i = 0; i < n; i++)  {
        bzero((void *) indx, sizeof(char) * (n_features * 4 + 1));
        for (k = 0; k < n_features; k++) {
            aux_index = 0;
            if (sel_feat[k])  {
                //Search for the first interval of less value than data itself
                while (aux_index < (intervals[k]) && (data[i][k] > limits[k][aux_index]))
                    aux_index++;

                sprintf(feat_ind, "%04d", aux_index);

                strcat(indx, feat_ind);
            }
        }
        add_element(&ind, indx, labels[i], n_features, classes);
    }

    //Get inconsistency measure
    acc_max = get_inconsistent_elements(ind, classes);
    incons_res = (n - acc_max)/(n * 1.0);

    //Free memory
    delete_all(&ind);
    FREE(indx);

    return incons_res;
}

void add_element(t_Hash **ind, char *user_id, int label, int n_features, int classes)
{
    t_Hash *s, *find;

    find = find_element(*ind, user_id);
    //If element-id is new: create its entry in the hash table
    if (find == NULL)  {
        MALLOC(s, t_Hash, 1, "s: add_element");
        MALLOC(s->id, char, (n_features * 4 + 1), "s->id: add_element");
        CALLOC(s->labels, int, classes, "s->labels: add_element");
        strcpy(s->id, user_id);
        HASH_ADD_KEYPTR(hh, *ind, s->id, strlen(s->id), s);
        s->labels[label]++;
    } else
        find->labels[label]++;
}

t_Hash *find_element(t_Hash *ind, char *user_id)
{
    t_Hash *s;

    HASH_FIND_STR(ind, user_id, s);    /* s: output pointer */
    return s;
}

int get_inconsistent_elements(t_Hash *ind, int classes)
{
    t_Hash *s;
    int max = 0, acc_max = 0;

    for (s = ind; s != NULL; s = s->hh.next) {
        max = 0;
        for (int j = 0; j < classes; j++)  {
            //Get the majority class for the previous pattern
            if (max < s->labels[j])
                max = s->labels[j];
        }
        //Addition of the number of instances classified in the majority class
        acc_max += max;
    }
    return acc_max;
}

void delete_element(t_Hash **ind, t_Hash *user)
{
    HASH_DEL(*ind, user);     /* user: pointer to delete */
    free(user->id);
    free(user->labels);
    free(user);
}

void delete_all(t_Hash **ind)
{
    t_Hash *current_user;

    while (*ind) {
        current_user = *ind;          /* grab pointer to first item */
        HASH_DEL(*ind,current_user);     /* delete it (ind advances to next) */
        free(current_user->id);
        free(current_user->labels);
        free(current_user);            /* free it */
    }
}

void print_elements(t_Hash *ind, FILE *f, int classes)
{
    t_Hash *s;

    for (s=ind; s != NULL; s=s->hh.next) {
        for (int j = 0; j < classes; j++)  {
            fprintf(f, "%d ", s->labels[j]);
        }
    }
}
