/* THIS FILE MAINLY CONTAINS THE CODE FOR READING DATA FROM AN EXTERNAL DATA
                      FILE AND GENERATING RANDOM NUMBERS                     */

#include "fs.h"
#include "filter.h"

#define IB1 1                            /* Definitions for GetChangedSeed() */
#define IB2 2
#define IB5 16
#define IB18 131072

#define MBIG 1000000000                           /* Definitions for rand3() */
#define MSEED 161803398
#define MZ 0
#define FAC (1.0/MBIG)


/*----------------------------- Global variables ----------------------------*/
// scalar **DataSet;                                           /* dataset buffer */
// long ROW, COL;                      /* Number of row and column in a dataset */
// unsigned long Seed;                      /* Seed for random value generation */

/*---------------------------------------------------------------------------*/
void InitRandomize(unsigned long Seed)
{

    Seed = time(NULL);
}
/*---------------------------------------------------------------------------*/
int GetChangedSeed(unsigned long *seed)
{
    unsigned long newbit;

    newbit = (*seed & IB18) >> 17                                /* Get bit 18 */
             ^ (*seed & IB5) >> 4                                   /* XOR with bit 5 */
             ^ (*seed & IB2) >> 1                                   /* XOR with bit 2 */
             ^ (*seed & IB1);                                       /* XOR with bit 1 */
    *seed = (*seed << 1) | newbit;     /* Leftshit the seed and put the result */
    return (int) newbit;                          /* of the XOR's in its bit 1 */
}
/*---------------------------------------------------------------------------*/
/*---------------------Generates Random feature set--------------------------*/
/*---------------------------------------------------------------------------*/
void GenerateRandomFeatures(char *mask, unsigned long Seed, int COL)
{
    long i;

    for (i = 0; i < COL - 1; i++) {
        if (GetChangedSeed(&Seed))
            mask[ i ] = 1;
        else
            mask[ i ] = 0;
    }
}
/*---------------------------------------------------------------------------*/
scalar ran3(long *idum)
{
    static int inext, inextp;
    static long ma[56];
    static int iff=0;
    long mj,mk=0;
    int i,ii,k;

    if (*idum < 0 || iff == 0) { /* initialization */
        iff=1;
        mj=MSEED-(*idum < 0 ? -*idum : *idum);
        mj %= MBIG;
        ma[55]=mj;
        for (i=1; i<=54; i++) {
            ii=(21*i) % 55;
            ma[ii]=mk;
            mk=mj-mk;
            if (mk < MZ) mk += MBIG;
            mj=ma[ii];
        }
        for (k=1; k<=4; k++)
            for (i=1; i<=55; i++) {
                ma[i] -= ma[1+(i+30) % 55];
                if (ma[i] < MZ) ma[i] += MBIG;
            }
        inext=0;
        inextp=31;
        *idum=11;
    }
    if (++inext == 56) inext=1;
    if (++inextp == 56) inextp=1;
    mj=ma[inext]-ma[inextp];
    if (mj < MZ) mj +=MBIG;
    ma[inext]=mj;
    return mj*FAC;
}

/*---------------------------------------------------------------------------*/
/*-------------------------Random Function Generator-------------------------*/
/*---------------------------------------------------------------------------*/
long GetRandomNumber(long MaxRange)
{
    long value = 786;

    return (long)(ran3(&value) * MaxRange);
}
/*---------------------------------------------------------------------------*/
/*--------------------Reads data from the file-------------------------------*/
/*---------------------------------------------------------------------------*/
// ReadData(char *filename)
// {
//     FILE *fp;
//
//     /* checking file existance */
//
//     if ((fp = fopen(filename, "r")) == NULL) {
//         printf("\n %s not found in this path", filename);
//         exit(0);
//     }
//     fclose(fp);
//
//     /* get number of tuples and attrib  from the data file*/
//
//     COL = GetNumberofAttributes(filename);
//     ROW = GetNumberofTuples(filename);
//
//     /* allocate memory for the data set in the DataSet buffer */
//
//     AllocateDataSpace(ROW, COL);
//
//     /* Collect data into DataSet buffer from the file */
//
//     GetDataIntoBuffer(filename);
// }
/*---------------------------------------------------------------------------*/
// long GetNumberofTuples(char *filename)
// {
//     FILE *pfp;
//     char command[STRBUF];
//
//     sprintf(command, "cat %s | wc -l", filename);    /* use wc system command  */
//     pfp = popen(command, "r");                       /* open a pipe to read    */
//     fgets(command, STRBUF, pfp);                 /* get the number of line */
//     pclose(pfp);
//     return atol(command);
// }
/*---------------------------------------------------------------------------*/
// long GetNumberofAttributes(char *filename)
// {
//     long i, field;
//     FILE *fp;
//     char buffer[STRBUF];
//
//     fp = fopen(filename, "r");
//     fgets(buffer, STRBUF, fp);
//     fclose(fp);
//     field = 0;
//     for (i = strlen(buffer); i; i--) {
//         if (SEPARATOR(buffer[i]))
//             field ++;
//     }
//     return field + 1;                    /* number of field is = separator + 1 */
// }
/*---------------------------------------------------------------------------*/
// void AllocateDataSpace(long row, long col)
// {
//
//     DataSet = malloc(sizeof(scalar *) * row);
//     SUCCESS(DataSet);
//     while (row) {
//         DataSet[ row - 1] = (scalar *)malloc(sizeof(scalar) * col);
//         SUCCESS(DataSet[ row - 1 ]);
//         row --;
//     }
// }
/*---------------------------------------------------------------------------*/
// void GetDataIntoBuffer(char *database)
// {
//     FILE *fp;
//     long tuples, field, i, j;
//     char buffer[STRBUF], tempBuffer[STRBUF];
//
//
//     fp = fopen(database, "r");
//
//     for (tuples = 0; tuples < ROW; tuples ++) {
//         fgets(buffer, STRBUF, fp);                      /* collect one tuple */
//         i = 0;
//         for (field = 0; field < COL; field ++) {     /* collect attrib value */
//             strcpy(tempBuffer, buffer);
//             for (; i < strlen(buffer) && ! SEPARATOR(buffer[i]); i++);
//             tempBuffer[i++] = '\0';
//             for (j = strlen(tempBuffer)-1; j >= 0 &&
//                  ! SEPARATOR(tempBuffer[j]); j--);
//
//             DataSet[tuples][field] = atof(tempBuffer + j + 1);
//         }
//     }
//     fclose(fp);
// }
/*---------------------------------------------------------------------------*/
// void FreeDataBuffer()
// {
//
//     free(DataSet);
// }
/*---------------------------------------------------------------------------*/
// void DisplayDataBuffer()
// {
//     int i = ROW;
//     int j = COL;
//
//     for (i = 0 ; i < ROW; i ++) {
//         printf("%f", DataSet[i][0]);
//         for (j = 1; j < COL; j ++)
//             printf(",%f", DataSet[i][j]);
//         printf("\n");
//     }
// }
/*---------------------------------------------------------------------------*/
