/***************************************************************************
 *   Copyright (C) 2015 by Nicolás García-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "fs.h"
#include "filter.h"

/*
 * M. Dash, H. Liu, Hybrid search of feature subsets, in: Proceedings of Pacific Rim International Conference on Artificial Intelligence (PRICAI-98), Singapore,
 * 1998, pp. 238–249.
 */

void QuickBranchAndBound(const t_Parameters *parm, int *S, int n_features, scalar **data, int instances, int *labels, int classes)
{
    // Apply LVF algorithm
    LvfAlg(parm, S, n_features, data, instances, labels, classes, parm->lvf_iter);

    // Reduction for LVF
    int selected = 0;
    for (int i = 0; i < n_features; i++)
        if (S[i])
            selected++;
    fprintf(parm->out_file, "Reduction for LVF: %lf\n", 1.0 - ((scalar) selected)/n_features);

    // Apply ABB algorithm
    AutomaticBranchAndBound(data, instances, labels, S, n_features, parm->time_limit);

    // Reduction for ABB
    selected = 0;
    for (int i = 0; i < n_features; i++)
        if (S[i])
            selected++;
    fprintf(parm->out_file, "Reduction for ABB: %lf\n", 1.0 - ((scalar) selected)/n_features);
}
