/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR  PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#pragma once

// Definitions and prototypes
#include <stdio.h>
#include <time.h>
#include <stdbool.h>
#include "typex.h"
#include "common.h"
#include "defns.i"
#include "types.i"
#include "svm.h"
#include "classifiers.h"

// #define AUTOTEST     // Set auto-test code
// #define DEBUG        // Set debugging code
// #define VERY_VERBOSE

#define ZERO        0.000001
#define ZERO2       0.000000000001

// Structure for ordering the
typedef struct {
    int instance;
    scalar w;
} t_WeightInfo;

// Prototypes
/* boost.c */
scalar UpdateAdaBoostWeightsFS(const t_Parameters *, int *, scalar **, int, int *, scalar *, scalar *, int, int, bool, int, int);
void UpdateGentleBoostWeightsFS(const t_Parameters *, int *, scalar **, int, int *, scalar *, int, int, bool, int);
scalar UpdateeReweightBoostWeightsFS(const t_Parameters *, int **, scalar **, int, int *, scalar *, scalar *, int, int, int, int, t_Learner *, bool, int);
/* boostedfs.c */
int MarginBasedFeatureSelection(const t_Parameters *, int *selection, scalar **, int, int *, int, int, int, bool, int **);
scalar SubspaceUnweightedPerformance(const t_Parameters *, t_Classifier, int *, int *, int, int, scalar **, int, t_PerformanceMeasure, int, scalar **, int, int *, bool, int *, t_WeightInfo *, int);
int ConditionalExclusionMarginBasedFS(const t_Parameters *, int *, int, scalar **, int, int, int *, int, int, scalar **, int, int *, bool, int *, scalar **, int *, int, t_WeightInfo *, int, int **);
scalar GetFilterWeightedError(int, int *, bool *, scalar *, int *);
void NearestNeighborWithSemiCache(scalar **, int *, int, int, int *, scalar **, int *, int, int, int);
bool MarginBasedFeatureSelectionStop(const t_Parameters *, int, int, int, int, int);
int CompareWeight(const void *, const void *);
void NearestNeighborWithSemiCacheWihtWeights(scalar **, int *, int, int, int *, scalar **, int *, int, int, t_WeightInfo *, int, int, bool);
/* cascade.c */
int BalanceCascadeRemoveInstances(const t_Parameters *, int **, scalar **, int, int *, int, int, int, bool, int *, int *, scalar **, int *, scalar *, scalar *);
int GetEnsembleCategory(t_Classifier, scalar **, scalar *, int *, int, int **, int, int, int, struct svm_model **, Tree *, int, int, int, scalar *);
/* cache.c */
void UpdateCache(scalar **, int, int, scalar **, int, int **);
void ReserveAndInitializeCache(scalar ***, int);
void InitializeCacheToActualSqDistance(scalar **, int, scalar **, int, int **);
/* .c */
scalar GetWeightedErrorFS(const t_Parameters *, t_Classifier, int *, scalar **, int, int, int *, int, bool *, scalar *, bool, t_Learner *, int, scalar *, scalar **, int, int *, t_Ensemble *, bool);
/* ensemble.c */
scalar GetEnsembleInstanceMargin(const t_Parameters *, t_Classifier, scalar *, int, scalar **, int *, int **, int, int, int, int, t_Learner **, int, scalar *, bool *, int);
scalar GetEnsembleAccuracy(const t_Parameters *, t_Classifier, scalar **, int *, int **, int, int, int, int, t_Learner **, int, scalar *, bool *, int, bool, int, int);
/* evaluation.c */
scalar EvaluateSubspace(const t_Parameters *, scalar **, int, int *, int *, scalar *, int, int, bool);
/* floatboost.c */
scalar UpdateeFloatBoostWeightsFS(const t_Parameters *, int **, scalar **, int, int *, scalar *, scalar *, int, int, int, bool, t_Learner **, int);
int ConditionalExclusionFS(t_Parameters *, int **, int, scalar **, int, int, int, int *, int, scalar *, scalar *, scalar *, t_Learner **, int);
/* fsmutation.c */
int FeatureSelectionMutation(const t_Parameters *, int *, int, int *, scalar *, scalar **, int, t_FSAlgorithm, scalar *, int, bool);
int FocusMutation(const t_Parameters *, int *, int,  scalar **, int, int *, int, int *, int, scalar *, int, int, int);
/* fs-stub.c */
void FeatureSelectionAlgorithm(const t_Parameters *, int *, int, scalar **, int *, int, int, int, int **);
void FeatureSelection(const t_Parameters *, int *, int, scalar **, int *, int, int, int, int **);
/* fs-eval.c */
scalar EvaluateSubspaceWithCriterion(const t_Parameters *, scalar **, int *, int, int, int, int, t_FeatureSubsetEvaluation, int *, int *);
void GetBestSubspace(const t_Parameters *, scalar **, int, int *, int, int, scalar *, int *, scalar, int **, int);
/* margin.c */
scalar NearestNeighborMarginWithSemiCache(scalar **, int, int *, scalar **, int, int, t_WeightInfo *, int);
/* misc.c */
void RemoveConstantFeatures(scalar **, int, int *, scalar **, int, FILE *);
int CompareW(const void *, const void *);
int CompareRank(const void *, const void *);
int CompareFeat(const void *, const void *);
int CompareRankingScore(const void *, const void *);
/* multiboost.c */
scalar UpdateMultiBoostWeightsFS(const t_Parameters *, int *, scalar **, int, int *, scalar *, scalar *, int, int, bool, int, int, int);
/* .c */
void RandomPartitionKCDFS(int *, int *, int, int, int *, int);
/* quickselect.c */
scalar kSmallest(scalar *, int, int);
void swap(scalar *a, scalar *b);
int partition(scalar *, int, int);
scalar quickselect(scalar *, int, int, int);
/* random.c */
void RandomFeatureSelection(int *, int, scalar);
/* rank-to-fs.c */
void RankingToFeatureSubsetSelection(const t_Parameters *, scalar **, int *, int, int, int, t_RankInfo *, int *);
/* .c */
int GetSameSubspace(int *, int **s, int, int);
/* svm-real.c */
struct svm_model *TrainSVMClassifierRealInputs(const t_Parameters *, scalar **, int *, int, int *, int);
/* threshold.c */
scalar EvaluateThresholdFS (const t_Parameters*, scalar, scalar *, scalar **, int, int *, scalar *, scalar *, int, int, bool, int *, int, scalar *, scalar *, scalar *, scalar * auprc);
/* redundancy.c */
void RedundancyMeasureOfSeletedFeatures(const t_Parameters *parm, scalar **, int, int *, int *, int, int, scalar *, scalar *, scalar *);
scalar AverageMutualInformationAllPairs(const t_Parameters *, int, scalar **, int *, int, int, int *);
