/***************************************************************************
 *   Copyright (C) 2009 by Aida de Haro-García                             *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "macros.h"
#include "genAlg.h"
#include "fs.h"
#include "classifiers.h"

scalar EvaluateInd(const t_Parameters *parm, int *pop, int n_feat, scalar alpha, scalar **data, int n, int *labels, int classes, bool use_g_measure, int minority)
{
    scalar n_sel = 0.0, fitness = 0.0, suc_rate = 1;
    int **M;
    t_PerformanceInfo perf;

    for (int j = 0; j < n_feat; j++)
        n_sel += n_sel + pop[j];

    // Wrapper approach, evaluate training a classifier and getting the error
    if (parm->evaluator == eIncM)
        suc_rate = EvaluateSubspace(parm, data, n, labels, pop, NULL, n_feat, classes, use_g_measure);
    else {
        AllocateIntegerMatrix(&M, classes, classes);
        ClassifierEvaluation(parm, data, n, labels, data, n, labels, pop, n_feat, NULL, false, classes, (t_Classifier) parm->evaluator, NULL, minority, M, NULL, &perf, NULL, NULL, NULL, NULL,
                                   NULL, NULL, NULL, NULL);
        FreeIntegerMatrix(&M, classes);

        suc_rate = use_g_measure ? perf.gmean : perf.accuracy;
    }

    //Get individual success rate
    fitness = suc_rate * alpha + (1.0 - alpha) * (1.0 - n_sel / n_feat);
    return fitness;
}

void EvaluateAll(const t_Parameters *parm, int **pop, scalar *fitness, int pop_size, int n_feat, scalar alpha, scalar **data, int n, int *labels, int classes, bool use_g_measure, int minority)
{
    for (int i = 0; i < pop_size; i++)
        fitness[i] = EvaluateInd(parm, pop[i], n_feat, alpha, data, n, labels, classes, use_g_measure, minority);
}

void FitnessArrangement(int **pop, scalar *fitness, int pop_size, int n_feat)
{
    int best_ind = 0;
    scalar best_fitness, aux_best_fitness = 0.0;

    for (int l = 0; l < pop_size; l++) {
        best_fitness = -1.0;

        for (int m = l; m < pop_size; m++)
            if (fitness[m] > best_fitness) {
                best_fitness = fitness[m];
                best_ind = m;
            }

        // Exchange fitness
        aux_best_fitness = fitness[best_ind];
        fitness[best_ind] = fitness[l];
        fitness[l] = aux_best_fitness;

        // Exchange features
        int *aux_pt = pop[best_ind];
        pop[best_ind] = pop[l];
        pop[l] = aux_pt;
    }
}

void TwoPointCrossoverGA(int *child1, int *child2, int *father1, int *father2, int n_feat)
{
    int p1 = 0, p2 = 0, null_ind1 = 1, null_ind2 = 1;

    //Randomly select 2 points
    p1 = (int) RANDOM(0, n_feat-1);
    p2 = (int) RANDOM(p1 + 1, n_feat);

    for (int i = 0; i < p1; i++) {
        child1[i] = father1[i];
        child2[i] = father2[i];
    }
    for (int i = p1; i < p2; i++) {
        child1[i] = father2[i];
        child2[i] = father1[i];
    }
    for (int i = p2; i < n_feat; i++) {
        child1[i] = father1[i];
        child2[i] = father2[i];
    }

    //Do not allow null individuals, set one of its chromosomes to 1
    for (int i = 0; i < n_feat; i++) {
        if (child1[i] == 1)
            null_ind1 = 0;
        if (child2[i] == 1)
            null_ind2 = 0;
    }
    if (null_ind1) {
        p1 = (int) RANDOM(0, n_feat);
        child1[p1] = 1;
    }
    if (null_ind2) {
        p2 = (int) RANDOM(0, n_feat);
        child2[p2] = 1;
    }
}

//Get and store in the new_pop the two best individuals from both the parents and children
void InsertTwoBest(const t_Parameters *parm, int **new_pop, int **old_pop, scalar *comp_fitn, scalar *new_fitn, scalar *old_fitn, int father1, int father2, int *child1, int *child2, int n_feat,
                   scalar alpha, int index, scalar **data, int n, int *labels, int classes, int minority)
{
    int j, best_ind1, best_ind2, no_ind = 0;
    scalar fitness1 = 0.0, fitness2 = 0.0, best_fitness = 0.0;

    //Evaluate children
    fitness1 = EvaluateInd(parm, child1, n_feat, alpha, data, n, labels, classes, parm->use_g_measure, minority);
    fitness2 = EvaluateInd(parm, child2, n_feat, alpha, data, n, labels, classes, parm->use_g_measure, minority);

    comp_fitn[0] = old_fitn[father1];
    comp_fitn[1] = old_fitn[father2];
    comp_fitn[2] = fitness1;
    comp_fitn[3] = fitness2;
    best_fitness = comp_fitn[0];
    best_ind1 = 0;
    for (j = 1; j < 4; j++) {
        if (comp_fitn[j] > best_fitness) {
            best_ind1 = j;
            best_fitness = comp_fitn[j];
        }
    }

    //Reset values
    comp_fitn[best_ind1] = 0;
    best_fitness = comp_fitn[0];
    best_ind2 = 0;

    for (j = 1; j < 4; j++) {
        if (comp_fitn[j] > best_fitness) {
            best_ind2 = j;
            best_fitness = comp_fitn[j];
        }
    }

    no_ind = index;
    if (best_ind1 == 0 || best_ind2 == 0) {
        for (j = 0; j < n_feat; j++) {
            new_pop[no_ind][j] = old_pop[father1][j];
        }
        new_fitn[no_ind] = old_fitn[father1];
        no_ind++;
    }
    if (best_ind1 == 1 || best_ind2 == 1) {
        for (j = 0; j < n_feat; j++) {
            new_pop[no_ind][j] = old_pop[father2][j];
        }
        new_fitn[no_ind] = old_fitn[father2];
        no_ind++;
    }
    if (best_ind1 == 2 || best_ind2 == 2) {
        for (j = 0; j < n_feat; j++) {
            new_pop[no_ind][j] = child1[j];
        }
        new_fitn[no_ind] = fitness1;
        no_ind++;
    }
    if (best_ind1 == 3 || best_ind2 == 3) {
        for (j = 0; j < n_feat; j++) {
            new_pop[no_ind][j] = child2[j];
        }
        new_fitn[no_ind] = fitness2;
    }
}

void Mutation(const t_Parameters *parm, int **new_pop, int pop_size, int n_feat, scalar ind_mut_prob, scalar feat_mut_prob, scalar *fitness, scalar alpha, scalar **data, int n, int *labels,
              int classes,
              int minority)
{
    int i, j, p1 = 0;
    scalar local_pfeat = 0.0, local_pind = 0.0, null_ind;

    // Never mutate best i=1
    for (i = 1; i < pop_size; i++) {
        null_ind = 1;
        //Get scalar random value between (0,1) in order to decide if we are going to mutate an individual
        local_pind = RANDOM(0.0, 1.0);
        if (local_pind < ind_mut_prob) {
            for (j = 0; j < n_feat; j++) {
                //Get scalar random value between (0,1) in order to decide if we are going to mutate a feature
                local_pfeat = RANDOM(0.0, 1.0);
                //Mutate feature
                if (local_pfeat < feat_mut_prob) {
                    if (new_pop[i][j] == 1)
                        new_pop[i][j] = 0;
                    else {
                        new_pop[i][j] = 1;
                        null_ind = 0;
                    }
                } else if (new_pop[i][j] == 1)
                    null_ind = 0;
            }
            //Do not allow null individuals, set one of its chromosomes to 1
            if (null_ind) {
                p1 = (int) RANDOM(0, n_feat);
                new_pop[i][p1] = 1;
            }
            // Evaluate
            fitness[i] = EvaluateInd(parm, new_pop[i], n_feat, alpha, data, n, labels, classes, parm->use_g_measure, minority);
        }

    }
}

void Generation(const t_Parameters *parm, int ***new_pop, int ***old_pop, scalar **new_fitn, scalar **old_fitn, int pop_size, int elite_size, int n_feat, scalar ind_mut_prob, scalar feat_mut_prob,
                scalar alpha, scalar **data, int n, int *labels, int classes, int minority)
{
    int i, j, father1 = 0, father2 = 0;
    int *child1 = NULL, *child2 = NULL;
    scalar comp_fitn[4];

    //Memory allocation for children generated by 2 point crossover operator
    MALLOC(child1, int, n_feat, "Generation");
    MALLOC(child2, int, n_feat, "Generation");

    //Fitness arrangement: sort individual by decreasing fitness
    FitnessArrangement(*old_pop, *old_fitn, pop_size, n_feat);

    //Store the elite in the first elite size positions of the new population
    for (i = 0; i < elite_size; i++) {
        for (j = 0; j < n_feat; j++)
            (*new_pop) [i][j] = (*old_pop) [i][j];

        (*new_fitn) [i] = (*old_fitn) [i];
    }

    //Create the rest of the new population by tournament and 2-point crossover
    for (i = elite_size; i < pop_size; i += 2) {
        //Select parents by tournament selection
        father1 = TournamentSelectionGA((*old_fitn), pop_size);
        father2 = TournamentSelectionGA((*old_fitn), pop_size);
        while (father2 == father1)
            father2 = TournamentSelectionGA((*old_fitn), pop_size);

        //2 point crossover
        TwoPointCrossoverGA(child1, child2, (*old_pop) [father1], (*old_pop) [father2], n_feat);
        InsertTwoBest(parm, (*new_pop), (*old_pop), comp_fitn, (*new_fitn), (*old_fitn), father1, father2, child1, child2, n_feat, alpha, i, data, n, labels, classes, minority);
    }

    //Mutate population
    Mutation(parm, (*new_pop), pop_size, n_feat, ind_mut_prob, feat_mut_prob, (*new_fitn), alpha, data, n, labels, classes, minority);

    //In the next iteration: old population will be the current one
    int **aux_pop = *old_pop;
    (*old_pop) = (*new_pop);
    *new_pop = aux_pop;
    scalar *aux_fitn = (*old_fitn);
    (*old_fitn) = (*new_fitn);
    (*new_fitn) = aux_fitn;

    FREE(child1);
    FREE(child2);
}


void GeneticAlgorithm(const t_Parameters *parm, int *S, int pop_size, scalar elite, int n_feat, int n_gens, scalar feat_sel_prob, scalar ind_mut_prob,scalar feat_mut_prob, scalar alpha, scalar **data,
                      int n, int *labels, int classes, int minority)
{
    int i, curr_gen = 0, no_improvement = 0, elite_size;
    int **new_pop = NULL, **old_pop = NULL;
    scalar *new_fitn = NULL, *old_fitn = NULL, best_so_far = -1.0;
    bool stalled = false;

    elite_size = elite*pop_size;

    //If (n-gens * pop-size) is less than all possible feature combination: run genetic algorithm
    if (pow(2, n_feat) > (n_gens * pop_size))  {
        //Memory allocation for new and old population
        CALLOC(old_pop, int *, pop_size, "GeneticAlgorithm");
        CALLOC(new_pop, int *, pop_size, "GeneticAlgorithm");

        for (i = 0; i < pop_size; i++) {
            CALLOC(old_pop[i], int, n_feat, "GeneticAlgorithm");
            CALLOC(new_pop[i], int, n_feat, "GeneticAlgorithm");
        }

        //Memory allocation for new and old population
        CALLOC(old_fitn, scalar, pop_size, "GeneticAlgorithm");
        CALLOC(new_fitn, scalar, pop_size, "GeneticAlgorithm");

        GenerateRandomPopulation(old_pop, pop_size, n_feat, feat_sel_prob);

        //Evaluate all individuals from the population
        EvaluateAll(parm, old_pop, old_fitn, pop_size, n_feat, alpha, data, n, labels, classes, parm->use_g_measure, minority);

        while (curr_gen < n_gens && !stalled) {
            Generation(parm, &new_pop, &old_pop, &new_fitn, &old_fitn, pop_size, elite_size, n_feat, ind_mut_prob, feat_mut_prob, alpha, data, n, labels, classes, minority);
            curr_gen++;

            // If no improvement in 10 generation we go
            if (best_so_far < old_fitn[0]) {
                best_so_far = old_fitn[0];
                no_improvement = 0;
            } else
                no_improvement++;

            if (no_improvement > 10)
                stalled = true;
        }

        // Copy best individual to S
        int best = 0;

        for (i = 0; i < n_feat; i++)
            S[i] = old_pop[best][i];

        // Free memory
        for (i = 0; i < pop_size; i++) {
            FREE(new_pop[i]);
            FREE(old_pop[i]);
        }
        FREE(new_pop);
        FREE(old_pop);
        FREE(new_fitn);
        FREE(old_fitn);
    }  else  {
        //If (n-gens * pop-size) is more than all possible feature combination: run exhaustive search
        int *pop = NULL;
        scalar best_fist = 0.0, fitness = 0.0;
        int num, cont;

        CALLOC(pop, int, n_feat, "GeneticAlgorithm");

        //Get all possible feature combinations
        for (i = 1; i < pow(2, n_feat); i++)  {
            cont = n_feat - 1;
            num = i;
            while (num)  {
                if (num == 1)  {
                    pop[cont] = 1;
                    break;
                } else  {
                    pop[cont] = num%2;
                    num /= 2;
                    cont--;
                }
            }
            fitness = EvaluateInd(parm, pop, n_feat, alpha, data, n, labels, classes, parm->use_g_measure, minority);
            if (fitness > best_fist)  {
                // Copy current best to S
                for (int k = 0; k < n_feat; k++)
                    S[k] = pop[k];
                best_fist = fitness;
            }
        }

        FREE(pop);
    }

}
