/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "fs.h"
#include "common.h"
#include "macros.h"

void LoadGlobalParameters(t_Parameters *parm, char *f_name)
{
    FILE *f;

    // Open file
    if ((f = fopen(f_name, "r")) == NULL)
        OnErrorExit(1, "Cannot open parm file");

    fprintf(parm->out_file, "Parameters (* = default) \n----------\n");

    SetDefaultsParameterValues(parm);

    // [General]
    parm->verbose = LoadInteger(f, "Verbose", -1, parm->out_file);
    parm->bias = false;
    parm->seed = LoadInteger(f, "Seed", 0, parm->out_file);
    parm->typify = LoadBoolean(f, "Tipify", true, parm->out_file);
    parm->fasta_file = LoadBoolean(f, "FASTA file", false, parm->out_file);
    LoadString(f, "Data directory", "Data/MachineLearning/Classification", parm->datad, parm->out_file);

    // Data
    parm->positive_class = LoadInteger(f, "Positive class", -1, parm->out_file);

    // [SVM]
    parm->kernel = LoadInteger(f, "Kernel", 0, parm->out_file);
    // 0 = linear, 1 = poly, 2 = rbf, 3 = sigmoid
    parm->svm_type = LoadInteger(f, "SVM_type", 0, parm->out_file);
    // 0 = C-SVC, 1 = nu-SVc, 2 = one-class SVM, 3 = epsilon-SVR, 4 = nu-SVR
    parm->gamma = LoadReal(f, "Gamma", 1.0, parm->out_file);
    parm->svm_c = LoadReal(f, "C", 1.0, parm->out_file);
    parm->svm_d = LoadInteger(f, "Polynomial degree", 3, parm->out_file);
    parm->svm_r = LoadReal(f, "R", 0.0, parm->out_file);
    parm->tolerance = LoadReal(f, "Tolerance", 0.1, parm->out_file);
    parm->cache_size = LoadInteger(f, "Cache size", 10000, parm->out_file);

    fprintf(parm->out_file, "----------\n");
    fclose(f);
}
