/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <typex.h>
#include "fs.h"
#include "defns.i"
#include "types.i"
#include "extern.i"
#include "macros.h"
#include "common.h"
#include "svm.h"

// Margin for NN rule with a cache of distances to all but one feature added or remove
scalar NearestNeighborMarginWithSemiCache(scalar **data, int instances, int *labels, scalar **sq_dists, int add, int remove, t_WeightInfo *w_info, int top)
{
    scalar value, dmin_friend, dmin_enemy, margin = 0.0;

    if (!sq_dists)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "NN Margin without cache");

    // Add?
    if (add != -1) {
        // Get nearest neighbor
        for (int k = instances-top; k < instances; k++) {
            dmin_enemy = dmin_friend = INFINITE;

            // Next worse
            int j = w_info[k].instance;

            for (int i = 0; i < instances; i++) {
                value = sq_dists[j][i] + SQR(data[j][add] - data[i][add]);

                if (labels[i] == labels[j] && value < dmin_friend && i != j)
                    dmin_friend = value;
                else if (labels[i] != labels[j] && value < dmin_enemy)
                    dmin_enemy = value;
            }

            margin += dmin_enemy - dmin_friend;
        }
    } else if (remove != -1) {
        // Get nearest neighbor
        for (int k = instances-top; k < instances; k++) {
            dmin_enemy = dmin_friend = INFINITE;

            // Next worse
            int j = w_info[k].instance;

            for (int i = 0; i < instances; i++) {
                value = sq_dists[j][i] - SQR(data[j][remove] - data[i][remove]);

                if (labels[i] == labels[j] && value < dmin_friend && i != j)
                    dmin_friend = value;
                else if (labels[i] != labels[j] && value < dmin_enemy)
                    dmin_enemy = value;
            }

            margin += dmin_enemy - dmin_friend;
        }
    } else {    // Use current cache
        // Get nearest neighbor
        for (int k = instances-top; k < instances; k++) {
            dmin_enemy = dmin_friend = INFINITE;

            // Next worse
            int j = w_info[k].instance;

            for (int i = 0; i < instances; i++)
                if (labels[i] == labels[j] && sq_dists[j][i] < dmin_friend && i != j)
                    dmin_friend = sq_dists[j][i];
                else if (labels[i] != labels[j] && sq_dists[j][i] < dmin_enemy)
                    dmin_enemy = sq_dists[j][i];

            margin += dmin_enemy - dmin_friend;
        }
    }

    return margin/top;
}
