/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas   *
 *   npedrajas@uco.es   *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "fs.h"

// Remove features with zero variance
void RemoveConstantFeatures(scalar **train_data, int train_instances, int *n_features, scalar **test_data, int test_instances, FILE *out_file)
{
    scalar x, x2;
    int removed = 0, inputs;

    inputs = *n_features;

    // Get the variance of every feature
    for (int i = 0; i < *n_features; i++) {
        x = x2 = 0.0;
        for (int j = 0; j < train_instances; j++) {
            x += train_data[j][i];
            x2 += train_data[j][i]*train_data[j][i];
        }

        // Remove if constant
        if (x2/train_instances - pow(x/train_instances, 2.0) < ZERO) {
            for (int j = 0; j < train_instances; j++)
                train_data[j][i] = train_data[j][*n_features-1];
            for (int j = 0; j < test_instances; j++)
                test_data[j][i] = test_data[j][*n_features-1];

            (*n_features)--;
            i--;
            removed++;
        }
    }

    fprintf(out_file, "@Removed constant features: %d\n", removed);
    fprintf(out_file, "@Percentage of constant features: %lf\n", ((scalar)removed)/inputs);
}

// Auxiliary function for qsort
int CompareW(const void *x, const void *y)
{
    scalar diff;

    if ((diff = ((t_Feat *) y)->w - ((t_Feat *) x)->w) > 0)
        return +1;
    else {
        if (diff < 0)
            return -1;
        else
            return 0;
    }
}

// Auxiliary function for qsort.
int CompareRank(const void *x, const void *y)
{
    scalar diff;

    if ((diff = ((t_Feat *) y)->ranking - ((t_Feat *) x)->ranking) > 0)
        return +1;
    else {
        if (diff < 0)
            return -1;
        else
            return 0;
    }
}


// Auxiliary function for qsort: from best to worst weight
int CompareFeat(const void *x, const void *y)
{
    scalar diff;

    if ((diff = ((t_Feat *) y)->weight - ((t_Feat *) x)->weight) > 0)
        return +1;
    else {
        if (diff < 0)
            return -1;
        else
            return 0;
    }
}

// Auxiliary function for qsort: from best to worst ranking
int CompareRankingScore(const void *x, const void *y)
{
    scalar diff;

    if ((diff = ((t_RankInfo *) y)->score - ((t_RankInfo *) x)->score) > 0)
        return +1;
    else {
        if (diff < 0)
            return -1;
        else
            return 0;
    }
}
