/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "fs.h"
#include "macros.h"

// Functions that deal with partition which keep the class distribution
void RandomPartitionKCDFS(int *partition, int *count, int n, int n_subsets, int *label, int classes)
{
    int *list = NULL, **class_list, *class_count, *class_c;
    scalar size;

    MALLOC(list, int, n, "RandomPartitionKCDFS: list");
    MALLOC(class_list, int *, classes, "RandomPartitionKCDFS: class_list");
    MALLOC(class_count, int, classes, "RandomPartitionKCDFS: class_count");
    MALLOC(class_c, int, classes, "RandomPartitionKCDFS: class_c");
    // Number of instances per class
    for (int i = 0; i < classes; i++)
        class_count[i] = 0;
    for (int i = 0; i < n; i++)
        class_count[label[i]]++;
    for (int i = 0; i < classes; i++)
        MALLOC(class_list[i], int, class_count[i], "RandomPartitionKCDFS: class_list[i]");

    // List of instances in each class
    for (int i = 0; i < classes; i++)
        class_count[i] = 0;
    for (int i = 0; i < n; i++) {
        class_list[label[i]][class_count[label[i]]] = i;
        class_count[label[i]]++;
    }

    // Random partition
    for (int i = 0; i < classes; i++)
        RandomizeVector(class_list[i], class_count[i]);

    size = n / n_subsets;
    for (int i = 0; i < classes; i++)
        class_c[i] = 0;
    for (int i = 0; i < n_subsets; i++) {
        count[i] = 0;
        // For each class
        for (int k = 0; k < classes; k++) {
            for (int j = 0; j < (size * class_count[k]) / n - 1 && class_c[k] < class_count[k]; j++) {
                partition[class_list[k][class_c[k]]] = i;
                class_c[k]++;
                count[i]++;
            }
        }
    }

    // Assign remaining Round Robin
    for (int i = 0; i < classes; i++) {
        int subset = RANDOM(0, n_subsets);
        for (; class_c[i] < class_count[i]; class_c[i]++) {
            partition[class_list[i][class_c[i]]] = subset;
            count[subset]++;
            subset = (subset+1) % n_subsets;
        }
    }

    FREE(list);
    FREE(class_list);
    FREE(class_count);
    FREE(class_c);
}
