/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include "typex.h"
#include "fs.h"

/*
 * Algorithm quickselect() to get the k-th item in unordered array
 inputs...
 A: pointer to array
 n: size of array
 k: the item in question
 */

scalar kSmallest(scalar *A, int n, int k)
{

    int left = 0;
    int right = n-1;

    return quickselect(A, left, right, k);
}

// Swaps
void swap(scalar *a, scalar *b)
{
    scalar temp = *a;
    *a = *b;
    *b = temp;
}

int partition(scalar *A, int left, int right)
{

    int pivot = A[right], i = left, x;

    for (x = left; x < right; x++) {
        if (A[x] <= pivot) {
            swap(&A[i], &A[x]);
            i++;
        }
    }

    swap(&A[i], &A[right]);
    return i;
}

scalar quickselect(scalar *A, int left, int right, int k)
{

    // p is position of pivot in the partitioned array
    int p = partition(A, left, right);

    // k equals pivot got lucky
    if (p == k-1)
        return A[p];
    // k less than pivot
    else if (k - 1 < p)
        return quickselect(A, left, p - 1, k);
    // k greater than pivot
    else
        return quickselect(A, p + 1, right, k);
}
