/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "typex.h"
#include "common.h"
#include "macros.h"
#include "fs.h"
#include "isfs.h"

void RankingToFeatureSubsetSelection(const t_Parameters *parm, scalar **data, int *labels, int instances, int inputs, int classes, t_RankInfo *ranks, int *subset)
{
    scalar *scores, *coefficient, *complexity, *red;
    int best_f = 0;

    MALLOC(coefficient, scalar, inputs, "RankingToFeatureSubsetSelection: coefficient");
    MALLOC(red, scalar, inputs, "RankingToFeatureSubsetSelection: red");
    MALLOC(complexity, scalar, inputs, "RankingToFeatureSubsetSelection: complexity");
    MALLOC(scores, scalar, inputs, "RankingToFeatureSubsetSelection: scores");

    for (int i = 0; i < inputs; i++)
        scores[i] = ranks[i].score;

    switch (parm->rank_to_fs_threshold_evaluation) {
        case Thresholding:
            OnErrorExit(ERR_PARAM, "These thresholds methods are not suitable for ranking to feature subset selection");
            break;

        case Mfdr:
        case Voor:
        case Mfe:
        case Cf:
            MALLOC(coefficient, scalar, inputs, "main: coefficient");
            MALLOC(complexity, scalar, inputs, "main: complexity");
            MALLOC(red, scalar, inputs, "main: r");

            if (parm->rank_to_fs_threshold_evaluation == Mfdr) {
                MaximumFishersDiscriminantRatioEvaluation(parm, data, labels, instances, inputs, scores, coefficient, red, complexity, classes, scores, inputs, NULL);

                // Get best thresholds
                best_f = 0;
                for (int j = 1; j < inputs; j++)
                    if (coefficient[j] > coefficient[best_f] || (coefficient[j] == coefficient[best_f] && red[j] < red[best_f]))
                        best_f = j;
            } else if (parm->rank_to_fs_threshold_evaluation == Mfe) {
                MulticlassMaximumFeatureEfficiency(parm, data, labels, instances, inputs, scores, coefficient, red, complexity, classes, scores, inputs);

                // Get best thresholds
                best_f = 0;
                for (int j = 1; j < inputs; j++)
                    if (coefficient[j] > coefficient[best_f] || (coefficient[j] == coefficient[best_f] && red[j] > red[best_f]))
                        best_f = j;
            } else if (parm->rank_to_fs_threshold_evaluation == Voor) {
                MulticlassVolumeOfOverlapRegion(parm, data, labels, instances, inputs, scores, coefficient, red, complexity, classes, scores, inputs);

                // Get best thresholds
                best_f = 0;
                SHOWPROGRESS(parm->show_progress > 0, "Threshold %i VOOR coefficient: %lf\n", 1, coefficient[0]);
                for (int j = 1; j < inputs; j++) {
                    SHOWPROGRESS(parm->show_progress > 0, "Threshold %i VOOR coefficient: %lf\n", j+1, coefficient[j]);
                    if (coefficient[j] < coefficient[best_f] || (coefficient[j] == coefficient[best_f] && red[j] > red[best_f]))
                        best_f = j;
                }
            } else {
                ComplexityFusion(parm, data, labels, instances, inputs, scores, coefficient, red, complexity, classes, scores, inputs, NULL);

                // Get best thresholds
                best_f = 0;
                for (int j = 1; j < inputs; j++)
                    if (coefficient[j] > coefficient[best_f] || (coefficient[j] == coefficient[best_f] && red[j] > red[best_f]))
                        best_f = j;
            }
            break;

        case Loor:
        case Riic:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "LOOR/N2 for ranking to subset selection");
            break;
    }

    // Select subset
    for (int i = 0; i < inputs; i++)
        subset[i] = scores[i] >= scores[best_f] ? 1 : 0;

    FREE(coefficient);
    FREE(complexity);
    FREE(red);
    FREE(scores);
}
