/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <typex.h>
#include "common.h"
#include "macros.h"
#include "fs.h"
#include "ranking.h"

// Weston, J., Perez-Cruz, F., Bousquet, O., Chapelle, O., Elisseeff,
// A. B. Scholkopf (2003) Bioinformatics 19, 764–771. http://dx.doi.org/10.1093/bioinformatics/btg054

void UnbalancedCorrelationScore(const t_Parameters *parm, int inputs, scalar **data, int *labels, int classes, int instances, int minority, t_RankInfo *ranking, int **index)
{
    if (classes != 2)
        OnErrorExit(ERR_PARAM, "Unbalanced Correlation Score method only available for two-class problems");

    if (minority == -1)
        OnErrorExit(ERR_PARAM, "Minority class not specified");

    if (index) {
        // Sparse format
        scalar *sum_pos, *sum_neg;
        CALLOC(sum_pos, scalar, inputs, "UnbalancedCorrelationScore: sum_pos");
        CALLOC(sum_neg, scalar, inputs, "UnbalancedCorrelationScore: sum_neg");

        for (int i = 0; i < instances; i++) {
            int j = 0;
            while (index[i][j] != -1) {
                if (labels[i] == minority)
                    sum_pos[index[i][j]-1] += data[i][j];
                else
                    sum_neg[index[i][j]-1] += data[i][j];
                j++;
            }
        }

        for (int i = 0; i < inputs; i++) {
            ranking[i].feature = i;
            ranking[i].score = sum_pos[i] - parm->cub_lambda*sum_neg[i];

            SHOWPROGRESS(parm->show_progress > 0, "IG feature %*d score: %lf\n", WIDTH(inputs), i+1, ranking[i].score);
        }

        FREE(sum_neg);
        FREE(sum_pos);
    } else {
        for (int i = 0; i < inputs; i++) {
            scalar sum_pos = 0.0, sum_neg = 0.0;

            for (int j = 0; j < instances; j++)
                if (labels[j] == minority)
                    sum_pos += data[j][i];
                else
                    sum_neg += data[j][i];

            ranking[i].feature = i;
            ranking[i].score = sum_pos - parm->cub_lambda*sum_neg;

            SHOWPROGRESS(parm->show_progress > 0, "IG feature %*d score: %lf\n", WIDTH(inputs), i+1, ranking[i].score);
        }
    }
}
