/***************************************************************************
 *   Copyright (C) 2019 by Nicolás García-Pedrajas                         *
 *   adeharo@uco.es                                                        *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "macros.h"
#include "fs.h"
#include "filter.h"
#include "ranking.h"

// Euclidean Distance Based Entropy (edse)
// Ricardo W. Pino Urias et al., "IMMAN: free software for information theory-based chemometric analysis", Mol Divers (2015) 19:305–319

void EuclideanDistanceBasedEntropy(const t_Parameters *parm, int inputs, scalar **data, int instances, t_RankInfo *ranking, int **index)
{
    scalar max_distance, dij, vj, vk;

    // For every feature
    for (int i = 0; i < inputs; i++) {
        max_distance = -1.0;
        for (int j = 1; j < instances; j++)
            for (int k = 0; k < j; k++) {
                vj = index ? SparseFeature(data[j], index[j], i) : data[j][i];
                vk = index ? SparseFeature(data[k], index[k], i) : data[k][i];
                if (max_distance < fabs(vj - vk))
                    max_distance = fabs(vj - vk);
            }

        // Euclidean distance based entropy
        scalar edse = 0.0;
        for (int j = 1; j < instances; j++)
            for (int k = 0; k < j; k++) {
                vj = index ? SparseFeature(data[j], index[j], i) : data[j][i];
                vk = index ? SparseFeature(data[k], index[k], i) : data[j][k];

                dij = fabs(vj - vk)/max_distance;
                edse += (dij > ZERO && dij < 1.0) ? 2.0*(dij*log2(dij)+(1-dij)*log2(1.0-dij)) : 0.0;
            }

        ranking[i].feature = i;
        ranking[i].score = -edse;

        SHOWPROGRESS(parm->show_progress > 0, "EDSE feature %*d score: %lf (smallest best)\n", WIDTH(inputs), i+1, edse);
    }
}
