/***************************************************************************
 *   Copyright (C) 2019 by Nicolás García-Pedrajas                         *
 *   adeharo@uco.es                                                        *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "macros.h"
#include "fs.h"
#include "filter.h"

// F-Score
// Bahram Hemmateenejad, Ahmadreza Mehdipour, Omar Deeb, Mahmood Sanchooli, and Ramin Miri
// Toward an Optimal Approach for Variable Selection in Counter-Propagation Neural Networks: Modeling Protein- Tyrosine Kinase Inhibitory of Flavanoids Using Substituent Electronic Descriptors
// Mol. Inf. 2011, 30, 939 – 949

void FScore(const t_Parameters *parm, int inputs, scalar **data, int *labels, int classes, int instances, t_RankInfo *ranking)
{
    int pos, neg;
    if (classes != 2)
        OnErrorExit(ERR_PARAM, "F-Score only valid for two-class problems");

    for (int i = 0; i < inputs; i++) {
        scalar mean = 0.0, mean_pos = 0.0, mean_neg = 0.0;
        pos = neg = 0;
        for (int j = 0; j < instances; j++) {
            mean += data[j][i];

            if (labels[j] == 1) {
                mean_pos += data[j][i];
                pos++;
            } else {
                mean_neg += data[j][i];
                neg++;
            }
        }

        mean /= instances;
        mean_pos /= pos;
        mean_neg /= neg;

        // Obtain F-score of i-th feature
        // Dispersion of positive and negative instances
        scalar disp_pos = 0.0, disp_neg = 0.0;
        for (int j = 0; j < instances; j++)
            if (labels[j] == 1)
                disp_pos += pow(data[j][i] - mean_pos, 2.0);
            else
                disp_neg += pow(data[j][i] - mean_neg, 2.0);

        scalar f_score = (pow(mean_pos-mean, 2.0)+pow(mean_neg-mean, 2.0))/(disp_pos/(pos-1.0)+(disp_neg/(neg-1.0)));

        ranking[i].feature = i;
        ranking[i].score = f_score;

        SHOWPROGRESS(parm->show_progress > 0, "F-Score feature %*d score: %lf\n", WIDTH(inputs), i+1, f_score);
    }
}
