/***************************************************************************
 *   Copyright (C) 2010 Aida de Haro-García                                *
 *   adeharo@uco.es                                                        *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include "macros.h"
#include "fs.h"
#include "is.h"

void RFE(const t_Parameters *parm, t_Feat *feat_info, int n_feat, scalar **data, int n, int *labels, int *mask)
{
    int iter, k, *s = NULL;
    scalar sum = 0.0, **w = NULL;
    struct svm_model *svm;

    // Memory allocation
    CALLOC(s, int, n_feat, "RFE: s");

    // Initializations
    int count = 0;
    for (int i = 0; i< n_feat; i++)  {
        //At the beginning, RFE hasn't removed any feature yet
        if (mask[i]) {
            s[i] = 1;
            count++;
            feat_info[i].weight = 0.0;
            feat_info[i].w = 0.0;
        } else {
            s[i] = 0;
            feat_info[i].weight = -1.0;
            feat_info[i].w = -1.0;
        }

        //Reset weight accumulator
        feat_info[i].ranking = 0;
        feat_info[i].index = i;
    }

    // From 0 to n_feat iterations
    for (iter = 0; iter < count; iter ++)  {
        // Train SVM classifier
        svm = TrainSVMClassifierRealInputs(parm, data, s, n, labels, n_feat);
        w = GetSVMW(svm, n_feat);
        // k is the number of comparative classifiers built by svm to emulate multiclass problem
        k = svm->nr_class * (svm->nr_class - 1)/2;

        //Get absolute average weight for each feature among all classifiers
        for (int j = 0; j < n_feat ; j++)  {
            sum = 0.0;
            //If we set the element of mask: s[n_feat] to zero, then the output w[i][n_feat] will be zero
            for (int i = 0; i < k; i++)  {
                sum += fabs(w[i][j]);
            }
            feat_info[j].w = (sum/k);
            feat_info[j].weight += feat_info[j].w;
        }

        //Sort features from worst to best weight
        qsort(feat_info, n_feat, sizeof(t_Feat), CompareWInv);

        // Remove next worst feature
        // Get to first considered feature
        int j = 0;
        while (!s[feat_info[j].index] && j < n_feat)
            j++;

        // The sooner a feature is removed the lesser ranking value it receives
        feat_info[j].ranking = iter;
        s[feat_info[j].index] = 0;

        // Restore index order (ascending index order)
        qsort(feat_info, n_feat, sizeof(t_Feat), CompareIndex);

        // Destroy svm model
        svm_free_and_destroy_model(&svm);
        for (int i = 0; i < k; i++)
            FREE(w[i]);
        FREE(w);
    }

    FREE(s);
}

// Auxiliary function for qsort: from worst to best weight
int CompareRanking(const void *x, const void *y)
{
    scalar diff;

    if ((diff = ((t_Feat *) y)->ranking - ((t_Feat *) x)->ranking) > 0)
        return +1;
    else {
        if (diff > 0)
            return -1;
        else
            return 0;
    }
}

// Auxiliary function for qsort: from worst to best weight
int CompareWInv(const void *x, const void *y)
{
    scalar diff;

    if ((diff = ((t_Feat *) x)->w - ((t_Feat *) y)->w) > 0)
        return +1;
    else {
        if (diff > 0)
            return -1;
        else
            return 0;
    }
}

// Auxiliary function for qsort: sort by ascending index
int CompareIndex(const void *x, const void *y)
{
    scalar diff;

    if ((diff = ((t_Feat *) x)->index - ((t_Feat *) y)->index) > 0)
        return +1;
    else {
        if (diff > 0)
            return -1;
        else
            return 0;
    }
}
