#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 23 11:41:15 2021

@author: frodo

Instance Hardness Threshold
D. Smith, Michael R., Tony Martinez, and Christophe Giraud-Carrier. “An instance level analysis of data complexity.” Machine learning 95.2 (2014): 225-256.
"""

import argparse
from skmultilearn.dataset import load_from_arff
from imblearn.under_sampling import InstanceHardnessThreshold
import warnings
import sys
import numpy as np

warnings.filterwarnings("ignore")

parser = argparse.ArgumentParser()
parser.add_argument("dataset", help="Dataset (must be a -.train and a -.test)", type=str)

parser.add_argument("-sparse", help="File is in sparse format", action='store_true')
args = parser.parse_args()

X_train, y_train = load_from_arff(args.dataset, label_count=1, label_location='end', load_sparse=args.sparse)

# Random Over-sampling
sampler = InstanceHardnessThreshold(random_state=0)
X, y = sampler.fit_resample(X_train.A, y_train.A)

X_train = np.column_stack((X, y))

# Save to CSV file
with open(args.dataset, 'w') as f:
    f.write('[ONE LABEL]\n$ %d\n$ %d\n$ 2\n' % (X.shape[0], X.shape[1]))
    np.savetxt(f, X_train, fmt="%g", delimiter=" ")

sys.exit(1)
