#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 23 11:41:15 2021

@author: frodo

kMeans SMOTE
Felix Last, Georgios Douzas, Fernando Bacao, “Oversampling for Imbalanced Learning Based on K-Means and SMOTE” https://arxiv.org/abs/1711.00837

There are two available implementations: scikit learn contrib and pypi (not working)
"""



import argparse
from skmultilearn.dataset import load_from_arff
from imblearn.over_sampling import KMeansSMOTE
import warnings
import sys
import numpy as np

warnings.filterwarnings("ignore")

parser = argparse.ArgumentParser()
parser.add_argument("dataset", help="Dataset (must be a -.train and a -.test)", type=str)

parser.add_argument("-sparse", help="File is in sparse format", action='store_true')
args = parser.parse_args()

X_train, y_train = load_from_arff(args.dataset, label_count=1, label_location='end', load_sparse=args.sparse)

# Random Over-sampling
cluster_balance_threshold = 0.5

finished = False
while not finished and cluster_balance_threshold >= 0.0:
    sampler = KMeansSMOTE(random_state=0, cluster_balance_threshold=cluster_balance_threshold)

    try:
        X, y = sampler.fit_resample(X_train.A, y_train.A)
        finished = True
    except:
        cluster_balance_threshold -= 0.05
        print("Retrying with cluster balance threshold: %lf" % cluster_balance_threshold)

X_train = np.column_stack((X, y))

# Save to CSV file
with open(args.dataset, 'w') as f:
    f.write('[ONE LABEL]\n$ %d\n$ %d\n$ 2\n' % (X.shape[0], X.shape[1]))
    np.savetxt(f, X_train, fmt="%g", delimiter=" ")

sys.exit(1)
