#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 23 11:41:15 2021

@author: frodo

Smote Variants: https://github.com/analyticalmindsltd/smote_variants#id1

"""

import argparse
from skmultilearn.dataset import load_from_arff
import smote_variants as sv
import warnings
import sys
import numpy as np

warnings.filterwarnings("ignore")

parser = argparse.ArgumentParser()
parser.add_argument("variant", help="Variant (int [0, 84])", type=int)
parser.add_argument("dataset", help="Dataset (must be a -.train and a -.test)", type=str)
parser.add_argument("-sparse", help="File is in sparse format", action='store_true')
args = parser.parse_args()

X_train, y_train = load_from_arff(args.dataset, label_count=1, label_location='end', load_sparse=args.sparse)
inputs = X_train.toarray().shape[1]

# Variants
variant = [sv.SMOTE(), sv.SMOTE_TomekLinks(), sv.SMOTE_ENN(), sv.Borderline_SMOTE1(), sv.Borderline_SMOTE2(), sv.ADASYN(), sv.AHC(), sv.LLE_SMOTE(), sv.distance_SMOTE(), sv.SMMO(),
           sv.polynom_fit_SMOTE(), sv.Stefanowski(), sv.ADOMS(), sv.Safe_Level_SMOTE(), sv.MSMOTE(), sv.DE_oversampling(), sv.SMOBD(), sv.SUNDO(), sv.MSYN(), sv.SVM_balance(), sv.TRIM_SMOTE(),
           sv.SMOTE_RSB(), sv.ProWSyn(), sv.SL_graph_SMOTE(), sv.NRSBoundary_SMOTE(), sv.LVQ_SMOTE(), sv.SOI_CJ(), sv.ROSE(), sv.SMOTE_OUT(), sv.SMOTE_Cosine(), sv.Selected_SMOTE(), sv.LN_SMOTE(),
           sv.MWMOTE(), sv.PDFOS(), sv.IPADE_ID(), sv.RWO_sampling(), sv.NEATER(), sv.DEAGO(), sv.Gazzah(), sv.MCT(), sv.ADG(), sv.SMOTE_IPF(), sv.KernelADASYN(), sv.MOT2LD(), sv.V_SYNTH(), sv.OUPS(),
           sv.SMOTE_D(), sv.SMOTE_PSO(), sv.CURE_SMOTE(), sv.SOMO(), sv.ISOMAP_Hybrid(n_components=inputs), sv.CE_SMOTE(), sv.Edge_Det_SMOTE(), sv.CBSO(), sv.E_SMOTE(min_features=inputs),
           sv.DBSMOTE(), sv.ASMOBD(), sv.Assembled_SMOTE(), sv.SDSMOTE(), sv.DSMOTE(), sv.G_SMOTE(), sv.NT_SMOTE(), sv.Lee(), sv.SPY(), sv.SMOTE_PSOBAT(), sv.MDO(), sv.Random_SMOTE(), sv.ISMOTE(),
           sv.VIS_RST(), sv.GASMOTE(), sv.A_SUWO(), sv.SMOTE_FRST_2T(), sv.AND_SMOTE(), sv.NRAS(), sv.AMSCO(), sv.SSO(), sv.NDO_sampling(), sv.DSRBF(), sv.Gaussian_SMOTE(), sv.kmeans_SMOTE(),
           sv.Supervised_SMOTE(), sv.SN_SMOTE(), sv.CCR(), sv.ANS(), sv.cluster_SMOTE()
           ]


# Smote variant method
oversampler = variant[args.variant]
print(oversampler)

X, y = oversampler.sample(X_train.toarray(), y_train.toarray().flatten())

# Check features match
if inputs != X.shape[1]:
    print("Features do not match: %d != %d" % (inputs, X.shape[1]))
    sys.exit(1)

X_train = np.column_stack((X, y))

# Save to CSV file
with open(args.dataset, 'w') as f:
    f.write('[ONE LABEL]\n$ %d\n$ %d\n$ 2\n' % (X.shape[0], X.shape[1]))
    np.savetxt(f, X_train, fmt="%g", delimiter=" ")

sys.exit(1)
