/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <typex.h>
#include "isfs.h"
#include "defns.i"
#include "types.i"
#include "extern.i"
#include "macros.h"
#include "common.h"
#include "svm.h"
#include "classifiers.h"

// Get weighted error and vector of classified instances

scalar GetWeightedError(const t_Parameters *parm, t_Classifier classifier, int *S, scalar **data, int inputs, int instances, int *labels, int classes, bool *ok, scalar *w, bool cv, int **index,
                        int *subspace)
{
    scalar error = 0.0, sum = 0.0;
    t_Learner learner;
    t_Ensemble ensemble;

    if (index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse format for IS/FS boosting");

    // Validate?
    if (cv)
        SetCrossValidatedParameters((t_Parameters *) parm, data, instances, labels, subspace, inputs, classes, classifier, 0, S, index);

    // Train classifier
    ClassifierTraining(parm, classifier, data, instances, labels, inputs, classes, index, subspace, S, true, &learner, &ensemble);

    // Test classification over all instances
    for (int i = 0; i < instances; i++) {
        sum += w[i];
        if (ClassifierOutput(parm, classifier, data[i], inputs, classes, index ? index[i] : NULL, subspace, data, instances, labels, S, &learner, labels[i], NULL, -1, index, &ensemble) != labels[i]) {
            ok[i] = false;
            error += w[i];
        } else
            ok[i] = true;
    }

    ClassifierFree(parm, &learner, classifier, classes, inputs, NULL);

    return error/sum;
}
