/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "isfs.h"
#include "macros.h"
#include "classifiers.h"

// Margin of an ensemble of NN classifiers with selection

scalar kNearestNeighbourEnsembleMarginSubspace(int instance, int **S, int size, scalar **protos, int n_protos, int inputs, int k, int *nearest, int *label, int classes, int last, scalar *alpha,
        bool *correct, bool knn_with_weighted_votes)
{
    scalar margin, sum = 0.0;

    // Each selection votes for a class
    scalar votes[classes];

    ZeroVector(votes, NULL, classes);

    // Classify with each selection
    for (int i = 0; i < size; i++)
        sum += (votes[kNNCategory(protos[instance], protos, S[last-size+i]+n_protos, n_protos, k, inputs, label, classes, nearest, S[last-size+i], NULL, NULL, -1, NULL, NULL)] += alpha[i]);

    // Get maximum number of votes
    int category = 0;
    for (int i = 1; i < classes; i++)
        if (votes[i] > votes[category])
            category = i;

    // Correctly classified?
    if (category == label[instance]) {
        *correct = true;

        int second_best = (category == 0 ? 1 : 0);

        for (int i = 0; i < classes; i++)
            if (votes[second_best] < votes[i] && i != category)
                second_best = i;

        margin = (votes[category] - votes[second_best])/sum;
    } else {
        *correct = false;

        margin = (votes[label[instance]] - votes[category])/sum;
    }

    return margin;
}

// Ensembles of nearest neighbors with selection accuracy
scalar kNearestNeighbourEnsembleAccuracySubspace(int **S, int size, scalar **protos, int n_protos, int inputs, int k, int *label, int classes, int last, scalar *alpha, bool knn_with_weighted_votes)
{
    scalar accuracy = 0.0;
    int nearest[k];

    // Classify each instance
    for (int i = 0; i < n_protos; i++)
        if (kNearestNeighbourEnsembleSubspace(i, S, size, protos, n_protos, inputs, k, nearest, label, classes, last, alpha, knn_with_weighted_votes) == label[i])
            accuracy++;

    return accuracy/n_protos;
}

// Ensembles of nearest neighbors with selection
int kNearestNeighbourEnsembleSubspace(int instance, int **S, int size, scalar **protos, int n_protos, int inputs, int k, int *nearest, int *label, int classes, int last, scalar *alpha,
                                      bool knn_with_weighted_votes)
{
    // Each selection votes for a class
    scalar votes[classes];

    for (int i = 0; i < classes; i++)
        votes[i] = 0.;

    // Classify with each selection
    for (int i = 0; i < size; i++)
        votes[kNNCategory(protos[instance], protos, S[last-size+i]+n_protos, n_protos, k, inputs, label, classes, nearest, S[last-size+i], NULL, NULL, -1, NULL, NULL)] += alpha[i];

    // Get maximum number of votes
    int category = 0;
    for (int i = 1; i < classes; i++)
        if (votes[i] > votes[category])
            category = i;

    return category;
}
