/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// Implements functions needed to deal with features

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <math.h>
#include <stdbool.h>
#include "isfs.h"
#include "macros.h"
#include "common.h"

void FeaturePartition(int features, int n_subspaces, int *subspace_size, int **subspaces)
{
    int *list;

    if (n_subspaces == 1) {
        subspace_size[0] = features;
        for (int i = 0; i < features; i++)
            subspaces[0][i] = i;
        return;
    }

    ZeroVector(NULL, subspace_size, n_subspaces);

    // Randomly assign features to subspaces
    // Random partition
    MALLOC(list, int, features, "FeaturePartition: list");
    for (int i = 0; i < features; i++)
        list[i] = i;
    RandomizeVector(list, features);

    int subspace = 0;
    for (int i = 0; i < features; i++) {
        subspaces[subspace][subspace_size[subspace]] = list[i];
        subspace_size[subspace]++;

        subspace = (subspace+1) % n_subspaces;
    }

    FREE(list);
}
