/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Hash style functions for keeping track of evaluated individuals
#include <string.h>
#include "isfs.h"
#include "common.h"
#include "macros.h"

#define SEGMENT 100    // Entries are created in segments of this size

void CreateHashTable(t_HashTable *ht)
{
    MALLOC(ht->entries, t_HashEntry, SEGMENT, "CreateHashTable: ht->entries");

    ht->size = ht->hits = 0;
}

// Search for an entry
// Return NULL if not found
t_HashEntry *HashSearchEntry(t_HashTable ht, char *key)
{
    for (int i = 0; i < ht.size; i++)
        if (EqualKeys(key, ht.entries[i].key)) {
            ht.hits++;
            return &ht.entries[i];
        }

    // Not found
    return NULL;
}

// Add new entry
// The entry must not be in the hash table
void HashAddEntry(t_HashTable *ht, t_HashEntry *entry)
{
    // Test if more memory needed
    if (ht->size && (ht->size % SEGMENT) == 0) {
        REALLOC(ht->entries, t_HashEntry, ht->size + SEGMENT, "HasdAddEntry: ht.entries");
    }

    // Add entry
    MALLOC(ht->entries[ht->size].key, char, strlen(entry->key) + 1, "HashAddEntry: ht.entries[ht->size].key");
    strcpy(ht->entries[ht->size].key, entry->key);
    ht->entries[ht->size].ok = entry->ok;

    ht->size++;
}

void DestroyHashTable(t_HashTable *ht)
{
    for (int i = 0; i < ht->size; i++)
        FREE(ht->entries[i].key);

    FREE(ht->entries);
    ht->entries = NULL;
}

bool EqualKeys(char *key1, char *key2)
{
    int i = 0;
    while (key1[i] == key2[i] && key1[i] != '\0')
        i++;

    return key1[i] == '\0';
}
