/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Definitions and prototypes
#pragma once

#include <stdbool.h>
#include <stdio.h>
#include <time.h>
#include "typex.h"
#include "common.h"
#include "is.h"
#include "fs.h"
#include "svm.h"
#include "ranking.h"
#include <gsl/gsl_matrix_double.h>

// #define AUTOTEST             // Set auto-test code
// #define VERY_VERBOSE

// #define PARALLEL_IMPLEMENTATION // If defined some programs are run in parallel using PVM

// Objectives for individuals
#define N_OBJECTIVES    2
#define OK              0
#define REDUCTION       1

#define CACHE_USE_TRIANGULAR    // Use only lower triangular matrix for distance cache

// Keep class distribution in subsets
#define KEEP_CLASS_DISTRIBUTION

// For rotation of vectors
#define SET     1
#define INIT    2
#define NEXT    3
#define CLEAR   4

// Credit components for sub-individuals
#define N_CREDIT_COMPONENTS     6
#define DIFFERENCE              0
//#define REDUCTION             1
#define SUBSTITUTION            2
#define PERFORMANCE             3
#define FITNESS                 4
#define LOCAL_PERFORM           5

typedef struct {
    int nodes;
    int edges;
    bool use_dense_matrix, use_edge_list;
    float **adjacency_matrix; // Graph implemented as adjacency matrix
    t_Edge *edge_list;
    float *node_list;
    scalar size, max_size;  // In GB
} t_Graph;

typedef struct {
    int size;
    double *dec_value;
    cache_t ***nns;
    scalar mem_size;    // In Giga Bytes
} t_SubspaceCache;

// Prototypes
/* boost.c */
scalar UpdateAdaBoostWeights(const t_Parameters *, int *, scalar **, int, int *, scalar *, scalar *, int, int, bool, t_Classifier, int **, int *);
void UpdateGentleBoostWeightsISFS(const t_Parameters *, int *, scalar **, int, int *, scalar *, int, int, int **, scalar **, int *);
scalar kNearestNeighbourMarginSubspaceBoosting(int, int, int *, scalar **, int, int, int, int *, int, scalar *, scalar **, int **, int *);
scalar NearestNeighbourMarginSubspaceBoosting(int, int, int *, scalar **, int, int, int *, scalar *, scalar **, int **, int *);
scalar UpdateeReweightBoostWeights(const t_Parameters *, int *, int **, scalar **, int, int *, scalar *, scalar *, int, int, int, int, scalar *, int **, int *);
// scalar kNearestNeighbourMargin(int, int, int *, scalar **, int, int, int, int *, int, scalar *, scalar **, int **);
// scalar NearestNeighbourMargin(int, int, int *, scalar **, int, int, int *, scalar *, scalar **, int **);
/* boostis+fs.c */
int NumberOfInternalStrata(t_Parameters *, int, int);
/* chcisfs.c */
scalar CHCISplusFS(const t_Parameters *, int *, scalar **, int, int, int *, scalar *, bool,  int *, scalar *, int, scalar **, int, int *, scalar *, int **, scalar *);
void CopyIndividual(int *, int *, int);
bool AcceptIfBetter(const t_Parameters *, int *, int *, scalar *, scalar *, scalar *, scalar **, int, int, int *, scalar *, scalar *, int, cache_t ***, bool *, int, scalar *, int *, int *,
                    int **, t_SubspaceCache *);
int ConvertToDecimalAndUpdateCache(int *, bool *, cache_t ***, scalar **, int, int, bool, int *, int *, int **);
void FreeCache(cache_t ****, int, int);
void AllocateCache(cache_t ****, int, int);
void ObtainCache(cache_t **, scalar **, int, int, int *, int **);
/* ensemble.c */
scalar kNearestNeighbourEnsembleMarginSubspace(int, int **, int, scalar **, int, int, int, int *, int *, int, int, scalar *, bool *, bool);
scalar kNearestNeighbourEnsembleAccuracySubspace(int **, int, scalar **, int, int, int, int *, int, int, scalar *, bool);
int kNearestNeighbourEnsembleSubspace(int, int **, int, scalar **, int, int, int, int *, int *, int, int, scalar *, bool);
/* evalth-fs.c */
// scalar FSGraphDemocraticThresholdEvaluationForSelectionMethod(const t_Parameters *, scalar **, int *, int, int, int *, int, t_Graph *, int, int, scalar *, scalar *, t_ThresholdEvaluation);
// int SelectWithGraph(const t_Parameters *, t_Graph *, int, int *, scalar, scalar, t_ThresholdEvaluation);
void RemoveWithMinorThreshold(t_Graph *, int, int *, scalar, int *);
int CompareEdges(const void *, const void *);
scalar PerformanceAsPercentage(scalar, scalar);
// t_ThresholdEvaluation FSGraphDemocraticThresholdEvaluation(const t_Parameters *, scalar **, int *, int, int, int *, int, t_Graph *, int, int, scalar *, scalar *);
/* evalth-is.c */
// scalar ISGraphDemocraticThresholdEvaluationForSelectionMethod(const t_Parameters *, scalar **, int *, int, int, int *, int, t_Graph *, int, int, scalar *, scalar *, t_ThresholdEvaluation, int **);
// t_ThresholdEvaluation ISGraphDemocraticThresholdEvaluation(const t_Parameters *, scalar **, int *, int, int, int *, int, t_Graph *, int, int, scalar *, scalar *, int **);
int kNearestNeighbourWithListSubspaceSparse(scalar *, int *, scalar **, int, int, int, int *, int *, int *, int, bool, int *, int **);
/* evalth-isfs.c */
void ISFSDemocraticThresholdEvaluation(const t_Parameters *, scalar **, int *, int, int, int *, int, scalar *, scalar *, scalar **, int, int, scalar *, scalar *, int, int, int **);
/* features.c */
void FeaturePartition(int, int, int *, int **);
/* floatboost.c */
scalar UpdateeFloatBoostWeights(int **, scalar **, int, int *, scalar *, scalar *, int, int, int, t_Classifier, int, bool, int **);
int ConditionalExclusion(const t_Parameters *, int **, int, scalar **, int, int, int, int *, int, scalar *, scalar *, scalar *, FILE *, int *, int **);
/* graph.c */
bool Connected(t_Graph *, int);
void DFS(int, bool *, int, t_Graph *);
void PrintGraph(t_Graph *, int, FILE *, char *);
// bool TestIfGraphOfSelectedFeaturesIsConnected(const t_Parameters *, t_Graph *, int, int *, scalar, scalar, t_ThresholdEvaluation, FILE *);
int ConnectedComponents(t_Graph *, int, int *, int **, scalar, scalar);
void DFSWithThreshold(int, bool *, int, t_Graph *, int *, int *, scalar);
void AllocateAndInitializeGraph(t_Graph *, int, scalar);
void FreeGraph(t_Graph *);
void InitializeGraph(t_Graph *, int);
float GetValueFromEdge(t_Graph *, int, int);
float GetValueFromNode(t_Graph *, int);
void AssignValueToEdge(t_Graph *, int, int, scalar);
void AddValueToEdge(t_Graph *, int, int, scalar);
void AddValueToNode(t_Graph *, int, scalar);
void AssignValueToNode(t_Graph *, int, scalar);
scalar GraphMemorySizeInGB(t_Graph *);
void AddGraphToGraph(const t_Parameters *, t_Graph *, t_Graph *, scalar);
int GetListOfNodeValues(t_Graph *, scalar *);
int GetListOfEdgeValues(t_Graph *, float **, scalar);
void SortGraphByValue(t_Graph *, t_Graph *);
int CompareEdgesHighestToLowest(const void *, const void *);
void SaveGraph(t_Graph *, char *, int *, bool, bool, scalar, scalar);
void AddGraphToGraphWithConversion(const t_Parameters *, t_Graph *, t_Graph *, int *, scalar);
/* hash.c */
void CreateHashTable(t_HashTable *);
t_HashEntry *HashSearchEntry(t_HashTable, char *);
void HashAddEntry(t_HashTable *, t_HashEntry *);
void DestroyHashTable(t_HashTable *);
bool EqualKeys(char *, char *);
/* intelligent.c */
void IntelligentMOEA(const t_Parameters *, int *, int *, scalar **, int, int *, int, time_t, int);
void GPSIFF(scalar *, scalar **, int);
void IntelligentCrossover(const t_Parameters *, int **, int *[2], int *son, int *, int, int, int **, scalar **, scalar **, int, int *, int, int);
void EvaluateObjectives(const t_Parameters *, int *, scalar *, scalar **, int, int *, int, int, int );
bool DominatedBy(scalar *, scalar *, int);
bool TestIfDominated(scalar *, scalar **, int, int);
void OrthogonalArray(int, int **);
int RNNMutationIMOEA(const t_Parameters *, int *, scalar **, int, int *, int, int, int);
/* isfsindividual.c */
scalar EvaluateISFSIndividual(const t_Parameters *, int *, scalar *, scalar *, scalar **, int, int, int *, scalar *, scalar *, int, cache_t **, int, int, scalar *, int **);
int RNNMutationISFS(const t_Parameters *, int *, int, int *, scalar *, scalar **, int, bool, bool, bool, scalar *, int, int, int, scalar *);
scalar EvaluateISFS2ClassError(const t_Parameters *, int *, int, int, int *, scalar *, scalar *, scalar *, scalar **, int, cache_t **, int, scalar *, int **);
int InstanceSelectionMutation(const t_Parameters *, int *, int, int *, scalar *, scalar **, int, t_ISAlgorithm, scalar *, int, int **, int *);
int AddIorFMutation(const t_Parameters *, int *, int, int *, scalar *, scalar **, int, bool, bool, scalar *, int, int, int, scalar *);
scalar Fitness(const t_Parameters *, scalar, int, int, int, int, scalar *);
/* is+fs-stub.c */
void InstanceAndFeatureSelectionAlgorithm(const t_Parameters *, int **, scalar ***, int *, int, int **, int *, scalar *, int, int, int *, int **, scalar *);
/* isfsmutation.c */
int InstanceAndFeatureSelectionMutation(const t_Parameters *, int *, int, int *, scalar *, scalar **, int, t_ISFSAlgorithm, int);
/* learners-mutation.c */
int RNNMutationISFSForLearners(const t_Parameters *, int *, int, int *, scalar *, scalar **, int, bool, bool, bool, scalar *, int, int, int, bool, scalar *);
/* mfdr.c */
void MaximumFishersDiscriminantRatioEvaluation(const t_Parameters *, scalar **, int *, int, int, scalar *, scalar *, scalar *, scalar *, int, scalar *, int, int *);
scalar FishersDiscriminantRatio(scalar **, int *, int, int, int, int *);
// scalar AverageFishersDiscriminantRatio(scalar **, int *, int, int, int, int *);
scalar MaximumFishersDiscriminantRatio(scalar **, int *, int, int, int, int *, int *);
/* mfe.c */
void MulticlassMaximumFeatureEfficiency(const t_Parameters *, scalar **, int *, int, int, scalar *, scalar *, scalar *, scalar *, int, scalar *, int);
scalar MaximumFeatureEfficiency(scalar **, int, int *, int, int, int *, int, int *);
void ComplexityFusion(const t_Parameters *, scalar **, int *, int, int, scalar *, scalar *, scalar *, scalar *, int, scalar *, int, int *);
/* misc.c */
void AverageFeatureAndInstanceSelection(scalar *, scalar *, int *, int *, int, int, int);
int DuplicateFeatures(scalar **, int, int, int);
scalar Normal(scalar, scalar);
/* multiboost.c */
scalar UpdateMultiBoostWeights(const t_Parameters *, int *, scalar **, int, int *, scalar *, scalar *, int, int, bool, int, int, t_Classifier, int **, bool, int *);
/* parallel-is+fs.c */
time_t ParallelInstanceFeatureSelectionForAllStrata(const t_Parameters *, int, bool, bool, char *, scalar ***, int *, int **, int **, int **, int ***, int, int, char *, int, int, scalar, scalar *,
                                                    scalar *, int);
void SendDataToSlaveSubspace(int, int, int, int, int, int, scalar, int *, int **, scalar **, int, int);
int ReceiveSelection(int *, int *, int *, int *, int *, int *, int **, time_t *, int);
int RestartDeadSlaves(int *, int, char *, int, char *, int *, int *, int *, scalar ***, int **, int ***, int ***, int **, int, int, scalar, time_t *, int);
/* reduction.c */
void ObtainReductionISFS(const t_Parameters *, scalar **, int, int, scalar *, scalar *, scalar *, scalar *, int, int);
/* select.c */
// int SelectWithSortedGraph(const t_Parameters *, t_Graph *, int, int *, scalar, scalar, t_ThresholdEvaluation);
void SelectWithThreshold(int, int, int *, int *, int *, int *, scalar, scalar, scalar *, scalar *);
/* voor.c */
void MulticlassVolumeOfOverlapRegion(const t_Parameters *, scalar **, int *, int, int, scalar *, scalar *, scalar *, scalar *, int, scalar *, int);
void LOORThresholdEvaluation(const t_Parameters *, scalar **, int *, int, int, scalar *, scalar *, scalar *, scalar *, int, scalar *, int);
scalar MaxValueOfFeatureInClass(scalar **, int, int *, int, int, int *);
scalar MinValueOfFeatureInClass(scalar **, int, int *, int, int, int *);
scalar VolumeOfOverlapRegion(scalar **, int, int *, int, int, int, int *, int *);
/* wrapper-is+fs.c */
void ReceiveDataFromMaster(int, int *, int *, int *, int *, int *, int *, scalar *, int **, int, int ***, scalar ***, int);
void SendSelection(int, int, int, int *, int *, int, int, time_t, int);
/* classifier.c */
scalar GetWeightedError(const t_Parameters *, t_Classifier, int *, scalar **, int, int, int *, int, bool *, scalar *, bool, int **, int *);
/* n2.c */
scalar RIICComplexityMeasure(scalar **, int, int *, int, int *, int *);
void RIICThresholdEvaluation(const t_Parameters *, scalar **, int *, int, int, scalar *, scalar *, scalar *, scalar *, int, scalar *, int);
/* graph-is+fs.c */
// scalar ISFSGraphDemocraticThresholdEvaluationForSelectionMethod(const t_Parameters *, scalar **, int *, int, int, int *, int, t_Graph *, int, int, scalar *, scalar *, scalar *, t_ThresholdEvaluation);
void RemoveWithMinorThresholdInstancesAndFeatures(t_Graph *, int, int, int *s, scalar, int *);
// void ISplusFSSelectWithGraph(const t_Parameters *, t_Graph *, int, int, int *, scalar, scalar, scalar, t_ThresholdEvaluation, int *, int *);
int GetListOfEdgeValuesInstancesAndFeatures(t_Graph *, float **, int, int, scalar, scalar, bool);
int GetListOfNodeValuesInterval(t_Graph *, scalar *, int, int);
int ConnectedComponentsInstancesAndFeatures(t_Graph *, int, int, int *, int **, scalar, scalar, scalar, bool);
void PrintISGraphSummaryInfo(t_Graph *, int, int, FILE *);
/* subscache.c */
void FreeSubspaceCache(t_SubspaceCache *, int);
void AllocateSubspaceCache(t_SubspaceCache *);
double ConvertToDecimalAndUpdateSubspaceCache(int *, t_SubspaceCache *, scalar **, int, int, bool, int *, int *, int **, scalar);
cache_t **ReturnCache(bool, bool, cache_t ***, t_SubspaceCache *, double, bool *);
void AddNewSubspace(t_SubspaceCache *, int, scalar, scalar);
/* sifs2.c */
void CopyAndUpdateCache(scalar **, scalar **, int, int, scalar **, int, int **);
void ConstructNeighborsCacheFromDistances(int, cache_t **, scalar **);
scalar SIFS2Performance(const t_Parameters *, int *, int *, int *, int, int, int, t_PerformanceMeasure, int, cache_t **, scalar *, scalar *, scalar **, int **);
int ISFSWithFloatingSearch(const t_Parameters *, int *, int *, scalar **, int, int *, int, int, int, int **);
void DFSWithThresholdInstancesAndFeatures(int, bool *, int, int, t_Graph *, int *, int *, scalar);
int CompareAscending(const void *, const void *);
void CopyDistanceCacheWithList(scalar **, scalar **, int *, int);
void GetSqCacheWithSubsetSubspace(scalar **, int, scalar **, int, int **, int *, int *);
void ConstructNeighborsCacheFromDistancesWithSubset(int, cache_t **, scalar **, int *);
/* stratified.c */
scalar StratifiedEvaluationOfNextFeature(const t_Parameters *, scalar **, int, int, int *, int **, int *, int *, int, int, int, scalar *, scalar *, int *, int, int, t_ISAlgorithm);
/* rmhc-pf1.c */
scalar RMHC_PF1Algorithm(const t_Parameters *parm, int *, scalar **, int, int, int *, scalar *, int, scalar *, int **);
