/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include "macros.h"
#include "isfs.h"
#include "filter.h"

// This mutation is based on feature selection algorithms.
// Now implements: IMOEA

int InstanceAndFeatureSelectionMutation(const t_Parameters *parm, int *individual, int instances, int *label, scalar *w, scalar **data, int inputs, t_ISFSAlgorithm method,
                                        int classes)
{
    scalar **l_data;
    int selected_inputs = 0, selected_instances = 0, *S, *l_label, *l_individual, l_instances;
    char *name[] = { NULL, "IMOEA" };

    for (int i = 0; i < inputs; i++)
        if (individual[instances+i])
            selected_inputs++;

    // If no inputs, nothing to do
    if (selected_inputs == 0)
        return 0;

    // Create local data using selected features
    MALLOC(l_data, scalar *, instances, "InstanceAndFeatureSelectionMutation: l_data");
    ALLOCA(S, int, inputs, "InstanceAndFeatureSelectionMutation: S");

    // Modified 4/January/2011
    // Both possibilities are used
    if (parm->ls_only_use_selected) {
        ALLOCA(l_label, int, instances, "InstanceAndFeatureSelectionMutation: l_label");
        ALLOCA(l_individual, int, instances, "InstanceAndFeatureSelectionMutation: l_individual");

        // Only selected individuals are used
        for (int i = 0; i < instances; i++)
            if (individual[i]) {
                MALLOC(l_data[selected_instances], scalar, selected_inputs,
                       "InstanceAndFeatureSelectionMutation: l_data[selected_instances]");
                for (int j = 0, k = 0; j < inputs; j++)
                    if (individual[instances+j]) {
                        l_data[selected_instances][k] = data[i][j];
                        k++;
                    }
                l_label[selected_instances] = label[i];
                l_individual[selected_instances] = 1;

                selected_instances++;
            }

        l_instances = selected_instances;
    } else {
        l_label = label;
        l_individual = individual;

        // All individuals are used
        for (int i = 0; i < instances; i++) {
            if (individual[i])
                selected_instances++;

            MALLOC(l_data[i], scalar, selected_inputs, "InstanceAndFeatureSelectionMutation: l_data[i]");
            for (int j = 0, k = 0; j < inputs; j++)
                if (individual[instances+j]) {
                    l_data[i][k] = data[i][j];
                    k++;
                }
        }
        l_instances = instances;
    }

    // If no instances, nothing to do
    if (selected_instances == 0) {
        for (int i = 0; i < l_instances; i++)
            FREE(l_data[i]);
        FREE(l_data);
        return 0;
    }

    switch (method) {
        case IMOEA:
            IntelligentMOEA(parm, l_individual+l_instances, l_individual, l_data, l_instances, l_label, selected_inputs, parm->imoea_seconds, classes);
            break;

        default:
            OnErrorExit(ERR_PARAM, "Bad instance+feature selection mutation method");
            break;
    }

    // Restore individual
    int keptrain_instances = 0, kept_inputs = 0;
    for (int i = 0, j = 0; i < instances; i++)
        if (individual[i]) {
            individual[i] = l_individual[j];
            if (l_individual[j])
                keptrain_instances++;
            j++;
        }

    for (int i = 0, j = 0; i < inputs; i++)
        if (individual[instances+i]) {
            individual[instances+i] = l_individual[l_instances+j];
            if (l_individual[l_instances+j])
                kept_inputs++;
            j++;
        }

    for (int i = 0; i < l_instances; i++)
        FREE(l_data[i]);
    FREE(l_data);

    if (parm->verbose >= 3)
        fprintf(parm->out_file, "Instance+Feature selection method (%s) removed %d/%d instances and  %d/%d features\n", name[method], selected_instances -
                keptrain_instances, selected_instances, selected_inputs - kept_inputs, selected_inputs);

    return selected_instances - keptrain_instances+selected_inputs - kept_inputs;
}
