/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "isfs.h"
#include "macros.h"

void AverageFeatureAndInstanceSelection(scalar *average_is, scalar *average_fs, int *is, int *fs, int instances, int features, int c)
{
    int selected;
    scalar reduction;

    // Instances
    selected = 0;
    for (int i = 0; i < instances; i++)
        if (is[i])
            selected++;

    reduction = 1.0 - ((scalar) selected)/instances;

    *average_is = ((c-1)*(*average_is)+reduction)/c;

    // Features
    selected = 0;
    for (int i = 0; i < features; i++)
        if (fs[i])
            selected++;

    reduction = 1.0 - ((scalar) selected)/features;

    *average_fs = ((c-1)*(*average_fs)+reduction)/c;
}

int DuplicateFeatures(scalar **data, int instances, int inputs, int ratio)
{
    for (int i = 0; i < instances; i++) {
        REALLOC(data[i], scalar, inputs+(ratio*inputs), "DuplicateFeatures: data[i]");
        for (int j = 1; j <= ratio; j++)
            memcpy(data[i]+(j*inputs), data[i], sizeof(scalar)*inputs);
    }

    return inputs*(ratio+1);
}

// Generate a normal distributed value for N(m, s).
scalar Normal(scalar mean, scalar sigma)
{
    static scalar glgset;
    static int gliset = 0;
    scalar fac, r, v1, v2, gasdev;

    if (gliset == 0) {
        do {
            v1 = 2*RANDOM(0.0, 1.0)-1.0;
            v2 = 2*RANDOM(0.0, 1.0)-1.0;
            r = v1*v1 + v2*v2;
        } while (r >= 1);
        fac = sqrt(-2.0*log10(r) /r);
        glgset = v1*fac;
        gasdev = v2*fac;
        gliset = 1;
    } else {
        gasdev = glgset;
        gliset = 0;
    }
    return ((gasdev*sigma) + mean);
}
