/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas and                     *
 *   Juan Antonio Romero del Castillo                                      *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "isfs.h"
#include "macros.h"

#define SIGMOID(x, y)        (tanh(10.*exp(-.025*(y))*(x))/tanh(10.*exp(-.025*(y))))

void ObtainReductionISFS(const t_Parameters *parm, scalar **r, int instances, int features, scalar *votes_is, scalar *votes_fs, scalar *thresholds_is, scalar *threshold_fs, int n_thresholds_is,
                         int n_thresholds_fs)
{
    for (int is = 0; is < n_thresholds_is; is++) {
        // Count instances
        int sel_ins = 0;

        scalar th_is = thresholds_is[is];

        for (int i = 0; i < instances; i++)
            if (votes_is[i] >= th_is)
                sel_ins++;

        for (int fs = 0; fs < n_thresholds_fs; fs++) {
            // Count features
            int sel_feat = 0;

            scalar th_fs = threshold_fs[fs];

            for (int i = 0; i < features; i++)
                if (votes_fs[i] >= th_fs)
                    sel_feat++;

            if (!sel_feat || !sel_ins) {
                r[is][fs] = -1.0;
                continue;
            }

            // Get actual reduction
            scalar ired = 1.0 - ((scalar) sel_ins)/instances;
            scalar fred = 1.0 - ((scalar) sel_feat)/features;

            if (parm->eval_exp_redinstance)
                ired = SIGMOID(ired, instances);

            if (parm->eval_exp_redfeature)
                fred =  SIGMOID(fred, features);

            r[is][fs] = (parm->alpha_redfeat*fred + parm->alpha_redins*ired)/(parm->alpha_redfeat+parm->alpha_redins);
        }
    }
}
