/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es   *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "is.h"
#include "macros.h"

#define ACCEPTANCE      0.9
#define DROPPING        0.7

// Implements IB3 acceptance criterion

bool IsAcceptable(scalar attemps, scalar accuracy, scalar freq, scalar n)
{
    scalar lower_acc, upper_freq;

    // Lower bound of the accuracy of the instance
    lower_acc = (accuracy+(ACCEPTANCE*ACCEPTANCE)/(2.0*attemps)-ACCEPTANCE*sqrt((accuracy*(1.0-accuracy))/attemps +
                 (ACCEPTANCE*ACCEPTANCE)/(4.0*attemps*attemps)))/(1.0+(ACCEPTANCE*ACCEPTANCE)/attemps);

    upper_freq = (freq+(ACCEPTANCE*ACCEPTANCE)/(2.0*n)+ACCEPTANCE*sqrt((freq*(1.0-freq))/n +
                  (ACCEPTANCE*ACCEPTANCE)/(4.0*n*n)))/(1.0+(ACCEPTANCE*ACCEPTANCE)/n);

    if (lower_acc > upper_freq)
        return true;

    return false;
}

bool IsDroppable(scalar attemps, scalar accuracy, scalar freq, scalar n)
{
    scalar lower_freq, upper_acc;

    // Lower bound of the accuracy of the instance
    upper_acc = (accuracy+(DROPPING*DROPPING)/(2.0*attemps)+DROPPING*sqrt((accuracy*(1.0-accuracy))/attemps +
                 (DROPPING*DROPPING)/(4.0*attemps*attemps)))/(1.0+(DROPPING*DROPPING)/attemps);

    lower_freq = (freq+(DROPPING*DROPPING)/(2.0*n)-DROPPING*sqrt((freq*(1.0-freq))/n+(DROPPING*DROPPING)/(4.0*n*n)))/(1.0+(DROPPING*DROPPING)/n);

    if (upper_acc < lower_freq)
        return true;

    return false;
}
