/***************************************************************************
 *   Copyright (C) 2015 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Functions that implement Nearest Neighbor classification
#include <math.h>
#include "is.h"
#include "macros.h"

// Implements:
// J Wang, P. Neskovic and L. N. Cooper, "Improving nearest neighbor rule with a simple adaptive distance measure", Pattern Recognition Letters 28 (2007) 207–213

#define EPSILON 0.000001

void AdaptiveDistanceMeasureGetRadius(scalar **data, int instances, int *labels, int inputs, scalar *r)
{
    scalar nearest_enemy_dist, value;

    // Get distance to nearest enemy for every instance
    for (int i = 0; i < instances; i++) {
        nearest_enemy_dist = INFINITE;

        for (int j = 0; j < instances; j++)
            // Compute nearest enemy
            if (labels[i] != labels[j] && (value = EuclideanDistance(data[i], data[j], inputs)) < nearest_enemy_dist)
                nearest_enemy_dist = value;


        r[i] = nearest_enemy_dist - EPSILON;
    }
}
