/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <stdbool.h>
#include "is.h"
#include "macros.h"
#include "common.h"

// Add the k neighbors of the selected instances
int AddKNeighbors(scalar **data, int instances, int *labels, int *list, scalar **sel_data, int *sel_labels, int *sel_instances, int *sel_list, int k, cache_t **neighbors_cache,
                  int **sel_index, int **index)
{
    bool *to_add, *selected;
    int n;

    if (!neighbors_cache)
        OnErrorExit(1, "Add k neighbors without cache not yet implemented");

    MALLOC(to_add, bool, instances, "AddKNeighbors: to_add");
    MALLOC(selected, bool, instances, "AddKNeighbors: selected");
    for (int i = 0; i < instances; i++)
        to_add[i] = false;

    // Get k neighbors of sampled instances
    for (int l = 0; l < instances; l++) {
        selected[l] = false;
        for (int i = 0; i < *sel_instances && !selected[l]; i++)
            if (list[l] == sel_list[i])
                selected[l] = true;

        if (selected[l])
            for (int j = 0; j < k+1; j++)
                to_add[neighbors_cache[l][j]] = true;
    }

    // Count new size
    int new_instances = *sel_instances;
    for (int i = 0; i < instances; i++)
        if (to_add[i] && !selected[i])
            new_instances++;

    // Add remaining if not already selected
    n = 0;
    for (int i = 0; i < instances; i++)
        if (to_add[i] && !selected[i]) {
            sel_data[*sel_instances] = data[i];
            sel_labels[*sel_instances] = labels[i];
            sel_list[*sel_instances] = list[i];
            if (index)
                sel_index[*sel_instances] = index[i];
            (*sel_instances)++;
            n++;
        }

    FREE(to_add);
    FREE(selected);

    return n;
}
