/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "is.h"
#include "common.h"
#include "macros.h"

// WARNING: Binary files are always scalar = scalar

// Functions that deal with binary files
void GetInstanceFromBinaryFile(scalar *data, int *label, int index, int inputs, FILE *binary)
{
    static long previous = 0;

    // Check
    if (sizeof(scalar) != sizeof(float))
        OnErrorExit(1, "Binary files must be float");

    long position = ((long) index) * (sizeof(scalar) *inputs+sizeof(int));

    // Seek to instance
    if (previous == 0 || previous > position) {
        fseek(binary, position, SEEK_SET);
        previous = position+(sizeof(scalar) *inputs+sizeof(int));
    } else {
        fseek(binary, position - previous, SEEK_CUR);

        previous = position+(sizeof(scalar) *inputs+sizeof(int));
    }

    // Read data
    if (!fread(data, sizeof(scalar), inputs, binary))
        OnErrorExit(1, "GetInstanceFromBinaryFile: Error reading binary file");
    if (!fread(label, sizeof(int), 1, binary))
        OnErrorExit(1, "GetInstanceFromBinaryFile: Error reading binary file");

#ifdef AUTOTEST
    if (previous != 0 && previous != ftell(binary)) {
        printf("Previous: %ld\tftell(): %ld\n", previous, ftell(binary));
        OnErrorExit(1, "Bad fseek");
    }
#endif
}

// Read random block of instances
int ReadBlockFromBinaryFile(scalar **data, int *label, int inputs, int start, int end, int instances, FILE *binary, int *list)
{
    int next;

    // Check
    if (sizeof(scalar) != sizeof(float))
        OnErrorExit(1, "Binary files must be float");

    int count = (start < end) ? end - start : end+instances - start;
    long position = ((long) start) * (sizeof(scalar) * inputs+sizeof(int));

    next = start;

    // Seek to first instances
    fseek(binary, position, SEEK_SET);

    // Read data
    for (int i = 0; i < count; i++) {
        if (!fread(data[i], sizeof(scalar), inputs, binary))
            OnErrorExit(1, "ReadBlockFromBinaryFile: Error reading block from binary file");
        if (!fread(&label[i], sizeof(int), 1, binary))
            OnErrorExit(1, "ReadBlockFromBinaryFile: Error reading block from binary file");
#ifdef AUTOTEST
        if (feof(binary))
            OnErrorExit(1, "Read after end of file");
#endif
        // If end of file rewind
        if (i+start+1 == instances)
            rewind(binary);
        if (list != NULL) {
            list[i] = next;
            next = (next+1) % instances;
        }
    }

    return count;
}

// Read all minority class instances
int ReadMinorityClass(scalar **data, int minority, int inputs, FILE *binary, int instances, int minority_instances)
{
    int label;

    // Check
    if (sizeof(scalar) != sizeof(float))
        OnErrorExit(1, "Binary files must be float");

    // Seek to first instance
    rewind(binary);

    // Read data
    int next = 0;
    for (int i = 0; i < instances && next < minority_instances; i++) {
        if (!fread(data[next], sizeof(scalar), inputs, binary))
            OnErrorExit(1, "ReadMinorityClass: Error reading minority class");
        if (!fread(&label, sizeof(int), 1, binary))
            OnErrorExit(1, "ReadMinorityClass: Error reading minority class");

        // Minority class?
        if (label == minority)
            next++;
    }

    return next;
}
