/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include "is.h"
#include "common.h"
#include "macros.h"

#define kFOLDS  10

scalar EvaluateSelectionCICoefficient(const t_Parameters *parm, scalar **data, int instances, int *labels, int *S, int minority, int inputs, int classes)
{
    int *l_labels, *l_list, selected = 0, nearest[parm->k], *partition;
    scalar TP, TN, FP, FN, **l_data;

    // Number of subsets
    int n_subsets = MAX(1, instances/parm->eval_subset_size);
    int count[n_subsets];

    // Obtain random partition for the evaluation
    MALLOC(partition, int, instances, "EvaluateSelectionCICoefficient: partition");
    RandomPartition(partition, count, instances, n_subsets, labels, classes);

    // Evaluate for each subset using k-fold cv
    TP = TN = FP = FN = 0.0;

    // Allocate memory
    int max_size = count[0];
    for (int i = 1; i < n_subsets; i++)
        if (max_size < count[i])
            max_size = count[i];

    ALLOCA(l_data, scalar *, max_size, "EvaluateSelectionCICoefficient: l_data");
    ALLOCA(l_labels, int, max_size, "EvaluateSelectionCICoefficient: l_labels");
    ALLOCA(l_list, int, max_size, "EvaluateSelectionCICoefficient: l_list")

    for (int subset = 0; subset < n_subsets; subset++) {
        int cursor = 0, l_selected = 0;
        for (int j = 0; j < instances; j++)
            if (partition[j] == subset) {
                l_data[cursor] = data[j];
                l_labels[cursor] = labels[j];

                // Selected?
                if (S[j]) {
                    l_list[l_selected] = cursor;
                    l_selected++;
                }

                cursor++;
            }

        selected += l_selected;

        for (int fold = 0; fold < kFOLDS; fold ++) {
            int first = (fold * cursor)/kFOLDS;
            // Account for remaining instances
            int last = ((fold == kFOLDS - 1) ? cursor : ((fold+1) * cursor)/kFOLDS);

            // Remove selected from fold
            int fold_selected = l_selected;
            for (int i = 0; i < fold_selected; i++)
                if (l_list[i] >= first && l_list[i] < last) {
                    // Exchange with last
                    fold_selected--;
                    int sw = l_list[i];
                    l_list[i] = l_list[fold_selected];
                    l_list[fold_selected] = sw;
                }

            // Evaluate fold
            for (int i = first; i < last; i++) {
                bool r = (kNearestNeighbourWithList(l_data[i], l_list, l_data, fold_selected, inputs, parm->k, nearest, l_labels, classes, parm->knn_with_weighted_votes, NULL, NULL) == l_labels[i]);

                if (r) {
                    if (labels[i] == minority)
                        TP++;
                    else
                        TN++;
                } else {
                    if (labels[i] == minority)
                        FN++;
                    else
                        FP++;
                }
            }
        }
    }

    // Check
    if (TN+TP+FN+FP != instances)
        OnErrorExit(1, "Error evaluating thresholds with G-mean");

    scalar specificity = (TN+FP) > ZERO ? TN/(TN+FP) : 0.0;
    scalar sensitivity = (TP+FN) > ZERO ? TP/(TP+FN) : 0.0;
    scalar G = pow(sensitivity*specificity, 0.5);
    scalar reduction = 1.0 - ((scalar) selected)/instances;

    FREE(partition);

    return parm->alpha_acc * G+(1.0 - parm->alpha_acc) * reduction;
}
