/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include "is.h"
#include "macros.h"
#include "common.h"

int EditedkNN(const t_Parameters *parm, scalar ***data, int *instances, int **labels, int classes, int inputs, scalar *r)
{
    bool *remove;
    int nearest[parm->k], count[classes], removed = 0, *S;

    MALLOC(remove, bool, *instances, "EditedkNN: remove");
    MALLOC(S, int, *instances, "EditedkNN: S");

    for (int i = 0; i < classes; i++)
        count[i] = 0;

    for (int i = 0; i < *instances; i++) {
        remove[i] = false;
        S[i] = 1;
        count[(*labels)[i]]++;
    }

    for (int i = 0; i < *instances; i++) {
        S[i] = 0;
        if (kNearestNeighbour((*data)[i], S, *data, *instances, inputs, parm->k, nearest, *labels, classes, parm->knn_with_weighted_votes, r, NULL, NULL, NULL) != (*labels)[i])
            remove[i] = true;
        S[i] = 1;
    }

    // Remove all marked
    for (int i = 0; i < *instances; i++)
        // Do not let empty classes
        if (remove[i] && (count[(*labels)[i]] > 1)) {
            removed++;
            count[(*labels)[i]]--;

            (*labels)[i] = (*labels)[*instances-1];
            (*data)[i] = (*data)[*instances-1];
            remove[i] = remove[*instances-1];
            (*instances)--;
            i--;
        }

    fprintf(parm->out_file, "Edited k-NN removed %d instances\n", removed);

    if (removed) {
        REALLOC(*data, scalar *, *instances, "EditedkNN: *data");
        REALLOC(*labels, int, *instances, "EditedkNN: *labels");
    }

    FREE(remove);
    FREE(S);

    return removed;
}
