/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// Implements instance selection based on entropy following:
// S-H. Son and J-Y. Kim, "Data reduction for instance-based learning using entropy-based partitioning",
// Lecture Notes on Computer Science, no. 3982, pp. 590-599, 2006.

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "is.h"
#include "macros.h"
#include "common.h"

int EntropyInstanceSelection(int *S, scalar **data, int instances, int *labels, int inputs, int classes, int k, bool knn_with_weighted_votes, int **index)
{
    int *partition, n_subsets, *subspace;

    k = 1;

    // Initially all in the same subset
    MALLOC(partition, int, instances, "main: partition");
    for (int i = 0; i < instances; i++)
        partition[i] = 0;

    // Initially consider all inputs
    MALLOC(subspace, int, inputs, "main: subspace");
    for (int i = 0; i < inputs; i++)
        subspace[i] = true;

    n_subsets = 1;
    PerformPartition(data, instances, partition, 0, &n_subsets, subspace, labels, NULL, inputs, classes);

    // Obtain subspace considering the variables used for data partition
    for (int i = 0; i < inputs; i++)
        subspace[i] = !subspace[i];

    // For each subset select center and k nearest instances
    // Reorder subset labels
    n_subsets = ReorderSubsetsLables(partition, instances);

    // Un-select all instances
    for (int i = 0; i < instances; i++)
        S[i] = 0;

    // Select representatives of each subset
    for (int i = 0; i < n_subsets; i++)
        RepresentativesOfSubset(data, instances, partition, i, subspace, S, labels, inputs, classes, knn_with_weighted_votes, index);

    // Count selected instances
    int selected = 0;
    for (int i = 0; i < instances; i++)
        if (S[i])
            selected++;

    return selected;
}
