/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "is.h"
#include "common.h"
#include "macros.h"


// Random partition for huge datasets. Block of subset-size are contiguous
void RandomHugePartition(int *partition, int *count, int instances, int n_subsets, int size)
{
    // Adjust size to number of subsets
    size = instances/n_subsets;
    for (int i = 0; i < n_subsets; i++)
        count[i] = size;

    // Distribute remainder equally
    if (instances % n_subsets)
        for (int i = 0; i < instances - n_subsets*size; i++)
            count[(int) RANDOM(0, n_subsets)]++;

    // Next, chunks of size "size"
    partition[0] = RANDOM(0, instances);
    for (int i = 1; i < n_subsets; i++)
        partition[i] = (partition[i-1]+count[i-1]) % instances;

    partition[n_subsets] = partition[0];

    // Check count of instances
    for (int i = 0; i < n_subsets; i++)
        if (count[i] != ((partition[i] < partition[i+1]) ? partition[i+1] - partition[i] : partition[i+1]+instances - partition[i]))
            OnErrorExit(1, "Subset count does not match");
}

// Undersample majority class without loading the whole dataset in memory
void UndersampleHugeData(scalar ***data, int *instances, int inputs, int **labels, char *file_name, char *datad)
{
    int CompareInt(const void *, const void *);

    FILE *binary;
    char file[1000];

    char *home = getenv("HOME");
    sprintf(file, "%s.train.bin", file_name);
    if (access(file, R_OK))
        sprintf(file, "%s/%s/%s.train.bin", home, datad, file_name);
    if ((binary = fopen(file, "rb")) == NULL)
        OnErrorExit(1, "Cannot open binary training file");

    // Count classes
    int count[] = {0, 0};
    for (int i = 0; i < *instances; i++)
        count[(*labels) [i]]++;

    int minority = (count[0] < count[1]) ? 0 : 1;

    // List of instances
    int M = 0, m = 0;
    int *list;
    MALLOC(list, int, *instances, "UndesampleHugeData: list");
    for (int i = 0; i < *instances; i++)
        if ((*labels) [i] == minority) {
            list[m] = i;
            m++;
        } else {
            list[M+count[minority]] = i;
            M++;
        }

    if (M+m != *instances)
        OnErrorExit(1, "Bad instances count");

    // Randomize majority samples
    RandomizeVector(list+count[minority], *instances - count[minority]);

    // New number of instances
    *instances = 2*count[minority];

    // Sort to access disk faster
    qsort(list, *instances, sizeof(int), CompareInt);

    // Read data
    if (*labels)
        FREE(*labels);
    MALLOC(*labels, int, *instances, "UndesampleHugeData: *labels");
    MALLOC(*data, scalar *, *instances, "UndesampleHugeData: *data");

    for (int i = 0; i < *instances; i++) {
        MALLOC((*data) [i], scalar, inputs, "UndesampleHugeData: (*data)[i]");
        GetInstanceFromBinaryFile((*data) [i], & ((*labels) [i]), list[i], inputs, binary);
    }

    fclose(binary);

    FREE(list);
}
