/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "is.h"
#include "common.h"
#include "macros.h"

// Measures of dependence/independence of a partition

// Percentage of times classification is the same within partition and with all instances
scalar Independence(scalar **data, int instances, int *partition, int k, int *labels, int inputs, int n_subsets, int classes, bool knn_with_weighted_votes)
{
    scalar indep = 0.0;
    int nn[k], *S, **S_partition;

    MALLOC(S, int, instances, "Independence: S");
    MALLOC(S_partition, int *, n_subsets, "Independence: S_partition");
    for (int i = 0; i < n_subsets; i++)
        CALLOC(S_partition[i], int, instances, "Independence: S_partition[i]");
    for (int i = 0; i < instances; i++) {
        S[i] = 1;
        S_partition[partition[i]][i] = 1;
    }

    // Check whether classification is the same within partition and with all instances
    for (int i = 0; i < instances; i++) {
        // Get classification
        S[i] = 0;
        int withall = kNearestNeighbour(data[i], S, data, instances, inputs, k, nn, labels, classes, knn_with_weighted_votes, NULL, NULL, NULL, NULL);
        S[i] = 1;

        S_partition[partition[i]][i] = 0;
        int within = kNearestNeighbour(data[i], S_partition[partition[i]], data, instances, inputs, k, nn, labels, classes, knn_with_weighted_votes, NULL, NULL, NULL, NULL);
        S_partition[partition[i]][i] = 1;

        if (withall == within)
            indep++;
    }

    FREE(S);
    for (int i = 0; i < n_subsets; i++)
        FREE(S_partition[i]);
    FREE(S_partition);

    indep /= instances;

    return indep;
}
