/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Definitions and prototypes
#pragma once

#include <stdbool.h>
#include <stdio.h>
#include <time.h>
#include "typex.h"
#include "common.h"
#include "svm.h"
#include "fs.h"
#include "ranking.h"

// #define AUTOTEST             // Set auto-test code
// #define VERY_VERBOSE

// #define PARALLEL_IMPLEMENTATION // If defined some programs are run in parallel using PVM

// Objectives for individuals
#define N_OBJECTIVES    2
#define OK              0
#define REDUCTION       1

#define CACHE_USE_TRIANGULAR    // Use only lower triangular matrix for distance cache

// Keep class distribution in subsets
#define KEEP_CLASS_DISTRIBUTION

// Credit components for sub-individuals
#define N_CREDIT_COMPONENTS     6
#define DIFFERENCE              0
//#define REDUCTION             1
#define SUBSTITUTION            2
#define PERFORMANCE             3
#define FITNESS                 4
#define LOCAL_PERFORM           5

typedef struct {
     int mi;
     scalar **vectors;
     int *labels;
} t_LVQ_net;

typedef struct {
     int init, end, size;
} t_Range;

// For hash tables in simultaneous is+fs
typedef struct {
     char *key;
     scalar ok;
} t_HashEntry;

typedef struct {
     int size, hits;
     t_HashEntry *entries;
} t_HashTable;

// Sending data to slave
typedef struct {
     int elector, subset;
     int tid;
     int max_subset_size;
     int *partition;
     FILE *binary;
} t_Arguments;

// For vote ordering
typedef struct {
     int instance;
     int votes;
} t_VoteRank;

// Prototypes
/* accept.c */
bool IsAcceptable(scalar, scalar, scalar, scalar);
bool IsDroppable(scalar, scalar, scalar, scalar);
/* adaptive.c */
void AdaptiveDistanceMeasureGetRadius(scalar **, int, int *, int, scalar *);
/* addk.c */
int AddKNeighbors(scalar **, int, int *, int *, scalar **, int *, int *, int *, int, cache_t **, int **, int **);
/* best_round.c */
int GetBestRound(const t_Parameters *, byte **, int, int, FILE *, char *, scalar *, int, int, int, scalar **, int *);
void ParallelDemocraticRoundEvaluation(const t_Parameters *, scalar **, int *, int, int *, int, byte **, scalar *, scalar *, scalar *, int, int *, char *, FILE *, int, int, int);
/* binary.c */
void GetInstanceFromBinaryFile(scalar *, int *, int, int, FILE *);
int ReadBlockFromBinaryFile(scalar **, int *, int, int, int, int, FILE *, int *);
int ReadMinorityClass(scalar **, int, int, FILE *, int, int);
/* boost.c */
scalar UpdateAdaBoostWeightsIS(const t_Parameters *, int *, scalar **, int, int *, scalar *, scalar *, int, int, bool, t_Classifier, int **);
scalar Poisson(void);
void UpdateGentleBoostWeightsIS(const t_Parameters *, int *, scalar **, int, int *, scalar *, int, int, int **, scalar **);
scalar UpdateeReweightBoostWeightsIS(const t_Parameters *, int *, int **, scalar **, int, int *, scalar *, scalar *, int, int, int, int, cache_t **, scalar *, int **);
/* boostedis.c */
int BoostBasedInstanceSelection(const t_Parameters *, int *selection, scalar **, int, int *, bool, int, int, bool, scalar *, scalar *, int, bool, int **, scalar **, cache_t **, int *);
scalar EvaluatekWithWeights(int, int *, cache_t **, int *, int, int, scalar **, int, bool, scalar *, scalar *, scalar *, bool *, bool, int *, bool, scalar **, scalar *, int, int, int *);
void UpdateWeights(const t_Parameters *, int, scalar *, scalar, bool *, bool, t_EnsembleMethod, scalar **, int *, int *, int, int, int, scalar **, int **, bool, bool, scalar *, t_Classifier, int *);
scalar PerformanceWithCache(int, int *, cache_t **, int *, int, int, scalar **, int, bool, scalar *, scalar *, int *, bool, t_PerformanceMeasure, int, int *);
/* boostis.c */
int NumberOfInternalStrata(t_Parameters *, int, int);
/* cache.c */
void ConstructAllNeighborsCache(scalar **, int, int, scalar ***, cache_t ***, scalar *, int **, int *, int *, bool);
void ConstructkNeighborsCache(scalar **, int, int, cache_t ***, scalar *, int **, int, int *);
/* cfalgo.c */
int ConsensusFilter(const t_Parameters *, int *, scalar **, int, int *, int, int, int **, int *);
/* chcindivs.c */
scalar EvaluateCHCIndividual(const t_Parameters *, int *, scalar *, scalar *, scalar **, int, int *, scalar *, int, cache_t **, int, int, bool, scalar *, int **, scalar, scalar, int *, int);
scalar Evaluate2ClassError(const t_Parameters *, int *, int, int *, scalar *, int, scalar **, cache_t **, int, int, bool, scalar *, int **);
scalar EvaluateAUC(const t_Parameters *, int *, int, int *, scalar *, int, cache_t **, int, bool, scalar *);
int RNNMutationCHC(const t_Parameters *, int *, int, int, int *, scalar *, int, cache_t **, scalar **, int, int, scalar *, bool, scalar *, int *);
int RNNMutationCHCUsingGMeasure(const t_Parameters *, int *, int, int, int *, scalar *, int, cache_t **, int, scalar **, bool, scalar *);
/* ci.c */
int RandomUndersampling(int *, int, int *, int, scalar);
/* ci-subsample.c */
int ClassImbalanceRandomSubsampling(int, int *, scalar, int, int [2], int *);
int ClassImbalanceSubsampling(int, int *, scalar, int, int [2], int *, scalar *);
/* ci-threshold.c */
int SelectInstancesCIUsingThreshold(const t_Parameters *, int *, int, int *, int *, int, int, int *, int, int *, int *);
int GetThresholdForImbalancedClass(const t_Parameters *, scalar **, int, int *, int, int classtoget, int *, int other_threshold, scalar *, int, int);
void GetBothThresholdsForImbalancedClass(const t_Parameters *, scalar **, int, int *, int, int *, int *, int *majority_th, scalar *, int, int, scalar *rr);
int CompareVotes(const void *, const void *);
/* cnn.c */
int CondensedNearestNeighborRule(int *, scalar **, int, int *, int, bool, int, cache_t **, bool *, int, int, int **, int *);
int CNNAndTomekLinks(int *, scalar **, int, int *, int, int, int, int, int **);
int OneSidedSelection(int *, scalar **, int, int *, int, int, int, int, int **, cache_t **, bool);
void GetTomekLinks(int *, scalar **, int, int *, int, int *, int *, int *, int, int **, cache_t **);
/* coeff.c */
scalar EvaluateSelectionCICoefficient(const t_Parameters *, scalar **, int, int *, int *, int, int, int);
/* cpm.c */
int ClassPurityMaximization(int *, scalar **, int, int *, int, int);
void RecursiveCPM(scalar imp, int *parent);
/* decoupled.c */
void DecoupledDataPartitioning(int, int *, int, int *, scalar, int n_classes, int *, int **, bool *);
/* drop3algo.c */
int Drop3(const t_Parameters *, int *, scalar **, int, int *, bool, scalar *, int, int, cache_t **, scalar **, bool *, int **, int *, bool);
/* editedknn.c */
int EditedkNN(const t_Parameters *, scalar ***, int *, int **, int, int, scalar *);
/* eisalgo.c */
scalar EvolutionaryInstanceSelection(const t_Parameters *, int *, scalar **, int, int *, scalar *, bool, bool, bool, bool, int, int *, scalar *, int, int, t_EvolutionaryScheme, bool, int, int,
                                     scalar *, int **, cache_t **, int *);
void CheckMultiselectionOfMajorityClass(int *, int, int, int *);
/* ensemble.c */
int kNearestNeighbourEnsemble(int, int **, int, scalar **, int, int, int, int *, int *, int, int, cache_t **, scalar *, bool, int **);
scalar kNearestNeighbourEnsembleMargin(int, int **, int, scalar **, int, int, int, int *, int *, int, int, cache_t **, scalar *, bool *, bool, int **);
scalar kNearestNeighbourEnsembleAccuracy(int **, int, scalar **, int, int, int, int *, int, int, cache_t **, scalar *, bool, int **);
/* entropy.c */
int EntropyInstanceSelection(int *, scalar **, int, int *, int, int, int, bool, int **);
/* eval-threshold.c */
void DemocraticThresholdEvaluation(const t_Parameters *, scalar **, int *, int, int *, int, scalar *, scalar *, scalar *, scalar *coef, int, int);
int ParallelDemocraticThresholdEvaluation(const t_Parameters *, scalar **, int *, int, int *, int, scalar *, scalar *, scalar *, scalar *coef, int, int *, char *, FILE *, int, char *out, int, int);
/* evolocalk.c */
void EvolutionaryLocalK(t_Parameters *, int *, scalar **, int, int *, scalar *, bool, bool, bool, int, int *, scalar *, int, int, t_EvolutionaryScheme, scalar *, int **);
scalar EvaluateLocalKIndividual(const t_Parameters *, int *, scalar **, int, int *, scalar *, int, int **, int, int, scalar *, scalar *, scalar **, scalar *, scalar *, int **);
int kNearestNeighbourLargeCache(const t_Parameters *, int, int *, int, int *, int *, int, int **, scalar **, int, scalar *, int, int, int, scalar **, scalar *, int **);
bool RandomIntegerMutation(int *, int, scalar, int, int);
scalar EvaluateLocalK(const t_Parameters *, int *, scalar **, int, int *, int, int, scalar **, int, int *, int, scalar *, int **, int **);
void Crossover(const t_Parameters *, int *son, int *, int *, int *, int s);
void MemeticIntegerMutation(const t_Parameters *, int *, int, scalar, int, int, scalar **, int, scalar *, int *, int **, int, int, int, scalar *, scalar **, int, scalar *, int **);
int PseudoHammingDistance(int *, int *, int);
/* fasta2binary.c */
void ConvertToFloat(char *sequence, float *, int);
/* floatboost.c */
scalar UpdateeFloatBoostWeightsIS(int **, scalar **, int, int *, scalar *, scalar *, int, int, int, cache_t **, t_Classifier, int, bool, int **);
int ConditionalExclusionIS(const t_Parameters *, int **, int, scalar **, int, int, int, int *, int, scalar *, cache_t **, scalar *, scalar *, FILE *, int *t, int **);
/* gcnn.c */
int GeneralizedCondensedNearestNeighborRule(int *, scalar **, int, int *, int, cache_t **, scalar, int, scalar **, int, int **, int *);
/* hmnei.c */
int HitMissNetworkForEditingIterate(const t_Parameters *, int *, int, cache_t **, int, int *, scalar **, int, int, int **, int *);
int HitMissNetworkForEditing(int *, int, cache_t **, int, int *, scalar);
bool ***HitMissNetwork(int, cache_t **, int, int *, int *);
/* huge.c */
void RandomHugePartition(int *, int *, int, int, int);
void UndersampleHugeData(scalar ***, int *, int, int **, char *, char *);
/* ib3algo.c */
int IB3Algorithm(int *, scalar **, int, int *, int, scalar *, int, cache_t **, scalar **, int **, int *);
/* icfalgo.c */
int ICFAlgorithm(const t_Parameters *, int *, scalar **, int, int *, bool, int, int, cache_t **, scalar **, bool *, int **, int *);
/* independence.c */
scalar Independence(scalar **, int, int *, int, int *, int, int, int, bool);
/* individuals.c */
void CreateInitialPopulation(int ***, int, int, int);
scalar EvaluateIndividual(int *, int indiv, cache_t **, int, int *);
void MutateIndividual(int *, int, int, scalar);
scalar EvaluateAccuracy(int *, scalar **, int, int *, int);
/* isde.c */
scalar DifferentialEvolutionOfMultiselectedInstancesAndK(const t_Parameters *, scalar **, int, int, int *, int, int *, int, scalar, scalar, scalar, time_t, int *, int *, scalar *, bool, scalar *);
// void DifferentialMutation(const t_Parameters *, scalar *, scalar *, scalar *, scalar *, int, scalar);
// void Recombination(const t_Parameters *, scalar *, scalar *, scalar *, scalar, int);
scalar EvaluateIndividualDE(const t_Parameters *, int *, scalar *, scalar **, int, int, int *, int, int, cache_t **, int *, scalar *);
void ConvertToSelectionAndk(const t_Parameters *, int *, scalar *, int, int *, scalar *, int *, int);
bool LocalSearchIS(const t_Parameters *, int *, scalar *, int, int *, int, cache_t **, scalar **, int, int, scalar *, int, scalar *);
scalar LocalSearchK(const t_Parameters *, int *, scalar *, int, int *, int, cache_t **, scalar **, int, int, int *, scalar *);
int RandomUndersamplingDE(const t_Parameters *, int *, scalar *, int, int *, int, int);
/* is-stub.c */
int InstanceSelectionAlgorithm(const t_Parameters *, t_ISAlgorithm, int **, scalar ***, int *, int, int **, int, int **, cache_t **, scalar **, int, int *, scalar, scalar *, bool *, scalar *, scalar *,
                               bool, scalar *, bool, int *);
/* localk.c */
void OptimalLocalK(const t_Parameters *, int *, scalar **, int, int *, scalar *, bool, int, int, int, int *, bool, scalar *, int **);
scalar KappaForLocalK(const t_Parameters *, scalar **, int, int *, int *, int, int, scalar **, int, int *, FILE *, scalar *, scalar *, scalar *, int *, int **, int **, int **);
/* lvq.c */
t_LVQ_net *LVQCreate(int, int);
void LVQLearning(t_LVQ_net *lvq, scalar **, int, int *, int, int, int *, int, int);
void NearestNeighbourAndEnemy(scalar *, scalar **, int, int, int *, int *, int *, int);
scalar TestLVQNetwork(t_LVQ_net *lvq, scalar **, int, int *, int);
int OutputLVQNetwork(t_LVQ_net *lvq, scalar *, int);
/* meme.c */
void MemeOptimization(int *, cache_t **, int, int *);
/* misc.c */
int VectorMaxIndex(int *, int);
void CopyInstance(scalar *, scalar *, int);
scalar AngleBetweenVectors(scalar *, scalar *, int);
scalar VectorModule(scalar *, int);
void RandomizeForValidation(int *, int, int, int *, int);
/* mssalgo.c */
int MSSAlgorithm(int *, scalar **, int, int *, int, cache_t **, scalar **, int **, int *);
/* mutation.c */
bool RandomMutation(const t_Parameters *, int *, int, scalar, int *, int, bool, int, int, int, int);
void RNNMutation(int, int, cache_t **, scalar **, int, int, int *);
void NonUniformMutation(int *, int, int, int, double, int, int);
double DeltaNUM(int, double, int);
/* nnauc.c */
scalar ObtainAUCForNearestNeighbor(scalar **, int, int *, int *, scalar **, int, int *, int, int, bool, scalar *);
scalar f(scalar, scalar **, int);
scalar trapzd(scalar **, scalar, scalar, int, int);
scalar qtrap(scalar **, scalar, scalar, int);
/* operators.c */
void TwoPointCrossover(int *, int *, int *, int *, int);
void HuxCrossover(int *, int *, int *, int *, int);
int HammingDistance(int *, int *, int);
void UniformCrossover(int *, int *, int *, int *, int);
void BiasedHuxCrossover(int *, int *, int *, int *, int, scalar);
/* parallel-is-.c */
time_t ParallelInstanceSelectionForAllStrata(const t_Parameters *, int, bool, bool, char *, scalar ***, int *, int **, int **, int ***, int, int, char *, int, int, scalar);
void SendDataToSlave(int, int, int, int, int, int, scalar, int *, int **, scalar **);
/* parallel-knn.c */
scalar ParallelkNN(const t_Parameters *, int *, int, int, int, int, char *, char *, scalar *, scalar *, scalar *, scalar *, char *out, int, int);
/* parallel-.c */
void ParallelOptimalLocalK(const t_Parameters *, int *, scalar **, int, int *, scalar *, bool, int, int, int, int *, bool, scalar *, int, char *, char *);
/* par-ci-threshold.c */
void ParallelGetBothThresholdsForImbalancedClass(scalar **, int, int *, int, int *, int *, int *, char *, int, scalar *, char *, FILE *, int, int);
/* par-knn-.c */
void GetNNConfusionMatrix(const t_Parameters *, int **, scalar **, int, int *, scalar **, int, int *, int, int, int, char *, int **, int **, int *);
int IndexOfMinimumValue(scalar *, int);
/* part3.c */
int *Partition(int, int, scalar);
/* partial.c */
scalar PartialEvaluationOfLocalKIndividual(const t_Parameters *, int *, scalar **, int, int *, scalar *, int, int **, int, int, scalar *, scalar *, scalar **, scalar *, int, int, int *, scalar *, int **);
/* .c */
int LowestEntropy(scalar **, int, int *, int *, int, int *, int, int);
void PerformPartition(scalar **, int, int *, int, int *, int *, int *, int *, int, int);
int ReorderSubsetsLables(int *, int);
void RepresentativesOfSubset(scalar **, int, int *, int, int *, int *, int *, int, int, bool, int **);
/* populate.c */
void SMOTEPopulate(scalar ***, int **, int *, scalar **, int, int *, int, int, int, int, int, bool *, bool, scalar *);
void EvaluateWithArtificialPopulation(const t_Parameters *, int *, scalar **, int, int *, int, scalar *, scalar *, int, scalar *, int, int, int *, int, scalar *);
/* pruning.c */
t_LVQ_net *LVQPruning(const t_Parameters *, scalar **, int, int *, int, int);
void CondensedNN(int *, scalar **, int, int *, int, int);
/* psa.c */
int PatternSelectionAlgorithm(int *, scalar **, int, int *, int, int, int, bool, int *, int **);
scalar Logarithm(scalar, scalar);
/* psrcgalgo.c */
int PSRCGAlgorithm(const t_Parameters *, int *, scalar **, int, int *, int, int, int, int **, int *);
scalar ComputeRCG(unsigned char **, int, int *, scalar *, int *, int);
/* randomalgo.c */
int RandomSelection(int *, scalar **, int, int *, scalar, bool, int, int);
int SelectOneRandomSampleFromEveryClass(int *, int,  int *, int);
/* randomis+fs.c */
scalar RandomISFSWithTimeLimit(const t_Parameters *, int *, int *, scalar **, int, int *, scalar, scalar, scalar, scalar, int, int, int, int *, int *, scalar);
/* rc45algo.c */
int RobustC45(const t_Parameters *, int *, scalar **, int, int *, int, int, int *, int **);
int RobustDT(const t_Parameters *, int *, scalar **, int, int *, int, int, int *, int **);
/* reach.c */
void ComputeReachableAndCoverage(const t_Parameters *, scalar *, scalar *, int *, scalar **, int *, int, int, int, cache_t **, scalar **, bool *, int, scalar *, int **, int *);
/* rmhc.c */
int RandomMutationHillClimbingInstanceSelection(const t_Parameters *, int *, scalar **, int *, int, int, cache_t **, int, int, int **, int *);
/* rnge.c */
int RelativeNeighborhoodGraphEditing(t_ProximityGraph, int *, scalar **, int, int *, int, int, int **, scalar **, int *);
bool **GabrielGraph(scalar **, int, int, int **, scalar **, int *);
bool **RelativeNeighborhoodGraph(scalar **, int, int, int **, scalar **, int *);
bool GabrielLink(int i, int j, int, scalar **);
bool RelativeNeighborhoodLink(int i, int j, int, scalar **);
/* rnnalgo.c */
int ReducedNearestNeighbor(const t_Parameters *, int *, scalar **, int, int *, int value, int, int, cache_t **, int **, int *);
/* sampling.c */
bool TestOneClass(int *, int, int);
int *RandomSampling(scalar ***, int **, int *, int, scalar, int ***, int, bool);
/* selection.c */
int OrderedSelection(scalar *, int, int);
int TournamentSelection(scalar *, int no_rival, int);
void OrderedList(scalar *, int *, int, bool);
int BinaryTournament(scalar *, int);
void NBestOrderedSelection(scalar *, int, int, int *);
/* stop.c */
bool StopCriterionMet(const t_Parameters *, time_t, scalar, int);
/* threshold.c */
scalar EvaluateThresholdIS(const t_Parameters *, scalar threshold, scalar *, scalar **, int, int *, scalar *, scalar *, int, int, bool, int *, int, cache_t **, int **);
/* thselect.c */
int SelectInstancesUsingThreshold(scalar *, int, int *, int *, int *, int, bool, int);
/* tsparknn.c */
scalar TrainingSetParallelkNN(int, int, int, char *, char *, int, scalar *, scalar *, scalar *, scalar *, char *, int, int);
/* w.c */
scalar **GetSVMW(struct svm_model *, int);
void Product(scalar *, scalar **, scalar *, int, int);
/* multiboost.c */
scalar UpdateMultiBoostWeightsIS(const t_Parameters *, int *, scalar **, int, int *, scalar *, scalar *, int, int, bool, int, int, t_Classifier, int **, bool);
/* mcs.c */
int MinimalConsistentSet(const t_Parameters *, int *, scalar **, int, int *, int, cache_t **, int, int **, int *, scalar **);
