/*
   Clustering steps.
*/
#include <stdio.h>
#include <malloc.h>
#include "macros.h"
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "kclustering.h"

//List functions: add an element into the inverted list
nodeINV *list_add_inv(nodeINV **head, int index, int n_number)
{
    nodeINV *new_node;
    CALLOC(new_node, nodeINV, 1, "list_add_inv");

    new_node->next = *head;
    *head = new_node;
    new_node->index = index;
    new_node->n_number = n_number;
    return new_node;
}

//List functions: print elements of the inverted list
void list_print_inv(nodeINV *n)
{
    if (n == NULL) {
        printf("Empty list\n");
    }
    while (n != NULL) {
        printf("Index: %d, n_number: %d\n", n->index, n->n_number);
        n = n->next;
    }
}

//List functions: recursively delete the elements of the inverted list
void recursiveFreeInvList(nodeINV **head)
{
    nodeINV *next_elem;

    if (*head != NULL) {
        next_elem = (*head)->next;
        recursiveFreeInvList(&next_elem);
        FREE(*head);
        *head = NULL;
    }
}

//List functions: add an element into the similarity list
nodeSIM *list_add_sim(nodeSIM **head, int index)
{
    nodeSIM *node;
    nodeSIM *new_node = NULL;

    //Look if the element already exists
    node = list_search_sim(*head, index);

    //Element has been already created
    if (node != NULL)
        node->sim = node->sim + 1;
    //If not: create new element in the list
    else {
        CALLOC(new_node, nodeSIM, 1, "list_add_sim");
        new_node->next = *head;
        *head = new_node;
        new_node->index = index;
        new_node->sim = 1;
    }
    return new_node;
}

//List functions: look for an element in the similarity list
nodeSIM *list_search_sim(nodeSIM *n, int index)
{
    while (n != NULL) {
        if (n->index == index)
            return n;
        n = n->next;
    }
    return NULL;
}

//List functions: print elements
void list_print_sim(nodeSIM *n)
{
    if (n == NULL) {
        printf("Empty list\n");
    }
    while (n != NULL) {
        printf("Index: %d, sim: %d\n", n->index, n->sim);
        n = n->next;
    }
}

//List functions: recursively delete the elements of the similarity list
void recursiveFreeSimList(nodeSIM **head)
{
    nodeSIM *next_elem;

    if (*head != NULL) {
        next_elem = (*head)->next;
        recursiveFreeSimList(&next_elem);
        FREE(*head);
        *head = NULL;
    }
}

//List functions: add an element into the similarity list
void list_add_id(nodeID **head_trgt, nodeID **head_src, int target, int element)
{
    nodeID *current = NULL;

    //Point current pointer to the head of the list
    current = *head_trgt;

    //Place current pointer in the last element of the target list
    while (current->next != NULL) {
        //fprintf(f, "E%d R%d\n", current->index, current->root->index);
        //      fflush(f);
        current = current->next;
    }
    //Copy the source list into the target
    current->next = *head_src;
    //The root pointer of the copied list, now points to the root of the target list
    current->next->root = (*head_trgt)->root;
    while (current->next != NULL) {
        current->root = (*head_trgt)->root;
        //fprintf(f, "*E%d R%d\n", current->index, current->root->index);
        //      fflush(f);
        //fprintf(f, "Sigte de E%d es E%d\n", current->index, current->next->index);
        //      fflush(f);
        current = current->next;
    }
    //Label the last element
    current->root = (*head_trgt)->root;
    //fprintf(f, "**E%d R%d\n", current->index, current->root->index);
    //     fflush(f);
}

//List functions: add the first element of the identity list
void list_add_first_ident(nodeID **head, int index)
{
    nodeID *new_node;

    CALLOC(new_node, nodeID, 1, "list_add_first_ident");
    *head = new_node;
    new_node->index = index;
    new_node->root = *head;
    new_node->next = NULL;
}

//List functions: recursively deletes the elements of the identity list
void recursiveFreeIdList(nodeID **head)
{
    nodeID *next_elem;

    if (*head != NULL) {
        next_elem = (*head)->next;
        recursiveFreeIdList(&next_elem);
        FREE(*head);
        *head = NULL;
    }
}

//First step of the clustering algorithm: Function that creates the matrix of neighbours
void CreateNeighbourhoodTable(scalar **data_instances, nodePREVNN **head_prev, tNN  ***neigh_matrix, tNN *near_neighbours, tKDTREE *root_node, int n_patt,
                              int k_near, int inputs)
{
    int i, j, k, n_number = 0, n_added;
    scalar min_dist = (scalar) INFINITE;
    scalar *hr_max, *hr_min, *left_max, *left_min, *right_max, *right_min;


    for (i = 0; i < n_patt; i++) {
        MALLOC((*neigh_matrix)[i], tNN, k_near, "CreateNeighbourhoodTable");
        MALLOC(left_max, scalar, inputs, "CreateNeighbourhoodTable");
        MALLOC(left_min, scalar, inputs, "CreateNeighbourhoodTable");
        MALLOC(right_max, scalar, inputs, "CreateNeighbourhoodTable");
        MALLOC(right_min, scalar, inputs, "CreateNeighbourhoodTable");
        MALLOC(hr_max, scalar, inputs, "CreateNeighbourhoodTable");
        MALLOC(hr_min, scalar, inputs, "CreateNeighbourhoodTable");

        //Initialize distances to infinite and reset neighbour number
        for (j = 0; j < k_near; j++) {
            (*neigh_matrix)[i][j].dist = INFINITE;
            (*neigh_matrix)[i][j].n_number = -1;
        }

        //#####Initialize max and min values
        for (j = 0; j < inputs; j++) {
            left_max[j] = (scalar) INFINITE;
            left_min[j] = (scalar) -INFINITE;
            right_max[j] = (scalar) INFINITE;
            right_min[j] = (scalar) -INFINITE;
            hr_max[j] = (scalar) INFINITE;
            hr_min[j] = (scalar) -INFINITE;
        }

        n_added = 0;
        //      fprintf(f, "\n\n//////////////////////////////////////////////////\n");
        //      fprintf(f, "\nCall FindNearNeighbours: %d\n", i);
        //      fprintf(f, "//////////////////////////////////////////////////\n\n");
        FindNearNeighbours(data_instances[i], hr_max, hr_min, (*neigh_matrix)[i], root_node, &n_added, k_near);

        //Set neighbor number according to the distance value
        for (j = 0; j < k_near; j++) {
            for (k = 0; k < k_near; k++) {
                //Not set before and less distance
                if ((*neigh_matrix)[i][k].n_number == -1 && (*neigh_matrix)[i][k].dist < min_dist) {
                    (*neigh_matrix)[i][k].n_number = n_number;
                    min_dist = (*neigh_matrix)[i][k].dist;
                }
            }
            n_number++;
            min_dist = (scalar) INFINITE;
        }
        n_number = 0;
    }
}

//Second step of the clustering algorithm: Function that creates the inverted list of the matrix of neighbours
void CreateInvertedList(nodeINV  ***head_inv, tNN **neigh_matrix, int n_patt, int k_near)
{
    int i, j, index = 0;

    for (i = 0; i < n_patt; i++) {
        for (j = 0; j < k_near; j++) {
            index = neigh_matrix[i][j].index;
            list_add_inv(&(*head_inv)[index], i, neigh_matrix[i][j].n_number);
        }
    }
}

//Third step of the clustering algorithm: Function that creates the inverted list of the matrix of neighbours
void CreateSimilarityList(nodeSIM  ***head_sim, nodeINV **head_inv, int n_patt)
{
    int i, element0, element1;
    nodeINV *n = NULL, *temp = NULL;

    for (i = 0; i < n_patt; i++) {
        n = head_inv[i];
        //Add similarity if there is at least two elements in the inverted list
        while (n != NULL && n->next != NULL) {
            element0 = n->index;
            temp = n->next;
            while (temp != NULL) {
                element1 = temp->index;
                temp = temp->next;
                //Similarities are symmetrical so we store them in the list of the element earlier in memory
                if (element1 < element0)
                    list_add_sim(&(*head_sim)[element1], element0);
                else
                    list_add_sim(&(*head_sim)[element0], element1);
            }
            n = n->next;
        }
    }
}

//Fourth step of the clustering algorithm: Function that creates the clustering identity lists
void CreateIdentityList(nodeID  ***head_id, nodeINV  ***head_inv, nodeSIM **head_sim, int n_patt, int k_t)
{
    int i, element;
    nodeSIM *n = NULL;

    //Create the first element of each list in the identity list
    for (i = 0; i < n_patt; i++)
        list_add_first_ident(&(*head_id)[i], i);

    //Build the clusters
    for (i = 0; i < n_patt; i++) {
        n = head_sim[i];
        while (n != NULL) {
            //Add element to cluster if there is a similarity list element with a value higher than k_t
            if (n->sim > k_t) {
                element = n->index;
                //Copy the list of the element
                if (i > element) {
                    if ((*head_id)[i]->root->index != (*head_id)[element]->root->index) {
                        list_add_id(&(*head_id)[i], &(*head_id)[element]->root, i, element);
                    }
                } else {
                    if ((*head_id)[element]->root->index != (*head_id)[i]->root->index) {
                        list_add_id(&(*head_id)[element], &(*head_id)[i]->root, element, i);
                    }
                }
            }
            n = n->next;
        }
    }
}

//Fifth step of the clustering algorithm: Function that assigns a cluster number to each element
void LabelElements(nodeID **head_id, int *partition, tCLST **clst_info, int *n_subsets, int *count, int **match, int **new_clst_root, int n_patt)
{
    int i, elem = 0, act_root = 0, clst_ind;
    nodeID *n = NULL;

    //Label cluster elements & count number of elements per cluster
    for (i = 0; i < n_patt; i++) {
        //fprintf(f, "\n\nExamining identity list %d\n", i);
        n = head_id[i];
        //Examine list only if its root pointer point to itself (it's the beginning of a cluster)
        if (n->root->index == n->index) {
            (*n_subsets)++;
            while (n != NULL) {
                //We refer to the match vector in order to maintain global indexes and do not delete previous values
                act_root = (*match)[n->root->index];
                partition[(*match)[n->index]] = act_root;

                //Count number of elements in the cluster
                count[act_root]++;
                //fprintf(f, "Set element %d/%d to cluster %d\n", n->index, (*match)[n->index], act_root);

                n = n->next;
            }
        }
    }

    //Store cluster elements and newly root_indexes
    clst_ind = 0;
    //Memory allocation
    CALLOC(*new_clst_root, int, *n_subsets, "LabelElements");
    for (i = 0; i < n_patt; i++) {
        n = head_id[i];

        //Examine list only if its root pointer point to itself (it's the beginning of a cluster)
        if (n->root->index == n->index) {
            //Store the local root indexes of the newly added clusters
            (*new_clst_root)[clst_ind] = n->root->index;
            //fprintf(f, "Store local root%d\n", (*new_clst_root)[clst_ind]);
            clst_ind++;

            elem = 0;
            act_root = (*match)[n->root->index];

            (*clst_info)[act_root].index = act_root;
            (*clst_info)[act_root].number = count[act_root];

            //Memory allocation
            MALLOC((*clst_info)[act_root].components, int, count[act_root], "LabelElements");

            while (n != NULL) {
                //We refer to the match vector in order to maintain global indexes and do not delete previous values
                (*clst_info)[act_root].components[elem] = (*match)[n->index];
                elem++;

                n = n->next;
            }
        }
    }
}
