/*
   Definitions and prototypes for individual networks.
 */
#ifndef _KD_H
#define _KD_H

#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include "common.h"

//Basic kdtree information
typedef struct tree {
        //Data element number
        int index;
        //Index of the largest spread dimension
        int dim;
        //Median value of data (considering the specified dimension).
        scalar median;
        //Dispersion value of data (considering the specified dimension).
        scalar disp;
        //Data values
        scalar *node_data;
        //Data classes
        int node_class;
        //Node type, 't': terminal, 'n':non terminal
        char type;
        //Data element less than median value
        struct tree *left_tree;
        //Data element greater than / or equal to median value
        struct tree *right_tree;
} tKDTREE;


//Near neighbour information (elements of neighbourhood table)
typedef struct {
        //Data element number
        int index;
        //Neighbour number
        int n_number;
        //Distance to element.
        scalar dist;
} tNN;


//Cluster information
typedef struct {
        //Root index
        int index;
        //Number of elements
        int number;
        //Indexes of the cluster elements.
        int *components;
} tCLST;


//Previous limit values of the search of near neighbours
typedef struct prev_list {
        //Modified dimension
        int dim;
        //Min value in left tree.
        scalar left_min;
        //Max value in left tree.
        scalar left_max;
        //Min value in right tree.
        scalar right_min;
        //Max value in right tree.
        scalar right_max;
        //Pointer to the next element in the list
        struct prev_list *next;
} nodePREVNN;

//Basic information of elements in the inverted list
typedef struct inv_list {
        //Data element number
        int index;
        //Neighbour number
        int n_number;
        //Pointer to the next element in the list
        struct inv_list *next;

} nodeINV;


//Basic information of elements in the similarity list
typedef struct sim_list {
        //Data element number
        int index;
        //Similarity measure between elements
        int sim;
        //Pointer to the next element in the list
        struct sim_list *next;

} nodeSIM;


//Basic information of elements in the similarity list
typedef struct id_list {
        //Data element number
        int index;
        //Pointer to the root element in the list
        struct id_list *root;
        //Pointer to the next element in the list
        struct id_list *next;

} nodeID;
#endif                          /* _KD_H */
