
/*
   Kdtree functions.
*/
#include <stdio.h>
#include <malloc.h>
#include "macros.h"
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "kd.h"

#define BAL_MED         0.1

//Qsort auxiliary function
int CompareGreatestLeast(const void *x, const void *y)
{
    scalar diff;

    diff = (*(scalar *) y - *(scalar *) x);
    if (diff > 0)
        return +1;
    else {
        if (diff < 0)
            return -1;
        else
            return 0;
    }
}

int CompareLeastGreatest(const void *x, const void *y)
{
    scalar diff;

    diff = (*(scalar *) y - *(scalar *) x);
    if (diff > 0)
        return -1;
    else {
        if (diff < 0)
            return +1;
        else
            return 0;
    }
}

scalar HyperrectangDistance(scalar *v, scalar *min, scalar *max, int n)
{
    scalar dist = 0.0;
    int i;

    for (i = 0; i < n; i++) {
        if (v[i] <= min[i]) {
            dist += (v[i] - min[i]) * (v[i] - min[i]);
            //          fprintf(f, "First hyper-case. Caract.%d. V[i] = %lf. Min[i] = %lf. Dist = %lf\n", i, v[i], min[i], dist);
        } else if ((v[i] > min[i]) && (v[i] < max[i])) {
            dist += 0;
            //          fprintf(f, "Second hyper-case. Caract.%d. V[i] = %lf. Min[i] = %lf. Max[i] = %lf. Dist = %lf\n",i, v[i], min[i], max[i], dist);
        } else if (v[i] >= max[i]) {
            dist += (v[i] - max[i]) * (v[i] - max[i]);
            //          fprintf(f, "Third hyper-case. Caract.%d. V[i] = %lf. Max[i] = %lf. Dist = %lf\n", i, v[i], max[i],dist);
        }
    }

    return sqrt(dist);
}

void GetAllEuclideanDistance(scalar *data_element, scalar **data_instances, scalar *eucl, int n_patt, int inputs)
{
    int i;
    scalar eucl_dist = 0;

    for (i = 0; i < n_patt; i++) {
        eucl_dist = EuclideanDistance(data_element, data_instances[i], inputs);
        eucl[i] = eucl_dist;
        //fprintf(f, "Eucl dist to pattern %d = %lf\n", i, eucl[i]);
    }
}

void TestNeighbours(scalar **data_instances, tNN **neigh_matrix, scalar *eucl, int n_patt, int k_near, int inputs)
{
    int i, j;
    scalar max = 0;

    for (i = 0; i < n_patt; i++) {
        //fprintf(f, "Program neighbours for pattern: %d\n", i);
        for (j = 0; j < k_near; j++) {
            //fprintf(f, "Index %d, %lf\n", neigh_matrix[i][j].index, neigh_matrix[i][j].dist);
            if (neigh_matrix[i][j].dist > max)
                max = neigh_matrix[i][j].dist;
        }
        //fprintf(f, "Max %lf\n", max);
        //fprintf(f, "\n\n");

        GetAllEuclideanDistance(data_instances[i], data_instances, eucl, n_patt, inputs);
        qsort(eucl, n_patt, sizeof(scalar), CompareLeastGreatest);

        //                 dif = max - eucl[k_near - 1];
        //                 //      if (max - eucl[k_near - 1] != 0) {
        //                 //          fprintf(f, "Max: %lf, eucl: %lf\n", max, eucl[k_near - 1]);
        //                 //          fprintf(f, "Error calculating neighbour of instance: %d\n", i);
        //                 //          for (j = 0; j < k_near; j++)
        //                 //              fprintf(f, "Program = %lf. Eucl = %lf\n", neigh_matrix[i][j].dist, eucl[j]);
        //                 //      }
        //                 max = 0;
    }

}



//List functions: add an element into the previous limits list
nodePREVNN *list_add_prev(nodePREVNN **head, int dimension, scalar left_min, scalar left_max,
                          scalar right_min, scalar right_max)
{
    nodePREVNN *new_node;
    CALLOC(new_node, nodePREVNN, 1, "list_add_prev");

    new_node->next = *head;
    *head = new_node;
    new_node->dim = dimension;
    new_node->left_min = left_min;
    new_node->left_max = left_max;
    new_node->right_min = right_min;
    new_node->right_max = right_max;
    return new_node;
}

//List functions: print elements of the previous limits list
void list_print_prev(nodePREVNN *n)
{
    if (n == NULL) {
        printf("Empty list\n");
    }
    while (n != NULL) {
        printf("Dimension: %d, left_min: %f, left_max: %f, right_min: %f, right_max: %f\n", n->dim,
               (scalar) n->left_min, (scalar) n->left_max, (scalar) n->right_min, (scalar) n->right_max);
        n = n->next;
    }
}

//Function that creates the kd-tree
tKDTREE *MakeTree(scalar **tree_data, scalar ***data_less, scalar ***data_greater, int n_patt, int n_patt_other, int dispersion, int *n_terminal, int inputs)
{
    int i, j, n_less = 0, n_greater = 0, greatest_dim = 1;
    scalar act_least_disp, act_greatest_disp, greatest_disp = 0.0;
    //         scalar greatest = 0.0;
    scalar *disp, *aux;
    scalar **data_greater_copy = NULL;
    tKDTREE *new_node;

    if (data_greater_copy == NULL) {
        ALLOCA(data_greater_copy, scalar *, n_patt, "MakeTree");
    }
    //Malloc for the new terminal node
    MALLOC(new_node, tKDTREE, 1, "MakeTree");
    MALLOC(new_node->node_data, scalar, inputs, "MakeTree");
    MALLOC(new_node->left_tree, tKDTREE, 1, "MakeTree");
    MALLOC(new_node->right_tree, tKDTREE, 1, "MakeTree");

    new_node->left_tree = NULL;
    new_node->right_tree = NULL;

    //Terminal node (only one element remaining in 'data'
    if (n_patt == 1 || (n_patt_other == 0 && dispersion == 0)) {
        //fprintf(f, "\nAdding terminal node:\n");

        //fprintf(f, "Data:\t");
        //      fflush(stdin);

        if (dispersion == 0) {
            //        if (n_patt > 1)
            //          fprintf(stdout, "All values must be the same, n_patt = %d, n_patt_other = %d\n", n_patt, n_patt_other);
            for (i = 0; i < n_patt - 1; i++) {
                //View values in terminal-nodes
                for (j = 0; j < inputs; j++) {
                    //fprintf(f, "%lf\t", tree_data[i][j]);
                    //                  fflush(stdin);
                }
                //fprintf(f, "\n");
            }
        }
        //Set data values in terminal-nodes
        for (i = 0; i < inputs; i++) {
            new_node->node_data[i] = tree_data[n_patt - 1][i];
            //fprintf(f, "%lf\t", new_node->node_data[i]);
            //          fflush(stdin);
        }

        //Convert binary class into integer class
        j = 0;
        while (tree_data[n_patt - 1][j + inputs] == 0)
            j++;
        new_node->node_class = j + 1;
        //fprintf(f, "%d\n", new_node->node_class);
        //      fflush(stdin);

        //Set node type
        new_node->type = 't';
        //Set node index
        new_node->index = tree_data[n_patt - 1][-1];
        //fprintf(f, "Adding terminal node, index: %d\n", new_node->index);
        //      fflush(stdin);
        (*n_terminal)++;

        n_patt = 0;
    }
    //Non-terminal node
    else if (n_patt > 1) {
        //fprintf(f, "Adding non-terminal node:\n");
        MALLOC(disp, scalar, inputs, "MakeTree");

        for (i = 0; i < inputs; i++) {
            //fprintf(f, "Dimension %d\n", i);
            act_least_disp = tree_data[0][i];
            act_greatest_disp = tree_data[0][i];
            for (j = 0; j < n_patt; j++) {
                if (tree_data[j][i] > act_greatest_disp) {

                    //fprintf(f, "New greatest: %lf, least: %lf\n", tree_data[j][i], act_greatest_disp);
                    act_greatest_disp = tree_data[j][i];

                } else if (tree_data[j][i] < act_least_disp) {
                    act_least_disp = tree_data[j][i];
                }
            }

            disp[i] = act_greatest_disp - act_least_disp;
            //fprintf(f, "Greatest: %lf, least: %lf\n", act_greatest_disp, act_least_disp);

            if (disp[i] > greatest_disp) {
                greatest_dim = i;
                greatest_disp = disp[i];
                //                                 greatest = act_greatest_disp;
            }
        }

        //Set greatest dimension of the non terminal node
        new_node->dim = greatest_dim;
        //Set dispersion of the greatest dimension of the non terminal node
        new_node->disp = disp[new_node->dim];
        //fprintf(f, "Greatest dim: %d. Max value: %lf. Disp: %lf\n", new_node->dim, greatest, new_node->disp);
        //      fflush(stdin);
        FREE(disp);

        //Malloc aux vector
        MALLOC(aux, scalar, n_patt, "MakeTree");

        //Sort by qsort greatest dimension values
        for (i = 0; i < n_patt; i++)
            aux[i] = tree_data[i][greatest_dim];
        qsort(aux, n_patt, sizeof(scalar), CompareGreatestLeast);
        //fprintf(f, "Data vector sorted by dimension %d:\n", greatest_dim);
        for (i = 0; i < n_patt; i++)
            //fprintf(f, "%lf\t", aux[i]);
            //fprintf(f, "\n");

            //Median value of data (considering the specified dimension) will be in the middle of the sorted vector
            if (n_patt % 2 == 0)
                new_node->median = (aux[n_patt / 2] + aux[n_patt / 2 - 1]) / 2;
            else
                new_node->median = aux[n_patt / 2];
        FREE(aux);
        //fprintf(f, "Median value in dimension %d is: %lf\n", new_node->dim, new_node->median);
        //      fflush(stdin);

        //Create two vectors: data_less and data_greater, with the values less than and greater than the median value
        for (i = 0; i < n_patt; i++) {
            //fprintf(f, "Element compared: %lf\n", tree_data[i][greatest_dim]);
            if (tree_data[i][greatest_dim] < new_node->median) {
                (*data_less)[n_less] = tree_data[i];
                n_less++;
            } else {
                (*data_greater)[n_greater] = tree_data[i];
                data_greater_copy[n_greater] = tree_data[i];
                n_greater++;
            }
        }

        //Modify median in order to balance node distribution (if all of them are greater or less than median)
        if (new_node->disp != 0 && (n_less == 0 || n_greater == 0)) {
            if (n_less == 0)
                new_node->median = new_node->median + BAL_MED;
            else if (n_greater == 0)
                new_node->median = new_node->median - BAL_MED;

            //          fprintf(f, "New adjusted median value in dimension %d is: %lf\n", new_node->dim, new_node->median);
            //          fflush(stdin);

            n_less = 0;
            n_greater = 0;

            //Create two vectors: data_less and data_greater, with the values less than and greater than the median value
            for (i = 0; i < n_patt; i++) {
                //fprintf(f, "Element compared: %lf\n", tree_data[i][greatest_dim]);
                if (tree_data[i][greatest_dim] < new_node->median) {
                    (*data_less)[n_less] = tree_data[i];
                    n_less++;
                } else {
                    (*data_greater)[n_greater] = tree_data[i];
                    data_greater_copy[n_greater] = tree_data[i];
                    n_greater++;
                }
            }
        }

        //Set node type
        new_node->type = 'n';

        if (n_less != 0) {
            //fprintf(f, "Going to the left with median: %lf\n", new_node->median);
            new_node->left_tree =
                MakeTree(*data_less, data_less, data_greater, n_less, n_greater, new_node->disp, n_terminal, inputs);
            //fprintf(f, "\n");
        }
        if (n_greater != 0) {
            //fprintf(f, "Going to the right with median: %lf\n", new_node->median);
            new_node->right_tree =
                MakeTree(data_greater_copy, data_less, data_greater, n_greater, n_less, new_node->disp,
                         n_terminal, inputs);
            //fprintf(f, "\n");
        }
    }                           //else if (n_patt > 1)

    //Return pointer to node just added
    return new_node;
}

//Function that counts the number of terminal nodes
void CountTerminals(tKDTREE *node, int *n_terminal)
{
    //Non - terminal node
    if (node->type == 'n') {
        if (node->left_tree != NULL)
            CountTerminals(node->left_tree, n_terminal);
        if (node->right_tree != NULL)
            CountTerminals(node->right_tree, n_terminal);
    }
    if (node->type == 't') {
        (*n_terminal)++;
    }
}

//Function that searches for the k near neighbours
void FindNearNeighbours(scalar *data_element, scalar *hr_max, scalar *hr_min, tNN *near_neighbours, tKDTREE *node, int *n_added, int k_near, int inputs)
{
    int i, fur_neigh = 0;
    //    int prev_dimension;
    scalar lim = 0, max_dist, eucl_dist = 0;
    //scalar prev_left_max, prev_left_min, prev_right_max, prev_right_min;
    int further_is_left = 1;
    scalar left_max[inputs], left_min[inputs], right_max[inputs], right_min[inputs];

    //Non - terminal node
    if (node->type == 'n') {
        //Divide 'hr' vector into 'left' and 'right' vectors
        memcpy(left_max, hr_max, sizeof(scalar) * inputs);
        memcpy(left_min, hr_min, sizeof(scalar) * inputs);
        memcpy(right_max, hr_max, sizeof(scalar) * inputs);
        memcpy(right_min, hr_min, sizeof(scalar) * inputs);

        //      //Store previous values
        //      list_add_prev(/*head_prev,*/ node->dim, left_min[node->dim], left_max[node->dim], right_min[node->dim],
        //                    right_max[node->dim]);

        //#####
        //Update limit values
        if (left_max[node->dim] > node->median)
            left_max[node->dim] = node->median;
        if (right_min[node->dim] < node->median)
            right_min[node->dim] = node->median;

        //Search the nearest node
        if (data_element[node->dim] < node->median && node->left_tree != NULL) {
            //fprintf(f, "*Left branch -> type: %c\n",  node->left_tree->type);
            //fprintf(f, "*Left branch -> dim: %d\n",  node->left_tree->dim);
            //fprintf(f, "*Left branch -> median: %lf\n\n", node->left_tree->median);
            //fprintf(f, "*Element vs Left branch -> %lf\n",  data_element[node->dim]);

            //Further neighbours are in the right branch
            further_is_left = 0;

            FindNearNeighbours(data_element, /*head_prev, */ left_max, left_min,        /* left_max, left_min, right_max,
                                                                                           right_min, */ near_neighbours, node->left_tree, n_added, k_near, inputs);
        } else if (data_element[node->dim] >= node->median && node->right_tree != NULL) {
            //fprintf(f, "*Right branch -> type: %c\n",  node->right_tree->type);
            //fprintf(f, "*Right branch -> dim: %d\n",  node->right_tree->dim);
            //fprintf(f, "*Right branch -> median: %lf\n\n", node->right_tree->median);
            //fprintf(f, "*Element vs Right branch -> %lf\n",  data_element[node->dim]);

            //Further neighbours are in the left branch
            further_is_left = 1;

            FindNearNeighbours(data_element, /*head_prev, */ right_max, right_min,      /* left_max, left_min, right_max,
                                                                                           right_min, */ near_neighbours, node->right_tree, n_added, k_near, inputs);
        } else {
            //          fprintf(f, "*****WARNING: DOES NOT DESCEND IN THE KD-TREE*****\n");
            //          fprintf(f, "\nElement: %lf, Median: %lf***\n", data_element[node->dim], node->median);
        }

        //fprintf(f, "\n###Evaluating if we will go up in the tree with median value: %lf.\n", node->median);
        //Search for the next nearest nodes
        if (data_element[node->dim] < node->median && node->right_tree != NULL) {

            //If we haven't added the k near neighbours yet, we'll explore that branch independtly of the limit measure
            if ((*n_added) < k_near) {
                //fprintf(f, "\n###Going up directly in the kd-tree with median %lf. To the right\n", node->median);
                FindNearNeighbours(data_element, /*head_prev, */ right_max, right_min,  /* left_max, left_min,
                                                                                           right_max, right_min, */ near_neighbours, node->right_tree, n_added,
                                   k_near, inputs);
            } else {
                //Calculate if it is worth to search in an upper branch
                if (further_is_left) {
                    /*              fprintf(f, "Left_min\n");
                                    for (i = 0; i < inputs; i++)
                                        fprintf(f, "Min[%d]=%lf\t", i, left_min[i]);
                                    fprintf(f, "\nLeft_max\n");
                                    for (i = 0; i < inputs; i++)
                                        fprintf(f, "Max[%d]=%lf\t", i, left_max[i]);
                                    fprintf(f, "\n");
                    */
                    lim = HyperrectangDistance(data_element, left_min, left_max, inputs);
                    //fprintf(f, "\n###Further root is the left one\n");
                } else {
                    /*              fprintf(f, "Right_min\n");
                                    for (i = 0; i < inputs; i++)
                                        fprintf(f, "Min[%d]=%lf\t", i, right_min[i]);
                                    fprintf(f, "\nRight_max\n");
                                    for (i = 0; i < inputs; i++)
                                        fprintf(f, "Max[%d]=%lf\t", i, right_max[i]);
                                    fprintf(f, "\n");
                    */
                    lim = HyperrectangDistance(data_element, right_min, right_max, inputs);
                    //fprintf(f, "\n###Further root is the right one,\n");

                }

                //Get furthest distance of the near neighbours already added
                max_dist = 0;
                for (i = 0; i < k_near; i++)
                    if (near_neighbours[i].dist > max_dist)
                        max_dist = near_neighbours[i].dist;

                //fprintf(f, "Right branch -> Limit value: %lf. Furthest neighbour: %lf\n", lim, max_dist);
                if (lim < max_dist) {

                    FindNearNeighbours(data_element, /*head_prev, */ right_max, right_min,      /* left_max, left_min,
                                                                                   right_max, right_min, */
                                       near_neighbours,
                                       node->
                                       right_tree,
                                       n_added,
                                       k_near, inputs);
                } else {
                    //                  fprintf(f, "We don't go up in the kd-tree with median %lf.\n", node->median);
                }
            }
        } else if (data_element[node->dim] >= node->median && node->left_tree != NULL) {
            //If we haven't added the k near neighbours yet, we'll explore that branch independtly of the limit measure
            if ((*n_added) < k_near) {
                //fprintf(f, "\n###Going up directly in the kd-tree with median %lf. To the left\n",node->median);
                FindNearNeighbours(data_element, /*head_prev, */ left_max, left_min,    /* left_max, left_min, right_max,
                                                                                           right_min, */ near_neighbours, node->left_tree, n_added, k_near, inputs);
            } else {
                //Calculate if it is worth to search in an upper branch
                if (further_is_left) {
                    /*              fprintf(f, "Left_min\n");
                                    for (i = 0; i < inputs; i++)
                                        fprintf(f, "Min[%d]=%lf\t", i, left_min[i]);
                                    fprintf(f, "\nLeft_max\n");
                                    for (i = 0; i < inputs; i++)
                                        fprintf(f, "Max[%d]=%lf\t", i, left_max[i]);
                                    fprintf(f, "\n");
                    */
                    lim = HyperrectangDistance(data_element, left_min, left_max, inputs);
                    //fprintf(f, "\n###Further root is the left one\n");
                } else {
                    /*              fprintf(f, "Right_min\n");
                                    for (i = 0; i < inputs; i++)
                                        fprintf(f, "Min[%d]=%lf\t", i, right_min[i]);
                                    fprintf(f, "\nRight_max\n");
                                    for (i = 0; i < inputs; i++)
                                        fprintf(f, "Max[%d]=%lf\t", i, right_max[i]);
                                    fprintf(f, "\n");
                    */
                    lim = HyperrectangDistance(data_element, right_min, right_max, inputs);
                    //fprintf(f, "\n###Further root is the right one,\n");
                }

                //Get furthest distance of the near neighbours already added
                max_dist = 0;
                for (i = 0; i < k_near; i++)
                    if (near_neighbours[i].dist > max_dist)
                        max_dist = near_neighbours[i].dist;

                //fprintf(f, "Left branch -> Limit value: %lf. Furthest neighbour: %lf\n", lim, max_dist);


                if (lim < max_dist) {
                    //fprintf(f, "Passed the limit value: %lf is nearer than the furthest neighbour: %lf\n",lim, max_dist);
                    //fprintf(f, "Left branch -> type: %c\n",  node->left_tree->type);
                    //fprintf(f, "Left branch -> dim: %d\n",  node->left_tree->dim);
                    //fprintf(f, "Left branch -> median: %lf\n",  node->left_tree->median);
                    //fprintf(f, "Element vs Left branch -> %lf\n",  data_element[node->dim]);
                    //fprintf(f, "\n###Going up in the kd-tree with median %lf. To the left\n", node->median);
                    FindNearNeighbours(data_element, /*head_prev, */ left_max, left_min,        /* left_max, left_min,
                                                                                                   right_max, right_min, */ near_neighbours, node->left_tree, n_added,
                                       k_near, inputs);
                } else {
                    //fprintf(f, "We don't go up in the kd-tree with median %lf.\n", node->median);
                }
            }
        }
    }
    //Terminal node
    else if (node->type == 't' && node != NULL) {
        eucl_dist = EuclideanDistance(data_element, node->node_data, inputs);

        //fprintf(f, "\nN_added: %d", (*n_added));
        //If we haven't added the k near neighbours yet, we'll add it without any checking
        if ((*n_added) < k_near) {
            near_neighbours[*n_added].index = node->index;
            near_neighbours[*n_added].n_number = (*n_added);
            near_neighbours[*n_added].dist = eucl_dist;
        } else {
            //fprintf(f, "Euclidean distance: %lf\n", eucl_dist);
            //Get furthest distance of the near neighbours already added
            max_dist = 0;
            for (i = 0; i < k_near; i++) {
                if (near_neighbours[i].dist > max_dist) {
                    max_dist = near_neighbours[i].dist;
                    fur_neigh = i;
                }
            }
            //If new element is nearer than a previous neighbour: replace
            if (eucl_dist < max_dist) {
                near_neighbours[fur_neigh].index = node->index;
                //We keep the same neighbour number (n_number variable)
                near_neighbours[fur_neigh].dist = eucl_dist;
            }

            for (i = 0; i < k_near; i++) {
                //fprintf(f, "\n\nNeighbour, index %d: \n", i);
                //fprintf(f, "Index: %d\n", near_neighbours[i].index);
                //fprintf(f, "Distance: %lf\n", near_neighbours[i].dist);
            }

        }
        // New element,
        (*n_added)++;
    }
}
