#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include "common.h"
#include "macros.h"
#include "kd.h"
#include "kclustering.h"


// #define FORCE_CLST   189
void kdClust(scalar **data_instances, int n_patt, int *partition, int *n_subsets, int *count, tCLST **clst_info, int **new_clst_root, int **match, int k_near, int k_t,
             int inputs, int classes)
{
    int i, j, /*node_class = 0,*/ n_patt_term = 0;
    //int n_terminal = 0;
    scalar **data_less = NULL, **data_greater = NULL;
    tNN *near_neighbours = NULL;
    tNN **neigh_matrix = NULL;
    nodeSIM **head_sim = NULL;
    nodeID **head_id = NULL;
    tKDTREE *root_node;
    nodeINV **head_inv = NULL;
    nodePREVNN *head_prev = NULL;
    //Euclidean distance vector
    //scalar *eucl;

    //Memory allocation.
    MALLOC(data_less, scalar *, n_patt, "kdClust");
    MALLOC(data_greater, scalar *, n_patt, "kdClust");
    for (i = 0; i < n_patt; i++) {
        MALLOC(data_less[i], scalar, inputs + classes, "kdClust");
        MALLOC(data_greater[i], scalar, inputs + classes, "kdClust");
    }

    MALLOC(near_neighbours, tNN, k_near, "kdClust");
    MALLOC(neigh_matrix, tNN *, n_patt, "kdClust");
    CALLOC(head_inv, nodeINV *, n_patt, "kdClust");
    CALLOC(head_sim, nodeSIM *, n_patt, "kdClust");
    CALLOC(head_id, nodeID *, n_patt, "kdClust");
    CALLOC(head_prev, nodePREVNN, n_patt, "kdClust");

    root_node = MakeTree(data_instances, &data_less, &data_greater, n_patt, n_patt, -1, &n_patt_term);

    //Create neighborhood table
    CreateNeighbourhoodTable(data_instances, &head_prev, &neigh_matrix, near_neighbours, root_node, n_patt, k_near, inputs);

    CreateInvertedList(&head_inv, neigh_matrix, n_patt, k_near);

    CreateSimilarityList(&head_sim, head_inv, n_patt);

    CreateIdentityList(&head_id, &head_inv, head_sim, n_patt, k_t);

    *n_subsets = 0;
    LabelElements(head_id, partition, clst_info, n_subsets, count, match, new_clst_root, n_patt);

    //fprintf(f, "\n\nElement   /   Cluster   /   Class\n");
    for (i = 0; i < n_patt; i++) {
        //Convert binary class into integer class
        j = 0;
        //                 node_class = 0;
        while (data_instances[i][j + inputs] == 0)
            j++;
        //                 node_class = j + 1;

    }

    //Free memory allocation.
    FREE(data_less);
    FREE(data_greater);
    FREE(near_neighbours);
    //FREE(eucl);
    FREE(neigh_matrix);
    //Recursively free memory space of the lists
    for (j = 0; j < n_patt; j++) {
        recursiveFreeInvList(&head_inv[j]);
        recursiveFreeSimList(&head_sim[j]);
    }
    for (j = 0; j < *n_subsets; j++) {
        //Free root lists of the newly added clusters
        //fprintf(f, "*Free cluster %d\n", (*new_clst_root)[j]);
        fflush(stdin);
        recursiveFreeIdList(&head_id[(*new_clst_root)[j]]);
    }

    FREE(head_inv);
    FREE(head_sim);
    FREE(head_id);

}

void kdTreeClusteringAlgorithm(const t_Parameters *parameters, int *partition, int *count, int *n_subsets, scalar **data, int instances, int k_near, int k_t, int inputs,
                               int classes)
{
    int i, j, k, group = 0, comp = 0, prev_elem = 0, new_root = 0, rem = 0, real_new, stck_ind = 0, index = 0, root_index = 0, n_elem = 0;
    tCLST *clst_info = NULL;
    int *new_clst_root;
    int *clst_stack;
    //Reduced pattern matrix
    scalar **red_data;

    //Matching between 'data' and 'red_data' indexes
    int *match;

    MALLOC(clst_info, tCLST, instances, "kdTreeClusteringAlgorithm");
    CALLOC(clst_stack, int, 10000, "kdTreeClusteringAlgorithm");
    MALLOC(match, int, instances, "kdTreeClusteringAlgorithm");


    //Initialize variables
    for (i = 0; i < instances; i++) {
        //Set initial matching between 'data' and 'red_data' indexes
        match[i] = i;
        //Initialize number of cluster elements
        count[i] = 0;
        partition[i] = 0;
    }

    //Store index elements in first column
    for (i = 0; i < instances; i++) {
        data[i][0] = i;
        //Shift columns 1 position to the right. Now indexes would be in data[i][-1]
        data[i]++;
    }
    inputs--;

    //First execution of the clustering algorithm
    kdClust(data, instances, partition, n_subsets, count, &clst_info, &new_clst_root, &match, k_near, k_t, inputs, classes);
    FREE(new_clst_root);

    for (i = 0; i < instances; i++) {
        //Stack cluster with count[i] > parameters->subset_size
        if (count[i] > parameters->subset_size) {
            if (stck_ind > 10000) {
                REALLOC(clst_stack, int, 10000, "kdTreeClusteringAlgorithm");
            }
            clst_stack[stck_ind] = i;
            //fprintf(stdout, "Push: %d. N%d: %d\n", stck_ind, i, count[i]);
            stck_ind++;
        }
    }

    //Process the stack
    while (stck_ind > 0) {
        root_index = clst_stack[stck_ind - 1];
        //fprintf(stdout, "%d.Pop: R%d, n%d\n", stck_ind - 1, root_index, count[root_index]);
        stck_ind--;

        //Recluster
        //Memory allocation
        MALLOC(red_data, scalar *, count[root_index], "Sim");

        for (j = 0; j < count[root_index]; j++) {
            index = clst_info[root_index].components[j];
            red_data[j] = data[index];

            //Update absolute index in red_data matrix to local index
            red_data[j][-1] = j;

            //Store matching between 'data' and 'red_data' indexes
            match[j] = index;
            //fprintf(f, "Renombro de indice %d a %d\n", index, j);
        }

        //Reset number of elements of the cluster we are going to reclusterize
        n_elem = clst_info[root_index].number;
        count[root_index] = 0;

        kdClust(red_data, n_elem, partition, n_subsets, count, &clst_info, &new_clst_root, &match, k_near, k_t, inputs, classes);

        //Restore original indexes
        for (j = 0; j < instances; j++) {
            data[j][-1] = match[j];
        }

        //Free memory allocation.
        FREE(red_data);


        //fprintf(stdout, "\n*%d produced %d clusters: ", root_index, *n_subsets);
        if (*n_subsets == 1) {
            //Conversion needed as new_clst_root has local indexes
            real_new = match[new_clst_root[0]];

            //If it is has less elements than the maximum allowed: stop reclustering
            if (count[real_new] < parameters->subset_size) {
                //Delete previous cluster
                clst_info[root_index].number = 0;
                count[root_index] = 0;

                //fprintf(stdout, "Let cluster %d as it is (n_elem: %d)\n", real_new, count[real_new]);
                fflush(stdin);
            }
            //If it is has more elements than the maximum allowed: force division
            else {
                //Delete previous cluster
                clst_info[root_index].number = 0;
                count[root_index] = 0;

                //We divide the original cluster into clusters with parameters->subset_size elements (as maximum)
                prev_elem = count[real_new];
                group = count[real_new] / parameters->subset_size;
                //fprintf(stdout, "\nForce division of cluster %d (n_elem: %d)\n", real_new, count[real_new]);
                //fprintf(stdout, "%d cluster of parameters->subset_size elems\n", group);

                //Update number of elements of the original cluster
                count[real_new] = parameters->subset_size;
                clst_info[real_new].number = parameters->subset_size;


                comp = 0;
                //We start in j=1 (not 0) because the first parameters->subset_size elements remain in the original cluster
                for (j = 1; j < group; j++) {
                    //Create a new cluster, its root would be the first one different from the current root
                    while (real_new == clst_info[real_new].components[comp]) {
                        comp++;
                        //fprintf(stdout, "Following component %d\n", clst_info[real_new].components[comp]);
                    }
                    new_root = clst_info[real_new].components[comp];
                    //fprintf(stdout, "New cluster %d\n", new_root);
                    clst_info[new_root].index = new_root;
                    clst_info[new_root].number = parameters->subset_size;
                    count[new_root] = parameters->subset_size;

                    //Memory allocation
                    MALLOC(clst_info[new_root].components, int, parameters->subset_size, "Sim");

                    for (k = 0; k < parameters->subset_size; k++) {
                        clst_info[new_root].components[k] = clst_info[real_new].components[parameters->subset_size * j + k];
                        partition[clst_info[new_root].components[k]] = new_root;
                    }
                    comp++;
                }

                rem = prev_elem % parameters->subset_size;
                //fprintf(stdout, "Remains: %d\n", rem);
                if (rem != 0) {
                    //Create a new cluster, its root would be the first one different from the current root
                    while (real_new == clst_info[real_new].components[comp]) {
                        comp++;
                    }
                    new_root = clst_info[real_new].components[comp];
                    //fprintf(stdout, "New cluster %d\n", new_root);
                    clst_info[new_root].index = new_root;
                    clst_info[new_root].number = rem;
                    count[new_root] = rem;
                    //Memory allocation
                    MALLOC(clst_info[new_root].components, int, clst_info[new_root].number, "Sim");

                    for (k = 0; k < clst_info[new_root].number; k++) {
                        clst_info[new_root].components[k] =
                            clst_info[real_new].components[parameters->subset_size * group + k];
                        partition[clst_info[new_root].components[k]] = new_root;
                    }
                }
            }                   //end_if > FORCE_CLST
        }                       //end_if n_subsets = 1
        //More than one cluster produced, consider reclusterization (adding index to the stack)
        else {
            for (j = 0; j < *n_subsets; j++) {
                //Conversion needed as new_clst_root has local indexes
                real_new = match[new_clst_root[j]];
                //              fprintf(stdout, "%d (n_elemen: %d )\t", real_new, count[real_new]);
                if (count[real_new] > parameters->subset_size) {
                    if (stck_ind > 10000) {
                        REALLOC(clst_stack, int, 10000, "Sim");
                    }
                    clst_stack[stck_ind] = real_new;
                    //                  fprintf(stdout, " \n%d.Push: N%d: %d\n", stck_ind, real_new, count[real_new]);
                    stck_ind++;
                }
            }
        }
        FREE(new_clst_root);

    }                           //end_while

    //Cluster indexes: translate instance indexes into sequential indexes
    int aux = 0, clst_ind = 0;
    int *transl;
    MALLOC(transl, int, instances, "Sim");
    for (i = 0; i < instances; i++)
        transl[i] = -1;

    for (i = 0; i < instances; i++) {
        aux = partition[i];
        if (transl[aux] == -1) {
            transl[aux] = clst_ind;
            clst_ind++;
        }
        partition[i] = transl[aux];
    }
    FREE(transl);

    *n_subsets = clst_ind;

    // Count instances per cluster
    for (int i = 0; i < *n_subsets; i++)
        count[i] = 0;

    for (int i = 0; i < instances; i++)
        count[partition[i]]++;

    // Restore bias in first column
    for (i = 0; i < instances; i++) {
        // Unshift columns
        data[i]--;

        data[i][0] = 1.0;
    }
    // Restore number of inputs
    inputs++;

    FREE(clst_stack);
    FREE(clst_info);
    FREE(match);
}
