/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "macros.h"
#include "is.h"

// Margin for test instance

scalar NearestNeighbourMarginTestInstance(scalar *instance, int instance_class, int *S, scalar **protos, int n_protos, int inputs, int *labels, scalar *r, scalar **dist_cache, int  **index)
{
    scalar ok_nearest = INFINITE, wrong_nearest = INFINITE, value, margin;
    bool at_least_one_ok, at_least_one_wrong;

    if (r || dist_cache || index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "R/Cache/Index");

    at_least_one_ok = at_least_one_wrong = false;

    // Iterate for every prototype
    for (int i = 0; i < n_protos; i++)
        if (S[i]) {
            if (labels[i] == instance_class) {
                if ((value = EuclideanDistanceSparseData(instance, protos[i], inputs, NULL, NULL)) < ok_nearest) {
                    ok_nearest = value;
                    at_least_one_ok = true;
                }
            } else {
                if ((value = EuclideanDistanceSparseData(instance, protos[i], inputs, NULL, NULL)) < wrong_nearest) {
                    wrong_nearest = value;
                    at_least_one_wrong = true;
                }
            }
        }

    if (at_least_one_ok && at_least_one_wrong)
        margin = (wrong_nearest - ok_nearest);
    else if (at_least_one_ok)
        margin = 1.0/ok_nearest;
    else if (at_least_one_wrong)
        margin = -1.0/wrong_nearest;
    else
        margin = 0.0;

    return margin;
}
