/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "is.h"
#include "macros.h"
#include "common.h"

// Meme optimization: Salvador García, José Ramón Cano and Francisco Herrera, "A memetic algorithm for evolutionary prototype selection: A scaling up approach", Pattern Recognition 41 (2008),
// 2693 – 2709

void MemeOptimization(int *s, cache_t **neighbors_cache, int instances, int *labels)
{
    bool *r;
    int selected, threshold = 0, *list;

    MALLOC(r, bool, instances, "MemeOptimization: r");
    MALLOC(list, int, instances, "MemeOptimization: list");

    selected = 0;
    for (int i = 0; i < instances; i++) {
        r[i] = false;
        if (s[i]) {
            list[selected] = i;
            selected++;
        }
    }

    while (selected > 1) {
        // Check all currently selected instances are not in R
        bool all_in_r = true;
        for (int i = 0; i < selected && all_in_r; i++)
            if (!r[list[i]])
                all_in_r = false;

        // Finish?
        if (all_in_r)
            break;

        int j, index;

        // Chose j in S not in R to avoid duplicate attempts
        do {
            index = RANDOM(0, selected);
        } while (r[list[index]]);

        int gain = 0;

        j = list[index];

        for (int i = 0; i < instances; i++) {
            // Check if nearest neighbor of i is j
            int u, k = 0;
            do {
                u = neighbors_cache[i][k];
                k++;
            } while (!s[u]);

            if (u == j) {
                // Update neighbor and check whether there is a gain
                s[j] = 0;

                int u_prime;
                do {
                    u_prime = neighbors_cache[i][k];
                    k++;
                } while (!s[u_prime]);

                if (labels[i] == labels[u] && labels[i] != labels[u_prime])
                    gain--;

                if (labels[i] != labels[u] && labels[i] == labels[u_prime])
                    gain++;
            }

            s[j] = 1;
        }

        if (gain >= threshold) {
            // Remove
            s[j] = 0;
            selected--;
            list[index] = list[selected];
            for (int i = 0; i < instances; i++)
                r[i] = false;
        } else {
            s[j] = 1;
            r[j] = true;
        }
    }

    FREE(r);
    FREE(list);
}
