/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <time.h>
#include "is.h"
#include "macros.h"

bool RandomMutation(const t_Parameters *parm, int *individual, int size, scalar bit_mutation, int *label, int minority, bool mutate_local_k, int min_k, int max_k, int g, int max_g)
{
    bool modified = false;

    for (int j = 0; j < size; j++) {
        // Mutation from 1 to 0 might be more likely than 0 to 1
        if (RANDOM(0.0, 1.0) < bit_mutation) {
            modified = true;

            if (individual[j])
                individual[j] = 0;
            else
                individual[j] = (!parm->multiselection || (parm->multiselect_only_minority_class && label [j] != minority) ? 1 : RANDOM(1, (parm->k/2)+2));
        }
    }

    // Non uniform mutation
    if (mutate_local_k) {
        modified = true;

        NonUniformMutation(individual+size, size, g, max_g, bit_mutation, min_k, max_k);
    }

    return modified;
}

void RNNMutation(int individual, int subpop, cache_t **dist_cache, scalar **data, int inputs, int classes, int *train_labels)
{
    /*scalar all_ok, ok;
    int nearest, removed = 0, selected, *sample = NULL;*/

    OnErrorExit(1, "This function must be updated to account for the new definition of NearestNeighbourCache");
    OnErrorExit(1, "This function must be updated to account for sample being now a local variable");

    /*// Create a sample with the current instances
    selected = 0;
    for (int i = gl_limits[subpop].init; i < gl_limits[subpop].end; i++)
        if (gl_subpopulation[subpop][individual][i - gl_limits[subpop].init]) {
            sample[selected] = i;
            selected++;
        }

    if (selected <= 1)
        return;

    // Test if removing each instance affects the classification of any other instance
    all_ok = 0.0;
    for (int i = gl_limits[subpop].init; i < gl_limits[subpop].end; i++)
        if (kNearestNeighbourCache(i, sample, selected, &nearest, train_labels, gl_k, dist_cache, data, inputs, classes) == train_labels[i])
            all_ok++;

    // Remove each instance and test
    for (int i = 0; i < selected && selected > 1; i++) {
        // Assign to next
        int aux = sample[i];

        sample[i] = (i == selected - 1) ? sample[i - 1] : sample[i+1];

        ok = 0.0;
        for (int j = gl_limits[subpop].init; j < gl_limits[subpop].end; j++)
            if (kNearestNeighbourCache(j, sample, selected, &nearest, train_labels, gl_k, dist_cache, data, inputs, classes) == train_labels[j])
                ok++;

        if (ok >= all_ok) {
            // Remove definitively
            gl_subpopulation[subpop][individual][aux - gl_limits[subpop].init] = 0;
            removed++;

            // Remove from sample
            for (int j = i+1; j < selected - 1; j++)
                sample[j] = sample[j+1];
            selected--;
            i--;
        } else
            sample[i] = aux;
    }

    if (gl_verbose > 2)
        fprintf(gl_out_file, "Subpop: %02d, individual: %02d. RNN removed %3d instances\n", subpop, individual, removed);*/
}


void NonUniformMutation(int *individual, int n, int generation, int max_generations, double bit_mutation_rate, int min_value, int max_value)
{
    scalar delta;

    for (int i = 0; i < n; i++)
        if (RANDOM(0.0, 1.0) < bit_mutation_rate) {
            if (RANDOM(0.0, 1.0) < 0.5)
                individual[i] += (delta = DeltaNUM(generation, (double) max_value - individual[i], max_generations));
            else
                individual[i] -= (delta = DeltaNUM(generation, (double) individual[i] - min_value, max_generations));

            if (individual[i] < min_value)
                individual[i] = min_value;
            else if (individual[i] > max_value)
                individual[i] = max_value;
        }
}

double DeltaNUM(int t, double y, int max_g)
{
    double r, b = 5.0;

    r = RANDOM(0.0, 1.0);

    return y * (1.0 - pow(r, pow(1.0 - ((double)t)/((double)max_g), b)));
}
