/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Functions that implement genetic operators
#include "is.h"
#include "macros.h"

void TwoPointCrossover(int *offspring1, int *offspring2, int *father, int *mother, int n)
{
    int first_point, second_point;

    first_point = RANDOM(0, n);
    second_point = RANDOM(first_point, n);

    // Perform crossover.
    for (int k = 0; k < first_point; k++) {
        offspring1[k] = father[k];
        offspring2[k] = mother[k];
    }

    for (int k = first_point; k < second_point; k++) {
        offspring2[k] = father[k];
        offspring1[k] = mother[k];
    }

    for (int k = second_point; k < n; k++) {
        offspring1[k] = father[k];
        offspring2[k] = mother[k];
    }
}

void HuxCrossover(int *father, int *mother, int *off1, int *off2, int n)
{
    bool turn = true;
    int next1, next2;

    for (int i = 0; i < n; i++)
        if (father[i] == mother[i]) {
            off1[i] = off2[i] = mother[i];
        } else {
            if (turn) {
                next1 = father[i];
                next2 = mother[i];
                turn = false;
            } else {
                next1 = mother[i];
                next2 = father[i];
                turn = true;
            }

            off1[i] = next1;
            off2[i] = next2;
        }
}

int HammingDistance(int *x, int *y, int n)
{
    int dist = 0;

    for (int i = 0; i < n; i++)
        if (x[i] != y[i])
            dist++;

    return dist;

}

void UniformCrossover(int *father, int *mother, int *off1, int *off2, int n)
{
    for (int i = 0; i < n; i++)
        // Assign one of the parents with a 0.5 probability
        if (RANDOM(0.0, 1.0) < 0.5)  {
            off1[i] = father[i];
            off2[i] = mother[i];
        } else {
            off2[i] = father[i];
            off1[i] = mother[i];
        }
}

// Biased HUX crossover for sparse binary representations
// 1's and 0's are inherited with different probability

void BiasedHuxCrossover(int *father, int *mother, int *off1, int *off2, int n, scalar one_bias)
{
    bool turn = true;
    int next1, next2;

    for (int i = 0; i < n; i++)
        if (father[i] == mother[i]) {
            off1[i] = off2[i] = mother[i];
        } else {
            if (turn) {
                next1 = father[i] ? father[i] : (RANDOM(0.0, 1.0) < one_bias ? mother[i] : father[i]);
                next2 = mother[i] ? mother[i] : (RANDOM(0.0, 1.0) < one_bias ? father[i] : mother[i]);
                turn = false;
            } else {
                next1 = mother[i] ? mother[i] : (RANDOM(0.0, 1.0) < one_bias ? father[i] : mother[i]);
                next2 = father[i] ? father[i] : (RANDOM(0.0, 1.0) < one_bias ? mother[i] : father[i]);
                turn = true;
            }

            off1[i] = next1;
            off2[i] = next2;
        }
}
