/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "is.h"
#include "macros.h"

// Random selection of instances

int RandomSelection(int *S, scalar **data, int instances, int *label, scalar ratio, bool with_replacement, int inputs, int classes)
{
    int selected;
    bool class_concious = false;

    // First un-select all instances
    selected = 0;
    for (int i = 0; i < instances; i++)
        S[i] = 0;

    if (!class_concious)  {
        if (with_replacement) {
            // Select instances with replacement
            for (int i = 0; i < ratio * instances; i++) {
                S[(int) RANDOM(0, instances) ] = 1;
                selected++;
            }
        } else {
            // Select instances without replacement
            while (selected < (ratio * instances)) {
                int this_one = RANDOM(0, instances);
                if (!S[this_one]) {
                    S[this_one] = 1;
                    selected++;
                }
            }
        }
    } else {
        // Class conscious
        // Select the same ratio for each class
        int count, list[instances];
        for (int i = 0; i < classes; i++) {
            // List of instances of the class
            count = 0;
            for (int j = 0; j < instances; j++)
                if (label[j] == i) {
                    list[count] = j;
                    count++;
                }

            RandomizeVector(list, count);

            for (int j = 0; j < ratio * count; j++) {
                S[list[j]] = 1;
                selected++;
            }
        }
    }

    return selected;
}

// Select one random sample from every class

int SelectOneRandomSampleFromEveryClass(int *S, int instances, int *labels, int classes)
{
    // First un-select all instances
    for (int i = 0; i < instances; i++)
        S[i] = 0;

    // Select one random instance from every class
    int count, list[instances];
    for (int i = 0; i < classes; i++) {
        // List of instances of the class
        count = 0;
        for (int j = 0; j < instances; j++)
            if (labels[j] == i) {
                list[count] = j;
                count++;
            }

        if (count)
            S[list[(int) RANDOM(0, count)]] = 1;
    }


    return classes;
}
