/***************************************************************************
 *   Copyright (C) 2012 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program *is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program *is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "is.h"
#include "macros.h"

// Random selection of instances and feature with time limit for evaluation

scalar RandomISFSWithTimeLimit(const t_Parameters *parm, int *S, int *subspace, scalar **data, int instances, int *label, scalar m_i_ratio, scalar M_i_ratio,
                               scalar m_f_ratio, scalar M_f_ratio, int features, int classes, int time_limit, int *is, int *fs, scalar alpha)
{
    scalar acc, coef, best_coef = -1.0, red;
    int *do_not_repeat, *list_is, *list_fs, instances_selected = 0, features_selected = 0;
    time_t init_time;

    // Repeat until time exhausted
    init_time = time(NULL);

    // Memory
    MALLOC(list_is, int, instances, "RandomISFSWithTimeLimit: list_is");
    MALLOC(list_fs, int, features, "RandomISFSWithTimeLimit: list_fs");
    MALLOC(do_not_repeat, int, instances+features, "RandomISFSWithTimeLimit: do_not_repeat");

    do {
        // First un-select all instances and features
        instances_selected = features_selected = 0;
        for (int i = 0; i < instances+features; i++)
            do_not_repeat[i] = 0;

        // Select instances without replacement
        scalar i_ratio = RANDOM(m_i_ratio, M_i_ratio);
        while (instances_selected < (i_ratio * instances)) {
            int this_one = RANDOM(0, instances);
            if (!do_not_repeat[this_one]) {
                do_not_repeat[this_one] = 1;
                list_is[instances_selected] = this_one;
                instances_selected++;
            }
        }

        // Select features without replacement
        scalar f_ratio = RANDOM(m_f_ratio, M_f_ratio);
        while (features_selected < (f_ratio * features)) {
            int this_one = RANDOM(0, features);
            if (!do_not_repeat[this_one+instances]) {
                do_not_repeat[this_one+instances] = 1;
                list_fs[features_selected] = this_one;
                features_selected++;
            }
        }

        // Evaluate selection
        // Final results
        acc = EvaluateWithList((t_Parameters *) &parm, data, instances, label, list_is, data, instances_selected, label, list_fs, NULL, features, classes, kNN, NULL, NULL, NULL, NULL, NULL);
        red = 1.0 - (((scalar)instances_selected)/instances)*(((scalar)features_selected)/features);

        coef = alpha*acc+(1.0-alpha)*red;

        // Best? Store
        if (coef > best_coef) {
            best_coef = coef;
            *fs = features_selected;
            *is = instances_selected;
            for (int i = 0; i < instances; i++)
                S[i] = do_not_repeat[i];
            for (int i = 0; i < features; i++)
                subspace[i] = do_not_repeat[i+instances];
        }

    } while (time(NULL) - init_time < time_limit);

    // Copy best
    FREE(list_fs);
    FREE(list_is);

    return 1.0 - (((scalar)*is)/instances)*(((scalar)*fs)/features);
}
