/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "defns.i"
#include "types.i"
#include "extern.i"
#include "classifiers.h"

// RobustC45 algorithm
// G. H. John, "Robust decision trees: removing outliers from databases", First International Conference on Knowledge Discovery and Data Mining, pp. 174-179,
// 1995.

int RobustC45(const t_Parameters *parm, int *S, scalar **data, int instances, int *labels, int inputs, int classes, int *subspace, int **index)
{
    Tree tree;
    bool finish;
    int Np;
    t_Parameters local_parameters;

    for (int i = 0; i < instances; i++)
        S[i] = 1;

    Np = instances;

    // Always prune data
    local_parameters = *parm;
    local_parameters.prune_tree = true;
    local_parameters.soften_thresholds = false;
    TRIALS = 1;

    do {
        // Train tree with the current subset
        // Learn model
        tree = TrainC45(&local_parameters, data, subspace, instances, labels, inputs, true, S, classes);

        // Remove misclassified instances
        finish = true;
        for (int i = 0; i < instances; i++)
            if (S[i] && (C45Output(tree, data[i], subspace, inputs) != labels[i])) {
                S[i] = 0;
                finish = false;

                Np--;
            }

        ReleaseTree(tree);

        // While there are misclassified instances
    } while (!finish);


    return Np;
}

int RobustDT(const t_Parameters *parm, int *S, scalar **data, int instances, int *labels, int inputs, int classes, int *subspace, int **index)
{
    t_DTree dt;
    bool finish;
    int Np;
    scalar probabilities[classes];

    for (int i = 0; i < instances; i++)
        S[i] = 1;

    Np = instances;

    do {
        // Train tree with the current subset
        // Learn model
        dt = DecisionTreeTrain(parm, data, instances, labels, inputs, S, subspace, classes, false, index, imGainRatio, false, false);

        // Remove misclassified instances
        finish = true;
        for (int i = 0; i < instances; i++)
            if (S[i] && (DecisionTreeClassify(dt, data[i], probabilities, classes, index ? index[i] : NULL) != labels[i])) {
                S[i] = 0;
                finish = false;

                Np--;
            }

        FreeDecisionTree(dt);

        // While there are misclassified instances
    } while (!finish);


    return Np;
}
