/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Compute reachable and coverage sets

#include <stdio.h>
#include <string.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "maths.h"
#include "classifiers.h"

void ComputeReachableAndCoverage(const t_Parameters *parm, scalar *reachable, scalar *coverage, int *S, scalar **data, int *labels, int instances, int inputs, int classes,
                                 cache_t **neighbors_cache, scalar **distances_cache, bool *considerlist, int k, scalar *r, int **index, int *subspace)
{
    scalar enemy, value;
    int nearest[k];

    // Initially empty
    for (int i = 0; i < instances; i++)
        coverage[i] = reachable[i] = 0.0;

    // Compute both sets
    for (int i = 0; i < instances; i++)
        if (S[i]) {
            // Compute nearest enemy
            enemy = INFINITE;
            if (distances_cache == NULL) {
                for (int j = 0; j < instances; j++)
                    if (S[j] && labels[j] != labels[i] && (value = EuclideanDistanceSparseDataSubspace(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL, subspace)) < enemy)
                        enemy = value;
            } else {
                for (int j = 0; j < instances; j++)
                    if (S[j] && labels[j] != labels[i] && (value = distances_cache[i][j]) < enemy)
                        enemy = value;
            }

            // Add to reachable instances nearer than enemy
            // If no enemy bound by k neighbors
            if (enemy >= INFINITE) {
                kNearestNeighbourCache(i, S, instances, nearest, labels, k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, index, subspace);
                for (int i = 0; i < k; i++)
                    if (nearest[i] != -1) {
                        if (considerlist == NULL || considerlist[i]) {
                            reachable[i]++;
                            coverage[nearest[i]]++;
                        }
                    }
            } else {
                if (distances_cache == NULL) {
                    for (int j = 0; j < instances; j++)
                        if (S[j] && EuclideanDistanceSparseDataSubspace(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL, subspace) < enemy)  {
                            if (considerlist == NULL || considerlist[j]) {
                                coverage[j]++;
                                reachable[i]++;
                            }
                        }
                } else {
                    for (int j = 0; j < instances; j++)
                        if (S[j] && distances_cache[i][j] < enemy)  {
                            if (considerlist == NULL || considerlist[j]) {
                                coverage[j]++;
                                reachable[i]++;
                            }
                        }
                }
            }
        }

}
