/***************************************************************************
 *   Copyright (C) 2012 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include "is.h"
#include "macros.h"
#include "common.h"

// Functions related to sampling

// Test whether all the instances are from the same class
bool TestOneClass(int *labels, int instances, int classes)
{
    int count[classes], present = 0;

    for (int i = 0; i < classes; i++)
        count[i] = 0;

    for (int i = 0; i < instances; i++)
        count[labels[i]]++;

    // Check more than one class
    for (int i = 0; i < classes; i++)
        if (count[i] != 0) {
            present++;

            if (present > 1)
                return false;
        }

    return true;
}

int *RandomSampling(scalar ***data, int **labels, int *instances, int subset_size, scalar ratio, int ***index, int classes, bool stratified)
{
    int *list, size = 0;

    // New size
    if (subset_size)
        size = subset_size;
    else if (ratio > ZERO)
        size = *instances*ratio;
    else
        OnErrorExit(ERR_PARAM, "Pre-sample with null size");

    MALLOC(list, int, *instances, "RandomSampling: list");

    if (size > *instances) {
        for (int i = 0; i < *instances; i++)
            list[i] = i;

        return list;
    }

    if (!stratified) {
        // Random list of instances
        for (int i = 0; i < *instances; i++)
            list[i] = i;
        RandomizeVector(list, *instances);
    } else {
        scalar r = size/((scalar) *instances);

        // Random selection per class
        size = 0;
        for (int i = 0; i < classes; i++) {
            int count = 0;
            for (int j = 0; j < *instances; j++)
                if ((*labels)[j] == i) {
                    list[count+size] = j;
                    count++;
                }

            // Randomize class
            RandomizeVector(list+size, count);

            size += count*r;
        }
    }

    // Sort selected size
    qsort(list, size, sizeof(int), CompareInt);

    // Random selection of first instances
    for (int i = 0; i < size; i++) {
        scalar *sw_s_ptr = (*data)[i];
        (*data)[i] = (*data)[list[i]];
        (*data)[list[i]] = sw_s_ptr;

        int sw_i = (*labels)[i];
        (*labels)[i] = (*labels)[list[i]];
        (*labels)[list[i]] = sw_i;

        if (index) {
            int *sw_i_ptr = (*index)[i];
            (*index)[i] = (*index)[list[i]];
            (*index)[list[i]] = sw_i_ptr;
        }
    }

    // Free and resize
    for (int i = size; i < *instances; i++)
        FREE((*data)[i]);
    REALLOC(*data, scalar *, size, "RandomSampling: *data");
    REALLOC(*labels, int, size, "RandomSampling: *labels");
    if (index) {
        for (int i = size; i < *instances; i++)
            FREE((*index)[i]);
        REALLOC(*index, int *, size, "RandomSampling: *index");
    }
    *instances = size;

    // Return list of selected instances
    REALLOC(list, int, *instances, "RandomSampling: list");

    return list;
}
