/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
/*
 * Strategies of selection of individuals from the population.
 */
#include <stdlib.h>
#include "macros.h"
#include "is.h"

// Select n-th individual ordered by fitness.
int OrderedSelection(scalar *fitness, int n, int no_indiv)
{
    int i, max_index;
    RANK ranking[no_indiv];

    // If the best just search the maximum.
    if (n == 0) {
        max_index = 0;
        for (i = 1; i < no_indiv; i++)
            if (fitness[i] > fitness[max_index])
                max_index = i;

        return max_index;
    } else {
        // Ordered by fitness
        for (i = 0; i < no_indiv; i++) {
            ranking[i].being = i;
            ranking[i].fitness = fitness[i];
        }

        qsort(ranking, no_indiv, sizeof(RANK), Compare);

        // n-th individual
        return (ranking[n].being);
    }
}

// N individuals are chosen at random, the one with highest fitness is selected.
int TournamentSelection(scalar *fitness, int no_rival, int no_indiv)
{
    int best, beings[no_rival];

    // N individuals are chosen.
    InitVectorRandomly(NULL, beings, no_rival, 0.0, (scalar) no_indiv);

    // The most fitted is returned
    best = beings[0];
    for (int i = 1; i < no_rival; i++)
        if (fitness[beings[i]] > fitness[best])
            best = beings[i];

    return best;
}

// Return the ordered list of indexes
void OrderedList(scalar *fitness, int *list, int no_indiv, bool MaxToMin)
{
    RANK ranking[no_indiv];

    // Ordered by fitness.
    for (int i = 0; i < no_indiv; i++) {
        ranking[i].being = i;
        ranking[i].fitness = (MaxToMin ? fitness[i] : -fitness[i]);
    }

    qsort(ranking, no_indiv, sizeof(RANK), Compare);

    // Ordered list
    for (int i = 0; i < no_indiv; i++)
        list[i] = ranking[i].being;

}

// Binary tournament
int BinaryTournament(scalar *fitness, int n)
{
    // Random selection of two individuals
    int one = RANDOM(0, n);
    int another = RANDOM(0, n);

    if (fitness[one] > fitness[another])
        return (one);
    else
        return (another);
}

// Select N best individuals sorted by fitness.
void NBestOrderedSelection(scalar *fitness, int n, int n_best, int *list)
{
    RANK ranking[n];

    // Sorted by fitness
    for (int i = 0; i < n; i++) {
        ranking[i].being = i;
        ranking[i].fitness = fitness[i];
    }

    qsort(ranking, n, sizeof(RANK), Compare);

    for (int i = 0; i < n_best; i++)
        list[i] = ranking[i].being;
}
