#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 23 11:41:15 2021

@author: frodo

Michael Rapp, Eneldo Loza Mencía, Johannes Fürnkranz Vu-Linh Nguyen and Eyke Hüllermeier. Learning Gradient Boosted Multi-label Classification Rules.
In: Proceedings of the European Conference on Machine Learning and Knowledge Discovery in Databases (ECML-PKDD), 2020, Springer.

pip install mlrl-testbed

"""
# BOOMER multi-label algorithm

from numpy import asarray
import sys
import warnings
import argparse
import numpy as np
from skmultilearn.dataset import load_from_arff
import scipy
import sklearn.metrics as metrics
from mlrl.boosting import Boomer

warnings.filterwarnings("ignore")

parser = argparse.ArgumentParser()
parser.add_argument("dataset", help="Dataset (must be a -.train and a -.test)", type=str)
parser.add_argument("nlabels", help="Number of labels", type=int)

parser.add_argument("-sparse", help="File is in sparse format", action='store_true')
parser.add_argument("-cv", help="Cross-validate hyper-parameters", action='store_true')
args = parser.parse_args()

# Get the BOOMER model
clf = Boomer(time_limit=3600, prediction_format='dense')

if args.cv:
    print("BOOMER ignores hyper-parameters cross-validation")

# Load dataset
X_train, Y_train = load_from_arff(args.dataset+".training", label_count=args.nlabels, label_location='end', load_sparse=args.sparse)
X_test, Y_test = load_from_arff(args.dataset+".testing", label_count=args.nlabels, label_location='end', load_sparse=args.sparse)

# Test in train and test columns do not match. That could happen for sparse matrices
if X_train.shape[1] > X_test.shape[1]:
    toadd = scipy.sparse.lil_matrix((X_test.shape[0], X_train.shape[1] - X_test.shape[1]))
    X_test = scipy.sparse.hstack([X_test, toadd])
elif X_train.shape[1] < X_test.shape[1]:
    toadd = scipy.sparse.lil_matrix((X_train.shape[0], X_test.shape[1] - X_train.shape[1]))
    X_train = scipy.sparse.hstack([X_train, toadd])

n_inputs, n_outputs = X_train.shape[1], Y_train.shape[1]

clf.fit(X_train, Y_train)

# Test error
preds = clf.predict(X_test)
np.savetxt(args.dataset+".testing.preds", preds, fmt='%d')
preds = clf.predict_proba(X_test)
np.savetxt(args.dataset+".testing.probs", preds, fmt='%g')

sys.exit(1)
