#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 23 11:41:15 2021

@author: frodo
"""

import argparse
import sklearn
from keras.models import Sequential
from keras.layers import Dense
import arff
from skmultilearn.dataset import load_from_arff
import warnings
import sklearn.metrics as metrics
import numpy as np
import sys
import random
import scipy

from parameters import rfparam

warnings.filterwarnings("ignore")

parser = argparse.ArgumentParser()
parser.add_argument("dataset", help="Dataset (must be a -.train and a -.test)", type=str)
parser.add_argument("nlabels", help="Number of labels", type=int)

parser.add_argument("-sparse", help="File is in sparse format", action='store_true')
parser.add_argument("-cv", help="Cross-validate hyper-parameters", action='store_true')
parser.add_argument("-r", help="Use random order of labels", action='store_true')
parser.add_argument("-v", help="Verbose mode", action='store_true')
args = parser.parse_args()

# Hyper-parameters to optimize
parameters = {}

X_train, Y_train = load_from_arff(args.dataset+".training", label_count=args.nlabels, label_location='end', load_sparse=args.sparse)
X_test, Y_test = load_from_arff(args.dataset+".testing", label_count=args.nlabels, label_location='end', load_sparse=args.sparse)

# Test in train and test columns do not match. That could happen for sparse matrices
if X_train.shape[1] > X_test.shape[1]:
    toadd = scipy.sparse.lil_matrix((X_test.shape[0], X_train.shape[1] - X_test.shape[1]))
    X_test = scipy.sparse.hstack([X_test, toadd])
elif X_train.shape[1] < X_test.shape[1]:
    toadd = scipy.sparse.lil_matrix((X_train.shape[0], X_test.shape[1] - X_train.shape[1]))
    X_train = scipy.sparse.hstack([X_train, toadd])

n_inputs = X_train.shape[1]
n_outputs = Y_train.shape[1]

# Create model
model = Sequential()
model.add(Dense(50, input_dim=n_inputs, kernel_initializer='he_uniform', activation='relu'))
model.add(Dense(20, input_dim=20, kernel_initializer='he_uniform', activation='relu'))
model.add(Dense(n_outputs, activation='sigmoid'))
model.compile(loss='binary_crossentropy', optimizer='Adam')

if args.cv:
    print("WARNING: No cross-validation of hyper-parameters for DL")

model.fit(X_train.A, Y_train.A)

# Test error
probs = model.predict(X_test)
preds = probs.round()
np.savetxt(args.dataset+".testing.preds", preds, fmt='%d')
if args.v:
    print("F1: %f" % metrics.f1_score(Y_test, preds, average='samples', zero_division=0))

np.savetxt(args.dataset+".testing.probs", probs, fmt='%g')

sys.exit(1)
