#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 23 11:41:15 2021

@author: frodo
"""
# Use a MLP for prediction on multi-label classification
# Taken from https://machinelearningmastery.com/multi-label-classification-with-deep-learning/

from numpy import asarray
import sys
from sklearn.datasets import make_multilabel_classification
from keras.models import Sequential
from keras.layers import Dense
import warnings
import argparse
import numpy as np
from skmultilearn.dataset import load_from_arff
import scipy
import sklearn.metrics as metrics

warnings.filterwarnings("ignore")

parser = argparse.ArgumentParser()
parser.add_argument("dataset", help="Dataset (must be a -.train and a -.test)", type=str)
parser.add_argument("nlabels", help="Number of labels", type=int)
parser.add_argument("nodes", help="Number of hidden nodes", type=int)

parser.add_argument("-sparse", help="File is in sparse format", action='store_true')
parser.add_argument("-cv", help="Cross-validate hyper-parameters", action='store_true')
args = parser.parse_args()

# Get the MLP model
# Number of hidden nodes is bounded by 1,000
def get_model(n_inputs, n_outputs):
	model = Sequential()
	model.add(Dense(n_inputs if args.nodes == 0 else args.nodes, input_dim=n_inputs, kernel_initializer='he_uniform', activation='relu'))
	model.add(Dense(n_outputs, activation='sigmoid'))
	model.compile(loss='binary_crossentropy', optimizer='adam')
	return model

if args.cv:
    print("ML MLP ignores hyper-parameters cross-validation")

# Load dataset
X_train, Y_train = load_from_arff(args.dataset+".training", label_count=args.nlabels, label_location='end', load_sparse=args.sparse)
X_test, Y_test = load_from_arff(args.dataset+".testing", label_count=args.nlabels, label_location='end', load_sparse=args.sparse)

# Test in train and test columns do not match. That could happen for sparse matrices
if X_train.shape[1] > X_test.shape[1]:
    toadd = scipy.sparse.lil_matrix((X_test.shape[0], X_train.shape[1] - X_test.shape[1]))
    X_test = scipy.sparse.hstack([X_test, toadd])
elif X_train.shape[1] < X_test.shape[1]:
    toadd = scipy.sparse.lil_matrix((X_train.shape[0], X_test.shape[1] - X_train.shape[1]))
    X_train = scipy.sparse.hstack([X_train, toadd])

n_inputs, n_outputs = X_train.shape[1], Y_train.shape[1]

model = get_model(n_inputs, n_outputs)
model.fit(X_train.A, Y_train.A, verbose=0, epochs=10000, batch_size=16)

# Test error
preds = model.predict(X_test)
np.savetxt(args.dataset+".testing.preds", preds, fmt='%d')
np.savetxt(args.dataset+".testing.probs", preds, fmt='%g')

sys.exit(1)
