#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 23 11:41:15 2021

@author: frodo
"""

import argparse
from skmultilearn.adapt import MLTSVM
from skmultilearn.dataset import load_from_arff
from sklearn.model_selection import GridSearchCV
import numpy as np
import sys
import scipy
import warnings
import sklearn.metrics as metrics

warnings.filterwarnings("ignore")

parser = argparse.ArgumentParser()
parser.add_argument("dataset", help="Dataset (must be a -.train and a -.test)", type=str)
parser.add_argument("nlabels", help="Number of labels", type=int)

parser.add_argument("-sparse", help="File is in sparse format", action='store_true')
parser.add_argument("-cv", help="Cross-validate hyper-parameters", action='store_true')
args = parser.parse_args()

X_train, Y_train = load_from_arff(args.dataset+".training", label_count=args.nlabels, label_location='end', load_sparse=args.sparse)
X_test, Y_test = load_from_arff(args.dataset+".testing", label_count=args.nlabels, label_location='end', load_sparse=args.sparse)

# Test in train and test columns do not match. That could happen for sparse matrices
if X_train.shape[1] > X_test.shape[1]:
    toadd = scipy.sparse.lil_matrix((X_test.shape[0], X_train.shape[1] - X_test.shape[1]))
    X_test = scipy.sparse.hstack([X_test, toadd])
elif X_train.shape[1] < X_test.shape[1]:
    toadd = scipy.sparse.lil_matrix((X_train.shape[0], X_test.shape[1] - X_train.shape[1]))
    X_train = scipy.sparse.hstack([X_train, toadd])

classifier = MLTSVM(c_k = 2**-1, max_iteration=100, threshold=.001)

# It does not work without grid search
if args.cv:
    parameters = {'c_k': [2**i for i in range(-5, 5, 2)]}
    clf = GridSearchCV(classifier, parameters, scoring='f1_micro', cv=3, n_jobs=-1, verbose=3)
    clf.fit(X_train.todense(), Y_train.todense())
    f = open(args.dataset+'.parameters', 'w')
    f.write(str(clf.best_params_))
    f.close()
else:
    classifier.fit(X_train, Y_train)
    clf = classifier

# Prediction
preds = clf.predict(X_test)
np.savetxt(args.dataset+".testing.preds", preds, fmt='%d')
print("F1: %f" % metrics.f1_score(Y_test, preds, average='samples', zero_division=0))

sys.exit(1)
