#!	/usr/bin/python3

""""
Jesse Read, Concha Blieza, Pedro Larranaga
IEEE Transactions on Knowledge and Data Engineering
Multi-Dimensional Classification with Super-Classes
2013
"""

from skmultilearn.ext import download_meka
from skmultilearn.dataset import load_dataset
from sklearn.metrics import hamming_loss
from skmultilearn.ext import Meka
import argparse
import warnings
from skmultilearn.dataset import load_from_arff
import scipy
import numpy as np
import sys
import random

warnings.filterwarnings("ignore")

parser = argparse.ArgumentParser()
parser.add_argument("dataset", help="Dataset (must be a -.train and a -.test)", type=str)
parser.add_argument("nlabels", help="Number of labels", type=int)
parser.add_argument("method", help="Base method (as defined in 'typex.h'", type=int)

parser.add_argument("-sparse", help="File is in sparse format", action='store_true')
parser.add_argument("-cv", help="Cross-validate hyper-parameters", action='store_true')
args = parser.parse_args()

if args.cv:
    print("BCC ignores hyper-parameters cross-validation")

classifiers = [
    "weka.classifiers.functions.Logistic",
    "weka.classifiers.lazy.IBk",
    "weka.classifiers.trees.RandomForest",
    "weka.classifiers.trees.J48",
    "weka.classifiers.bayes.NaiveBayes",
    "weka.classifiers.functions.SMO",
    "weka.classifiers.functions.SMO",
    "weka.classifiers.meta.AdaBoost.M1",
    "weka.classifiers.functions.MultilayerPerceptron"
    ]

meka_classpath = download_meka()

# Load dataset
X_train, Y_train = load_from_arff(args.dataset+".training", label_count=args.nlabels, label_location='end', load_sparse=args.sparse)
X_test, Y_test = load_from_arff(args.dataset+".testing", label_count=args.nlabels, label_location='end', load_sparse=args.sparse)

# Test in train and test columns do not match. That could happen for sparse matrices
if X_train.shape[1] > X_test.shape[1]:
    toadd = scipy.sparse.lil_matrix((X_test.shape[0], X_train.shape[1] - X_test.shape[1]))
    X_test = scipy.sparse.hstack([X_test, toadd])
elif X_train.shape[1] < X_test.shape[1]:
    toadd = scipy.sparse.lil_matrix((X_train.shape[0], X_test.shape[1] - X_train.shape[1]))
    X_train = scipy.sparse.hstack([X_train, toadd])

meka = Meka(
        meka_classifier = "meka.classifiers.multitarget.SCC",
        weka_classifier = "meka.classifiers.multitarget.CC",
        meka_classpath = meka_classpath, # obtained via download_meka
        java_command = '/usr/bin/java', # path to java executable
        add_args = ['-I', "100", '-P', "2", '-N', "2"]
)

# If any class is empty add one label randomly to avoid computational problems
# cs = Y_train.sum(axis=0)
# for c in range(0, cs.shape[1]):
#     if cs[0, c] == 0:
#         Y_train[random.randint(0, Y_train.shape[0]-1), c] = 1
#         print("Add random label to class %d" % (c+1))

meka.fit(X_train, Y_train)

# Test error
preds = meka.predict(X_test).todense()
np.savetxt(args.dataset+".testing.preds", preds, fmt='%d')

sys.exit(1)
