/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <omp.h>
#include <stdbool.h>
#include <stdio.h>
#include <math.h>
#include "macros.h"
#include "common.h"
#include "ml.h"
#include "is.h"
#include "isfs.h"

// Modify labels of instance following 3 way partial instance selection
// 0, 1 Keep as it
// -1   Change relevant <-> irrelevant

void CheatLabels(int **originals, int *n_originals, int **labels, int *n_labels, int instances, int classes, int **sselection, bool reallocate, bool cheat_neg_to_pos_only)
{
    int *list;

    MALLOC(list, int, classes, "CheatLabels: list");

    for (int i = 0; i < instances; i++) {
        // Convert to binary
        ZeroVector(NULL, list, classes);
        for (int j = 0; j < n_originals[i]; j++)
            list[originals[i][j]] = 1;

        // Flip -1 selections
        for (int j = 0; j < classes; j++)
            if (sselection[j][i] == -1)
                list[j] = cheat_neg_to_pos_only ? 1 : !list[j];

        // It no labels, assign one of the originals randomly
        int count = 0;
        for (int j = 0; j < classes; j++)
            if (list[j])
                count++;

        // Reallocate memory to new length?
        if (reallocate)
            REALLOC(labels[i], int, MAX(count, 1), "CheatLabels: labels[i]");

        if (!count) {
            labels[i][0] = originals[i][(int) RANDOM(0, n_originals[i])];
            n_labels[i] = 1;
        } else {
            // Assign to new labels
            n_labels[i] = 0;
            for (int j = 0; j < classes; j++)
                if (list[j])
                    labels[i][n_labels[i]++] = j;
        }
    }

    FREE(list);
}

// Initialization of population
void ThreeWayIndividualInitialization(const t_Parameters *parm, int *individual, int length, bool undersampling, int instances, int **labels, int *n_labels, int classes)
{
    if (!undersampling)
        for (int j = 0; j < length; j++)
            individual[j] = (RANDOM(0.0, 1.0) < parm->chc_init_prob) ? (RANDOM(0.0, 1.0) < parm->three_way_minus1_prob ? -1 : 1) : 0;
    else {
        // First: Initialize with under-sampling
        for (int j = 0; j < classes; j++)
            MultiSelectionIndividulaWithRandomUndersampling(parm, individual+j*instances, labels, n_labels, j, parm->chc_init_prob, instances, false);

        // Second: Flip signs according to probability
        for (int i = 0; i < length; i++)
            if (individual[i] && RANDOM(0.0, 1.0) < parm->three_way_minus1_prob)
                individual[i] = -1;
    }
}
