/***************************************************************************
 *   Copyright (C) 2020 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "macros.h"
#include "maths.h"
#include "common.h"
#include "ml.h"

// Cross-validate hyper-parameters

void CrossValidateHyperParametersForMultiLabel(t_Parameters *parm, t_MLClassifier method, scalar **data, int instances, int inputs, int *n_labels, int **labels, int classes, int **index, int *s,
                                               int *subspace, int folds)
{
    // Number of parameters, or combinations of parameters, to test per method
    int n_parms[60] = {13, 0, 0, 21, 0, 0, 9, 5, 0, 0, 5 /*10*/, 5, 5, 5, 7, 7, 7, 0, 0, 0, 0 /*20*/, 13, 12, 5, 7, 7, 0, 0, 0, 0, 0 /*30*/, 30, 30, 0, 25, 13, 0, 13, 13, 13, 13 /*40*/, 13, 13, 13,
                       13, 13, 13, 0, 7, 7, 15 /*50*/, 5, 5, 10, 20, 7, 7, 7, 7, 7};

    // Parameters, or combinations of parameters, to test per method
    int ks[] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 25, 50, 100};
    int ps[] = {1, 2, 3, 4, 5};
    int kcs[][2] = {{1, 3}, {1, 5}, {1, 7}, {2, 3}, {2, 5}, {2, 7}, {3, 3}, {3, 5}, {3, 7}, {5, 3}, {5, 5}, {5, 7}, {10, 3}, {10, 5}, {10, 7}, {25, 3}, {25, 5}, {25, 7}, {50, 3}, {50, 5}, {50, 7}};
    int kks[][2] = {{1, 1}, {1, 3}, {1, 5}, {1, 7}, {1, 10}, {3, 1}, {3, 3}, {3, 5}, {3, 7}, {3, 10}, {5, 1}, {5, 3}, {5, 5}, {5, 7}, {5, 10}, {7, 1}, {7, 3}, {7, 5}, {7, 7}, {7, 10}, {10, 1},
                    {10, 3}, {10, 5}, {10, 7}, {10, 10}};
    scalar lcs[][2] = {{5, 0.5}, {5, 0.6}, {5, 0.7}, {5, 0.8}, {5, 0.9}, {5, 1.0}, {10, 0.5}, {10, 0.6}, {10, 0.7}, {10, 0.8}, {10, 0.9}, {10, 1.0}, {20, 0.5}, {20, 0.6}, {20, 0.7}, {20, 0.8},
                       {20, 0.9}, {20, 1.0}, {30, 0.5}, {30, 0.6}, {30, 0.7}, {30, 0.8}, {30, 0.9}, {30, 1.0}, {40, 0.5}, {40, 0.6}, {40, 0.7}, {40, 0.8}, {40, 0.9}, {40, 1.0}, {50, 0.5}, {50, 0.6},
                       {50, 0.7}, {50, 0.8}, {50, 0.9}, {50, 1.0}};
    scalar rakelos[][2] = {{3, 0.6}, {3, 1.0}, {3, 2.0}, {4, 0.6}, {4, 1.0}, {4, 2.0}, {5, 0.6}, {5, 1.0}, {5, 2.0}};
    int rakelds[] = {2, 3, 4, 5, 6};
    int dts[] = {0, 5, 10, 20, 50};
    scalar epss[] = {0.1, 0.2, 0.3, 0.4, 0.5};
    scalar thresholds[] = {0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7};
    int pss[][2] = {{1, 1}, {1, 2}, {1, 3}, {1, 4}, {1, 5}, {2, 1}, {2, 2}, {2, 3}, {2, 4}, {2, 5}, {3, 1}, {3, 2}, {3, 3}, {3, 4}, {3, 5}};
    int pptns[] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10};
    scalar pptcs[][2] = {{3, 0.1}, {3, 0.2}, {3, 0.3}, {3, 0.4}, {3, 0.5}, {5, 0.1}, {5, 0.2}, {5, 0.3}, {5, 0.4}, {5, 0.5}, {10, 0.1}, {10, 0.2}, {10, 0.3}, {10, 0.4}, {10, 0.5}, {15, 0.1},
                         {15, 0.2}, {15, 0.3}, {15, 0.4}, {15, 0.5}};
    scalar mlrfs[] = {0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7};
    scalar nodess[] = {0.1, 0.25, 0.5, 1.0, 2.0, 5.0, 10.0};
    int homers[] = {2, 3, 4, 5, 6};

    // Test whether there are parameters to be optimized
    extern char *ml_method[];
    switch (method) {
        case MLkNN:
        case PPT:
        case LAMLkNN:
        case SMLKNN:
        case IBLRML:
        case IBLRMLplus:
        case DMLKNN:
        case DWKNND:
        case DWKNNM:
        case DWKNNZ:
        case FVkNN:
        case FRONEC1H:
        case FRONEC2H:
        case FRONEC3H:
        case FRONEC1LD:
        case FRONEC2LD:
        case FRONEC3LD:
        case MLRS:
        case MLRSLC:
        case RAkEL_o:
        case RAkEL_d:
        case MLDT:
        case PCT:
        case EPS:
        case PRUNEDSET:
        case PPTN:
        case PPTC:
        case MLRF:
        case EPCT:
        case RFPCT:
        case EDT:
        case EXTRAPCT:
        case EXTRAMLDT:
        case MLBASICBP:
        case BPMLL:
        case HOMER_B:
        case HOMER_K:
        case HOMER_R:
        case BAGGING:
        case REALADABOOSTRE:
        case DISCRETEADABOOSTRE:
        case REALADABOOSTMH:
        case DISCRETEADABOOSTMH:
            SHOWPROGRESS(parm->show_progress > 0, "Cross-validating hyper-parameters for %s\n", ml_method[method]);
            break;

        case CC:
        case BR:
        case LP:
        case CLR:
        case NS:
        case NSSC:
        case TSVM:
        case TSCCM:
        case TSPCCM:
        case STA:
        case DBR:
        case LAMLksNN:
        case MLksNN:
        case ECC:
            SHOWPROGRESS(parm->show_progress > 0, "NO hyper-parameters for %s\n", ml_method[method]);
            return;
            break;

        case MLTREE:
        case MLFOREST:
        case OPCT:
            SHOWPROGRESS(parm->show_progress > 0, "Method %s TOO computationally expensive for hyper-parameters tuning\n", ml_method[method]);
            return;
            break;

        case MrKNN:
            SHOWPROGRESS(parm->show_progress > 0, "Hyper-parameters set in TRAINING function%c", '\n');
            return;
            break;

        default:
            OnErrorExit(ERR_UNSPECTED_RESULT, "I should not be here");
            break;
    }

    // If only subset reduce size temporarily
    if (parm->cv_sampling_size && parm->cv_sampling_size < instances) {
        SHOWPROGRESS(parm->show_progress > 0, "Cross-validating hyper-parameters with %d instances (%4.2lf%% of total)\n", instances, 100.0*parm->cv_sampling_size/((scalar) instances));
        SAVEPROGRESS(parm->show_progress > 0, "Cross-validating hyper-parameters with %d instances (%4.2lf%% of total)\n", instances, 100.0*parm->cv_sampling_size/((scalar) instances));
        instances = parm->cv_sampling_size;
    }

    int best_index = -1, v_instances = instances/folds;
    scalar best = -1.0;

#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int p = 0; p < n_parms[method]; p++) {
        t_Parameters l_parm = *parm;
        int *selection, *mask;

        MALLOC(mask, int, instances, "CrossValidateHyperParametersForMultiLabel: mask");
        ZeroVector(NULL, mask, instances);

        MALLOC(selection, int, instances, "CrossValidateHyperParametersForMultiLabel: selection");

        // No useless info
        l_parm.show_progress = -1;
        l_parm.time_all = false;

        // Assign parameters depending on method
        switch (method) {
            case MLkNN:
            case DMLKNN:
            case DWKNND:
            case DWKNNM:
            case DWKNNZ:
            case FVkNN:
            case FRONEC1H:
            case FRONEC2H:
            case FRONEC3H:
            case FRONEC1LD:
            case FRONEC2LD:
            case FRONEC3LD:
                l_parm.mlknn_k = ks[p];
                break;

            case IBLRMLplus:
            case IBLRML:
                l_parm.iblr_k = ks[p];
                break;

            case SMLKNN:
                l_parm.mlknn_k = kks[p][0];
                l_parm.smlknn_k = kks[p][1];
                break;

            case PPT:
                l_parm.ps_p = ps[p];
                break;

            case LAMLkNN:
                l_parm.mlknn_k = kcs[p][0];
                l_parm.lamlknn_n_clusters = kcs[p][1];
                break;

            case MLRS:
            case MLRSLC:
                l_parm.mlrs_k = (int) lcs[p][0];
                l_parm.mlrs_beta = lcs[p][1];
                break;

            case RAkEL_o:
                l_parm.rakel_k = (int) rakelos[p][0];
                l_parm.rakel_m = (int) (rakelos[p][1]*classes);
                break;

            case RAkEL_d:
                l_parm.rakel_k = rakelds[p];
                break;

            case MLDT:
            case PCT:
                l_parm.dt_max_depth = dts[p];
                break;

            case EPS:
                l_parm.eps_t = epss[p];
                break;

            case PRUNEDSET:
                l_parm.ps_b = pss[p][0];
                l_parm.ps_p = pss[p][1];
                break;

            case PPTN:
                l_parm.ps_p = pptns[p];
                break;

            case PPTC:
                l_parm.pptc_k = (int) pptcs[p][0];
                l_parm.pptc_t = pptcs[p][1];
                break;

            case MLRF:
            case EPCT:
            case RFPCT:
            case EDT:
            case EXTRAPCT:
            case EXTRAMLDT:
                l_parm.mlrf_t = mlrfs[p];
                break;

            case MLBASICBP:
                l_parm.nodes = nodess[p]*inputs;
                break;

            case BPMLL:
                l_parm.bpmll_nodes = nodess[p]*inputs;
                break;

            case HOMER_B:
            case HOMER_K:
            case HOMER_R:
                l_parm.homer_k = homers[p];
                break;

            case BAGGING:
            case REALADABOOSTRE:
            case DISCRETEADABOOSTRE:
            case REALADABOOSTMH:
            case DISCRETEADABOOSTMH:
                l_parm.ml_ensemble_t = thresholds[p];
                break;

            default:
                OnErrorExit(ERR_UNSPECIFIED, "I should not be here");
                break;
        }

        scalar fitness = 0.0;
        for (int fold = 0; fold < folds; fold++) {
            t_MLLearner learner;
            t_MLEnsemble ensemble;

            // Set fold mask
            FillVector(NULL, mask+(fold*v_instances), v_instances, 1.0);
            if (s)
                CopyIntegerVector(s, selection, instances);
            else
                FillVector(NULL, selection, instances, 1.0);
            FillVector(NULL, selection+(fold*v_instances), v_instances, 0.0);

            TrainMultiLabelClassifier((const t_Parameters *) &l_parm, method, data, instances, labels, n_labels, classes, inputs, selection, NULL, index, NULL, NULL, &learner, l_parm.k, ks, NULL,
                                      l_parm.lamlknn_n_clusters, subspace, &ensemble);

            // Evaluate fold
            scalar l_fitness = EvaluateFoldPerformance((const t_Parameters *) &l_parm, instances, labels, n_labels, classes, inputs, data, selection, NULL, index, mask, &learner, method, subspace,
                                                       &ensemble);
            fitness += l_fitness;

            SHOWPROGRESS(parm->show_progress > 1, "Evaluated %2d fold of %2d iteration with performance: %6.4lf, accumulated: %6.4lf\r", fold+1, p+1, l_fitness, fitness/(fold+1));

            FreeMLClassifier((const t_Parameters *) &l_parm, &learner, method, classes, inputs, instances, &ensemble);

            // Unset fold
            FillVector(NULL, mask+(fold*v_instances), v_instances, 0.0);
        }

        SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');

        fitness /= folds;

#ifdef _OPENMP
        #pragma omp critical
#endif
        if (fitness > best) {
            best = fitness;
            best_index = p;
        }

        FREE(selection);
        FREE(mask);
    }

    SHOWPROGRESS(parm->show_progress > 0, "BEST %d-th with performance: %lf\n", best_index+1, best);
    SAVEPROGRESS(parm->show_progress > 0, "BEST %d-th with performance: %lf\n", best_index+1, best);

    // Assign to permanent parameters depending on  method
    switch (method) {
        case MLkNN:
        case DMLKNN:
        case DWKNND:
        case DWKNNM:
        case DWKNNZ:
        case FVkNN:
        case FRONEC1H:
        case FRONEC2H:
        case FRONEC3H:
        case FRONEC1LD:
        case FRONEC2LD:
        case FRONEC3LD:
            parm->mlknn_k = ks[best_index];
            SAVEPROGRESS(true, "@ML-kNN/DML-kNN/DW-kNN/FV-kNN/FRONEC cross-validated k: %d\n", ks[best_index]);
            break;

        case SMLKNN:
            parm->mlknn_k = kks[best_index][0];
            parm->smlknn_k = kks[best_index][1];
            SAVEPROGRESS(true, "@SML-kNN cross-validated k: %d\n", kks[best_index][0]);
            SAVEPROGRESS(true, "@SML-kNN second layer cross-validated k: %d\n", kks[best_index][1]);
            break;

        case PPT:
            parm->ps_p = ps[best_index];
            SAVEPROGRESS(true, "@PPT cross-validated p: %d\n", ps[best_index]);
            break;

        case LAMLkNN:
            parm->mlknn_k = kcs[best_index][0];
            parm->lamlknn_n_clusters = kcs[best_index][1];
            SAVEPROGRESS(true, "@LAML-kNN cross-validated k: %d\n", kcs[best_index][0]);
            SAVEPROGRESS(true, "@LAML-kNN cross-validated number of clusters: %d\n", kcs[best_index][1]);
            break;

        case IBLRMLplus:
        case IBLRML:
            parm->iblr_k = ks[best_index];
            SAVEPROGRESS(true, "@IBRL-ML/IBRL-ML+ cross-validated k: %d\n", ks[best_index]);
            break;

        case MLRS:
        case MLRSLC:
            parm->mlrs_k = (int) lcs[best_index][0];
            parm->mlrs_beta = lcs[best_index][1];
            SAVEPROGRESS(true, "@MLRS/MLRS-LC cross-validated k: %d\n", (int) lcs[best_index][0]);
            SAVEPROGRESS(true, "@MLRS/MLRS-LC cross-validated beta: %lf\n", lcs[best_index][1]);
            break;

        case RAkEL_o:
            parm->rakel_k = (int) rakelos[best_index][0];
            parm->rakel_m = (int) (rakelos[best_index][1]*classes);
            SAVEPROGRESS(true, "@RAkELo cross-validated k: %d\n", (int) rakelos[best_index][0]);
            SAVEPROGRESS(true, "@RAkELo cross-validated m: %d\n", (int) (rakelos[best_index][1]*classes));
            break;

        case RAkEL_d:
            parm->rakel_k = rakelds[best_index];
            SAVEPROGRESS(true, "@RAkELd cross-validated k: %d\n", rakelds[best_index]);
            break;

        case MLDT:
        case PCT:
            parm->dt_max_depth = dts[best_index];
            SAVEPROGRESS(true, "@DT cross-validated maximum depth: %d\n", dts[best_index]);
            break;

        case EPS:
            parm->eps_t = epss[best_index];
            SAVEPROGRESS(true, "@EPS cross-validated threshold: %lf\n", epss[best_index]);
            break;

        case PRUNEDSET:
            parm->ps_b = pss[best_index][0];
            parm->ps_p = pss[best_index][1];
            SAVEPROGRESS(true, "@PS cross-validated b: %d\n", pss[best_index][0]);
            SAVEPROGRESS(true, "@PS cross-validated p: %d\n", pss[best_index][1]);
            break;

        case PPTN:
            parm->ps_p = pptns[best_index];
            SAVEPROGRESS(true, "@PPT-n cross-validated p: %d\n", pptns[best_index]);
            break;

        case PPTC:
            parm->pptc_k = (int) pptcs[best_index][0];
            parm->pptc_t = pptcs[best_index][1];
            SAVEPROGRESS(true, "@PPT-c cross-validated k: %d\n", (int) pptcs[best_index][0]);
            SAVEPROGRESS(true, "@PPT-c cross-validated t: %lf\n", pptcs[best_index][1]);
            break;

        case MLRF:
        case EPCT:
        case RFPCT:
        case EDT:
        case EXTRAPCT:
        case EXTRAMLDT:
            parm->mlrf_t = mlrfs[best_index];
            SAVEPROGRESS(true, "@ML-RF/E-PCT/RF-PCT/E-DT/Extra-PCT cross-validated t: %lf\n", mlrfs[best_index]);
            break;

        case MLBASICBP:
            parm->nodes = nodess[best_index]*inputs;
            SAVEPROGRESS(true, "@MLBASICBP cross-validated nodes: %d\n", (int) (nodess[best_index]*inputs));
            break;

        case BPMLL:
            parm->bpmll_nodes = nodess[best_index]*inputs;
            SAVEPROGRESS(true, "@BPMLL cross-validated nodes: %d\n", (int) (nodess[best_index]*inputs));
            break;

        case HOMER_B:
        case HOMER_K:
        case HOMER_R:
            parm->homer_k = homers[best_index];
            SAVEPROGRESS(true, "@%s cross-validated k: %d\n", ml_method[method], homers[best_index]);
            break;

        case BAGGING:
        case REALADABOOSTRE:
        case DISCRETEADABOOSTRE:
        case REALADABOOSTMH:
        case DISCRETEADABOOSTMH:
            parm->ml_ensemble_t = thresholds[best_index];
            SAVEPROGRESS(true, "@%s cross-validated threshold: %lf\n", ml_method[method], thresholds[best_index]);
            break;

        default:
            OnErrorExit(ERR_UNSPECIFIED, "I should not be here");
            break;
    }
}

// Evaluate the performance of a fold masked with a query mask
scalar EvaluateFoldPerformance(const t_Parameters *parm,  int instances, int **labels, int *n_labels, int classes, int inputs, scalar **data, int *s, cache_t **neighbors_cache, int **index,
                               int *query_mask, t_MLLearner *learner, t_MLClassifier method, int *subspace, t_MLEnsemble *ensemble)
{
    scalar *TP, *FP, *TN, *FN, fitness = -1.0, optimal;
    int *chain;
    bool classification, ranks;

    // Classify testing set
    scalar **f  = NULL;
    int **h = NULL, *n_h = NULL;

    AllocateScalarMatrix(&f, instances, classes);
    AllocateIntegerMatrix(&h, instances, classes);
    MALLOC(n_h, int, instances, "main: n_h");
    MALLOC(chain, int, *n_labels, "EvaluateFoldPerformance: chain");

    // Classifier must be already trained
    ClassifyDataset(parm, method, data, h, f, &classification, &ranks, data, instances, labels, n_labels, classes, n_h, inputs, index, index, instances, s, neighbors_cache, NULL, learner,
                    subspace, ensemble);

    // Data for evaluation
    int q_instances = 0, **q_labels, *q_n_labels, **q_h, *q_n_h;
    scalar **q_f;

    MALLOC(q_labels, int *, instances, "EvaluatePerformance: q_labels");
    MALLOC(q_n_labels, int, instances, "EvaluatePerformance: q_n_labels");
    MALLOC(q_h, int *, instances, "EvaluatePerformance: q_h");
    MALLOC(q_n_h, int, instances, "EvaluatePerformance: q_n_h");
    MALLOC(q_f, scalar *, instances, "EvaluatePerformance: q_f");
    for (int i = 0; i < instances; i++)
        if (query_mask[i]) {
            q_labels[q_instances] = labels[i];
            q_n_labels[q_instances] = n_labels[i];
            q_h[q_instances] = h[i];
            q_n_h[q_instances] = n_h[i];
            q_f[q_instances] = f[i];
            q_instances++;
        }

    // Get fitness
    fitness = parm->eb_subset_accuracy_w > ZERO ? parm->eb_subset_accuracy_w * SubsetAccuracy(q_labels, q_n_labels, q_h, q_n_h, q_instances) : 0.0;
    fitness += parm->eb_hamming_loss_w  > ZERO ? parm->eb_hamming_loss_w * (1.0 - HammingLoss(q_labels, q_n_labels, q_h, q_n_h, q_instances, classes)) : 0.0;
    fitness += parm->eb_jaccard_coeff_w > ZERO ? parm->eb_jaccard_coeff_w * JaccardCoefficient(q_labels, q_n_labels, q_h, q_n_h, q_instances) : 0.0;
    fitness += parm->eb_accuracy_w > ZERO ? parm->eb_accuracy_w * ExampleBasedAccuracy(q_labels, q_n_labels, q_h, q_n_h, q_instances) : 0.0;
    fitness += parm->eb_precision_w > ZERO ? parm->eb_precision_w * ExampleBasedPrecision(q_labels, q_n_labels, q_h, q_n_h, q_instances) : 0.0;
    fitness += parm->eb_recall_w > ZERO ? parm->eb_recall_w * ExampleBasedRecall(q_labels, q_n_labels, q_h, q_n_h, q_instances) : 0.0;
    fitness += parm->eb_f1_w > ZERO ? parm->eb_f1_w * ExampleBasedF1(q_labels, q_n_labels, q_h, q_n_h, q_instances, -1.0, -1.0) : 0.0;

    if (parm->lb_accuracy_macro_w > ZERO || parm->lb_precision_macro_w > ZERO || parm->lb_recall_macro_w > ZERO || parm->lb_f1_macro_w > ZERO || parm->lb_accuracy_micro_w > ZERO ||
        parm->lb_precision_micro_w > ZERO || parm->lb_recall_micro_w > ZERO || parm->lb_f1_micro_w > ZERO) {

        ALLOCA(TP, scalar, classes, "EvaluateFoldPerformance: TP");
        ALLOCA(FP, scalar, classes, "EvaluateFoldPerformance: FP");
        ALLOCA(TN, scalar, classes, "EvaluateFoldPerformance: TN");
        ALLOCA(FN, scalar, classes, "EvaluateFoldPerformance: FN");

        BinaryPerformancePerLabel(TP, FP, TN, FN, q_labels, q_n_labels, q_h, q_n_h, q_instances, classes);

        fitness += parm->lb_accuracy_macro_w > ZERO ? parm->lb_accuracy_macro_w * AccuracyMacro(TP, TN, classes, q_instances) : 0.0;
        fitness += parm->lb_precision_macro_w > ZERO ? parm->lb_precision_macro_w * PrecisionMacro(TP, FP, classes) : 0.0;
        fitness += parm->lb_recall_macro_w > ZERO ? parm->lb_recall_macro_w * RecallMacro(TP, FN, classes) : 0.0;
        fitness += parm->lb_f1_macro_w > ZERO ? parm->lb_f1_macro_w * F1Macro(TP, FN, FP, classes) : 0.0;
        fitness += parm->lb_accuracy_micro_w > ZERO ? parm->lb_accuracy_micro_w * AccuracyMicro(TP, TN, FP, FN, q_instances, classes) : 0.0;
        fitness += parm->lb_precision_micro_w > ZERO ? parm->lb_precision_micro_w * PrecisionMicro(TP, FP, classes) : 0.0;
        fitness += parm->lb_recall_micro_w > ZERO ? parm->lb_recall_micro_w * RecallMicro(TP, FN, classes) : 0.0;
        fitness += parm->lb_f1_micro_w > ZERO ?  parm->lb_f1_micro_w * F1Micro(TP, FN, FP, classes) : 0.0;
    }

    fitness += parm->eb_f1_alt_w > ZERO ? parm->eb_f1_alt_w * F1Instance(q_labels, q_n_labels, q_h, q_n_h, q_instances) : 0.0;
    fitness += parm->eb_one_error_w > ZERO ? parm->eb_one_error_w * (1.0 - OneError(q_labels, q_n_labels, q_f, q_instances, classes)) : 0.0;
    fitness += parm->eb_ranking_loss_w > ZERO ? parm->eb_ranking_loss_w * (1.0 - RankingLoss(q_labels, q_n_labels, q_f, q_instances, classes)) : 0.0;
    fitness += parm->eb_average_precision_w > ZERO ? parm->eb_average_precision_w * AveragePrecision(q_labels, q_n_labels, q_f, q_instances, classes) : 0.0;
    fitness += parm->lb_auc_macro_w > ZERO ? parm->lb_auc_macro_w * AUCMacro(q_labels, q_n_labels, q_f, q_instances, classes) : 0.0;
    fitness += parm->lb_auc_micro_w > ZERO ? parm->lb_auc_micro_w * AUCMicro(q_labels, q_n_labels, q_f, q_instances, classes) : 0.0;
    fitness += parm->eb_auc_w > ZERO ? parm->eb_auc_w * AUCInstance(q_labels, q_n_labels, q_f, q_instances, classes) : 0.0;

    if (parm->eb_coverage_w > ZERO) {
        scalar coverage = Coverage(q_labels, q_n_labels, q_f, q_instances, classes, &optimal);
        fitness += coverage > ZERO ? optimal/coverage : 0.0;
    }

    FreeScalarMatrix(&f, instances);
    FreeIntegerMatrix(&h, instances);
    FREE(n_h);
    FREE(q_labels);
    FREE(q_n_labels);
    FREE(q_h);
    FREE(q_f);
    FREE(q_n_h);
    FREE(chain);

    if (isnan(fitness) || isinf(fitness))
        OnErrorExit(ERR_NUMERICAL, "Fitness invalid value (NaN or infinite)");

    return fitness;
}

