/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Definitions and prototypes
#pragma once

#include <stdbool.h>
#include <stdio.h>
#include <time.h>
#include <mpi.h>
#include "typex.h"
#include "common.h"
#include "classifiers.h"
#include "is.h"

#define NEIGHBORS_CACHE_SIZE(instances) ((sizeof(cache_t) * (instances)*(instances)+sizeof(cache_t *) * (instances))/pow(2.0, 30.0))

typedef struct performance {
    // Classification performance: Example based
    scalar eb_subset_accuracy, eb_hamming_loss, eb_accuracy, eb_precision, eb_recall, eb_f1, eb_jaccard;

    // Classification performance: Label based
    scalar lb_accuracy_macro, lb_precision_macro, lb_recall_macro, lb_f1_macro, lb_accuracy_micro, lb_precision_micro, lb_recall_micro, lb_f1_micro, eb_f1_alt;

    // Ranking performance: Example based
    scalar eb_one_error, eb_coverage, eb_ranking_loss, eb_average_precision;

    // Ranking performance: Label based
    scalar lb_auc_macro, lb_auc_micro, eb_auc;

    // auROC and auPRC macro-averaged label based
    scalar auROC_macro, auPRC_macro;

} t_Performance;

typedef struct {
    int *N, *subset;
    int *pairs, **labels;
    scalar **w;
} t_Individual;

// Prototypes
/* chcis.c */
void kInfo(const t_Parameters *, int ***, int, scalar *);
int CHCInstanceSelectionForML(const t_Parameters *, int, scalar **, int **, int *, int, int, int *, int **, int **, int **, int *, t_MLClassifier, int **, int *, scalar *, scalar, int *);
bool RandomMLMutation(const t_Parameters *, int *, int, scalar, int, int, bool, int);
void MultiSelectionMutation(const t_Parameters *, int *, scalar **, int, int **, int *, int, cache_t **, t_MLClassifier, scalar *, int **, int, int, int **, t_Performance *, int *);
void BLXAlphaCrossover(int *, int *, int *, int *, int, int, int);
void RandomMutationForPermutations(int *, int);
void OnePointPermutationCrossover(int *, int *, int *, int *, int);
void GetPerformanceWithAllInstances(const t_Parameters *, t_MLClassifier, int *, int, int, scalar **, int **, int, int **,  int, cache_t **, t_Performance *, int *, int);
/* classify.c */
void ClassifyInstance(const t_Parameters *, t_MLClassifier, scalar *, int *, scalar *, bool *, bool *, scalar **, int, int **, int *, int, int *, int, int *, int **, int *, cache_t **, int,
                      int **, t_MLLearner *, int *, int **, int *, t_MLEnsemble *);
void ClassifyDataset(const t_Parameters *, t_MLClassifier, scalar **, int **, scalar **,  bool *, bool *, scalar **, int, int **, int *, int, int *, int, int **, int **, int, int *,
                     cache_t **, int **, t_MLLearner *, int *, t_MLEnsemble *);
void TestAllMultiLabelClassifiers(const t_Parameters *, scalar **, int **, int *, scalar **, int, int **, int *, int n_labels, int, int **, int **, int, int *, cache_t **, int **, char *, int,
                                  int *, int **, int *, int *);
void TrainMultiLabelClassifier(const t_Parameters *, t_MLClassifier, scalar **, int, int **, int *, int, int, int *, int *, int **, cache_t **, int **, t_MLLearner *, int, int *, int **, int,
                               int *, t_MLEnsemble *);
void FreeMLClassifier(const t_Parameters *, t_MLLearner *, t_MLClassifier, int, int, int, t_MLEnsemble *);
void SaveMLOutputs(char *, int, int, int **, int *, int **, int *);
/* cooperative.c */
scalar CooperativeCoevolutiveInstanceSelection(const t_Parameters *, scalar **, int **, int *, int, int, int, int **, int *, int **, t_MLClassifier);
void CCMutateIndividual(const t_Parameters *, int *, int, int, scalar, scalar *, int, int *);
scalar Delta(scalar, scalar, int);
void CCMemeticMutation(const t_Parameters *, int *, scalar *, int, scalar **, int, int **, int *, int **, scalar *, scalar *, int ***, cache_t **, int, int, scalar **, t_MLClassifier);
void CCMemeticMutationForSubpopulations(const t_Parameters *, int *, scalar *, int, scalar **, int, int **, int *, int **, scalar *, scalar *, int ***, cache_t **, int, int, int **, bool, scalar **,
                                        t_MLClassifier);
void kMemeticMutation(const t_Parameters *, int *, scalar **, int, int **, int *, int, cache_t **, t_MLClassifier, scalar *, int **, int, int, int **, t_Performance *, int *);
/* data.c */
void ReadMultiLabelData(const t_Parameters *, scalar ***, int ***, int **, int *, int *, int *, char *, int ***);
void ReadNextInstanceDenseFormat(char *, scalar **, int **, int *, int, int);
void ReadNextInstanceSparseFormat(char *, scalar **, int **, int *, int, int);
void ConvertToSparseFormat(scalar **, int **, int);
void RemoveInstancesWithNoLabelsOrNull(scalar ***, int ***, int **, int *, int ***);
void ConvertDatasetToSparseFormat(scalar **, int ***, int, int);
void ConvertToDenseFormat(scalar *, int *, int, scalar *);
void MLDatasetInfo(int, int **, int *, int, int **, int *, int, scalar *, scalar *, int *, scalar *, FILE *, scalar *, scalar *, scalar *, int *, int, scalar []);
void SaveSelectedDataset(scalar **, int **, int, int, int *, int, int **, int *, char *);
/* evaluation.c */
scalar EvaluateSelectionOneLabel(const t_Parameters *, scalar **, int, int **, int, t_Classifier, int *, cache_t **, int **, int, int *, int, int **, int, int *, t_PerformanceMeasure, int *, int);
scalar EvaluatekNNCache(int, int *, int *S, scalar **, int, int *, int, int, int, bool, scalar *r, scalar *, int **, cache_t **, t_PerformanceMeasure, int *, int);
scalar kNNSingleLabelCategoryAndMargin(scalar *, scalar **, int *, int, int, int, int *, int, int *, int *, cache_t **,  int, int **, int *, int *);
/* is-stub.c */
scalar InstanceSelectionML(const t_Parameters *, scalar **, int **, int *, int, int, int, int **, int **, int ***, int *, t_MLClassifier, int **, scalar **, int **, int *, int, int **, scalar *,
                         scalar, int **);
scalar InstanceSelectionAlgorithmML(const t_Parameters *, scalar **, int **, int *, int, int, int, int **, int **, int ***, int *, t_MLClassifier, int **, scalar **, int **, int *, int, int **,
                                    int *, scalar *, scalar, int **);
void ShowMLInstanceSelectionInfo(const t_Parameters *, int, int, int *, int **, scalar);
/* labels.c */
bool LabelInList(int, int *, int);
bool LabelInListUnsorted(int, int *, int);
/* metrics.c */
scalar SubsetAccuracy(int **, int *, int **, int *, int);
scalar HammingLoss(int **, int *, int **, int *, int, int);
scalar ExampleBasedAccuracy(int **, int *, int **, int *, int);
scalar ExampleBasedPrecision(int **, int *, int **, int *, int);
scalar ExampleBasedRecall(int **, int *, int **, int *, int);
scalar ExampleBasedF1(int **, int *, int **, int *, int, scalar, scalar);
void ExampleBased(int **, int *, int **, int *, int, scalar *, scalar *, scalar *, scalar *);
scalar AccuracyMacro(scalar *, scalar *, int, int);
scalar PrecisionMacro(scalar *, scalar *, int);
scalar RecallMacro(scalar *, scalar *, int);
scalar F1Macro(scalar *, scalar *, scalar *, int);
scalar F1Instance(int **, int *, int **, int *, int);
scalar AccuracyMicro(scalar *, scalar *, scalar *, scalar *, int, int);
scalar PrecisionMicro(scalar *, scalar *, int);
scalar RecallMicro(scalar *, scalar *, int);
scalar F1Micro(scalar *, scalar *, scalar *, int);
scalar OneError(int **, int *, scalar **, int, int);
scalar Coverage(int **, int *, scalar **, int, int, scalar *);
scalar RankingLoss(int **, int *, scalar **, int, int);
scalar AveragePrecision(int **, int *, scalar **, int, int);
scalar AUCMacro(int **, int *, scalar **, int, int);
scalar AUCInstance(int **, int *, scalar **, int, int);
scalar AUCMicro(int **, int *, scalar **, int, int);
scalar JaccardCoefficient(int **, int *, int **, int *, int);
/* metrics-class.c */
void PrintClassificationMetrics(t_Performance *, int **, int *, int **, int *, int, int, bool, FILE *, char *, bool);
scalar HammingDistanceListOfIntegers(int *, int *, int, int);
scalar IntersectionCardinality(int *, int *, int, int);
scalar UnionCardinality(int *, int *, int, int);
void BinaryPerformancePerLabel(scalar *, scalar *, scalar *, scalar *, int **, int *, int **, int *, int, int);
scalar Average(scalar *, int);
scalar VectorSum(scalar *, int *, int);
scalar JaccardDistanceListOfIntegers(int *, int *, int, int);
/* metrics-rank.c */
void PrintRankingMetrics(t_Performance *, int **, int *, scalar **, int, int, FILE *, char *, bool);
scalar Maximum(scalar *, int, int *);
scalar Minimum(scalar *, int, int *);
bool ValueInVector(int, int *, int);
int OrderOfValue(scalar, scalar *, int, int);
int ComplementVector(int *, int *, int);
/* mlknn.c */
void MultiLabelkNearestNeighbors(const t_Parameters *, scalar *, scalar **, int, int **, int *, int, int *, int *, int *, scalar *, int, int *, int **, cache_t **, int, t_MLkNN *, int **, int *,
                                 int *, int, int *);
void ComputePosteriorProbabilities(const t_Parameters *, scalar **, int, int **, int *, int k, int, int *S, int, int **, scalar **, scalar **, scalar, cache_t **, int **, int *, int *, int, int *);
void ComputePriorProbabilities(const t_Parameters *, scalar *, int, int *, int **, int *, int **, int);
void GetkNeighborsForML(scalar *, scalar **, int, int **, int, int *, int k, int *, int *, cache_t **, int, bool, int **, int, int *, int, int *);
void GetkNeighborsForMLWithSelectionPerLabel(scalar *, scalar **, int, int **, int, int *, int, int **, int, cache_t **, int, int **, int *, int *);
bool NullColumn(int **, int, int);
void TrainMultiLabelkNearestNeighbors(const t_Parameters *, scalar **, int, int **, int *, int, int *, int, int **, cache_t **, t_MLkNN *, int **, int, bool, int *);
void FreeMLkNN(t_MLkNN *, int);
void CreateMLkNNClassifier(t_MLkNN *, int, int);
/* newsubpop.c */
void GenerateNewSubpopulation(const t_Parameters *, int, cache_t **, scalar **, int, int, int *);
/* partition.c */
void DataPartitionML(const t_Parameters *, scalar **, int, int **, int *, scalar ****, int **, int ****, int ***, int **, int ****, int, int *, int ***, int);
/* population.c */
scalar EvaluateIndividualFitness(const t_Parameters *, int *, int ***, int, int, int **, int *, int, int, scalar **, int *, cache_t **, t_HashTable *, scalar *, int **, scalar *, int **, scalar **,
                                 t_MLClassifier, t_MLLearner *, t_Performance *, int, int *, int, int *, int, int *, t_MLEnsemble *, int);
scalar AverageNumberOfPredictedLabelsDifference(int *, int *, int/*, int*/);
/* subpopulation.c */
void PerformGeneration(const t_Parameters *, int **, int, int, int *, int, int **, scalar **, int **, int *, int, int, int, scalar *, int ***, cache_t **, scalar *, bool *, t_HashTable *,
                       t_HashTable *, scalar *, scalar *, int **, int *, int *, scalar *, t_MLClassifier);
scalar EvaluateSubindividualFitness(const t_Parameters *, int *, int, int **, int, scalar **, int **, int *, int, int, int, int, scalar *, int ***, cache_t **, t_HashTable *,
                                    t_HashTable *, scalar *, scalar *, int **, int *, scalar, scalar, scalar, scalar, scalar, scalar, scalar *, t_MLClassifier);
int CreateSelectionFromSubpopulations(const t_Parameters *, int *, int ***, int, int *, int, scalar, scalar **);
void InitializeSubpopulations(const t_Parameters *, int ****, int, bool **, int, int **, int *);
/* br.c */
void TrainBinaryRelevanceModels(const t_Parameters *, scalar **, int, int **, int *, int, int, t_BR *, int *, bool, t_Classifier, int *, int **, int **);
void BinaryRelevanceClassifyInstance(const t_Parameters *, scalar *, int *, scalar *,  bool *, bool *, scalar **, int, int, int *, int, int *, t_BR *, int *, int *, int **, int **);
void UndersampleDataWithLabel(const t_Parameters *, int *, int *, int);
void FreeBinaryRelevance(const t_Parameters *, t_BR *, int, int);
/* cc.c */
void TrainClassifierChainModels(const t_Parameters *, scalar **, int, int **, int *, int, int, t_CC *, int *, bool, t_Classifier, int *, int *, int **, int **, bool);
void ClassifierChainClassifyInstance(const t_Parameters *, scalar *, int *, scalar *,  bool *, bool *, scalar **, int, int, int *, int , int *, t_CC *, int *, int *, int **, int **, bool);
void FreeClassifierChain(const t_Parameters *, t_CC *, int, int);
void GetListOfTrainingPowersets(int, int **, int *, int ***, int *, int);
void GetClosestSetOfLabels(int *, int *, int **, int, int);
// int HammingDistanceForIntegers(int, int);
/* mpi-cooperative.c */
int MPICooperativeCoevolutiveInstanceSelection(const t_Parameters *, scalar **, int **, int *, int, int, int, int **, int *);
MPI_Comm *MPIRunSlaves(const t_Parameters *, int, int *, char *);
void MPIEvaluatePopulationFitness(const t_Parameters *, int **, int ***, int, int, scalar *, int *, MPI_Comm, int, scalar *, scalar *);
/* mpi-comm.c */
void SendAndReceiveGenerationInfo(MPI_Comm, const t_Parameters *, int, int, int **, int ***, int, scalar *, scalar *, int *, scalar *, int **);
void ReceiveData(MPI_Comm, int *, scalar ***, int ***, int **, int *, int ***, int, int *, int *, int *);
void SendData(MPI_Comm, int, scalar **, int **, int *, int, int **, int, int, int, int);
/* coop-best.c*/
scalar CooperativeCoevolutiveInstanceSelectionWithoutPopulation(const t_Parameters *, scalar **, int **, int *, int, int, int, int **, int *, int **, t_MLClassifier);
/* partial.c */
void PartialSelectionMutation(const t_Parameters *, int **, int **, int *, int, int, cache_t **, t_HashTable *, int **, int, scalar **, scalar *, t_MLClassifier, void *, int *, int, int *, int);
void PartialInstanceSelectionOneLabelOptimization(const t_Parameters *, int **, int, scalar **, int, int, int **, int *, int **, cache_t **, t_MLClassifier, void *, int *, int, int *, int);
/* kperlabel.c */
void EvolveKPerLabel(const t_Parameters *, int, scalar **, int **, int *, int, int, int **, int *, cache_t **);
void CrossValidateKPerLabel(const t_Parameters *, int, scalar **, int **, int *, int, int, int **, int *, int *, cache_t ***, scalar **, int);
void GetKPerLabel(const t_Parameters *, int, scalar **, int **, int *, int, int, int **, int *, int *);
void MaskLabels(int *, int **, int *, int **, int, int *, int);
void GreedyKPerLabel(const t_Parameters *, int, scalar **, int **, int *, int , int, int **, int *, int *, cache_t **, int);
void PerpendicularSearchPerLabel(const t_Parameters *, int, scalar **, int **, int *, int, int, int **, int *, int *, scalar **, cache_t **, int);
scalar RegularizationTerm(const t_Parameters *, int *, /*int,*/ int);
scalar EvaluateNFoldsCrossValidation(const t_Parameters *, int, int *, int *, scalar **, int, cache_t **, int **, int *, int **, int, int, t_MLLearner *, t_MLClassifier);
int RegularizeMLksNN(const t_Parameters *, int, scalar **, int **, int *, int, int, int **, int *, int *, cache_t **, int);
scalar MLksNNMemeticMutation(const t_Parameters *, int *, scalar **, int, int **, int *, int **, int *, int, int, cache_t **);
scalar MLksNNLocalSearchMutation(const t_Parameters *, int *, scalar **, int, int **, int *, int **, int *, int, int, cache_t **);
void UpdateAdaBoostMHMLWeights(const t_Parameters *, scalar **, scalar **, int, int *, int **, int **, int, int, int *, cache_t **, int **, t_MLClassifier, void *, int *);
void MLksNNSampling(const t_Parameters *, scalar **, int, int **, int *, int **, int *, scalar **, int *, int **, int *, int **, int *, scalar **, int, int, cache_t ***, int);
/* cv.c */
void CrossValidateHyperParametersForMultiLabel(t_Parameters *, t_MLClassifier, scalar **, int, int, int *, int **, int, int **, int *, int *, int);
scalar EvaluateFoldPerformance(const t_Parameters *,  int, int **, int *, int, int, scalar **, int *, cache_t **, int **, int *, t_MLLearner *, t_MLClassifier, int *, t_MLEnsemble *);
/* deks.c */
void DifferentialEvolutionMLksNN(const t_Parameters *, int, scalar **, int **, int *, int, int, int **, int *);
void DifferentialMutation(scalar *, scalar *, scalar *, scalar *, int, scalar, scalar, scalar);
void Recombination(scalar *, scalar *, scalar *, scalar, int, scalar, scalar);
/* window.c */
int BestWindowedValue(int, int, int, scalar *);
/* lamlknn.c */
void kMeansSparse(scalar **, int, int, int, int, int **, scalar **, int **, int *, int *);
void LocallyAdaptiveMultiLabelkNearestNeighbors(const t_Parameters *, scalar *, scalar **, int, int **, int *, int, int *, int *, int *, scalar *, int, int *, int **, cache_t **, int, int **,
                                                t_LAMLkNN *, int **);
void TrainLocallyAdaptiveMultiLabelkNearestNeighbors(const t_Parameters *, scalar **, int , int **, int *, int, int *, int, int **, int ** ,  t_LAMLkNN *, cache_t **, int, int, bool);
void FreeLAMLkNN(t_LAMLkNN *, int);
void TranslateCache(int *, cache_t **, cache_t **, int);
void CreateLAMLkNNClassifier(t_LAMLkNN *, int, int **, int, int, int, int);
void GetkNeighborsForMLWithPartition(scalar *, scalar **, int, int **, int, int *, int, int *, int *, cache_t **, int, bool, int **, int, int *, int, int *);
/* mlksnn.c */
void TrainMultiLabelKsNearestNeighbors(const t_Parameters * , scalar **, int, int **, int *, int, int *, int, int **, cache_t **, t_MLksNN *, int **, int *, bool);
void FreeMLksNN(t_MLksNN *, int);
void CreateMLksNNClassifier(const t_Parameters *, t_MLksNN *, int, int *, scalar **, int, int *, int **, int, int **);
/* performance.c */
scalar EvaluateClassifierPerformance(const t_Parameters *, t_MLClassifier, t_MLLearner *, int *, int *, cache_t **, int, int **, int *, int, scalar **, int **, int, int **, bool, int *, int,
                                     int *, int, int *, t_MLEnsemble *);
/* lamlksnn.c */
void CreateLAMLksNNClassifier(const t_Parameters *, t_LAMLksNN *, int, int **, scalar **, int, int *, int **, int, int **, int);
void FreeLAMLksNN(t_LAMLksNN *, int);
void TrainLocallyAdaptiveMultiLabelkNearestsNeighbors(const t_Parameters *, scalar **, int, int **, int *, int, int *, int, int **, int **, t_LAMLksNN *, cache_t **);
void LocallyAdaptiveMultiLabelkNearestsNeighbors(const t_Parameters *, scalar *, scalar **, int, int **, int *, int, int *, int *, int *, scalar *, int, int *, int **, cache_t **, int, int **,
                                                 t_LAMLksNN *);
void ReallocateCopyLAMLksNNClassifier(const t_Parameters *, t_LAMLksNN *, t_LAMLksNN *, int, int, int);
void EvolveAndTrainLAMLksNN(const t_Parameters *, int, scalar **, int **, int *, int, int, int **, cache_t **, t_LAMLksNN *, int *, bool);
/* is-lamlknn.c */
scalar LAMLkNNInstanceSelection(const t_Parameters *, int, scalar **, int **, int *, int, int, int **, int *, scalar **, int **, int *, int, int **, int *);
scalar EvaluateLAMLkNNFitness(const t_Parameters *, t_LAMLkNN *, int *, cache_t **, int, int **, int *, int, scalar **, int **, int);
/* sampling.c */
void MultiSelectionIndividulaWithRandomUndersampling(const t_Parameters *, int *, int **, int *, int, scalar, int, bool);
void MultiSelectionWithRoulette(const t_Parameters *, int *, scalar, int);
void MultiSelectionIndividualWithCardinality(const t_Parameters *, int *, int *, int, scalar, int );
/* rcga-ms.c */
void SimulatedBinaryCrossover(scalar *father, scalar *mother, scalar *son, scalar *daughter, int length, int nc, scalar, scalar);
int RCGAMultiSelectionForML(const t_Parameters *, int, scalar **, int **, int *, int, int, int *, int **, int *, t_MLClassifier, int **, int *);
void RCGANonUniformMutation(scalar *, int, int, int, double, int, int);
/* de-is.c */
scalar DifferentialEvolutionInstanceSelection(const t_Parameters *, int, scalar **, int **, int *, int, int, int **, int *, t_MLClassifier);
scalar Reduction(int *, scalar *, int);
/* ci-stub.c */
void ClassImbalanceMethodMultiLabel(const t_Parameters *, int *, scalar ***, int ***, int **, int, int, int ***, scalar *, scalar *, int *, scalar *, int *, int *, int ***);
void ClassImbalanceMethodMultiLabelAlgorithm(const t_Parameters *, int *, scalar ***, int ***, int **, int, int, int ***, scalar *, scalar *, int *, scalar *, int *, int *, int ***, int *);
/* mlsmote.c */
void AddNewInstance(scalar **, int **, int **, int *, int, int, int, int *, int, int, int);
void MultiLabelSmote(const t_Parameters *, int *, scalar ***, int ***, int **, int, int, int ***, scalar *, scalar *, scalar *, int *, int *);
/* classexternal.c */
void TestAllExternalMultiLabelClassifiers(const t_Parameters *, scalar **, int **, int *,  scalar **, int, int **, int *, int, int, int **, int **, int, int *, char *, int **, int *);
void ClassifyExternalDataset(const t_Parameters *, t_MLExternalClassifier, scalar **, int **, scalar **, scalar **, int, int **, int *, int, int *, int, int **, int **, int, int *, bool *, bool *,
                             char *, int **, int *, bool, int **, int *);
void SaveDatasetAsMLARFF(char *, int, scalar **, int, int **, int **, int *, int, int *, int **, int *);
/* 3way.c */
void CheatLabels(int **, int *, int **, int *, int, int, int **, bool, bool);
void ThreeWayIndividualInitialization(const t_Parameters *, int *, int, bool, int, int **, int *, int);
/* rakel.c */
void TrainRAkELModels(const t_Parameters *, scalar **, int, int **, int *, int, int, t_RAkEL *, int *, bool, t_Classifier, int *, int **, int **, bool);
void GetSingleLabelFromPowerset(int **, int *, int, int, int *, int *);
void DecodifyAndAddVotes(int, int, int *, int *);
void ClassifierRAkELInstance(const t_Parameters *, scalar *quey, int *, scalar *,  bool *, bool *, scalar **, int, int, int *, int, int *, t_RAkEL *, int *, int *, int **);
void FreeRAkEL(const t_Parameters *, t_RAkEL *, int);
void GeneratePowersetsForRakel(const t_Parameters *, t_RAkEL *, int, bool);
/* lp.c */
void DecodifyLabelPowerset(scalar, int, int *);
void TrainLabelPowersetModel(const t_Parameters *, scalar **, int, int **, int *, int, int, t_LP *, int *, bool, t_Classifier, int *, int **, int **, bool, bool, bool);
void LabelPowersetClassifyInstance(const t_Parameters *, scalar *, int *, scalar *, bool *, bool *, scalar **, int, int, int *, int, int *, t_LP *, int *, int *, int **);
void FreeLabelPowerset(const t_Parameters *, t_LP *, int);
scalar LabelsToPowerset(int *, int);
void GetSubCombinations(int *, int, int **, int *);
int CompareInfos(const void *, const void *);
void AddInstance(scalar ***, int, int ***, int **, int ***, int *, int, int, int, int **);
int PruneSubsets(int, int, int ***, int **, int, int **, scalar ***, int ***, t_PS_Strategy, int, int, bool, bool);
void CopyMultiLabelDataset(scalar **, int **, int *, int **, scalar **, int **, int *, int **, int, int);
/* clr.c */
void TrainCalibratedLabelRankingModels(const t_Parameters *, scalar **, int, int **, int *, int, int, t_CLR *, int *, bool, t_Classifier, int *, int **, int **, bool, bool);
void CalibratedLabelRankingClassifyInstance(const t_Parameters *, scalar *, int *, scalar *, bool *, bool *, scalar **, int, int, int *, int, int *, t_CLR *, int *, int **);
void FreeCalibratedLabelRanking(const t_Parameters *, t_CLR *, int, int);
int TriangularOrder(int, int);
/* homer.c */
void RecursiveHierarchy(struct t_Leaf *, int, const t_Parameters *, scalar **, int, int **, int *, int, int *, bool, t_Classifier, int *, int **, bool, bool);
void CreateHierarchy(struct t_Leaf **hierarchy, int, int, const t_Parameters *, scalar **, int, int **, int *, int, int *, bool, t_Classifier, int *, int **, bool, bool);
void TrainHOMERModels(const t_Parameters *, scalar **, int, int **, int *, int, int, t_HOMER *, int *, bool, t_Classifier, int *, int **, int **, bool, bool);
void TrainModelForLevel(const t_Parameters *, scalar **, int, int **, int *, int, struct t_Leaf *, int *, bool, t_Classifier, int *, int **);
void FreeHOMER(const t_Parameters *, t_HOMER *, int);
void RecursiveFreeHiearchy(const t_Parameters *, struct t_Leaf *, int, t_Classifier, int);
void RecursiveClassifierHOMERInstance(const t_Parameters *, scalar *, scalar **, int, int, int, int *, struct t_Leaf *, int *, int *, int **, scalar *, int *, t_Classifier);
void ClassifierHOMERInstance(const t_Parameters *, scalar *, int *, scalar *, bool *, bool *, scalar **, int, int, int *, int, int *, t_HOMER *, int *, int *, int **);
void IntegerToBinary(int, int *, int);
/* bkmeans.c */
scalar DistanceFromLabelToCluster(int, scalar *, int **, int);
void BalancedKMeans(int, int *, int, int **, int *, int, int, int **, int *, bool);
/* mldt.c */
void MLGetLabelsFrequency(int **, int *, int *, int, int, scalar *, int **);
void MLDecisionTreeBuild(const t_Parameters *, struct mldt_node *, scalar **, int **, int, int *, int, int *, int, int, int **, int *, bool, bool, int **, bool);
int MLDTCountNodes(t_MLDTree *);
void MLDTCountAllNodes(struct mldt_node *, int *);
void FreeMLDecisionTree(t_MLDTree *);
void MLDecisionTreeTrain(const t_Parameters *, scalar **, int, int **, int, t_MLDTree *, int *, int *, int, int **, int *, bool, bool, int **, bool);
scalar MLDTClassEntropy(int, scalar *, int, int *);
void MLDecisionTreeClassify(struct mldt_node *, scalar *, scalar *, int, int *, int *, int *, bool *, bool *, scalar *);
void FreeMLDecisionTreeNodes(struct mldt_node *);
void MLDTGetThresholdWeights(t_MLDTree *, scalar **, int, int *, int **, int *, int, int **);
/* mlsplit.h */
int CompareMLFeature(const void *, const void *);
void MLGetBestCupPoint(scalar **, int **, int, int *, int, scalar *, scalar *, int, int **, int *, bool, int **, bool);
scalar MLInformationGain(int, int, scalar, scalar *, scalar *, int, int *);
scalar ClassLabelVariance(int, int, int **, int *, int *);
scalar VarianceGain(int, scalar, int *, int *, int, int, int **, int *);
/* mlmlp.c */
void TrainMLBasicBP(const t_Parameters *, t_MLMLP *, scalar **, int, int, int, int **, int *, int *, int *);
void FreeMLBasicBP(t_MLMLP *, int);
void MLBasciBPClassifyInstance(t_MLMLP *, scalar *, scalar *, int *, int *, int, int, int *, bool *, bool *);
/* bpmll.c */
void BackPropForBPMLL(const t_Parameters *, t_BPMLL *, int, scalar **, int, scalar **, int, int, int *);
void FreeBPMLL(t_BPMLL *, int);
void BPMLLClassifyInstance(t_BPMLL *, scalar *, scalar *, int *, int *, int, int, int *, bool *, bool *);
void TrainBPMLL(const t_Parameters *, t_BPMLL *, scalar **, int, int, int, int **, int *, int *, int *);
void GetBPMLLPOutput(t_BPMLL *, int, int *, scalar *, int, scalar *, scalar *);
scalar GetOptimumThreshold(scalar *, int, int *, int);
void GetThresholdWeights(t_BPMLL *, scalar **, int, int **, int *, int, int, int *);
void RemoveRepeatedLinesWithVector(gsl_matrix *, gsl_vector *);
/* mlrf.c */
void MLRandomForestClassify(t_MLRandomForest *, scalar *, scalar *, int, int *, int *, int *, bool *, bool *, bool);
void MLRandomForestTrain(const t_Parameters *, scalar **, int, int **, int, t_MLRandomForest *, int *, int *, int, int **, int *, bool, bool, int **, bool);
void FreeMLRandomForest(t_MLRandomForest *);
int MLRFCountNodes(t_MLRandomForest *);
/* fs-stub.c */
scalar FeatureSelectionML(const t_Parameters *, scalar **, int **, int *, int, int, int, int **, int **, int **, t_MLClassifier, scalar **, int **, int *, int, int **, int **);
scalar FeatureSelectionAlgorithmML(const t_Parameters *, scalar **, int **, int *, int, int, int, int **, int **, int **, t_MLClassifier, scalar **, int **, int *, int, int **, int *, int **);
/* chcfs.c */
int CHCFeatureSelectionForML(const t_Parameters *, int, scalar **, int **, int *, int, int, int *, int **, t_MLClassifier, int *, int *);
/* dbr.c */
void TrainDependentBinaryRelevanceModels(const t_Parameters *, scalar **, int, int **, int *, int, int, t_DBR *, int *, bool, t_Classifier, int *, int **, int **);
void DependentBinaryRelevanceClassifyInstance(const t_Parameters *, scalar *, int *, scalar *,  bool *, bool *, int, int *, int, int *, t_DBR *, int *, int *, int **, int **);
void FreeDependentBinaryRelevance(const t_Parameters *, t_DBR *, int, int);
void CopySparseIndex(int **, int **, int);
void CopySparseData(int **, int **, scalar **, scalar **, int);
/* stacking.c */
void TrainMLStackingModels(const t_Parameters *, scalar **, int, int **, int *, int, int, t_DBR *, int *, bool, t_Classifier, int *, int **, int **);
void MLStackingClassifyInstance(const t_Parameters *, scalar *, int *, scalar *,  bool *, bool *, scalar **, int, int, int *, int, int *, t_DBR *, int *, int *, int **, int **);
void FreeMLStacking(const t_Parameters *, t_DBR *, int, int);
/* ecc.c */
void FreeECC(const t_Parameters *, t_ECC *, int, int);
void TrainECCModel(const t_Parameters *, scalar **, int, int **, int *, int, int, t_ECC *, int *, bool, t_Classifier, int *, int **, int **);
void ECCClassifyInstance(const t_Parameters *, scalar *, int *, scalar *,  bool *, bool *, scalar **, int, int, int *, int, int *, t_ECC *, int *, int *, int **, int **);
/* iblrml.c */
void FreeIBLRML(t_IBLR *, int);
void TrainIBLRMLModels(const t_Parameters *, scalar **, int, int **, int *, int, int, t_IBLR *, int *, int *, int **, int **, bool);
void IBLRMLClassifyInstance(scalar *, int *, scalar *,  bool *, bool *, scalar **, int, int, int *, int, int *, t_IBLR *, int *, int *, int **, int **, int **, int *, bool);
/* tsa.c */
void TwoStageVotingMethodClassifyInstance(const t_Parameters *, scalar *, int *, scalar *, bool *, bool *, scalar **, int, int, int *, int, int *, t_CLR *, int *, int **, bool, bool);
/* mltree.c */
int SplitTest(const t_Parameters *, struct MLTreeNode *, scalar **, int, int **, int *, int, int **, int *, int *, bool, int, int *, int **, scalar *, int *);
void MLTreeBuild(const t_Parameters *, struct MLTreeNode *, scalar **, int, int **, int *, int, int **, int *, int *, bool, int, int *, int);
void TrainMLTreeModels(const t_Parameters *, scalar **, int, int **, int *, int, int, t_MLTree *, int *, bool, t_Classifier, int *, int **, int **);
void RecursiveFreeMLTree(const t_Parameters *, struct MLTreeNode *, int, t_Classifier);
void FreeMLTree(const t_Parameters *, t_MLTree *, int);
void RecursivelyMLTreeClassifyInstance(const t_Parameters *, scalar *, scalar *,  bool *, bool *, scalar **, int, int, int *, int, int *, struct MLTreeNode *, int *, int *, int **, int **, t_Classifier);
void MLTreeClassifyInstance(const t_Parameters *, scalar *, int *, scalar *,  bool *, bool *, scalar **, int, int, int *, int, int *, t_MLTree *, int *, int *, int **, int **);
void MLTreeCountAllNodes(struct MLTreeNode *, int *);
int MLTreeCountNodes(t_MLTree *);
/* mlforest.c */
int MLForestCountNodes(t_MLForest *);
void MLForestClassify(const t_Parameters *, t_MLForest *, scalar *, scalar *, int, int *, int *, int *, bool *, bool *, scalar **, int, int, int *, int **);
void MLForestTrain(const t_Parameters *, scalar **, int, int **, int, t_MLForest *, int *, int *, int, int **, int *);
void FreeMLForest(const t_Parameters *, t_MLForest *, int);
scalar MLCut(scalar *, int);
/* mlrs.c */
void TrainMLBasedOnNeighborhoodRoughSets(const t_Parameters *, t_MLRS *, int **, int *, int, int, int *, scalar **, int **, int *, int);
void MLRSClassify(t_MLRS *, scalar *, scalar *, int, int **, int *, int *, bool *, bool *, scalar **, int, int, int *, int *, int *, int **, int *);
void MLRSFree(t_MLRS *, int);
/* mlrslc.c */
void MLRSLCClassify(t_MLRSLC *, scalar *, scalar *, int, int **, int *, int *, bool *, bool *, scalar **, int, int, int *, int *, int *, int **, int *);
void TrainMLBasedOnNeighborhoodRoughSetsLocalCorrelation(const t_Parameters *, t_MLRSLC *, int **, int *, int , int, int *, scalar **, int **, int *, int);
void MLRSLCFree(t_MLRSLC *, int);
/* omldt.c */
void FreeOptionMLDecisionTree(t_OMLDTree *);
void FreeOptionMLDecisionTreeNodes(struct omldt_node *);
int OptionMLDTCountNodes(t_OMLDTree *);
void OptionMLDTCountAllNodes(struct omldt_node *, int *);
void OptionMLDecisionTreeBuild(const t_Parameters *, struct omldt_node *, scalar **, int **, int, int *, int, int *, int, int, int **, int *, bool, bool);
int CompareSplitInfo(const void *, const void *);
void OptionMLDecisionTreeClassify(struct omldt_node *, scalar *, scalar *, int, int *, int *, int *, bool *, bool *);
void OptionMLDecisionTreeTrain(const t_Parameters *, scalar **, int, int **, int, t_OMLDTree *, int *, int *, int, int **, int *, bool, bool);
/* smlknn.c */
void TrainStackedMLkNN(const t_Parameters *, scalar **, int, int **, int *, int, int *, int, int **, cache_t **, t_StackedMLkNN *, int **, int, bool, int *);
void MultiLabelStackedkNearestNeighbors(const t_Parameters *, scalar *, scalar **, int,  int **, int *, int, int *, int *, int *, scalar *, int, int *, int **, cache_t **, int, t_StackedMLkNN *,
                                        int **, int *);
void FreeStackedMLkNN(t_StackedMLkNN *, int, int);
/* dmlknn.c */
void CreateDMLkNNClassifier(t_DMLkNN *, int, int, int);
void DMultiLabelkNearestNeighbors(const t_Parameters *, scalar *, scalar **, int, int **, int *, int, int *, int *, int *, scalar *, int, int *, int **, cache_t **, t_DMLkNN *, int **, int *);
void TrainDMultiLabelkNearestNeighbors(const t_Parameters *, scalar **, int, int **, int *, int, int *, int, int **, cache_t **, t_DMLkNN *, int **, int, bool, int *);
void FreeDMLkNN(t_DMLkNN *, int);
/* mrknn.c */
void MultiLabelFuzzyCMeans(scalar **, int, int **, int *, int, int, scalar **, int *, int, scalar, int *, int **);
void TrainMrKNN(const t_Parameters *, t_MrKNN *, scalar **, int, int **, int *, int **, int, int *, int *, int, cache_t **);
void MrKNNClassify(t_MrKNN *, scalar *, int *, scalar **, int, int **, int, int *, int *, int, cache_t **, scalar *,  int *, int *);
void FreeMrKNN(t_MrKNN *, int);
/* bagging.c */
void TrainMultiLabelBagging(const t_Parameters *, t_MLEnsemble *, scalar **, int, int **, int *, int, int, int *, int **, int **, int *, int, t_MLClassifier, bool);
void BaggingClassifyInstance(const t_Parameters *, t_MLEnsemble *, scalar *, int *, scalar *, bool *, bool *, scalar **, int, int **, int *, int, int *, int, int *, int **, int *, int **,  int *,
                             int **, int *);
void FreeBaggingEnsemble(const t_Parameters *, t_MLEnsemble *, int, int, int);
/* adaboostre.c */
void NormalizeMatrix(scalar **, int, int);
void ResampleForML(scalar **, int *, int, int, t_MLResampling);
void TrainAdaBoostByResampling(const t_Parameters *, t_MLEnsemble *, scalar **, int, int **, int *, int, int, int *, int **, int **, int *, bool);
void AdaBoostREUpdateWeights(const t_Parameters *, int, t_MLLearner *, scalar **, int **, int *, int, int, int **, scalar **, int *, int *, int **, bool, t_MLClassifier, scalar);
/* adaboostmh.c */
void AdaBoostMHUpdateWeights(const t_Parameters *, int, t_Learner *, scalar **, int **, int *, int, int, int **, scalar **, int *, int *, int **, bool, t_Classifier, scalar);
void TrainAdaBoostMH(const t_Parameters *, t_MLEnsemble *, scalar **, int, int **, int *, int, int, int *, int **, int **, int *, bool, bool);
void FreeAdaBoostMHEnsemble(const t_Parameters *, t_MLEnsemble *, int, int);
void AdaBoostMHClassifyInstance(const t_Parameters *, scalar *, int *, t_MLEnsemble *, scalar **, int, int, int **, int *, int *, int **, int *, scalar *, int *, int, bool *, bool *);
/* dwknn.c */
scalar Dudani(scalar, scalar, scalar);
scalar Macleod(scalar, scalar, scalar);
scalar Zavrel(scalar);
void DWkNNClassify(t_MLClassifier, t_DWkNN *, scalar *, scalar **, int **, int, int *, cache_t **, int **, int *, int **, int *, int, int *, int *, int *, scalar *, int);
void TrainDWkNN(const t_Parameters *, t_DWkNN *);
/* fronec.c */
scalar R(scalar *, scalar *, int, scalar *);
scalar RdHammingBased(int *, int, int *, int, int);
void TrainFRONEC(const t_Parameters *, t_FRONEC *, int, int **, int *, int, int, scalar **);
void FRONECClassifyInstance(t_FRONEC *, scalar *, scalar **, int, int **, int, int *, int *, cache_t **, int **, int, int *, int **, int *, t_MLClassifier, scalar *, int *, int *);
scalar RdLabelDistributionBased(int *, int, int *, int, int, scalar *);
void FreeFRONEC(t_FRONEC *);
scalar I(scalar, scalar);
scalar T(scalar, scalar);
int CompareDoubleReverse(const void *, const void *);
/* fvknn.c */
void TrainFVkNN(t_FVkNN *, const t_Parameters *, scalar **, int, int **, int *, int, cache_t **, int *, int **, int *, int **, int);
void FVkNNClassify(t_FVkNN *, scalar *, scalar **, int, int **, int *, int, int *, int **, cache_t **, int, int *, scalar *, int *, int *);
void FreeFVkNN(t_FVkNN *, int);
/* mleso.c */
void InitializeMLPopulation(const t_Parameters *, t_Individual *, int, int, int, int *, int **, scalar **, int, int, int, int, int **, int);
int AddMLInstances(const t_Parameters *, t_Individual *, scalar **, int, int **, int *, int, int **, int, int, int **, scalar *, scalar);
int RemoveUnselectedMLInstances(scalar ***, int *, int ***, int **, int ***, int *);
void MultiLabelEvolutionarySyntheticOversampling(const t_Parameters *, scalar ***, int *, int ***, int, int, int ***, int *, int *, int *, int **, scalar *, scalar);
void GetkNeighborsLOOForML(const t_Parameters *, int, scalar **, int, int **, int, int, int *, int *);
void MLESOCrossover(t_Individual *, t_Individual *, t_Individual *, t_Individual *, int, int, int);
void CacheOfNeghborsFromDistances(int, scalar **, cache_t **);
void CombineSparseVector(scalar *, int *, scalar *, int, scalar *, int *, scalar *, int *, bool);
/* mlrus.c */
int CalculateMinorityInstances(int, bool *, int **, int *, scalar *, scalar, int);
void MLRandomUndersampling(const t_Parameters *, int *, scalar ***, int ***, int **, int, int, int ***, scalar *, scalar *, scalar);
/* mlros.c */
void MLRandomOversampling(const t_Parameters *, int *, scalar ***, int ***, int **, int, int, int ***, scalar *, scalar *, scalar, int *);
/* ci.c */
void PartialRandomUndersampling(const t_Parameters *, int **, int *, int, int, int **, scalar, bool);
void PartialRandomOversampling(const t_Parameters *, int ***, int **, int, int *, int ***, scalar, scalar ***, int ***, int);
/* pevus.c */
scalar UndersamplingFitness(int **, int, int, int **, int *, scalar);
void PartialEvolutionaryUndersampling(const t_Parameters *, int ***, int **, int, int *, int **, scalar ***, int ***);
void GetMinorityTemplate(int **, int *, int, int, int **);
/* pptc.c */
int IndexOfMaximumValue(scalar *, int);
void PPTcClassifyInstance(const t_Parameters *, scalar *, int *, scalar *, bool *, bool *, scalar **, int, int, int *, int, int *, t_LP *, int *, int *, int **);
/* repop.c */
void PopulateLabels(int **, int *, int **, int *, int *, int, int);
scalar EvaluateRepopulatedLabels(const t_Parameters *, int *, int, int **, int *, int, int, scalar **, cache_t **, int **, t_MLClassifier, t_MLLearner *, t_Performance *, int, int *, int, int *, int,
                                 int *, t_MLEnsemble *, scalar *);
void EvolutionaryLabelRepopulation(const t_Parameters *, scalar **, int , int **, int,  int, int **, int *, int *,int *, int *);
scalar LabelRepopulationRatio(int *, int *, int);
