/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
#include <string.h>
#include <unistd.h>
#include "ml.h"
#include "macros.h"
#include "is.h"
#include "isfs.h"

/*
 * Cooperative co-evolution for multi-label instance selection
 */

#define BLX_ALPHA 0.5

int MPICooperativeCoevolutiveInstanceSelection(const t_Parameters *parm, scalar **data, int **labels, int *n_labels, int instances, int inputs, int classes, int **index, int *selection)
{
    //     MPI_Status status;
    //     int ***subpops, **population, neighbors_cache_size = 0, best_k = -1;
    //     scalar *fitness, *cummulative, best_fitness = -1.0, best_qs = -1.0, *performance;
    //     cache_t **neighbors_cache = NULL;
    //
    //     // Check parameters
    //     if (parm->ml_onesubpop_per_cycle)
    //         OnErrorExit(ERR_PARAM, "One subpopulation per cycle incompatible with MPI");
    //     if (parm->ml_use_hash_table)
    //         OnErrorExit(ERR_PARAM, "Has table incompatible with MPI");
    //
    //     // As many subpopulations as labels
    //     int n_subpops = classes;
    //
    //     if (parm->ml_ins_sel_per_label)
    //         OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "MPI version for CC selection for posterior probabilities");
    //
    //
    //     // For evaluations of individuals
    //     t_MLkNN mlknn;
    //     t_MLksNN mlksnn;
    //     t_LAMLkNN lamlknn;
    //     void *classifier[N_ML_CLASSIFIERS] = {(void *) &mlknn, NULL, NULL, (void *) &lamlknn, (void *) &mlksnn, NULL};
    //
    //     CreateMultiLabelClassifier(parm, classifier[method], method, classes, instances, index, inputs, parm->ml_max_k, parm->lamlknn_max_n_clusters, NULL, data, n_labels, labels, NULL);
    //
    //     // Create and initialize global population
    //     MALLOC(fitness, scalar, parm->population_size, "MPICooperativeCoevolutiveInstanceSelection: fitness");
    //     MALLOC(performance, scalar, parm->population_size, "MPICooperativeCoevolutiveInstanceSelection: performance");
    //     MALLOC(cummulative, scalar, parm->population_size, "MPICooperativeCoevolutiveInstanceSelection: cummulative");
    //     MALLOC(population, int *, parm->population_size, "MPICooperativeCoevolutiveInstanceSelection: population");
    //     for (int i = 0; i < parm->population_size; i++) {
    //         MALLOC(population[i], int, n_subpops, "MPICooperativeCoevolutiveInstanceSelection: population[i]");
    //         // Allow negative values to ignore the population?
    //         if (parm->ml_allow_negative_subpops)
    //             for (int j = 0; j < n_subpops; j++)
    //                 population[i][j] = (int) RANDOM(-parm->subpop_size, parm->subpop_size);
    //         else
    //             for (int j = 0; j < n_subpops; j++)
    //                 population[i][j] = (RANDOM(0.0, 1.0) < parm->chc_pop_init_prob) ? (int) RANDOM(0, parm->subpop_size) : -1;  // Allow -1 to not consider a subpopulation
    //     }
    //
    //     MALLOC(subpops, int **, n_subpops, "MPICooperativeCoevolutiveInstanceSelection: subpops");
    //     for (int i = 0; i < n_subpops; i++) {
    //         MALLOC(subpops[i], int *, parm->subpop_size, "MPICooperativeCoevolutiveInstanceSelection: subpops[i]");  // Space for duplication of population during evolution
    //         for (int j = 0; j < parm->subpop_size; j++) {
    //             MALLOC(subpops[i][j], int, instances, "MPICooperativeCoevolutiveInstanceSelection: subpops[i][j]");
    //             for (int k = 0; k < instances; k++)
    //                 subpops[i][j][k] = RANDOM(0.0, 1.0) < parm->chc_subpop_init_prob ? 1 : 0;
    //         }
    //     }
    //
    //     // Cache of neighbors?
    //     SHOWPROGRESS(parm->show_progress > 1, "BEGIN Cache construction finished%c", '\n');
    //     if (parm->use_neighbors_cache && NEIGHBORS_CACHE_SIZE(instances) < MAXIMUM_CACHE_SIZE) {
    //         ConstructAllNeighborsCache(data, instances, inputs, NULL, &neighbors_cache, NULL, NULL, &neighbors_cache_size, NULL, parm->ml_loo);
    //         SHOWPROGRESS(parm->show_progress > 1, "DONE Cache construction finished%c", '\n');
    //     } else
    //         SHOWPROGRESS(parm->show_progress > 0, "Neighbors cache DISABLED%c", '\n');
    //
    //     // For CHC subpopulation individuals
    //     int *minimum_distance;
    //     MALLOC(minimum_distance, int, n_subpops, "MPICooperativeCoevolutiveInstanceSelection: minimum_distance");
    //     for (int i = 0; i < n_subpops; i++)
    //         minimum_distance[i] = instances/parm->chc_subpop_minimum_distance_factor;
    //
    //     // Difference for sub-population selection
    //     scalar *difference, *cprobability;
    //     MALLOC(difference, scalar, n_subpops, "MPICooperativeCoevolutiveInstanceSelection: difference");
    //     MALLOC(cprobability, scalar, n_subpops, "MPICooperativeCoevolutiveInstanceSelection: cprobability");
    //     for (int i = 0; i < n_subpops; i++)
    //         difference[i] = 1.0;
    //
    //     // Evolution of quorum sensing threshold if apply
    //     scalar *qs = NULL;
    //     MALLOC(qs, scalar, parm->population_size, "MPICooperativeCoevolutiveInstanceSelection: qs");
    //     for (int i = 0; i < parm->population_size; i++)
    //         qs[i] = RANDOM(parm->quorum_sensing_threshold_min, parm->quorum_sensing_threshold_max);
    //
    //     // Evolution of k
    //     int *k = NULL;
    //     MALLOC(k, int, parm->population_size, "MPICooperativeCoevolutiveInstanceSelection: k");
    //     for (int i = 0; i < parm->population_size; i++)
    //         k[i] = (int) RANDOM(parm->ml_min_k, parm->ml_max_k+1);
    //
    //     // For protection of best individuals
    //     int *elite, **elite_members;
    //     int *elite_list;
    //     MALLOC(elite, int, parm->population_size, "MPICooperativeCoevolutiveInstanceSelection: elite");
    //     MALLOC(elite_members, int *, n_subpops, "MPICooperativeCoevolutiveInstanceSelection: elite_members");
    //     MALLOC(elite_list, int, parm->population_size, "MPICooperativeCoevolutiveInstanceSelection: elite_list");
    //     for (int i = 0; i < n_subpops; i++)
    //         MALLOC(elite_members[i], int, parm->subpop_size, "MPICooperativeCoevolutiveInstanceSelection: elite_members[i]");
    //
    //     // Star slaves
    //     int argc = 0, s_root, f_root, sparse;
    //     char **argv = NULL;
    //     MPI_Init(&argc, &argv);
    //
    //     sparse = index ? 1 : 0;
    //
    //     // Subpopulation evolution
    //     MPI_Comm *s_comm = MPIRunSlaves(parm, n_subpops, &s_root, "/home/frodo/Run/bin/mpi-subpopulation");
    //     SendData(*s_comm, instances, data, labels, n_labels, inputs, index, s_root, n_subpops, classes, sparse);
    //
    //     // Population fitness evaluation
    //     MPI_Comm *f_comm = MPIRunSlaves(parm, parm->population_size, &f_root, "/home/frodo/Run/bin/mpi-evaluation");
    //     SendData(*f_comm, instances, data, labels, n_labels, inputs, index, f_root, n_subpops, classes, sparse);
    //
    //     // Main loop
    //     for (int g = 0; g < parm->generations; g++) {
    //         SHOWPROGRESS(parm->show_progress > 1, "Generation: %*d/%d\n", WIDTH(parm->generations), g+1, parm->generations);
    //
    //         // Evaluate the fitness of the current population
    //         MPIEvaluatePopulationFitness(parm, population, subpops, n_subpops, instances, qs, k, *f_comm, f_root, fitness, performance);
    //
    //         // Np loops of population
    //         // Steady state genetic algorithm
    //         for (int np = 0; np < parm->np; np++) {
    //             SHOWPROGRESS(parm->show_progress > 1, "Np generation: %*d/%d\r", WIDTH(parm->np), np+1, parm->np);
    //
    //             cummulative[0] = fitness[0];
    //             for (int k = 1; k < parm->population_size; k++)
    //                 cummulative[k] = cummulative[k-1]+fitness[k];
    //
    //             // Select worsts individuals
    //             int worst = OrderedSelection(fitness, parm->population_size - 1, parm->population_size);
    //             int sworst = OrderedSelection(fitness, parm->population_size - 2, parm->population_size);
    //
    //             // Select two parents
    //             int father, mother;
    //             do {
    //                 father = RouletteSelection(cummulative, parm->population_size);
    //                 mother = RouletteSelection(cummulative, parm->population_size);
    //             } while (mother == worst || father == worst || mother == sworst || father == sworst);
    //
    //             // Perform crossover
    //             TwoPointCrossover(population[worst], population[sworst], population[father], population[mother], n_subpops);
    //
    //             // Perform BLX-alpha crossover for quorum sensing threshold
    //             scalar minimum = MIN(qs[father], qs[mother]);
    //             scalar maximum = MAX(qs[father], qs[mother]);
    //             scalar I = maximum - minimum;
    //             qs[worst] = AssignWithinBoundsScalar(parm->quorum_sensing_threshold_min, RANDOM(minimum - I * BLX_ALPHA, maximum + I * BLX_ALPHA), parm->quorum_sensing_threshold_max);
    //             qs[sworst] = AssignWithinBoundsScalar(parm->quorum_sensing_threshold_min, RANDOM(minimum - I * BLX_ALPHA, maximum + I * BLX_ALPHA), parm->quorum_sensing_threshold_max);
    //
    //             // Perform BLX-alpha crossover for k
    //             int min = MIN(k[father], k[mother]);
    //             int max = MAX(k[father], k[mother]);
    //             I = max - min;
    //             k[worst] = (int) AssignWithinBoundsScalar(parm->ml_min_k, RANDOM(min - I * BLX_ALPHA, max + I * BLX_ALPHA), parm->ml_max_k);
    //             k[sworst] = (int) AssignWithinBoundsScalar(parm->ml_min_k, RANDOM(min - I * BLX_ALPHA, max + I * BLX_ALPHA), parm->ml_max_k);
    //
    //             // Evaluate new individuals
    //             fitness[worst] = EvaluateIndividualFitness(parm, population[worst], subpops, n_subpops, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, NULL, &qs[worst],
    //                              k[worst], index, &performance[worst], NULL, NULL, NULL, MLkNN);
    //             fitness[sworst] = EvaluateIndividualFitness(parm, population[sworst], subpops, n_subpops, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, NULL, &qs[sworst],
    //                               k[sworst], index, &performance[sworst], NULL, NULL, NULL, MLkNN);
    //
    //             // Protect best ones from mutation
    //             for (int i = 0; i < parm->population_size; i++)
    //                 elite[i] = 0;
    //             for (int i = 0; i < n_subpops; i++)
    //                 for (int j = 0; j < parm->subpop_size; j++)
    //                     elite_members[i][j] = 0;
    //             NBestOrderedSelection(fitness, parm->population_size, parm->population_size*parm->ml_elitism, elite_list);
    //             for (int i = 0; i < floor(parm->population_size*parm->ml_elitism); i++) {
    //                 elite[elite_list[i]] = 1;
    //                 for (int j = 0; j < n_subpops; j++)
    //                     if (population[elite_list[i]][j] >= 0)
    //                         elite_members[j][population[elite_list[i]][j]] = 1;
    //             }
    //
    //             // Mutation
    //             for (int j = 0; j < parm->population_size; j++)
    //                 if (!elite[j] && RANDOM(0.0, 1.0) < parm->ssga_mutation) {
    //                     CCMutateIndividual(parm, population[j], n_subpops, parm->subpop_size, parm->chc_pop_init_prob, &qs[j], g, &k[j]);
    //                     fitness[j] = EvaluateIndividualFitness(parm, population[j], subpops, n_subpops, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, NULL, &qs[j], k[j],
    //                                                            index, &performance[j], NULL, NULL, NULL, MLkNN);
    //                 }
    //         }
    //
    //         SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');
    //
    //         // Send information to subpopulation slaves
    //         // Slaves perform generation(s) and return evolved subpopulations
    //         SendAndReceiveGenerationInfo(*s_comm, parm, s_root, instances, population, subpops, n_subpops, fitness, difference, k, qs, elite_members);
    //
    //         // Receive evolved subpopulation
    //         for (int i = 0; i < n_subpops; i++) {
    //             if (i != s_root) {
    //                 int p_index;
    //                 MPI_Recv(&p_index, 1, MPI_INT, i, 0, *s_comm, &status);
    //                 for (int j = 0; j < parm->subpop_size; j++)
    //                     MPI_Recv(subpops[p_index][j], instances, MPI_INT, i, 0, *s_comm, &status);
    //             }
    //         }
    //
    //         // Broadcast end value
    //         int end = (g == parm->generations-1) ? 1 : 0;
    //         MPI_Bcast(&end, 1, MPI_INT, s_root, *s_comm);
    //
    //         SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');
    //
    //         if (parm->show_progress > 0) {
    //             int best = OrderedSelection(fitness, 0, parm->population_size);
    //
    //             SHOWPROGRESS(true, "%*d generation best individual FITNESS, QS, K: %6.4lf %4.2lf %2d\n", WIDTH(parm->generations), g+1, fitness[best], qs[best], k[best]);
    //             VERBOSE(true, parm->out_file, "%*d generation best individual FITNESS, QS, K: %6.4lf %4.2lf %2d\n", WIDTH(parm->generations), g+1, fitness[best], qs[best], k[best]);
    //         }
    //
    //         // Return best overall individual
    //         if (parm->ml_return_overall_best) {
    //             int best = OrderedSelection(fitness, 0, parm->population_size);
    //             if (fitness[best] > best_fitness)
    //                 CreateSelectionFromSubpopulations(parm, population[best], subpops, n_subpops, selection, instances, qs[best], NULL);
    //
    //             best_k = k[best];
    //             best_qs = qs[best];
    //             best_fitness = fitness[best];
    //         }
    //     }
    //
    //     if (!ISSKYNET()) {
    //         MPI_Barrier(*s_comm);
    //         MPI_Comm_free(s_comm);
    //     }
    //
    //     // Reevaluate and get best individual
    //     SHOWPROGRESS(TEST, "Last population start evaluation%c", '\n');
    //     MPIEvaluatePopulationFitness(parm, population, subpops, n_subpops, instances, qs, k, *f_comm, f_root, fitness, performance);
    //     SHOWPROGRESS(TEST, "Last population end evaluation%c", '\n');
    //
    //     // Send end to all slaves
    //     int *dummy;
    //     MALLOC(dummy, int, instances+3, "MPICooperativeCoevolutiveInstanceSelection: dummy");
    //     for (int i = 0, target = 0; i < parm->population_size; i++, target++) {
    //         if (target == f_root)
    //             target++;
    //
    //         dummy[instances] = -1;
    //         MPI_Send(dummy, instances+3, MPI_INT, target, 0, *f_comm);
    //     }
    //     FREE(dummy);
    //
    //     if (!ISSKYNET()) {
    //         MPI_Barrier(*f_comm);
    //         MPI_Comm_free(f_comm);
    //     }
    //
    //     if (!parm->ml_return_overall_best) {
    //         int best = OrderedSelection(fitness, 0, parm->population_size);
    //         CreateSelectionFromSubpopulations(parm, population[best], subpops, n_subpops, selection, instances, qs[best], NULL);
    //
    //         ((t_Parameters *) parm)->mlknn_k = k[best];
    //         fprintf(parm->out_file, "@CC best MLkNN k: %d\n", k[best]);
    //
    //         if (parm->merge_method == QuorumSensing)
    //             fprintf(parm->out_file, "@Quorum sensing threshold: %lf\n", qs[best]);
    //     } else {
    //         ((t_Parameters *) parm)->mlknn_k = best_k;
    //         fprintf(parm->out_file, "@CC best MLkNN k: %d\n", best_k);
    //
    //         if (parm->merge_method == QuorumSensing)
    //             fprintf(parm->out_file, "@Quorum sensing threshold: %lf\n", best_qs);
    //     }
    //
    //     // Free memory
    //     for (int i = 0; i < n_subpops; i++) {
    //         for (int j = 0; j < parm->subpop_size; j++)
    //             FREE(subpops[i][j]);
    //         FREE(subpops[i]);
    //         FREE(elite_members[i]);
    //     }
    //     FREE(subpops);
    //     for (int i = 0; i < parm->population_size; i++)
    //         FREE(population[i]);
    //     FREE(population);
    //     FREE(performance);
    //     FREE(fitness);
    //     FREE(cummulative);
    //     FREE(difference);
    //     FREE(cprobability);
    //     FREE(qs);
    //     FREE(k);
    //     FREE(elite);
    //     FREE(elite_members);
    //     FREE(elite_list);
    //
    //     int selected = 0;
    //     for (int i = 0; i < instances; i++)
    //         if (selection[i])
    //             selected++;
    //
    //     return selected;
}

// Create slaves and return intra-communicator
MPI_Comm *MPIRunSlaves(const t_Parameters *parms, int n_slaves, int *root, char *program)
{
    MPI_Comm everyone, *new;

    // Know where I am running
    char name[MAX_FILENAME_LENGTH];
    gethostname(name, MAX_FILENAME_LENGTH);

    SHOWPROGRESS(true, "Running master in host: '%s'\n", name);

    if (MPI_Comm_spawn(program, MPI_ARGV_NULL, n_slaves, MPI_INFO_NULL, 0, MPI_COMM_SELF, &everyone, MPI_ERRCODES_IGNORE) != MPI_SUCCESS)
        OnErrorExit(ERR_MPI, "MPI cannot start slaves");

    MALLOC(new, MPI_Comm, 1, "MPIRunSlaves: new");
    if (MPI_Intercomm_merge(everyone, 1, new) != MPI_SUCCESS)
        OnErrorExit(ERR_MPI, "Cannot merge inter communicators");

    // Get master rank and send it to all slaves
    MPI_Comm_rank(*new, root);
    for (int i = 0; i <= n_slaves; i++)
        if (i != *root)
            MPI_Send(root, 1, MPI_INT, i, 0, *new);

    MPI_Bcast((void *) parms, sizeof(t_Parameters), MPI_BYTE, *root, *new);

    return new;
}

void MPIEvaluatePopulationFitness(const t_Parameters *parm, int **population, int ***subpops, int n_subpops, int instances, scalar *qs, int *k, MPI_Comm communicator, int root, scalar *fitness,
                                  scalar *performance)
{
    int *to_send;
    MPI_Status status;
    scalar to_receive[3];

    MALLOC(to_send, int, instances+3, "MPIEvaluateIndividualFitness: selection");

    if (qs[0] < 0.0)
        OnErrorExit(ERR_MPI, "MPI cannot accept negative quorum sensing");

    // Send selections
    for (int i = 0, target = 0; i < parm->population_size; i++, target++) {
        // Get selection from individual
        CreateSelectionFromSubpopulations(parm, population[i], subpops, n_subpops, to_send, instances, qs[i], NULL);

        if (target == root)
            target++;

        // Send individual to any slave
        to_send[instances] = i;
        to_send[instances+1] = k[i];
        to_send[instances+2] = qs[1] *1000.0;   // This way it can be sent as an integer
        MPI_Send(to_send, instances+3, MPI_INT, target, 0, communicator);
    }

    FREE(to_send);

    SHOWPROGRESS(TEST, "%d individuals sent to slaves to evaluate\n", parm->population_size);

    // Receive fitness and index
    for (int i = 0; i < parm->population_size; i++) {
#ifdef SCALAR_IS_DOUBLE
        MPI_Recv(to_receive, 3, MPI_DOUBLE, MPI_ANY_SOURCE, 0, communicator, &status);
#else
        MPI_Recv(to_receive, 3, MPI_FLOAT, MPI_ANY_SOURCE, 0, communicator, &status);
#endif

        SHOWPROGRESS(TEST, "%d individual received from slave evaluated\n", status.MPI_SOURCE);

        fitness[(int) to_receive[0]] = to_receive[1];
        performance[(int) to_receive[0]] = to_receive[2];
    }
}
