/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdbool.h>
#include <mpi.h>
#include <string.h>
#include <unistd.h>
#include "macros.h"
#include "typex.h"
#include "ml.h"
#include "is.h"
#include "isfs.h"

// Global variables for C4.5 algorithm
unsigned short MaxClass, MaxDiscrVal = 2;
ItemNo MaxItem;
int MaxAtt;
Description *Item;
DiscrValue *MaxAttVal;
char *SpecialStatus;
String *ClassName, *AttName, * *AttValName, FileName;
short VERBOSITY, TRIALS;
Boolean GAINRATIO = true, SUBSET = false;
ItemNo MINOBJS = 2, WINDOW, INCREMENT;
float ConfLimit = 0.25;
Tree *Pruned;
Boolean AllKnown = true;

// MPI slave to evaluate one individual
int main(int argc, char *argv[])
{
    //     t_Parameters parm;
    //     int instances, **labels, *n_labels, classes, inputs, n_subpops, **index = NULL;
    //     scalar **data;
    //     cache_t **neighbors_cache = NULL;
    //     scalar to_send[3];
    //     int *to_receive;
    //
    //     // Get communicator with parent
    //     MPI_Comm communicator, parent;
    //
    //     if (MPI_Init(&argc, &argv) != MPI_SUCCESS)
    //         OnErrorExit(ERR_MPI, "MPI Initialize failed on slave");
    //     else
    //         printf("MPI Initialize successfully on slave\n");
    //     MPI_Comm_get_parent(&parent);
    //     if (MPI_Intercomm_merge(parent, 1, &communicator) != MPI_SUCCESS)
    //         OnErrorExit(ERR_MPI, "Cannot merge inter communicators");
    //
    //     // subpopulation index corresponds to rank
    //     int p_index;
    //     MPI_Comm_rank(MPI_COMM_WORLD, &p_index);
    //
    //     char name[MAX_FILENAME_LENGTH];
    //     gethostname(name, MAX_FILENAME_LENGTH);
    //     SHOWPROGRESS(true, "%d slave started execution in host: %s\n", p_index+1, name);
    //
    //     // Receive root rank
    //     int root;
    //     MPI_Status status;
    //     MPI_Recv(&root, 1, MPI_INT, MPI_ANY_SOURCE, 0, communicator, &status);
    //     SHOWPROGRESS(TEST, "%d slave received root rank %d\n", p_index+1, root);
    //
    //     // Receive parameters
    //     MPI_Bcast((void *) &parm, sizeof(t_Parameters), MPI_BYTE, root, communicator);
    //     SHOWPROGRESS(TEST, "%d slave received parameters\n", p_index+1);
    //
    //     if (parm.ml_use_hash_table)
    //         OnErrorExit(ERR_MPI, "Hash table incompatible with MPI");
    //
    //     // Initialize random number generator
    //     char state[256];
    //     if (parm.seed == 0) {
    //         FILE *randDev;
    //
    //         if ((randDev = fopen("/dev/urandom", "r")) == NULL)
    //             OnErrorExit(1, "main: Cannot open random device\n");
    //
    //         if (!fread(&parm.seed, 1, sizeof(int), randDev))
    //             OnErrorExit(ERR_READ, "main");
    //
    //         fclose(randDev);
    //     }
    //
    //     char file_name[100];
    //     strcpy(file_name, "/dev/null");
    //     if ((parm.out_file = fopen(file_name, "w")) == NULL)
    //         OnErrorExit(ERR_FILE, "Cannot open output file using /dev/null");
    //
    //     srandom(parm.seed);
    //     srand(parm.seed);
    //     for (int i = 0; i < 256; i++)
    //         state[i] = random();
    //     initstate(parm.seed, state, 256);
    //     setstate(state);
    //
    //     // Check parameters
    //     if (parm.reevaluate_individuals)
    //         OnErrorExit(ERR_PARAM, "Reevaluate incompatible with MPI");
    //
    //     // Allocate memory
    //     int sparse = 0;
    //     ReceiveData(communicator, &instances, &data, &labels, &n_labels, &inputs, &index, root, &n_subpops, &classes, &sparse);
    //     if (!sparse)
    //         index = NULL;
    //     SHOWPROGRESS(TEST, "EVALUATOR %d slave received data correctly\n", p_index+1);
    //
    //     // Cache of neighbors?
    //     SHOWPROGRESS(TEST, "BEGIN Cache construction finished%c", '\n');
    //     if (parm.use_neighbors_cache && NEIGHBORS_CACHE_SIZE(instances) < MAXIMUM_CACHE_SIZE) {
    //         int neighbors_cache_size = 0;
    //         ConstructAllNeighborsCache(data, instances, inputs, NULL, &neighbors_cache, NULL, NULL, &neighbors_cache_size, NULL, parm.ml_loo);
    //         SHOWPROGRESS(TEST, "DONE Cache construction finished%c", '\n');
    //     } else
    //         SHOWPROGRESS(TEST, "Neighbors cache DISABLED%c", '\n');
    //
    //     // Populations information
    //     int k, ind_index;
    //     scalar qs;
    //
    //     MALLOC(to_receive, int, instances+3, "main: to_receive");
    //
    //     // For evaluations of individuals
    //     t_MLkNN mlknn;
    //     t_MLksNN mlksnn;
    //     t_LAMLkNN lamlknn;
    //     void *classifier[N_ML_CLASSIFIERS] = {(void *) &mlknn, NULL, NULL, (void *) &lamlknn, (void *) &mlksnn, NULL};
    //
    //     CreateMultiLabelClassifier(parm, classifier[method], method, classes, instances, index, inputs, parm->ml_max_k, parm->lamlknn_max_n_clusters, NULL, data, n_labels, labels, NULL);
    //
    //     while (true) {
    //         SHOWPROGRESS(parm.show_progress > 2, "EVALUATOR %d slave waiting for individual info\n", p_index+1);
    //
    //         // Receive information of individual
    //         MPI_Recv(to_receive, instances+3, MPI_INT, root, 0, communicator, &status);
    //
    //         ind_index = to_receive[instances];
    //         k = to_receive[instances+1];
    //         qs = to_receive[instances+2]/1000.0;
    //
    //         SHOWPROGRESS(TEST, "%d EVALUATION slave received individual %d to evaluate\n", p_index+1, ind_index);
    //
    //         // A negative index means the end
    //         if (ind_index < 0)
    //             break;
    //
    //         SHOWPROGRESS(parm.show_progress > 2, "EVALUATOR %d slave received individual %d information\n", p_index+1, ind_index);
    //
    //         to_send[1] = EvaluateIndividualFitness((const t_Parameters *) &parm, NULL, NULL, 0, instances, labels, n_labels, classes, inputs, data, to_receive, neighbors_cache, NULL, &qs, k, index,
    //                                                &to_send[2], NULL, NULL, NULL, MLkNN);
    //
    //         SHOWPROGRESS(parm.show_progress > 2, "EVALUATOR %d slave finished evaluation of individual %d\n", p_index+1, ind_index);
    //
    //         // Send fitness and index
    //         to_send[0] = (scalar) ind_index;
    // #ifdef SCALAR_IS_DOUBLE
    //         MPI_Send(to_send, 3, MPI_DOUBLE, root, 0, communicator);
    // #else
    //         MPI_Send(to_send, 3, MPI_FLOAT, root, 0, communicator);
    // #endif
    //     }
    //
    //     FREE(to_receive);
    //
    //     MPI_Barrier(communicator);
    //
    //     SHOWPROGRESS(TEST, "%d EVALUATION slave ended correctly\n", p_index+1);
    //
    //     MPI_Finalize();
}
