/***************************************************************************
 *   Copyright (C) 2023 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"
#include "ensemble.h"

void PPTcClassifyInstance(const t_Parameters *parm, scalar *query, int *h, scalar *f, bool *classification, bool *ranks, scalar **data, int instances, int n_labels, int *n_predicted,
                                   int inputs, int *query_index, t_LP *lp, int *subspace, int *S, int **index)
{
    scalar *output;

    MALLOC(output, scalar, lp->n_classes, "PPTcClassifyInstance: output");

    // Output of classifier
    ClassifierProbabilities(parm, lp->method, query, inputs, lp->n_classes, query_index, subspace, data, instances, (lp->method == kNN || lp->method == NN) ? lp->model.knn->labels : NULL, S,
                            &(lp->model), -1, NULL, -1, index, output, &(lp->ensemble));

    // Top k predicted class cast a votes weighted by their probability
    ZeroVector(f, NULL, n_labels);
    for (int i = 0; i < parm->pptc_k; i++) {
        int category = IndexOfMaximumValue(output, lp->n_classes);

        scalar confidence = output[category]/2.0 + 0.5; // From [-1, 1] to [0, 1]
        output[category] = -10.0;   // Discard for next round

        // Cast weighted votes
        for (int i = 0; i < n_labels; i++)
            if (lp->decoded[category][i])
                f[i] += confidence;
    }

    // Relevant labels
    *n_predicted = 0;
    for (int i = 0; i < n_labels; i++)
        if (f[i] > parm->pptc_t) {
            h[*n_predicted] = i;
            (*n_predicted)++;
        }

    *classification = true;

    // Normalize probabilities
    NormalizeVector(f, n_labels);

    *ranks = true;

    FREE(output);
}

int IndexOfMaximumValue(scalar *v, int n)
{
    int index = 0;

    for (int i = 1; i < n; i++)
        if (v[index] < v[i])
            index = i;

    return index;
}
