/***************************************************************************
 *   Copyright (C) 2020 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "typex.h"
#include "macros.h"

// Returns the value centered around the best window

int BestWindowedValue(int min, int max, int window, scalar *fitness)
{
    int best = 0;
    scalar best_average = -1.0;

    for (int i = min; i <= max; i++) {
        scalar average = 0.0;

        int count = 0;
        for (int j = MAX(min, i-window/2); j <= MIN(max, i+window/2); j++) {
            count++;
            average += fitness[j];
        }

        average /= count;

        if (average > best_average) {
            best_average = average;
            best = i;
        }
    }

    SHOWPROGRESS(true, "Best window [%d, %d, %d] with averaged fitness: %lf\n", MAX(min, best-window/2), best, MIN(max, best+window/2), best_average);

    return best;
}
